# @file
# 
# - Find MPI
# This module looks for MPI support (generic lookup). It will set the following variables:
#  MPI_INCLUDE_DIR     = where mpi.h can be found
#  MPI_LIBRARY         = the mpi library to link in
#  MPI_EXTRA_LIBRARY   = additional libraries that you might need
#  MPI_LIBRARIES       = these are all the needed libraries ${MPI_LIBRARY};${MPI_EXTRA_LIBRARIES}
#
# Note: If MPI_HOME is set by the user this is the place where the FIND_* macros
#       will look first
#
# @author Blasius Czink
#

include (MarmotHelperMacros)

# show an option for MPICH_HOME
IF (MPI_HOME)
  # user selected something
  SET(MPI_HOME ${MPI_HOME} CACHE PATH "MPI_HOME")
  # set variables to ...-NOTFOUND to reset (Is there a more elegant way to do that?)
  BC_RESET(MPI_INCLUDE_DIR)
  BC_RESET(MPI_LIBRARY)
  BC_RESET(MPI_EXTRA_LIBRARY)
  BC_RESET(MPI_LIBRARIES)
ELSE (MPI_HOME)
  # nothing, so just create an empty field
  SET(MPI_HOME "" CACHE PATH "MPI_HOME")
ENDIF (MPI_HOME)

FIND_PATH(MPI_INCLUDE_DIR mpi.h
          PATHS
          "${MPI_HOME}/include"
          /usr/include 
          /usr/include/mpi
          /usr/local/include 
          /usr/local/mpi/include
          "$ENV{EXTERNLIBS}/mpi/include"
          "$ENV{ProgramFiles}/MPI/include"
          "$ENV{CCP_INC}"
          DOC "The MPI Header path"
          NO_DEFAULT_PATH
)

IF (CMAKE_CL_64)
  SET (MSMPI_LIBPATH "$ENV{CCP_LIB64}")
ELSE (CMAKE_CL_64)
  SET (MSMPI_LIBPATH "$ENV{CCP_LIB32}")
ENDIF (CMAKE_CL_64)

FIND_LIBRARY(MPI_LIBRARY
          NAMES mpi mpich2 mpich msmpi
          PATHS
          ${MPI_HOME}/lib
          /usr/lib
          /usr/local/lib
          /usr/local/mpi/lib
          "$ENV{EXTERNLIBS}/mpi/lib"
          "$ENV{ProgramFiles}/mpi/lib"
          ${MSMPI_LIBPATH}
)

IF (NOT MPI_EXTRA_LIBRARY)
  SET (MPI_EXTRA_LIBRARY "" CACHE STRING "Any additional mpi-libraries you might need" FORCE)
ENDIF (NOT MPI_EXTRA_LIBRARY)

SET(MPI_FOUND "NO")
# was include path found at least?
IF (MPI_INCLUDE_DIR)
  # adjust libvars & check
  BC_ADJUST_LIB_VARS(MPI)
  # check
  IF (MPI_LIBRARY)
    SET(MPI_FOUND YES)
    SET(MPI_LIBRARIES ${MPI_LIBRARY} ${MPI_EXTRA_LIBRARY} CACHE STRING "The MPI_LIBRARIES" FORCE)
    #  ADD_DEFINITIONS(-DMPICH_IGNORE_CXX_SEEK)
  ENDIF (MPI_LIBRARY)
  # dump the info
  #BC_DUMP_LIB_SETUP(MPI)
  #BC_DUMP_LIB_SETUP(MPI_EXTRA_LIBRARY)
  #BC_DUMP_LIB_SETUP(MPI_LIBRARIES)
ENDIF (MPI_INCLUDE_DIR)

IF (MPI_FOUND)
   IF (NOT MPI_FIND_QUIETLY)
      MESSAGE(STATUS "Found MPI")
   ENDIF (NOT MPI_FIND_QUIETLY)
ELSE (MPI_FOUND)
   IF (MPI_FIND_REQUIRED)
      MESSAGE(FATAL_ERROR "Could not find MPI")
   ENDIF (MPI_FIND_REQUIRED)
ENDIF (MPI_FOUND)

# MARK_AS_ADVANCED(MPI_...)
