# @file
# 
# - Find MPICH
# This module looks for MPICH support. It will set the following variables:
#  MPICH_INCLUDE_DIR = where mpi.h can be found
#  MPICH_LIBRARY     = the library to link in
#  MPICH_LIBRARIES   = all the needed libraries
#  
#  MPICH_mpichcxx_LIBRARY
#  MPICH_pmpich_LIBRARY  (if available - optional)
#  MPICH_fmpich_LIBRARY  (if available - optional)
#  For VisualStudio the debug versions of the libraries named above will be searched too.
#
# Note: If MPICH_HOME is set by the user this is the place where the FIND_* macros
#       will look first
#
# @author Blasius Czink
#

# include(MarmotHelperMacros)
include(FindPackageHandleStandardArgs)

# for compile-wrapper stuff (optional)
SET(MPICH_COMPILE_WRAPPER_PATHS
  "${MPICH_HOME}/bin"
  "$ENV{EXTERNLIBS}/mpich2/bin"
  "/usr/bin"
  "/usr/local/bin"
  "/opt/mpich/bin"
)

FIND_PROGRAM(MPI_CC 
  NAMES mpicc mpcc mpcc_r mpxlc mpxlc_r mpiicc 
  PATHS ${MPICH_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
FIND_PROGRAM(MPI_CXX 
  NAMES mpicxx mpcxx mpcxx_r mpxlcxx mpxlcxx_r mpiicxx
        mpic++ mpc++ mpc++_r mpxlc++ mpxlc++_r mpiic++
        mpiCC mpCC mpCC_r mpxlCC mpxlCC_r mpiiCC 
  PATHS ${MPICH_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
FIND_PROGRAM(MPI_F77 
  NAMES mpif77 mpf77 mpf77_r mpxlf77 mpxlf77_r mpiif77
  PATHS ${MPICH_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
FIND_PROGRAM(MPI_F90 
  NAMES mpif90 mpf90 mpf90_r mpxlf90 mpxlf90_r mpiif90
  PATHS ${MPICH_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
MARK_AS_ADVANCED(MPI_CC MPI_CXX MPI_F77 MPI_F90)


FIND_PATH(MPICH_INCLUDE_DIR mpi.h
  PATHS
  "${MPICH_HOME}/include"
  "$ENV{EXTERNLIBS}/mpich2/include"
  /usr/include 
  /usr/include/mpi
  /usr/local/include 
  /usr/local/mpi/include
  /opt/mpich/include
  "$ENV{ProgramFiles}/MPICH/SDK/Include"
  "$ENV{ProgramFiles}/MPICH2/include"
  "$ENV{ProgramW6432}/MPICH2/include"
  "C:/Program Files/MPICH/SDK/Include"
  [HKEY_LOCAL_MACHINE\\SOFTWARE\\MPICH2;Path]/include 
  NO_DEFAULT_PATH
)

FIND_LIBRARY(MPICH_LIBRARY
  NAMES libmpi mpich2 mpich mpi
  PATHS
  "${MPICH_HOME}/lib"
  "$ENV{EXTERNLIBS}/mpich2/lib"
  /usr/lib
  /usr/local/lib
  /usr/local/mpi/lib
  /opt/mpich/lib
  "$ENV{ProgramFiles}/MPICH/SDK/Lib"
  "$ENV{ProgramFiles}/MPICH2/Lib"
  "$ENV{ProgramW6432}/MPICH2/Lib"
  "C:/Program Files/MPICH/SDK/Lib" 
  [HKEY_LOCAL_MACHINE\\SOFTWARE\\MPICH2;Path]/lib
)

FIND_LIBRARY(MPICH_mpichcxx_LIBRARY
  NAMES libmpicxx mpichcxx cxx
  PATHS "${MPICH_HOME}/lib" "$ENV{EXTERNLIBS}/mpich2/lib"
  /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/mpich/lib
  "$ENV{ProgramFiles}/MPICH/SDK/Lib"
  "$ENV{ProgramFiles}/MPICH2/Lib"
  "$ENV{ProgramW6432}/MPICH2/Lib"
  "C:/Program Files/MPICH/SDK/Lib"
  [HKEY_LOCAL_MACHINE\\SOFTWARE\\MPICH2;Path]/lib
)

FIND_LIBRARY(MPICH_pmpich_LIBRARY
  NAMES pmpich
  PATHS "${MPICH_HOME}/lib" "$ENV{EXTERNLIBS}/mpich2/lib"
  /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/mpich/lib
  "$ENV{ProgramFiles}/MPICH2/Lib"
  "$ENV{ProgramW6432}/MPICH2/Lib"
  "$ENV{ProgramFiles}/MPICH/SDK/Lib"
  "C:/Program Files/MPICH/SDK/Lib" 
  [HKEY_LOCAL_MACHINE\\SOFTWARE\\MPICH2;Path]/lib
)

FIND_LIBRARY(MPICH_fmpich_LIBRARY
  NAMES libfmpich2g fmpich2
  PATHS "${MPICH_HOME}/lib" "$ENV{EXTERNLIBS}/mpich2/lib"
  /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/mpich/lib
  "$ENV{ProgramFiles}/MPICH2/Lib"
  "$ENV{ProgramW6432}/MPICH2/Lib"
  "$ENV{ProgramFiles}/MPICH/SDK/Lib"
  "C:/Program Files/MPICH/SDK/Lib" 
  [HKEY_LOCAL_MACHINE\\SOFTWARE\\MPICH2;Path]/lib
)
# mark some of the libs as advanced
MARK_AS_ADVANCED(MPICH_mpichcxx_LIBRARY MPICH_pmpich_LIBRARY MPICH_fmpich_LIBRARY)

# only for windows & MS-VisualStudio
IF (WIN32 AND MSVC)
  FIND_LIBRARY(MPICH_mpichcxx_LIBRARY_DEBUG
    NAMES mpichcxxd cxxd
    PATHS "${MPICH_HOME}/lib" "$ENV{EXTERNLIBS}/mpich2/lib"
    /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/mpich/lib
    "$ENV{ProgramFiles}/MPICH2/Lib"
    "$ENV{ProgramW6432}/MPICH2/Lib"
    "$ENV{ProgramFiles}/MPICH/SDK/Lib"
    "C:/Program Files/MPICH/SDK/Lib" 
    [HKEY_LOCAL_MACHINE\\SOFTWARE\\MPICH2;Path]/lib
  )
  FIND_LIBRARY(MPICH_pmpich_LIBRARY_DEBUG 
    NAMES pmpichd
    PATHS "${MPICH_HOME}/lib" "$ENV{EXTERNLIBS}/mpich2/lib"
    /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/mpich/lib
    "$ENV{ProgramFiles}/MPICH2/Lib"
    "$ENV{ProgramW6432}/MPICH2/Lib"
    "$ENV{ProgramFiles}/MPICH/SDK/Lib"
    "C:/Program Files/MPICH/SDK/Lib" 
    [HKEY_LOCAL_MACHINE\\SOFTWARE\\MPICH2;Path]/lib
  )
  MARK_AS_ADVANCED(MPICH_mpichcxx_LIBRARY_DEBUG MPICH_pmpich_LIBRARY_DEBUG)

  IF (MSVC_IDE)
    # VS-IDE
    IF (MPICH_LIBRARY)
       SET(MPICH_LIBRARIES general ${MPICH_LIBRARY})
    ENDIF (MPICH_LIBRARY)
    # MPICH2 packages upto v1.0.8p1 included a release and a debug version of this library
    IF (MPICH_mpichcxx_LIBRARY AND MPICH_mpichcxx_LIBRARY_DEBUG)
      SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} optimized ${MPICH_mpichcxx_LIBRARY} debug ${MPICH_mpichcxx_LIBRARY_DEBUG})
    ELSE (MPICH_mpichcxx_LIBRARY AND MPICH_mpichcxx_LIBRARY_DEBUG)
      # Version 1.1 only includes a release version
      SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} general ${MPICH_mpichcxx_LIBRARY})
    ENDIF (MPICH_mpichcxx_LIBRARY AND MPICH_mpichcxx_LIBRARY_DEBUG)
    # optional
    IF (MPICH_pmpich_LIBRARY AND MPICH_pmpich_LIBRARY_DEBUG)
       SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} optimized ${MPICH_pmpich_LIBRARY} debug ${MPICH_pmpich_LIBRARY_DEBUG})
    ENDIF (MPICH_pmpich_LIBRARY AND MPICH_pmpich_LIBRARY_DEBUG)
  ELSE (MSVC_IDE)
    # nmake
    STRING(TOLOWER ${CMAKE_BUILD_TYPE} CMAKE_BUILD_TYPE_TOLOWER)
    IF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
       SET(MPICH_LIBRARIES ${MPICH_LIBRARY})
       IF (MPICH_mpichcxx_LIBRARY_DEBUG)
         SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} ${MPICH_mpichcxx_LIBRARY_DEBUG})
       ELSE (MPICH_mpichcxx_LIBRARY_DEBUG)
         SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} ${MPICH_mpichcxx_LIBRARY})
       ENDIF (MPICH_mpichcxx_LIBRARY_DEBUG)
       IF(MPICH_pmpich_LIBRARY_DEBUG)
         SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} ${MPICH_pmpich_LIBRARY_DEBUG})
       ENDIF(MPICH_pmpich_LIBRARY_DEBUG)
    ELSE(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
       SET(MPICH_LIBRARIES ${MPICH_LIBRARY} ${MPICH_mpichcxx_LIBRARY})
       IF(MPICH_pmpich_LIBRARY)
         SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} ${MPICH_pmpich_LIBRARY})
       ENDIF(MPICH_pmpich_LIBRARY)
    ENDIF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
  ENDIF (MSVC_IDE)
  
  # optional fortran stuff
  IF (MARMOT_ENABLE_FORTRAN AND MPICH_fmpich_LIBRARY)
    SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} general ${MPICH_fmpich_LIBRARY})
  ENDIF (MARMOT_ENABLE_FORTRAN AND MPICH_fmpich_LIBRARY)

  FIND_PACKAGE_HANDLE_STANDARD_ARGS(MPICH "Could not locate MPICH" MPICH_LIBRARY MPICH_mpichcxx_LIBRARY MPICH_INCLUDE_DIR)

ELSE (WIN32 AND MSVC)  

  # nothing special to be done for the rest of the world
  SET(MPICH_LIBRARIES ${MPICH_LIBRARY} ${MPICH_mpichcxx_LIBRARY})
  # append optional libs
  IF(MPICH_pmpich_LIBRARY)
    SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} ${MPICH_pmpich_LIBRARY})
  ENDIF(MPICH_pmpich_LIBRARY)
  # for fortran
  IF (MARMOT_ENABLE_FORTRAN AND MPICH_fmpich_LIBRARY)
    SET(MPICH_LIBRARIES ${MPICH_LIBRARIES} ${MPICH_fmpich_LIBRARY})
  ENDIF (MARMOT_ENABLE_FORTRAN AND MPICH_fmpich_LIBRARY)
  
  FIND_PACKAGE_HANDLE_STANDARD_ARGS(MPICH "Could not locate MPICH" MPICH_LIBRARY MPICH_mpichcxx_LIBRARY MPICH_INCLUDE_DIR)

ENDIF (WIN32 AND MSVC)

IF (MPICH_FOUND)
  # set variables MPI_INCLUDE_DIR and MPI_LIBRARIES which will be used by marmot
  SET (MPI_INCLUDE_DIR ${MPICH_INCLUDE_DIR} CACHE PATH "The MPI-Header path" FORCE)
  SET (MPI_LIBRARIES ${MPICH_LIBRARIES} CACHE STRING "The MPI-Libraries" FORCE)
  MARK_AS_ADVANCED (MPI_LIBRARIES MPI_INCLUDE_DIR)
  IF (MINGW)
    ADD_DEFINITIONS("-DMPICH_SKIP_MPICXX")
  ENDIF (MINGW)
ENDIF (MPICH_FOUND)
