# @file
#
# - Find OpenMPI
# This module looks for OpenMPI support. It will set the following variables:
#  OPENMPI_INCLUDE_DIR = where mpi.h can be found
#  OPENMPI_LIBRARY     = the library to link in
#  OPENMPI_LIBRARIES   = all the needed libraries
#  OPENMPI_mpicxx_LIBRARY_[RELEASE/DEBUG]
#  OPENMPI_orte_LIBRARY_[RELEASE/DEBUG]
#  OPENMPI_pal_LIBRARY_[RELEASE/DEBUG]
#
# -lmpi_cxx -lmpi -lopen-rte -lopen-pal
# Note: If OPENMPI_HOME is set by the user this is the place where the FIND_* macros
#       will look first
#
# @author Blasius Czink
#

# include (MarmotHelperMacros)
include(FindPackageHandleStandardArgs)

# for compile-wrapper stuff (optional)
SET(OPENMPI_COMPILE_WRAPPER_PATHS
  "${OPENMPI_HOME}/bin"
  "$ENV{EXTERNLIBS}/openmpi/bin"
  "/usr/bin"
  "/usr/local/bin"
  "/usr/local/mpi/bin"
  "/opt/openmpi/bin"
  "/opt/mpi/bin"
  "$ENV{ProgramFiles}/OpenMPI/bin"
  "[HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_BINDIR]"
)

FIND_PROGRAM(MPI_CC 
  NAMES mpicc mpcc mpcc_r mpxlc mpxlc_r mpiicc 
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
FIND_PROGRAM(MPI_CXX 
  NAMES mpicxx mpcxx mpcxx_r mpxlcxx mpxlcxx_r mpiicxx
        mpic++ mpc++ mpc++_r mpxlc++ mpxlc++_r mpiic++
        mpiCC mpCC mpCC_r mpxlCC mpxlCC_r mpiiCC 
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
FIND_PROGRAM(MPI_F77 
  NAMES mpif77 mpf77 mpf77_r mpxlf77 mpxlf77_r mpiif77
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
FIND_PROGRAM(MPI_F90 
  NAMES mpif90 mpf90 mpf90_r mpxlf90 mpxlf90_r mpiif90
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  NO_SYSTEM_ENVIRONMENT_PATH
)
MARK_AS_ADVANCED(MPI_CC MPI_CXX MPI_F77 MPI_F90)

FIND_PATH(OPENMPI_INCLUDE_DIR mpi.h
          PATHS
          "${OPENMPI_HOME}/include"
          "$ENV{EXTERNLIBS}/openmpi/include"
          /usr/include 
          /usr/include/mpi
          /usr/local/include 
          /usr/local/mpi/include
          /opt/openmpi/include
          /opt/mpi/include
          "$ENV{ProgramFiles}/OpenMPI/include"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_INCLUDEDIR] 
          NO_DEFAULT_PATH
)

FIND_LIBRARY(OPENMPI_LIBRARY
          NAMES mpi libmpi
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
)

FIND_LIBRARY(OPENMPI_mpicxx_LIBRARY
          NAMES mpi_cxx mpicxx libmpi_cxx libmpicxx
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
)

FIND_LIBRARY(OPENMPI_orte_LIBRARY
          NAMES open-rte libopen-rte
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
)

FIND_LIBRARY(OPENMPI_pal_LIBRARY
          NAMES open-pal libopen-pal
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
)

FIND_LIBRARY(OPENMPI_mpif77_LIBRARY
          NAMES libmpi_f77 mpi_f77
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
)

# mark some of the libs as advanced
MARK_AS_ADVANCED(OPENMPI_mpicxx_LIBRARY OPENMPI_orte_LIBRARY OPENMPI_pal_LIBRARY OPENMPI_mpif77_LIBRARY)

# only for windows & MS-VisualStudio
IF (WIN32 AND MSVC)
  FIND_LIBRARY(OPENMPI_LIBRARY_DEBUG
          NAMES mpid libmpid
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
  )
  
  FIND_LIBRARY(OPENMPI_mpicxx_LIBRARY_DEBUG
          NAMES mpi_cxxd mpicxxd libmpi_cxxd libmpicxxd
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
  )
  
  FIND_LIBRARY(OPENMPI_orte_LIBRARY_DEBUG
          NAMES open-rted libopen-rted
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
  )

  FIND_LIBRARY(OPENMPI_pal_LIBRARY_DEBUG
          NAMES open-pald libopen-pald
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
  )
  
  FIND_LIBRARY(OPENMPI_mpif77_LIBRARY_DEBUG
          NAMES libmpi_f77d mpi_f77d
          PATHS ${OPENMPI_HOME}/lib "$ENV{EXTERNLIBS}/openmpi/lib"
          /usr/lib /usr/local/lib /usr/local/mpi/lib /opt/openmpi/lib /opt/mpi/lib
          "$ENV{ProgramFiles}/OpenMPI/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]
  )
  
  MARK_AS_ADVANCED(OPENMPI_mpicxx_LIBRARY_DEBUG OPENMPI_orte_LIBRARY_DEBUG OPENMPI_pal_LIBRARY_DEBUG OPENMPI_mpif77_LIBRARY_DEBUG)
  
  IF (MSVC_IDE)
    # VS-IDE
    IF (OPENMPI_LIBRARY AND OPENMPI_LIBRARY_DEBUG AND OPENMPI_mpicxx_LIBRARY AND OPENMPI_mpicxx_LIBRARY_DEBUG)
       SET(OPENMPI_LIBRARIES optimized ${OPENMPI_LIBRARY} optimized ${OPENMPI_mpicxx_LIBRARY} debug ${OPENMPI_LIBRARY_DEBUG} debug ${OPENMPI_mpicxx_LIBRARY_DEBUG})
    ENDIF (OPENMPI_LIBRARY AND OPENMPI_LIBRARY_DEBUG AND OPENMPI_mpicxx_LIBRARY AND OPENMPI_mpicxx_LIBRARY_DEBUG)
    #IF (OPENMPI_orte_LIBRARY AND OPENMPI_orte_LIBRARY_DEBUG)
    #   SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} optimized ${OPENMPI_orte_LIBRARY} debug ${OPENMPI_orte_LIBRARY_DEBUG})
    #ENDIF (OPENMPI_orte_LIBRARY AND OPENMPI_orte_LIBRARY_DEBUG)
    #IF (OPENMPI_pal_LIBRARY AND OPENMPI_pal_LIBRARY_DEBUG)
    #   SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} optimized ${OPENMPI_pal_LIBRARY} debug ${OPENMPI_pal_LIBRARY_DEBUG})
    #ENDIF (OPENMPI_pal_LIBRARY AND OPENMPI_pal_LIBRARY_DEBUG)
    # optional fortran stuff
    IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY AND OPENMPI_mpif77_LIBRARY_DEBUG)
      SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} optimized ${OPENMPI_mpif77_LIBRARY} debug ${OPENMPI_mpif77_LIBRARY_DEBUG})
    ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY AND OPENMPI_mpif77_LIBRARY_DEBUG)
  ELSE (MSVC_IDE)
    # nmake
    STRING(TOLOWER ${CMAKE_BUILD_TYPE} CMAKE_BUILD_TYPE_TOLOWER)
    IF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
      SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARY_DEBUG} ${OPENMPI_mpicxx_LIBRARY_DEBUG})
      #IF(OPENMPI_orte_LIBRARY_DEBUG)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_orte_LIBRARY_DEBUG})
      #ENDIF(OPENMPI_orte_LIBRARY_DEBUG)
      #IF(OPENMPI_pal_LIBRARY_DEBUG)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_pal_LIBRARY_DEBUG})
      #ENDIF(OPENMPI_pal_LIBRARY_DEBUG)
      IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY_DEBUG)
        SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_mpif77_LIBRARY_DEBUG})
      ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY_DEBUG)
    ELSE(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
      SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARY} ${OPENMPI_mpicxx_LIBRARY})
      #IF(OPENMPI_orte_LIBRARY)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_orte_LIBRARY})
      #ENDIF(OPENMPI_orte_LIBRARY)
      #IF(OPENMPI_pal_LIBRARY)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_pal_LIBRARY})
      #ENDIF(OPENMPI_pal_LIBRARY)
      IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
        SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_mpif77_LIBRARY})
      ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
    ENDIF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
  ENDIF (MSVC_IDE)

  FIND_PACKAGE_HANDLE_STANDARD_ARGS(OPENMPI "Could not locate OpenMPI" OPENMPI_LIBRARY OPENMPI_LIBRARY_DEBUG OPENMPI_mpicxx_LIBRARY OPENMPI_mpicxx_LIBRARY_DEBUG OPENMPI_INCLUDE_DIR)

ELSE (WIN32 AND MSVC)  

  # nothing special to be done for the rest of the world
  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARY} ${OPENMPI_mpicxx_LIBRARY})
  # append optional libs
  #IF(OPENMPI_orte_LIBRARY)
  #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_orte_LIBRARY})
  #ENDIF(OPENMPI_orte_LIBRARY)
  #IF(OPENMPI_pal_LIBRARY)
  #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_pal_LIBRARY})
  #ENDIF(OPENMPI_pal_LIBRARY)
  # for fortran
  IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
    SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_mpif77_LIBRARY})
  ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
  
  FIND_PACKAGE_HANDLE_STANDARD_ARGS(OPENMPI "Could not locate OpenMPI" OPENMPI_LIBRARY OPENMPI_mpicxx_LIBRARY OPENMPI_INCLUDE_DIR)

ENDIF (WIN32 AND MSVC)

IF (OPENMPI_FOUND)
  # set variables MPI_INCLUDE_DIR and MPI_LIBRARIES which will be used by marmot
  SET (MPI_INCLUDE_DIR ${OPENMPI_INCLUDE_DIR} CACHE PATH "The MPI-Header path" FORCE)
  SET (MPI_LIBRARIES ${OPENMPI_LIBRARIES} CACHE STRING "The MPI-Libraries" FORCE)
  MARK_AS_ADVANCED (MPI_LIBRARIES MPI_INCLUDE_DIR)
ENDIF (OPENMPI_FOUND)
