/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This file declares and implements the MPO_ResourceGathererDummy class.
 * @author Tobias Hilbrich
 * @date 10.01.2007
 * $Id: $ 
 */

#ifndef MPO_RESOURCEGATHERERDUMMY_H
#define MPO_RESOURCEGATHERERDUMMY_H

namespace MPO 
{
    /**
     * This implementation of the MPO_ResourceGatherer interface is 
     * used when resource logging/tracing is deactivated. It just does 
     * nothing. Thus it is not neccessary to put resource trace type
     * switches in the code. 
     */
    template < class T_RESOURCE_TYPE > class MPO_ResourceGathererDummy : public MPO_ResourceGatherer<T_RESOURCE_TYPE>
    {
    public:
		
        /**
         * Finds a resource object and sets its free info to the given
         * source location.
         * If there is no resource with the speciefied handle present the 
         * operation is aborted!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the resource object to find.
         * @param loc the location of the free call.
         * @return the resource trace found or NULL if the location is invalid 
         * 			or no trace with the given handle is present.
         */  
        inline void findAndSetFree (T_RESOURCE_TYPE handle, MPO_SourceLocation loc) {};
		
        /**
         * Finds a resource object and sets its commit info to the given
         * source location.
         * If there is no resource with the speciefied handle present the 
         * operation is aborted!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the resource object to find.
         * @param loc the location of the commit call.
         * @return the resource trace found or NULL if the location is invalid 
         * 			or no trace with the given handle is present.
         */  
        inline void findAndSetCommit (T_RESOURCE_TYPE handle, MPO_SourceLocation loc) {};
		
        /**
         * Finds a resource object and uses it to print its trace to the given 
         * ostream.
         * If there is no resource with the speciefied handle present the 
         * operation is aborted!
         * @param handle the handle of the resource object to find.
         * @param out the target for the print operation.
         */  
        inline void findAndPrintTrace (T_RESOURCE_TYPE handle, std::ostream & out) {};
		
        /**
         * Finds a resource object and uses it to print its trace to the given 
         * ostream.
         * If there is no resource with the speciefied handle present the 
         * operation is aborted!
         * @param handle the handle of the resource object to find.
         * @param out the target for the print operation.
         * @param argument_name the name of the associated argument.
         */
        inline void findAndPrintTraceNamed (T_RESOURCE_TYPE handle, std::ostream & out, std::string argument_name) {};
		
        /**
         * Finds a resource object and an resource object it depends on.
         * If both are found the second one's creation location is added to the 
         * first one's dependencies.
         * If one of the resources was not found the operation is aborted
         * i.e because the second resource was a standard resource that is not
         * traced !
         * @param handle the handle of the dependent resource object.
         * @param dependency the handle of the resource object representing the dependency.
         */  
        inline void findAndTryAddDependendency (T_RESOURCE_TYPE handle, T_RESOURCE_TYPE dependency) {};
		
        /**
         * This method searches the resource trace for handle. If found it adds 
         * loc to its usage list.
         * If not found it aborts the operation.
         * @param handle the handle of the resource to find.
         * @param loc the usage location that should be added to the resource trace of handle.
         * @return the resource trace found or NULL.
         */
        inline void findAndAddUsage (T_RESOURCE_TYPE handle, MPO_SourceLocation loc) {};
		
        /**
         * Adds a resource object and sets its creation info to the given
         * source location.
         * If there is already a resource with the same handle present, the old 
         * one will be removed!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the new resource object.
         * @param loc the source location of the creation call.
         * @return the newly created resource trace or null if the location is invalid.
         */  
        inline void addAndSetCreation (T_RESOURCE_TYPE handle, MPO_SourceLocation loc) {};
		
        /**
         * Removes the resource with the given handle.
         * @param handle the handle of the resource to be removed.
         */
        inline void remove (T_RESOURCE_TYPE handle) {};
		
        /**
         * This method returns the maximum amount of resources ever used.
         * @return the max amount of resources used.
         */
        inline int getMaxNumResourcesUsed (void) {return -1;};
		
        /**
         * Returns the amount of resources that are not freed.
         * If resources don`t have a free information the return 
         * value is 0.
         * @return 	amount of resources not freed or 0 if this information 
         * 			is not available.
         */
        inline int getNumResourcesNotFreed (void) {return -1;};
		
        /**
         * Prints information about the resources to the given ostream.
         * @param out ostream to store the information in.
         */
        inline void printResourcesNotFreed (std::ostream & out) {};
		
        /**
         * Prints information for the amount of Resources used.
         * The output is thougth to be processed by a tool like
         * gnuplot in order to gain information on resource usage
         * and its timing.
         * @param out the stream to store the data in.
         */
        inline void printResourcesUsedGraph(std::ostream & out) {};
		
        /**
         * Prints information for the amount of Resources used.
         * The output is thougth to be processed by a tool like
         * gnuplot in order to gain information on resource usage
         * and its timing.
         * The file won`t be created if no data is present.
         * @param file_name the name of the file to store the output in
         * 		  the file is only created if data is present.
         */
        inline void printResourcesUsedGraph(string file_name) {};
		
    }; //end of class MPO_ResourceGathererDummy
	
} //end of namespace MPO

#endif /*MPO_RESOURCEGATHERERDUMMY_H*/
