/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This file declares and implements the MPO_ResourceGathererGraph class.
 * @author Tobias Hilbrich
 * @date 10.01.2007
 * $Id: $
 */

#ifndef MPO_RESOURCEGATHERERGRAPH_H
#define MPO_RESOURCEGATHERERGRAPH_H

#include "MPO_ResourceGatherer.h"
#include <list>

namespace MPO
{
    /**
     * This implementation of the MPO_ResourceGatherer interface is
     * used when resource tracing is wanted but no source locations
     * are present.
     */
    template < class T_RESOURCE_TYPE > class MPO_ResourceGathererGraph : public MPO_ResourceGatherer<T_RESOURCE_TYPE>
    {
    protected:

        /** Map with all the resources that are active. */
        std::list < T_RESOURCE_TYPE > my_active_resources;

        /** The name used for the class of resources (i.e. MPI_Datatype). */
        std::string my_resource_name;

        /** The data for the graph. */
        std::map < double, int > my_graph;

        /** The maximum amount of resources known. */
        int my_max_num_ress;

        /** Creation time of this ResourceGatherer used for printing creation timings of resources. */
        double my_creation_time;

        /**
         * Searches wheter the handle is listed in my_active_resources.
         * @param handle the handle identifing the resource.
         * @return true if the handle was found false otherwise.
         */
        bool find (T_RESOURCE_TYPE handle);

    public:

        /**
         * Constructor.
         * @param ress_name the name of the associated resource.
         */
        MPO_ResourceGathererGraph (string ress_name);

        /**
         * Finds a resource object and sets its free info to the given
         * source location.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the resource object to find.
         * @param loc the location of the free call.
         * @return the resource trace found or NULL if the location is invalid
         * 			or no trace with the given handle is present.
         */
        void findAndSetFree (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

        /**
         * Finds a resource object and sets its commit info to the given
         * source location.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the resource object to find.
         * @param loc the location of the commit call.
         * @return the resource trace found or NULL if the location is invalid
         * 			or no trace with the given handle is present.
         */
        void findAndSetCommit (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

        /**
         * Finds a resource object and uses it to print its trace to the given
         * ostream.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * @param handle the handle of the resource object to find.
         * @param out the target for the print operation.
         */
        void findAndPrintTrace (T_RESOURCE_TYPE handle, std::ostream & out);

        /**
         * Finds a resource object and uses it to print its trace to the given
         * ostream.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * @param handle the handle of the resource object to find.
         * @param out the target for the print operation.
         * @param argument_name the name of the associated argument.
         */
        void findAndPrintTraceNamed (T_RESOURCE_TYPE handle, std::ostream & out, std::string argument_name);

        /**
         * Finds a resource object and an resource object it depends on.
         * If both are found the second one's creation location is added to the
         * first one's dependencies.
         * If one of the resources was not found the operation is aborted
         * i.e because the second resource was a standard resource that is not
         * traced !
         * @param handle the handle of the dependent resource object.
         * @param dependency the handle of the resource object representing the dependency.
         */
        void findAndTryAddDependendency (T_RESOURCE_TYPE handle, T_RESOURCE_TYPE dependency);

        /**
         * This method searches the resource trace for handle. If found it adds
         * loc to its usage list.
         * If not found it aborts the operation.
         * @param handle the handle of the resource to find.
         * @param loc the usage location that should be added to the resource trace of handle.
         * @return the resource trace found or NULL.
         */
        void findAndAddUsage (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

        /**
         * Adds a resource object and sets its creation info to the given
         * source location.
         * If there is already a resource with the same handle present, the old
         * one will be removed!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the new resource object.
         * @param loc the source location of the creation call.
         * @return the newly created resource trace or null if the location is invalid.
         */
        void addAndSetCreation (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

        /**
         * Removes the resource with the given handle.
         * @param handle the handle of the resource to be removed.
         */
        void remove (T_RESOURCE_TYPE handle);

        /**
         * This method returns the maximum amount of resources ever used.
         * @return the max amount of resources used.
         */
        int getMaxNumResourcesUsed (void);

        /**
         * Returns the amount of resources that are not freed.
         * If resources don`t have a free information the return
         * value is 0.
         * @return 	amount of resources not freed or 0 if this information
         * 			is not available.
         */
        int getNumResourcesNotFreed (void);

        /**
         * Prints information about the resources to the given ostream.
         * @param out ostream to store the information in.
         */
        void printResourcesNotFreed (std::ostream & out);

        /**
         * Prints information for the amount of Resources used.
         * The output is thougth to be processed by a tool like
         * gnuplot in order to gain information on resource usage
         * and its timing.
         * @param out the stream to store the data in.
         */
        void printResourcesUsedGraph(std::ostream & out);

        /**
         * Prints information for the amount of Resources used.
         * The output is thougth to be processed by a tool like
         * gnuplot in order to gain information on resource usage
         * and its timing.
         * The file won`t be created if no data is present.
         * @param file_name the name of the file to store the output in
         * 		  the file is only created if data is present.
         */
        void printResourcesUsedGraph(string file_name);

    }; //end of class MPO_ResourceGathererDummy

    //Method MPO_ResourceGathererGraph
    //========================================================================
    template < class T_RESOURCE_TYPE > MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::MPO_ResourceGathererGraph (string ress_name) :
        my_resource_name (ress_name),
        my_max_num_ress(0),
        my_creation_time (PMPI_Wtime())
    {
        //my_graph.insert(std::make_pair(my_creation_time,0));
    }

    //Method MPO_ResourceGathererGraph
    //========================================================================
    template < class T_RESOURCE_TYPE > bool MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::find (T_RESOURCE_TYPE handle)
    {
        //this is pretty slow we might want to enhance this
        for (typename std::list<T_RESOURCE_TYPE>::iterator i = my_active_resources.begin();i != my_active_resources.end();i++)
        {
            if (handle == *i)
                return true;
        }

        return false;
    }

    //Method findAndSetFree
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::findAndSetFree (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
        my_active_resources.remove(handle);
        int size = my_active_resources.size();

        my_graph.insert(std::make_pair(PMPI_Wtime(),size));
    }

    //Method findAndSetCommit
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::findAndSetCommit (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
        //nothing to do we don`t care about commits
    }

    //Method findAndPrintTrace
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::findAndPrintTrace (T_RESOURCE_TYPE handle, std::ostream & out)
    {
        //nothing to do we dont have the information to print resource traces
    }

    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::findAndPrintTraceNamed (T_RESOURCE_TYPE handle, std::ostream & out, std::string argument_name)
    {
        //nothing to do we dont have the information to print resource traces
    }

    //Method findAndTryAddDependendency
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::findAndTryAddDependendency (T_RESOURCE_TYPE handle, T_RESOURCE_TYPE dependency)
    {
        //nothing to do we dont care about dependencies
    }

    //Method findAndAddUsage
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::findAndAddUsage (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
        //nothing to do we dont care about usage
    }

    //Method addAndSetCreation
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::addAndSetCreation (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
        my_active_resources.push_back(handle);

        int size = my_active_resources.size();

        if (size > my_max_num_ress)
            my_max_num_ress = size;

        my_graph.insert(std::make_pair(PMPI_Wtime(),size));
    }

    //Method remove
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::remove (T_RESOURCE_TYPE handle)
    {
        my_active_resources.remove(handle);
    }

    //Method getMaxNumResourcesUsed
    //========================================================================
    template < class T_RESOURCE_TYPE > int MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::getMaxNumResourcesUsed (void)
    {
        return my_max_num_ress;
    }

    //Method getNumResourcesNotFreed
    //========================================================================
    template < class T_RESOURCE_TYPE > int MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::getNumResourcesNotFreed (void)
    {
        return my_active_resources.size();
    }

    //Method printResourcesNotFreed
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::printResourcesNotFreed (std::ostream & out)
    {
        out << "Can`t give detailed information about the resources as no source-locations are present.";
    }

    //Method printResourcesUsedGraph
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::printResourcesUsedGraph(std::ostream & out)
    {
        //add a last point
        int size = my_active_resources.size();
        my_graph.insert(std::make_pair(PMPI_Wtime(),size));

        //now we print the stuff to the ostream
    	out << "#Time\tResources-Used" << std::endl;

    	for (std::map < double, int >::iterator i = my_graph.begin();i != my_graph.end();i++)
    	{
            out << (double)(i->first) << "\t" << i->second << std::endl;
            //out << (double)(i->first-my_creation_time) << "\t" << i->second << std::endl;
    	}
    }

    //Method printResourcesUsedGraph
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererGraph<T_RESOURCE_TYPE>::printResourcesUsedGraph(string file_name)
    {
        if (my_max_num_ress <= 0)
            return;

        std::fstream out;
        out.open (file_name.c_str(), std::fstream::out);
        printResourcesUsedGraph(out);
        out.close();
    }

} //end of namespace MPO

#endif /*MPO_RESOURCEGATHERERDUMMY_H*/
