/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This file declares the MPO_ResourceTrace class.
 * @author Tobias Hilbrich
 * @date 28.11.2006
 * $Id: $ 
 */

#ifndef MPO_RESOURCETRACE_H
#define MPO_RESOURCETRACE_H

/**
 * @todo maybe we should create a environmental variable for this value.
 */
#define NUM_USAGE_CALLS_TRACED 5

#include "mpo_SourceLocation.h"
#include <deque>

namespace MPO 
{
	
    class MPO_ResourceTrace
    {
    protected:
	
        /** The source location of the free call of this resource. */
        MPO_SourceLocation my_free_location;
        /** The source location of the commit call of this resource. */
        MPO_SourceLocation my_commit_location;
        /** The source location of the creation call of this resource. */
        MPO_SourceLocation my_creation_location;
        /** The source locations of the last calls using this resource. */
        std::deque < MPO_SourceLocation > my_usage_locations;
        /** The source locations of resources from which this one is dependent (the location their creation). */
        std::deque < MPO_SourceLocation > my_dependency_locations;
        /** The creation time (recieved by MPI_Wtime) of this resource. 
         * Important: 	In fact the creation time of this object is stored, but that should in most
         * 				cases also be the creation time of the associated resource.
         */ 
        double my_creation_time;
        /** The time (recieved by MPI_Wtime) when this resource was freed. */
        double my_free_time;
		
        /** Flag used to describe whether usage of the associated resource is traced. */
        bool my_usage_flag;
        /** Flag used to describe whether freeing the resource associated is traced. */
        bool my_free_flag;
        /** Flag used to describe whether comiting the resource associated is traced. */
        bool my_commit_flag;
        /** Flag used to describe whether creating the resource associated is traced. */
        bool my_creation_flag;
        /** Flag used to describe whether dependencies of the resource associated are traced. */
        bool my_dependencies_flag;
        /** The name used for the class of resources (i.e. MPI_Datatype). */
        std::string my_resource_name;
					
    public:
	
        /**
         * Constructor which sets the behavior of the resource trace.
         * @param resource_name the name of the resource-type i.e. "MPI_Datatype".
         * @param has_usage flag that sets whether resource usage is traced or not. 
         * @param has_free flag that sets whether freeing resources is traced or not.
         * @param has_commit flag that sets whether commiting resources is traced or not.
         * @param has_creation flag that sets whether resource creation is traced or not.
         * @param has_dependencies flag that sets whether resource dependencies are traced or not.
         */	
        MPO_ResourceTrace (std::string resource_name,bool has_usage,bool has_free,bool has_commit,bool has_creation,bool has_dependencies);
		
        /**
         * Destructor that deletes this object.
         */
        ~MPO_ResourceTrace (void);
		
        /**
         * Adds a source location to the usage list.
         * Important: only the last few usage calls are stored.
         * @param location the location where the resource was used.
         */
        void addUsage (MPO_SourceLocation location);
		
        /**
         * Sets the source location for the call that freed this resource.
         * @param location the location where the resource was freed.
         */
        void setFree (MPO_SourceLocation location);
		
        /**
         * Returns a flag that indicates wether the resource was freed or not.
         * @return true if the resource was freed, false otherwise.
         */
        bool isFreed (void);
		
        /**
         * Sets the source location for the call that commited this resource.
         * @param location the location where the resource was commited.
         */
        void setCommit (MPO_SourceLocation location);
		
        /**
         * Sets the source location for the call that created this resource.
         * @param location the location where the resource was created.
         */
        void setCreation (MPO_SourceLocation location);
		
        /**
         * Gets the source location for the call that created this resource.
         * @return the source location of the call that created this resource.
         */
        MPO_SourceLocation getCreation ();
		
        /**
         * Adds a source location for another resource from which this one 
         * is dependent.
         * @param creation location of the resource from which this one is dependent.
         */
        void addDependency (MPO_SourceLocation location);
		
        /**
         * Prints the resource trace to a stream.
         * @param out the stream to which the trace is printed.
         * @return the stream (hint this is nonsense but done allover in the code so I just join in).
         */  
        std::ostream& printTrace (std::ostream & out);
		
        /**
         * Returns the creation time of this resource.
         * @return the time when the resource was creates (the value recieved with MPI_Wtime).
         */
        double getCreationTime (void);
		
        /**
         * Returns the free time of this resource.
         * @return the time when the resource was freed (the value recieved with MPI_Wtime).
         */
        double getFreeTime (void);
		
        /**
         * Tests wheter this resource was used at the given time.
         * @param timeToCheck the time which will be tested against this resources 
         * 		  creation and free time.
         * @return true when the resource was used at the given time.
         */
        bool wasUsedAtTime (double timeToCheck);
				
    }; //end of class MPO_ResourceTrace
	
} //end of namespace MPO

#endif /*MPO_RESOURCETRACE_H*/

