/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
  * \file MPO_ThreadSynchronisationSyncBase.h
  * This Headerfile describes the base class for all synchronized implementations
  * of MPO_IThreadSynchronisation.
  * @author Tobias Hilbrich
  * @date 27.06.2007
  * $ID$
  */

#ifndef MPO_THREADSYNCHRONISATIONSYNCBASE_H
#define MPO_THREADSYNCHRONISATIONSYNCBASE_H

#include "config.h"
#include "MPO_IThreadSynchronisation.h"
#include <map>
#include <vector>
#include <string>
#include <mpo.h>

namespace MPO
{
    /**
      * This class is an OpenMP implementation of the MPO_IThreadSynchronisation interface.
      * It provides an implmentation that supports threadsynchonisation for Threads created
      * with OpenMP.
      */
   class MPO_ThreadSynchronisationSyncBase : public MPO_IThreadSynchronisation
    {
    protected:

        /**
         * This struct is used to store the source information of a call.
         */
        typedef struct
        {
            MPO_SourceLocation loc; /**< Source code location. */
            int threadId; /**< Id of the thread making the call. */
            int callId; /**< Id of the call. */
            string callName; /**< Name of the call. */
        } sourceInfo;

        /**
         * Type for the map we will use to store requests, ids and call locations.
         */
        typedef std::map<int,std::pair<std::vector<MPI_Request*>, sourceInfo > > my_mapType;

        my_mapType my_registeredRequests;/**< This map is used to associated ids with a call location and requests. */

        /**
         * Type for the map we will use to store requests, ids and call locations.
         */
        typedef std::map<int,std::pair<MPI_Comm, sourceInfo > > my_mapTypeComm;

        my_mapTypeComm my_registeredComms;/**< This map is used to associated ids with a call location and requests. */

        int my_minThreadLevel; /**< Minimal level of thread support possible for this application. */
        std::map<int,MPO_ThreadViolationInfo> my_threadViolationInfos; /**< Infos for the first violation to a thread level. */

        /**
         * Used to specify whether a certain thread is already in MARMOT or not.
         * Values used:
         *      0    is not yet in MARMOT
         *      n>0  is currently in MARMOT and (n-1) further recursions are in MARMOT
         *
         * Its size must be checked on each call !
         */
        std::vector<int> my_MPO_in_profiles;

#ifdef MARMOT_TCHECK
        std::map<MPO_Request,int> my_requestToIntConversion; /**< List used to map requests to an index of a conflict variable. */
        int my_nextRequestId; /**< Next free id for request to index conversion. */

        std::map<MPI_Comm,int> my_commToIntConversion; /**< List used to map communicators to an index of a conflict variable. */
        int my_nextCommId; /**< Next free id for communicator to index conversion. */
#endif
        /**
         * Checks whether this is a new thread and resizes
         * internal arrays.
         * Must be called within beginCritical and endCritical.
         */
        void checkIsNewThread (void);
        /**
         * Hook method used in registerRequest, registerRequests, unregisterRequests.
         * This method has to start a protection for a critical section.
         * If it uses a lock (probably the implementation will do) it should use another
         * lock than in enterMARMOT !
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         */
        virtual void beginCritical () = 0;

        /**
         * Hook method used in registerRequest, registerRequests, unregisterRequests.
         * This method has to end the protection for a critical section.
         * If it uses a lock (probably the implementation will do) it should use another
         * lock than in enterMARMOT !
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         */
        virtual void endCritical () = 0;

        /**
         * Logs an error in Request usage between two calls.
         * @param pInfo call information of the first call.
         * @param sourceLine source file line number of second call.
         * @param sourceName source file name of second call.
         * @param threadNum id of the thread of second call.
         * @param callId id of the second call.
         * @param callName call name of the second call.
         */
        void logRequestUsageError (sourceInfo *pInfo, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * Logs an error in Communicator usage between two calls.
         * @param pInfo call information of the first call.
         * @param sourceLine source file line number of second call.
         * @param sourceName source file name of second call.
         * @param threadNum id of the thread of second call.
         * @param callId id of the second call.
         * @param callName call name of the second call.
         */
        void logCommUsageError (MPI_Comm comm,sourceInfo *pInfo, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * This method is used when a thread enters the marmot wrappers and after a
         * thread made its PMPI call in order to guarantie that only one thread
         * is using marmots global data at any given time.
         *
         * Its implementation depends upon the concrete threading implementation used.
         *
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         */
        virtual void enterMARMOTSync (int line,const char * file) = 0;

        /**
         * This method is used when a thread leaves the marmot wrappers and before a
         * thread makes the PMPI call to guarantie that only one thread is using
         * marmots global data at any given time.
         *
         * Its implementation depends upon the concrete threading implementation used.
         *
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         */
        virtual void leaveMARMOTSync (int line,const char * file) = 0;

        /**
         * This method returns the amount of threads currently used.
         * @return the amount of threads currently used.
         *
         * Its implementation depends upon the concrete threading implementation used.
         */
        virtual int getNumThreads(void) = 0;

        /**
         * Returns the thread id and source location of the mpi call associated with
         * the thread that passed enterMARMOT.
         * This methos should only be called by one thread at a time.
         * @return thread id and source location of the mpi call associated with
         * the thread that passed enterMARMOT.
         */
        virtual std::pair<int,MPO_SourceLocation> getInfoForThreadInMARMOT(void) = 0;

        /**
         * Returns true if another thread is somewhere enterMARMOTSync and
         * leaveMARMOTSync.
         * @return true if other thread(s) are in MARMOT otherwise false.
         */
        virtual bool otherThreadsInMARMOT (void) = 0;

    public:

        /**
         * Constructor responsible for necessary initialisations.
         */
        MPO_ThreadSynchronisationSyncBase(void);

        /**
         * Destructor responsible releasing memory and objects.
         */
        ~MPO_ThreadSynchronisationSyncBase(void);

        /**
         * This method is used when a thread enters the marmot wrappers and after a
         * thread made its PMPI call in order to guarantie that only one thread
         * is using marmots global data at any given time.
         * This method also has to check for violations to the used level of thread
         * support.
         *
         * This method calls the hook methods "enterMARMOTSync", "getNumThreads",
         * "getInfoForThreadInMARMOT" and "otherThreadsInMARMOT" that have to be
         * implemented by subclasses.
         * of this class. (Template Method pattern)
         *
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @return returns true if this thread is not already in MARMOT and false otherwise.
         *         Note: this might happen if the Implementation of an MPI call uses other MPI calls.
         */
        bool enterMARMOT (int line,const char * file);

        /**
         * This method is used when a thread leaves the marmot wrappers and before a
         * thread makes the PMPI call to guarantie that only one thread is using
         * marmots global data at any given time.
         *
         * This method calls the hook method "leaveMARMOTSync" that has to be
         * implemented by subclasses of this class. (Template Method pattern)
         *
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param enter_succeeded value returned by enterMARMOT.
         */
        void leaveMARMOT (int line,const char * file, bool enter_succeeded);

        /**
         * Swichtes wrapping of MPI calls on for this
         * thread.
         * Must not be called from while inside Marmot
         */
        void wrappingOn (void);

        /**
         * Swichtes wrapping of MPI calls off for this
         * thread.
         * Must not be called from while inside Marmot.
         */
        void wrappingOff (void);

        /**
         * Returns whether the enterMARMOT succeeded.
         * Must only be called between Enter- and Leave-
         * MARMOT.
         */
        bool didEnterSucceed (void);

        /**
         * Registers the given request as currently used.
         * Checks for conflicts with another MPI-Call using the same requests
         * and reports an error in this case.
         * @param request the request to be registered.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @return an id used to unregister the requests after usage.
         */
        int registerRequest (MPI_Request *request,  int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * Registers an array of Requests as currently used.
         * Checks for conflicts with another MPI-Call using one of the
         * given requests and reports an error in this case.
         * @param requests the request to be registered.
         * @param count the amount of requests in the given array.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @return an id used to unregister the requests after usage.
         */
        int registerRequests (MPI_Request *requests, int count,  int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * Removes the Requests associated with the given id from the currently
         * used list.
         * @param id the id of the requests to be unregistered.
         */
        void unregisterRequests (int id);

#ifdef MARMOT_TCHECK
        /**
         * Used to remove a request registered in the request to int conversion for
         * TCheck conflict Variables.
         * Before calling make sure you are not within MARMOT i.e. by calling enterPMPI
         * and leavePMPI after the call !
         */
        void unregisterRequestToIntEntry(MPI_Request *request);

        /**
         * Used to remove a comm registered in the request to int conversion for
         * TCheck conflict Variables.
         * Before calling make sure you are not within MARMOT i.e. by calling enterPMPI
         * and leavePMPI after the call !
         */
        void unregisterCommToIntEntry(MPI_Comm comm);
#endif

        /**
         * Registers the given communicator of a collective call in an internal list
         * in order to detect the usage of one communicator in multiple collective
         * calls simultanously. If the given communicator is already used a error is
         * printed.
         * @param comm the communicator to be registered.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @param threadNum the id of the calling thread.
         * @param callId the call id of the associated call.
         * @param callName the name of the associated call.
         * @return an id used to unregister the communicator after usage.
         */
        int registerCollComm (MPI_Comm comm, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * Removes the Communicator associated with the given id from the
         * list of currently used communicators.
         * @param id the id of the communicator to be unregistered.
         */
        void unregisterCollComm (int id);

        /**
         * Returns the minimal level of thread support that was needed by this
         * application and run.
         * @return the minimal level of thread support neded for this application and run.
         */
        int getMinThreadLvl (void);

        /**
         * Returns a MPO_ThreadViolationInfo for the violated thread level.
         * @param threadLevel one of MPI_THREAD_SINGLE, MPI_THREAD_FUNNELED,
         *                       MPI_THREAD_SERIALIZED where
         *                       threadLevel < getMinThreadLvl ()
         * @return if threadLevel < getMinThreadLvl () the thread violation
         *            info of the given thread level otherwise empty info
         *         (0 participants).
         */
        MPO_ThreadViolationInfo getViolationInfoForThreadLvl (int threadLevel);

    };//end of class MPO_ThreadSynchronisationSyncBase

}//end of namespace MPO

#endif







