/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**********************************************************************
 * @file   SocketLib.h
 * @brief  Declaration of the socket-classes
 * 
 * @author Blasius Czink
 * @date   02.01.2003
 **********************************************************************
 * Version History:
 * V 0.1  02.01.2003  : First Revision
 * V 0.2  07.02.2008  : Stripped UDP Stuff and added some fixes for MacOS
 **********************************************************************/
#ifndef SOCKET_LIB_H
#define SOCKET_LIB_H

#include <stdio.h>
#include <string>
#include <queue>
#include "NetUtil.h"
#include "SocketAddr.h"

/**
 * @class  CSocket
 * @brief  Baseclass for Sockets
 * 
 * @author  B.C.
 * @date    02.01.2003 
 * @version 0.1
 * @since   0.1
 */

class CSocket
{
public:
  //! Default-Constructor, initialize attributes and network subsystem (WSAStartup() on Windows) 
	CSocket();
  virtual ~CSocket() {};

  //! Test socket status. If there was an error it returns FALSE, otherwise TRUE.
  //! You may retrieve the exact error description by calling getError()
  virtual bool         isOk(void)        { return m_ok; }

  //! Get the error code
  virtual int          getErrorCode(void){ return m_error_code; }

  //! Get a descriptive error message
  virtual std::string  getError(void)    { return m_error; }

  //! Get socket descriptor
  virtual net_socket_t getSocketId(void) { return m_socket; }

  //! Get the number of bytes available at the socket
  virtual int          available(void);

  //! Make socket blocking or nonblocking
  virtual bool setBlocking(bool state);

  /**
   * @brief Set socket linger option
   * @param[in] do_linger  If set to FALSE the socket will not linger at close(). 
   * Outstanding data will be sent but the caller does not have a chance to know
   * whether successfully or not since close() invalidates the socket.
   * Setting TRUE here and if timeout is 0 (default) then a close() on the socket
   * will close it immediately and outstanding data will get discarded.
   * @param[in] timeout    If do_linger is TRUE and timeout is >0 a close()
   * will block until all data is sent or the timeout is reached.
   * @note On POSIX systems the timeout is in seconds but there are also systems where it's 1/100 seconds.
   */
  virtual bool setLinger(bool do_linger = true, int timeout = 0);

  //! Get linger-options of the socket
  virtual bool getLinger(bool &is_lingering, int &set_timeout);

  //! Closes the socket
  virtual void close(void);

  /**
   * @brief Read data from the socket. The read can block if the socket is a blocking one and there is not enough data available.
   * @param[out] pData pointer to a buffer the read data will be written to
   * @param[in]  iLen  the size of the buffer receiving the data
   * @param[in]  flags usually you want to set this to 0. Other possibilities are MSG_PEEK or MSG_OOB.
   * @return Number of bytes read or -1 if anything went wrong
   */
	virtual int read (void *pData, unsigned int iLen, int flags) = 0;

  /**
   * @brief Write data to the socket
   * @param[in]  pData pointer to a buffer containing the data to be written
   * @param[in]  iLen  number of bytes to be written
   * @param[in]  flags usually you want to set this to 0. Other possibilities are MSG_PEEK or MSG_OOB
   * @return the number of bytes sent or -1 if anything went wrong
   */
	virtual int write(void *pData, unsigned int iLen, int flags) = 0;

  /**
   * @brief  Checks whether a read would block on the socket (Set the timeouts to 0 to effect a quick poll)
   * @param  timeout_sec   Timeout value in seconds
   * @param  timeout_usec  Timeout value in microseconds
   * @return TRUE if the read access will not block, FALSE otherwise
   */
  bool checkSocketReadable(unsigned int timeout_sec = 1, unsigned int timeout_usec = 0);

  /**
   * @brief  Checks whether a write would block on the socket (Set the timeouts to 0 to effect a quick poll)
   * @param  timeout_sec   Timeout value in seconds
   * @param  timeout_usec  Timeout value in microseconds
   * @return TRUE if the write access will not block, FALSE otherwise
   */
  bool checkSocketWritable(unsigned int timeout_sec = 1, unsigned int timeout_usec = 0);

  /**
   * @brief  Checks whether there is a socket exception (Set the timeouts to 0 to effect a quick poll)
   * @param  timeout_sec   Timeout value in seconds
   * @param  timeout_usec  Timeout value in microseconds
   * @return TRUE if there are socket exceptions for this socket, FALSE otherwise
   */
  bool checkSocketException(unsigned int timeout_sec = 1, unsigned int timeout_usec = 0);
  
  //! Blocks till there is data available (uses select())
  bool blockTillReadable(void);

protected:
  void handleError(const char *usermsg);

  net_socket_t  m_socket;
  SOCKADDR_IN   m_sockaddr;
  std::string   m_error;
  int           m_error_code;
  bool          m_ok;
};


/**
 * @class  CTcpSocket
 * @brief  Class to handle TCP-sockets
 * 
 * @author  B.C.
 * @date    02.01.2003 
 * @version 0.1
 * @since   0.1
 */
class CTcpSocket : public CSocket
{
public:
  /**
   * @brief Constructs a TCP socket and tries to connect to the given address.
   * @param szRemoteAddr  remote machine's address as an IP or a hostname
   * @param iRemotePort   remote machine's listenning port
   */
	CTcpSocket(const char *szRemoteAddr, int iPort);

  /**
   * @brief Constructs a TCP socket, "binds" to it and "listens" for incoming connection attempts. 
   * @param iPort  local listenning port
   */
  CTcpSocket(int iPort);

  /**
   * @brief Create a TCP socket for data transfer (typically after accept)
   * @param sfd    the socket descriptor for this new object
   * @param saptr  the socket address (can be set to NULL)
   */
	CTcpSocket(net_socket_t sfd, SOCKADDR_IN *saptr);

  //! Destructor
	~CTcpSocket();

  /**
   * @brief Accept a connection attempt on a server listen-socket.
   * The returned socket-object can be used to communicate with the client and should be deleted by the user when done.
   * @return Pointer to a new CTcpSocket object for incomming client. Can be NULL if anything went wrong.
   */
	CTcpSocket *accept(void);

  //! @see CSocket::read()
  int  read (void * pData, unsigned int iLen, int flags=0);

  //! @see CSocket::write()
  int  write(void * pData, unsigned int iLen, int flags=0);

  /**
   * @brief Read data "line-by-line". Lines are separated by <CR><LF> ("\r\n") or simply by "\n".
   * @param[in] block  When set to TRUE this method will select/block till a line can be read.
   */
  std::string readln(bool block=true);

  //! Writes a line of data terminated by <CR><LF> to the socket.
  int  writeln(const std::string &str);

  //! Get the peer's address
  CSocketAddr getPeerAddress();

  //! Get the socket's address
  CSocketAddr getAddress();

private:
  // disallow standard constructor, copy-constr. and operator=
  CTcpSocket();
  CTcpSocket(const CTcpSocket &) {};
  CTcpSocket& operator=(const CTcpSocket &) { return *this; };

  // helpers for readln()
  bool read_and_breakdown(int available_bytes, bool &line_read);
  std::queue<std::string> m_readln_queue;
  std::string m_unfinished_line;
  char m_readln_buf[2000];
};


#endif // SOCKET_LIB_H
