/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of all classes representing MPI-2 calls for file I/O and
 *  one-sided communication.
 *
 *  The calls are currently in the same order as in the MPI-2 standard, I/O
 *  chapter 9, p. 209 pp, one-sided, chapter 6, 109pp. Unfortunately, calls
 *  for I/O and one-sided communication are also spread over the MPI-2
 *  standard document, e.g. errhandlers, etc.
 *
 *  @author Bettina Krammer
 *
 *
 *  $Id: mpo-2-objects.h 979 2009-07-17 12:54:04Z hpcdiche $
 */

#ifndef MPO_2_OBJECTS_H
#define MPO_2_OBJECTS_H

#include <iostream>
#include <string>
using std::string;

#include <mpi.h>
#include "mpi_undef_makros.h"
//#include "mpo-2.h"
#include "config.h"
#include "mpo-objects.h"


namespace MPO
{
    //=======================================================================
    // parallel file I/O
    //=======================================================================

#ifdef MARMOT_IO

    //************************************************************************
    //
    //                          MPO_File_open
    //

    /** int MPI_File_open(MPI_Comm comm, char *filename, int amode,
     *                    MPI_Info info, MPI_File *fh)
     *
     *  collective call
     *
     *  @param  comm      communicator (handle)
     *  @param  filename  name of file to open (string)
     *  @param  amode     file access mode (integer)
     *  @param  info      info object (handle)
     *  @param  fh        file handle (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NULL
     *           - MARMOT_ERR_COMM_NOT_VALID
     *           - MARMOT_ERR_COMM_INTERCOMM
     *           - MARMOT_ERR_FILE_DIFF
     *           - MARMOT_ERR_AMODE_DIFF
     *           - MARMOT_ERR_AMODE_RDONLY_RDWR_WRONLY_NOT_SET_EXACTLY_ONE
     *           - MARMOT_ERR_AMODE_RDONLY_AND_CREATE
     *           - MARMOT_ERR_AMODE_RDONLY_AND_EXCL
     *           - MARMOT_ERR_AMODE_RDWR_AND_SEQUENTIAL
     *           - MARMOT_WARN_FILE_OPEN_LEFT (MPI_Finalize)
     */



    //************************************************************************
    //
    //                          MPO_File_close
    //

    /** int MPI_File_close(MPI_File *fh)
     *
     *  collective call
     *
     *  @param  fh        file handle (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_WARN_REQUESTS_NONBLOCKING_LEFT
     *           - MARMOT_WARN_OPERATION_SPLIT_COLLECTIVE_LEFT
     *           - MARMOT_NOTE_AMODE_DELETE_ON_CLOSE
     */



    //************************************************************************
    //
    //                          MPO_File_delete
    //

    /** int MPI_File_delete(char *filename, MPI_Info info)
     *
     *  @param  filename  name of file to delete (string)
     *  @param  info      info object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_FILE_DOES_NOT_EXIST
     *           - MARMOT_WARN_FILE_IN_USE
     */



    //************************************************************************
    //
    //                          MPO_File_set_size
    //

    /** int MPI_File_set_size(MPI_File fh, MPI_Offset size)
     *
     *  collective call
     *
     *  @param  filename  name of file to delete (string)
     *  @param  info      info object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_FILE_SIZE_NEG
     *           - MARMOT_ERR_FILE_SIZE_DIFF
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_REQUESTS_NONBLOCKING_LEFT
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_LEFT
     *           - MARMOT_WARN_DATA_IN_NEW_REGION_UNDEFINED
     *           - MARMOT_NOTE_FILE_SIZE_LT_OR_EQ_CURRENT_SIZE
     */



    //************************************************************************
    //
    //                          MPO_File_preallocate
    //

    /** int MPI_File_preallocate(MPI_File fh, MPI_Offset size)
     *
     *  collective call
     *
     *  @param  fh        file handle (handle)
     *  @param  size      size to preallocate (nonnegative integer)
     *  @see  MPO_File_set_size
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_FILE_SIZE_NEG
     *           - MARMOT_ERR_FILE_SIZE_DIFF
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_REQUESTS_NONBLOCKING_LEFT
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_LEFT
     *           - MARMOT_WARN_DATA_IN_NEW_REGION_UNDEFINED
     *           - MARMOT_NOTE_FILE_SIZE_LT_OR_EQ_CURRENT_SIZE
     */



    //************************************************************************
    //
    //                          MPO_File_get_size
    //

    /** int MPI_File_get_size(MPI_File fh, MPI_Offset *size)
     *
     *  @param  fh        file handle (handle)
     *  @param  size      size of the file in bytes (nonnegative integer)
     *  @see  MPO_File_set_size
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */


    //************************************************************************
    //
    //                          MPO_File_get_group
    //

    /** int MPI_File_get_group(MPI_File fh, MPI_Group *group)
     *
     *  @param  fh        file handle (handle)
     *  @param  group     group that opened the file (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */



    //************************************************************************
    //
    //                          MPO_File_get_amode
    //

    /**  int MPI_File_get_amode(MPI_File fh, int *amode)
     *
     *  @param  fh        file handle (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */



    //************************************************************************
    //
    //                          MPO_File_set_info
    //

    /**  int MPI_File_set_info(MPI_File fh, MPI_Info info)
     *
     *  collective call
     *
     *  @param  fh        file handle (handle)
     *  @param  info      info object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_FILEHANDLE_DIFF
     *           - MARMOT_ERR_INFO_DIFF
     *           - MARMOT_NOTE_INFO_IGNORED
     */



    //************************************************************************
    //
    //                          MPO_File_get_info
    //

    /**  int MPI_File_get_info(MPI_File fh, MPI_Info *info_used)
     *
     *  @param  fh         file handle (handle)
     *  @param  info_used  info object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */



    //************************************************************************
    //
    //                          MPO_File_set_view
    //

    /**  int MPI_File_set_view(MPI_File fh, MPI_Offset disp,
     *                         MPI_Datatype etype,
     *                         MPI_Datatype filetype, char *datarep,
     *                         MPI_Info info)
     *
     *  collective call
     *
     *  @param  fh        file handle (handle)
     *  @param  disp      displacement (nonnegative integer)
     *  @param  etype     elementary datatype (handle)
     *  @param  filetype  filetype (handle)
     *  @param  datarep   data representation (string)
     *  @param  info      info object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_FILETYPE_DISPLACEMENTS_NEG
     *           - MARMOT_ERR_DATAREP_DIFF
     *           - MARMOT_ERR_AMODE_SEQUENTIAL_AND_NOT_DISPLACEMENT_CURRENT
     *           - MARMOT_ERR_AMODE_SEQUENTIAL_NOT_SET_BUT_DISPLACEMENT_CURRENT
     *           - MARMOT_ERR_ETYPE_NOT_VALID
     *           - MARMOT_ERR_ETYPE_OVERLAPPING
     *           - MARMOT_ERR_ETYPE_EXTENT_DIFF
     *           - MARMOT_ERR_FILETYPE_EXTENT_HOLE_NO_MULTIPLE_EXTENT_ETYPE
     *           - MARMOT_ERR_FILETYPE_NOT_VALID
     *           - MARMOT_ERR_FILETYPE_OVERLAPPING
     *           - MARMOT_ERR_REQUESTS_NONBLOCKING_LEFT
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_LEFT
     *           - MARMOT_NOTE_FILETYPE_HAS_HOLES
     */


    //************************************************************************
    //
    //                          MPO_File_get_view
    //

    /**  int MPI_File_get_view(MPI_File fh, MPI_Offset *disp,
     *                         MPI_Datatype *etype,
     *                         MPI_Datatype *filetype, char *datarep)
     *
     *  @param  fh        file handle (handle)
     *  @param  disp      displacement (nonnegative integer)
     *  @param  etype     elementary datatype (handle)
     *  @param  filetype  filetype (handle)
     *  @param  datarep   data representation (string)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_DATAREP_GT_MAX_DATAREP_STRING
     *           - MARMOT_WARN_DATAREP_TOO_SMALL
     */


    //************************************************************************
    //
    //                          MPO_File_read_at
    //

    /**  int MPI_File_read_at(MPI_File fh, MPI_Offset offset, void *buf,
     *                        int count, MPI_Datatype datatype,
     *                        MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  offset    file offset (nonnegative integer)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_read_at_all
    //

    /**  int MPI_File_read_at_all(MPI_File fh, MPI_Offset offset, void *buf,
     *                            int count, MPI_Datatype datatype,
     *                            MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  offset    file offset (nonnegative integer)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_write_at
    //

    /**  int MPI_File_write_at(MPI_File fh, MPI_Offset offset, void *buf,
     *                         int count, MPI_Datatype datatype,
     *                         MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  offset    file offset (nonnegative integer)
     *  @param  buf       initial address of buffer (choice)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */


    //************************************************************************
    //
    //                          MPO_File_write_at_all
    //

    /**  int MPI_File_write_at_all(MPI_File fh, MPI_Offset offset, void *buf,
     *                             int count, MPI_Datatype datatype,
     *                             MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  offset    file offset (nonnegative integer)
     *  @param  buf       initial address of buffer (choice)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_iread_at
    //

    /**  int MPI_File_iread_at(MPI_File fh, MPI_Offset offset, void *buf,
     *                         int count, MPI_Datatype datatype,
     *                         MPIO_Request *request)
     *
     *  @param  fh        file handle (handle)
     *  @param  offset    file offset (nonnegative integer)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  request   request object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_iwrite_at
    //

    /**  int MPI_File_iwrite_at(MPI_File fh, MPI_Offset offset, void *buf,
     *                          int count, MPI_Datatype datatype,
     *                          MPIO_Request *request)
     *
     *  @param  fh        file handle (handle)
     *  @param  offset    file offset (nonnegative integer)
     *  @param  buf       initial address of buffer (choice)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  request   request object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_read
    //

    /**  int MPI_File_read(MPI_File fh, void *buf, int count,
     *                     MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_read_all
    //

    /**  int MPI_File_read_all(MPI_File fh, void *buf, int count,
     *                         MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */


    //************************************************************************
    //
    //                          MPO_File_write
    //

    /**  int MPI_File_write(MPI_File fh, void *buf, int count,
     *                      MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_write_all
    //

    /**  int MPI_File_write_all(MPI_File fh, void *buf, int count,
     *                          MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh        file handle (handle)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_iread
    //

    /**  int MPI_File_iread(MPI_File fh, void *buf, int count,
     *                      MPI_Datatype datatype, MPIO_Request *request)
     *
     *  @param  fh        file handle (handle)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  request   request object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_iwrite
    //

    /**  int MPI_File_iwrite(MPI_File fh, void *buf, int count,
     *                       MPI_Datatype datatype, MPIO_Request *request)
     *
     *  @param  fh        file handle (handle)
     *  @param  count     number of elements in buffer (nonnegative integer)
     *  @param  datatype  datatype of each buffer element (handle)
     *  @param  buf       initial address of buffer (choice)
     *  @param  request   request object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_seek
    //

    /** int MPI_File_seek(MPI_File fh, MPI_Offset offset, int whence)
     *
     *  @param  fh         file handle (handle)
     *  @param  count      number of elements in buffer (nonnegative integer)
     *  @param  datatype   datatype of each buffer element (handle)
     *  @param  buf        initial address of buffer (choice)
     *  @param  request    request object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_OFFSET_TO_NEG_POS_IN_VIEW
     */



    //************************************************************************
    //
    //                          MPO_File_get_position
    //

    /** int MPI_File_get_position(MPI_File fh, MPI_Offset *offset)
     *
     *  @param  fh      file handle (handle)
     *  @param  offset  offset of individual file pointer (nonnegative integer)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     */


    //************************************************************************
    //
    //                          MPO_File_get_byte_offset
    //

    /** int MPI_File_get_byte_offset(MPI_File fh, MPI_Offset offset,
     *                               MPI_Offset *disp)
     *
     *  @param  fh      file handle (handle)
     *  @param  offset  offset (nonnegative integer)
     *  @param  disp    absolute byte position of offset (nonnegative integer)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG ???
     */


    //************************************************************************
    //
    //                          MPO_File_read_shared
    //

    /** int MPI_File_read_shared(MPI_File fh, void *buf, int count,
     *                           MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     *  @param status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL ???
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_write_shared
    //

    /** int MPI_File_write_shared(MPI_File fh, void *buf, int count,
     *                           MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     *  @param status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_iread_shared
    //

    /** int MPI_File_iread_shared(MPI_File fh, void *buf, int count,
     *                            MPI_Datatype datatype, MPIO_Request *request)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     *  @param request   request object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_iwrite_shared
    //

    /** int MPI_File_iwrite_shared(MPI_File fh, void *buf, int count,
     *                             MPI_Datatype datatype,
     *                             MPIO_Request *request)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     *  @param request   request object (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */



    //************************************************************************
    //
    //                          MPO_File_read_ordered
    //

    /** int MPI_File_read_ordered(MPI_File fh, void *buf, int count,
     *                            MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     *  @param status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL ???
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_write_ordered
    //

    /** int MPI_File_write_ordered(MPI_File fh, void *buf, int count,
     *                            MPI_Datatype datatype, MPI_Status *status)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     *  @param status    status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL ???
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_seek_shared
    //

    /** int MPI_File_seek_shared(MPI_File fh, MPI_Offset offset, int whence)
     *
     *  collective call
     *
     *  @param  fh      file handle (handle)
     *  @param offset   file offset (integer)
     *  @param whence   update mode (state)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_OFFSET_DIFF
     *           - MARMOT_ERR_OFFSET_TO_NEG_POS_IN_VIEW
     *           - MARMOT_ERR_WHENCE_DIFF
     */



    //************************************************************************
    //
    //                          MPO_File_get_position_shared
    //

    /** int MPI_File_get_position_shared(MPI_File fh, MPI_Offset *offset)
     *
     *  @param  fh      file handle (handle)
     *  @param offset   offset of shared file pointer (nonnegative integer)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */



    //************************************************************************
    //
    //                          MPO_File_read_at_all_begin
    //

    /** int MPI_File_read_at_all_begin(MPI_File fh, MPI_Offset offset,
     *                                 void *buf,
     *                                 int count, MPI_Datatype datatype)
     *
     *  collective call
     *
     *  @param  fh       file handle (handle)
     *  @param offset    file offset (nonnegative integer)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_read_at_all_end
    //

    /** int MPI_File_read_at_all_end(MPI_File fh, void *buf,
     *                               MPI_Status *status)
     *
     *  @param  fh      file handle (handle)
     *  @param buf      initial address of buffer (choice)
     *  @param status   status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_NO_MATCHING_BEGIN
     */



    //************************************************************************
    //
    //                          MPO_File_write_at_all_begin
    //

    /** int MPI_File_write_at_all_begin(MPI_File fh, MPI_Offset offset,
     *                                  void *buf,
     *                                  int count, MPI_Datatype datatype)
     *
     *  @param  fh       file handle (handle)
     *  @param offset    file offset (nonnegative integer)
     *  @param buf       initial address of buffer (choice)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OFFSET_NEG
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_write_at_all_end
    //

    /** int MPI_File_write_at_all_end(MPI_File fh, void *buf,
     *                                MPI_Status *status)
     *
     *  @param  fh      file handle (handle)
     *  @param buf      initial address of buffer (choice)
     *  @param status   status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_NO_MATCHING_BEGIN
     */



    //************************************************************************
    //
    //                          MPO_File_read_all_begin
    //

    /** int MPI_File_read_all_begin(MPI_File fh, void *buf, int count,
     *                              MPI_Datatype datatype)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_read_all_end
    //

    /** int MPI_File_read_all_end(MPI_File fh, void *buf, MPI_Status *status)
     *
     *  @param  fh      file handle (handle)
     *  @param buf      initial address of buffer (choice)
     *  @param status   status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_NO_MATCHING_BEGIN
     */



    //************************************************************************
    //
    //                          MPO_File_write_all_begin
    //

    /** int MPI_File_write_all_begin(MPI_File fh, void *buf, int count,
     *                               MPI_Datatype datatype)
     *
     *  @param  fh       file handle (handle)
     *  @param buf       initial address of buffer (choice)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_write_all_end
    //

    /** int MPI_File_write_all_end(MPI_File fh, void *buf, MPI_Status *status)
     *
     *  @param  fh      file handle (handle)
     *  @param buf      initial address of buffer (choice)
     *  @param status   status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_NO_MATCHING_BEGIN
     */



    //************************************************************************
    //
    //                          MPO_File_read_ordered_begin
    //

    /** int MPI_File_read_ordered_begin(MPI_File fh, void *buf, int count,
     *          			MPI_Datatype datatype)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_WRONLY
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE
     *           - MARMOT_WARN_COUNT_ZERO
     */



    //************************************************************************
    //
    //                          MPO_File_read_ordered_end
    //

    /** int MPI_File_read_ordered_end(MPI_File fh, void *buf,
     *                                MPI_Status *status)
     *
     *  @param  fh      file handle (handle)
     *  @param buf      initial address of buffer (choice)
     *  @param status   status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_NO_MATCHING_BEGIN
     */




    //************************************************************************
    //
    //                          MPO_File_write_ordered_begin
    //

    /** int MPI_File_write_ordered_begin(MPI_File fh, void *buf, int count,
     *   				 MPI_Datatype datatype)
     *
     *  @param  fh       file handle (handle)
     *  @param count     number of elements in buffer (nonnegative integer)
     *  @param datatype  datatype of each buffer element (handle)
     *  @param buf       initial address of buffer (choice)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_OVERLAPPING
     *           - MARMOT_ERR_AMODE_SEQUENTIAL
     *           - MARMOT_ERR_AMODE_RDONLY
     *           - MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE
     *           - MARMOT_WARN_COUNT_ZERO
     */




    //************************************************************************
    //
    //                          MPO_File_write_ordered_end
    //

    /** int MPI_File_write_ordered_end(MPI_File fh, void *buf,
     *                                 MPI_Status *status)
     *
     *  @param  fh      file handle (handle)
     *  @param buf      initial address of buffer (choice)
     *  @param status   status object (Status)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_NO_MATCHING_BEGIN
     */




    //************************************************************************
    //
    //                          MPO_File_get_type_extent
    //

    /** int MPI_File_get_type_extent(MPI_File fh, MPI_Datatype datatype,
     *                               MPI_Aint *extent)
     *
     *  @param  fh       file handle (handle)
     *  @param datatype  datatype (handle)
     *  @param extent    extent of the datatype (nonnegative integer)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_INVALID
     */




    //************************************************************************
    //
    //                          MPO_Register_datarep
    //

    /**
     *
     *  @param
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */




    //************************************************************************
    //
    //                          MPO_File_set_atomicity
    //

    /** int MPI_File_set_atomicity(MPI_File fh, int flag)
     *
     *  @param fh       file handle (handle)
     *  @param flag     true to set atomic mode, false to set nonatomic mode
     *                  (logical)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_FLAG_DIFF
     */




    //************************************************************************
    //
    //                          MPO_File_get_atomicity
    //

    /** int MPI_File_get_atomicity(MPI_File fh, int *flag)
     *
     *  @param  fh      file handle (handle)
     *  @param flag     true if atomic mode, false if nonatomic mode (logical)
     */

    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     */

#endif
// end of MARMOT_IO



    //=======================================================================
    // one-sided communication
    //=======================================================================

#ifdef MARMOT_1SIDED

    //************************************************************************
    //
    //                          MPO_Winhandler
    //
    /**  This abstract base class represents calls handling windows.
     *
     *  All calls for window-handling are derived from this.
     *  Correctnes of used window is checked in implemented
     *  member function.
     */
    class MPO_Winhandler : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Winhandler();

        /** Constructor taking proper arguments.
         *
         *  @param  comm    communicator
         */
        MPO_Winhandler(MPO_Win win);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Winhandler();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - win
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        MPO_Win my_win;

    };                          // end class MPO_Winhandler



    class MPO_WinCommunicator : public MPO_Winhandler
    {
    public:
      MPO_WinCommunicator();
      MPO_WinCommunicator(void *origin_addr, int origin_count, MPO_Datatype origin_datatype, int target_rank, MPI_Aint target_disp,
                          int target_count, MPO_Datatype target_datatype, MPO_Win win);
      virtual ~MPO_WinCommunicator();
      int getFamilyId() const;
      int checkPrerequisites2();
      int checkCountOrigin();
      int checkDatatypeOrigin();
      int checkRankTarget();
      int checkDispTarget();
      int checkCountTarget();
      int checkDatatypeTarget();

    protected:
      void        *my_origin_addr;
      int          my_origin_count;
      MPO_Datatype my_origin_datatype;
      int          my_target_rank;
      MPI_Aint     my_target_disp;
      int          my_target_count;
      MPO_Datatype my_target_datatype;
    };


    //************************************************************************
    //
    //                          MPO_Win_create
    //
    /** This class represents MPI_Win_create.
     *
     *  collective call
     */
    class MPO_Win_create : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_create();

        /** Constructor taking proper arguments.
         *
         *  @param  base       initial address of window (choice)
         *  @param  size       size of window in bytes (nonnegative integer)
         *  @param  disp_unit  local unit size for displacements,
         *                     in bytes (positive integer)
         *  @param  info       info argument (handle)
         *  @param  comm       communicator (handle)
         *  @param  win        window object returned by the call (handle)
         */
        MPO_Win_create(void *base, MPI_Aint size, int disp_unit,
                       MPI_Info info, MPO_Comm comm, MPI_Win *win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_create();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - size of window
         *  - displacements
         *  - info?
         *  - communicator
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.*/
        virtual int checkPost();

        /** Checks whether size of window is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_SIZE_NEG
         */
        virtual int checkWinSize();

        /** Checks whether unit size of displacements is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_DISP_UNIT_SIZE_NONPOS
         */
        virtual int checkWinDispUnit();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_create
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    protected:
        //==========================
        // Data
        //==========================
        void *my_base;
        MPI_Aint my_size;
        int my_disp_unit;
        MPI_Info my_info;
        MPO_Comm my_comm;
        MPI_Win *mpo_win_create_Win;

    };                          // end class MPO_Win_create



    //************************************************************************
    //
    //                          MPO_Win_free
    //
    /** This class represents MPI_Win_free.
     *
     *  collective call
     *  can only be called after communication is completed, i.e. after that
     *  process has called an MPI_WIN_ fence, MPI_Win_wait, MPI_Win_complete
     *  or MPI_Win_unlock call.
     */
    class MPO_Win_free : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_free();

        /** Constructor taking proper arguments.
         *  @param  win        window object (handle)
         */
        MPO_Win_free(MPI_Win *win);

        MPI_Win MPO_Win_getMPI_Win();
        void MPO_Win_setMPI_Win(MPI_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_free();

        //==========================
        // Methods
        //==========================

        /** Checks whether
         *  - win
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        virtual int checkPost();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_free
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================
        MPI_Win *mpo_win_free_Win;


    };                          // end class MPO_Win_free



    //************************************************************************
    //
    //                          MPO_Win_get_group
    //
    /** This class represents MPI_Win_get_group. */
    class MPO_Win_get_group : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_get_group();

        /** Constructor taking proper arguments.
         *
         *  @param  win        window object (handle)
         *  @param  group      group of processes which share access to
         *                     the window (handle)
         */
        MPO_Win_get_group(MPO_Win win, MPI_Group *group);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_get_group();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_get_group
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Win my_win;
        MPI_Group *my_group;

    };                          // end class MPO_Win_get_group




    //************************************************************************
    //
    //                          MPO_Put
    //
    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_WIN
     *           - MARMOT_ERR_TYPE...
     *           - MARMOT_ERR_COUNT
     *           - MARMOT_WARN_COUNT_ZERO
     */

    /** This class represents MPI_Put. */
    class MPO_Put : public MPO_WinCommunicator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Put();

        /** Constructor taking proper arguments.
         *
         *  @param  origin_addr     initial address of origin buffer (choice)
         *  @param  origin_count    number of entries in origin buffer
         *                          (nonnegative integer)
         *  @param  origin_datatype datatype of each entry
         *                          in origin buffer (handle)
         *  @param  target_rank     rank of target (nonnegative integer)
         *  @param  target_disp     displacement from start of window to
         *                          target buffer (nonnegative integer)
         *  @param  target_count    number of entries in target buffer
         *                          (nonnegative integer)
         *  @param  target_datatype datatype of each entry
         *                          in target buffer (handle)
         *  @parma  win             window object used for communication
         *                          (handle)
         */
        MPO_Put(void *origin_addr, int origin_count, MPO_Datatype
            origin_datatype, int target_rank, MPI_Aint target_disp,
            int target_count, MPO_Datatype target_datatype, MPI_Win
            win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Put();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Put
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // void *my_origin_addr;
        // int my_origin_count;
        // MPO_Datatype my_origin_datatype;
        // int my_target_rank;
        // MPI_Aint my_target_disp;
        // int my_target_count;
        // MPO_Datatype my_target_datatype;
        // MPO_Win my_win;

    };                          // end class MPO_Put



    //************************************************************************
    //
    //                          MPO_Get
    //
    /** This class represents MPI_Get. */
    class MPO_Get : public MPO_WinCommunicator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Get();

        /** Constructor taking proper arguments.
         *
         *  @param  origin_addr     address of the buffer in which to receive
         *                          the data
         *  @param  origin_count    number of entries in origin buffer
         *                          (nonnegative integer)
         *  @param  origin_datatype datatype of each entry
         *                          in origin buffer (handle)
         *  @param  target_rank     rank of target (nonnegative integer)
         *  @param  target_disp     displacement from window start to the
         *                          beginning of the target buffer
         *                          (nonnegative integer)
         *  @param  target_count    number of entries in target buffer
         *                          (nonnegative integer)
         *  @param  target_datatype datatype of each entry
         *                          in target buffer (handle)
         *  @param  win             window object used for communication (handle)
         */
        MPO_Get(void *origin_addr, int origin_count,
                MPO_Datatype origin_datatype, int target_rank,
                MPI_Aint target_disp, int target_count,
                MPO_Datatype target_datatype, MPI_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Get();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Get
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // void *my_origin_addr;
        // int my_origin_count;
        // MPO_Datatype my_origin_datatype;
        // int my_target_rank;
        // MPI_Aint my_target_disp;
        // int my_target_count;
        // MPO_Datatype my_target_datatype;
        // MPO_Win my_win;

    };                          // end class MPO_Get



    //************************************************************************
    //
    //                          MPO_Accumulate
    //
    /** This class represents MPI_Accumulate. */
    class MPO_Accumulate : public MPO_WinCommunicator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Accumulate();

        /** Constructor taking proper arguments.
         *
         *  @param  origin_addr     initial address of buffer (choice)
         *  @param  origin_count    number of entries in buffer
         *                          (nonnegative integer)
         *  @param  origin_datatype datatype of each buffer entry (handle)
         *  @param  target_rank     rank of target (nonnegative integer)
         *  @param  target_disp     displacement from window start to the
         *                          beginning of the target buffer
         *                          (nonnegative integer)
         *  @param  target_count    number of entries in target buffer
         *                          (nonnegative integer)
         *  @param  target_datatype datatype of each entry
         *                          in target buffer (handle)
         *  @param  op              predefined reduce operation (handle)
         *  @param  win             window object used for communication
         *                          (handle)
         */
        MPO_Accumulate(void *origin_addr, int origin_count,
                       MPO_Datatype origin_datatype, int target_rank,
                       MPI_Aint target_disp, int target_count,
                       MPO_Datatype target_datatype, MPO_Op op, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Accumulate();

        //==========================
        // Methods
        //==========================

//TODO
        virtual int checkOp();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Accumulate
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // void *my_origin_addr;
        // int my_origin_count;
        // MPO_Datatype my_origin_datatype;
        // int my_target_rank;
        // MPI_Aint my_target_disp;
        // int my_target_count;
        // MPO_Datatype my_target_datatype;
        MPO_Op my_op;
        // MPO_Win my_win;

    };                          // end class MPO_Accumulate



    //************************************************************************
    //
    //                          MPO_Win_fence
    //
    /** This class represents MPI_Win_fence. */
    class MPO_Win_fence : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_fence();

        /** Constructor taking proper arguments.
         *
         *  @param  assert          program assertion (integer)
         *  @param  win             window object used for communication
         *                          (handle)
         */
        MPO_Win_fence(int assert, MPI_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_fence();

        //==========================
        // Methods
        //==========================

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        virtual int checkWin();
        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPost();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_fence
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);


    private:
        //==========================
        // Data
        //==========================
        int my_assert;
        MPI_Win mpo_win_fence_Win;

    };                          // end class MPO_Win_fence



    //************************************************************************
    //
    //                          MPO_Win_start
    //
    /** This class represents MPI_Win_start. */
    class MPO_Win_start : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_start();

        /** Constructor taking proper arguments.
         *
         *  @param  group           group of target processes (handle)
         *  @param  assert          Used to optimize this call;
         *                          zero may be used as a default.
         *  @param  win             window object (handle)
         */
        MPO_Win_start(MPO_Group group, int assert, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_start();

        //==========================
        // Methods
        //==========================

        /** Checks whether group is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup();

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_start
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================
        MPO_Group my_group;
        int my_assert;
        // MPO_Win my_win;

    };                          // end class MPO_Win_start



    //************************************************************************
    //
    //                          MPO_Win_complete
    //
    /** This class represents MPI_Win_complete. */
    class MPO_Win_complete : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_complete();

        /** Constructor taking proper arguments.
         *  @param  win             window object (handle)
         */
        MPO_Win_complete(MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_complete();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_complete
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // MPO_Win my_win;

    };                          // end class MPO_Win_complete




    //************************************************************************
    //
    //                          MPO_Win_post
    //
    /** This class represents MPI_Win_post. */
    class MPO_Win_post : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_post();

        /** Constructor taking proper arguments.
         *
         *  @param  group           group of origin processes (handle)
         *  @param  assert          used to optimize this call;
         *                          zero may be used as a default.
         *  @param  win             window object (handle)
         */
        MPO_Win_post(MPO_Group group, int assert, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_post();

        //==========================
        // Methods
        //==========================

        /** Checks whether group is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup();

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_post
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Group my_group;
        int my_assert;
        // MPO_Win my_win;

    };                          // end class MPO_Win_post



    //************************************************************************
    //
    //                          MPO_Win_wait
    //
    /** This class represents MPI_Win_wait. */
    class MPO_Win_wait : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_wait();

        /** Constructor taking proper arguments.
         *
         *  @param  win             window object (handle)
         */
        MPO_Win_wait(MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_wait();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_wait
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // MPO_Win my_win;

    };                          // end class MPO_Win_wait




    //************************************************************************
    //
    //                          MPO_Win_test
    //
    /** This class represents MPI_Win_test. */
    class MPO_Win_test : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object.
         *
         *  @param  win             window object (handle)
         *  @param  flag            success flag (logical)
         */
        MPO_Win_test();

        /** Constructor taking proper arguments. */
        MPO_Win_test(MPO_Win win, int *flag);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_test();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_test
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // MPO_Win my_win;
        int *my_flag;

    };                          // end class MPO_Win_test



    //************************************************************************
    //
    //                          MPO_Win_lock
    //
    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_RANK
     *           - MARMOT_ERR_WIN
     */


    /** This class represents MPI_Win_lock. */
    class MPO_Win_lock : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_lock();

        /** Constructor taking proper arguments.
         *
         *  @param  lock_type       indicates whether other processes may
         *                          access the target window at the same time
         *                          (if 'MPI_LOCK_SHARED') or not
         *                          ('MPI_LOCK_EXCLUSIVE')
         *  @param  rank            rank of locked window (nonnegative integer)
         *  @param  assert          Used to optimize this call;
         *                          zero may be used as a default.
         *  @param  win             window object (handle)
         */
        MPO_Win_lock(int lock_type, int rank, int assert, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_lock();

        //==========================
        // Methods
        //==========================

        /** Checks whether locktype is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_LOCKTYPE_NOT_VALID
         */
        virtual int checkLockType();

        /** Checks whether rank is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_ERR_RANK_PROC_NULL
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRank();

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_lock
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_lock_type;
        int my_rank;
        int my_assert;
        // MPO_Win my_win;

    };                          // end class MPO_Win_lock



    //************************************************************************
    //
    //                          MPO_Win_unlock
    //
    /** This class represents MPI_Win_unlock. */
    class MPO_Win_unlock : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_unlock();

        /** Constructor taking proper arguments.
         *
         *  @param  rank            rank of window (nonnegative integer)
         *  @param  win             window object (handle)
         */
        MPO_Win_unlock(int rank, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_unlock();

        //==========================
        // Methods
        //==========================

        /** Checks whether rank is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_ERR_RANK_PROC_NULL
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRank();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_unlock
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_rank;
        // MPO_Win my_win;

    };                          // end class MPO_Win_unlock


#endif
// end of MARMOT_1SIDED


#ifdef MARMOT_THREADED

    //*************************************************************************
    //
    //                          MPO_Init_thread
    //
    /** This class represents MPI_Init_thread. */
    class MPO_Init_thread : public MPO_Init
      {
      public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object.
         *  It should never be called anyhow, because MPI_Init is only called
         *  once. And the corresponding objects are never transferred.
         * \note we can`t use the MPI_Init constructor here because it would call PMPI_Init
         */
        MPO_Init_thread();

        /** Constructor taking proper arguments.
         *
         *  @param   argc - Pointer to the number of arguments
         *  @param   argv - Pointer to the argument vector
         *  @see     mk_fortran_wrapper.sh takes care that argument list is
         *           recovered when using mpich-p4.
         */
        MPO_Init_thread(int *argc, char ***argv,int required, int *provided);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Init_thread();

        //==========================
        // Methods
        //==========================


// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  my_rvalue
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** We override callDebugServer(), because we need special treatment
         *  in MPI_Init_thread.
         *
         *  @return  MPI_SUCCESS
         */
        virtual int callDebugServer();

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    protected:
        //==========================
        // Data
        //==========================

        int *my_provided;
        int  my_required;

    };                          // end class MPO_Init_thread

#else

    //*************************************************************************
    //
    //                          MPO_Init_thread
    // WE will use this dummy class for MPI_Init_thread. This is done to
    // aviod the necessity of 2 different Wrapper libraries (one with
    // a MPI_Init_thread function and one without)
    /** This class represents MPI_Init_thread. */
    class MPO_Init_thread : public MPO_Init
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** Constructor taking proper arguments.
         *
         *  @param   argc - Pointer to the number of arguments
         *  @param   argv - Pointer to the argument vector
         *  @see     mk_fortran_wrapper.sh takes care that argument list is
         *           recovered when using mpich-p4.
         */
        MPO_Init_thread(void);
        MPO_Init_thread(int *argc, char ***argv,int required, int *provided);
    };                          // end class MPO_Init_thread

#endif
// end of MARMOT_THREADED

}                               // end of namespace MPO

#endif
