/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of ostream and istream.
 *
 *  Defines the > and < operator with the same behavior as >> and <<
 *  for binary i/o and basic types (including string).
 *  - For basic datatypes, this is implemented with macros.
 *  - For other datatypes, this is implemented with a template function.
 *  
 *  @see datatypes: MPI-2 standard, section 10.1.6, p. 275 ff
 *
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller
 *
 *  $Id: mpo-binaryIO.h 953 2009-04-30 13:44:56Z hpczink $
 */


#ifndef MPO_BINARYIO_H
#define MPO_BINARYIO_H

#include <iostream>

namespace MPO
{

//***************************************************************************
//
//                      ostream/istream for basic datatypes   
//

/** This macro defines ostream for basic datatypes. */
#define _MPO_BINARYOUT(TYPE) \
inline std::ostream &\
operator<(std::ostream &o, const TYPE &a) \
{\
    o.write( (const char *) &a, sizeof(TYPE) );\
    return o;\
}


/** This macro defines istream for basic datatypes. */
#define _MPO_BINARYIN(TYPE) \
inline std::istream &\
operator>(std::istream &i, TYPE &a) \
{\
    i.read( (char *) &a, sizeof(TYPE) );\
    return i;\
}

//===========================================
// floating point types
//===========================================

_MPO_BINARYOUT(double)
_MPO_BINARYOUT(float)
    //bk// Don't know why long double was not enumerated here.
    //bk// MPI-2 needs it.
_MPO_BINARYOUT(long double) 

_MPO_BINARYIN(double)
_MPO_BINARYIN(float)
    //bk// Don't know why long double was not enumerated here.
    //bk// MPI-2 needs it.
_MPO_BINARYIN(long double)

//===========================================
// integer types (signed is implicit)
//===========================================
   
_MPO_BINARYOUT(int)
_MPO_BINARYOUT(unsigned)
_MPO_BINARYOUT(short)
_MPO_BINARYOUT(unsigned short)
_MPO_BINARYOUT(long)
_MPO_BINARYOUT(unsigned long)
    //bk// optional datatypes
_MPO_BINARYOUT(long long)
_MPO_BINARYOUT(unsigned long long)

_MPO_BINARYIN(int)
_MPO_BINARYIN(unsigned)
_MPO_BINARYIN(short)
_MPO_BINARYIN(unsigned short)
_MPO_BINARYIN(long)
_MPO_BINARYIN(unsigned long)
    //bk// optional datatypes
_MPO_BINARYIN(long long)
_MPO_BINARYIN(unsigned long long)

//===========================================
// boolean type
//===========================================

_MPO_BINARYOUT(bool)

_MPO_BINARYIN(bool)

//===========================================
// character types
//===========================================
// Whether a char behaves signed or
// unsigned is implementation dependent, pointers to 
// signed, unsigned, and plain chars are possibly different entities 
// so we need all three types here.
    
_MPO_BINARYOUT(char)
_MPO_BINARYOUT(signed char) _MPO_BINARYOUT(unsigned char)
    //bk// Don't know why wchar_t was not enumerated here.
    //bk// MPI-2 needs it.
_MPO_BINARYOUT(wchar_t)

_MPO_BINARYIN(char)
_MPO_BINARYIN(signed char) _MPO_BINARYIN(unsigned char)
    //bk// Don't know why wchar_t was not enumerated here.
    //bk// MPI-2 needs it.
_MPO_BINARYIN(wchar_t)

#undef _MPO_BINARYOUT
#undef _MPO_BINARYIN


//***************************************************************************
//
//                      ostream/istream for other datatypes    
//
/*******************************************************************
 *  Other than the basic types in MPO should provide save/restore
 *  virtual member functions that should be called, when binary io is 
 *  requested. This could be implemented with a template function
 *  here. 
 *******************************************************************/

/** This template function defines ostream for other than basic datatypes. */
   
template < class T >
inline std::ostream & operator<(std::ostream & out, const T & a)
{
    return a.save(out);
}


/** This template function defines istream for other than basic datatypes. */
template < class T >
inline std::istream & operator>(std::istream & in, T & a)
{
    return a.restore(in);
}


}                               // end of namespace MPO
//  Local Variables:
//  mode: c++
//  End:

#endif
