/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of all the enums we use in our namespace MPO.
 *
 *  HINT FOR DEVELOPERS:
 *  ====================
 *  DON'T USE HARD-CODED VALUES, DON'T USE MAGIC NUMBERS,
 *  THEY ARE EVIL AND MAKE HORRIBLE CODE!
 *  Use descriptive enums or variable names and try to keep the complexity
 *  of check functions low.
 *  If you declare a new enum, put it into THIS file.
 *
 *  @author Bettina Krammer, Tobias Hilbrich
 *
 *  $Id: mpo-enums.h 979 2009-07-17 12:54:04Z hpcdiche $
 */


#ifndef MPO_ENUMS_H
#define MPO_ENUMS_H

#include <mpi.h>
#include "mpi_undef_makros.h"

namespace MPO
{
    //==========================
    // Defines
    //==========================
    /** Global defines
     * \note we might create an extra Header file for them
     */
    //defines the maxlength of sourcefilenames (reduces dynamical mem overhead)
#define FILENAME_LEN 128
#define TCHECK_NUM_REQUEST_SLOTS 100000
#define TCHECK_NUM_COMM_SLOTS 100000
#define TCHECK_NUM_PROBE_SLOTS 100000
#define MESSENGER_TAG_OFFSET 2

    //==========================
    // Debug modes
    //==========================

    /** enum for the different grades of DEBUG_MODE (levels of verbosity).
     *  This enum replaces the magic numbers that had been used previously.
     */
    enum MPO_DEBUG_MODE
    {
        // Errors are reported.
        MSG_ERR = 0,            // used to be magic number 0

        // Errors and warnings are reported.
        MSG_ERR_WARN = 1,       // used to be magic number 1

        // Errors, warnings and remarks are reported.
        MSG_ERR_WARN_NOTE = 2,  // used to be magic number 2

        MSG_LAST_DEBUG_MODE
    };


    //==========================
    // Interface modes
    //==========================

    /** enum for the different interface modes.
     *  This enum replaces the magic numbers that had been used previously.
     */
    enum MPO_INTERFACE_MODE
    {
        C_INTERFACE = 0,        // used to be magic number 0
        F_INTERFACE = 1,        // used to be magic number 1
        LAST_INTERFACE_MODE
    };


    //==========================
    // Tracecall modes
    //==========================

// TODO //
    /** enum for the different options for tracing calls.
     *  This enum replaces the magic numbers that had been used previously.
     */
    enum MPO_TRACE_CALLS_MODE
    {
        TRACE_CALLS_WITH_OUTPUT_VERBOSE,  // used to be magic number 2
        TRACE_CALLS_WITH_OUTPUT,          // used to be magic number 1
        TRACE_CALLS_WITHOUT_OUTPUT,       // used to be magic number 0
        TRACE_CALLS_NEVER,                // used to be magic number -1
        LAST_TRACE_CALLS_MODE
    };

    //==========================
    // Logfile types
    //==========================
    /** This enum describes the different logfile types.
     */
    enum MPO_LOGFILE_TYPE
    {
       LOGFILE_TYPE_ASCII = 0,
       LOGFILE_TYPE_HTML,
       LOGFILE_TYPE_CUBE,
       LOGFILE_TYPE_CUBE_DDT,
       LOGFILE_TYPE_QUEUE,
       LAST_LOGFILE_TYPE
    };

    //==========================
    // Log flush type
    //==========================
    /** This enum describes the different logfile types.
     */
    enum MPO_LOGFILE_FLUSH_TYPE
    {
       LOGFILE_FLUSH_TYPE_ON_ERROR = 0,
       LOGFILE_FLUSH_TYPE_IMMEDIATE
    };

    //==========================
    // Resource tracing
    //==========================
    /** This enum describes the different resource
     * 	tracing modes.
     *  Finer grained modes might appear in the future.
     */
    enum MPO_RESOURCE_TRACE_SELECTION
    {
       RESOURCE_TRACE_ON = 0,
       RESOURCE_TRACE_OFF
    };

    //==========================
    // Typs of Messages
    //==========================

    /**
      * This Enum describes the types of messages used for logging.
      */
    enum MPO_MSG_TYPE
    {
        MSG_TYPE_INFO = 0,
        MSG_TYPE_WARNING,
        MSG_TYPE_ERROR,
        MSG_TYPE_NOTE
    };

    //==========================
    // Typs of Message contents
    //==========================

    /**
      * This Enum describes the types of data a message can own.
      */
    enum MPO_OUTPUT_TYPE
    {
	OUTPUT_TYPE_PURE_TEXT = 0,
	OUTPUT_TYPE_PURE_CALLINFO,
	OUTPUT_TYPE_TEXT_AND_CALLINFO
    };

    //==========================
    // ID of objects
    //==========================

    /** enum for the different families of objects. */
    enum MPO_FAMILY_ID
    {
        MPO_CALLOBJECT_ID = 1,
        MPO_TYPEOBJECT_ID,
        MPO_STATEOBJECT_ID,
        MPO_MESSAGEOBJECT_ID,
        MPO_FINALIZE_READY,
        MPO_LAST_FAMILY_ID
    };


    /** enum used for id of MPO_Typeobjects. */
    enum MPO_TYPE_ID
    {
        MPO_COMM_ID = 1,
        MPO_GROUP_ID,
        MPO_DATATYPE_ID,
        MPO_OP_ID,
        MPO_REQUEST_ID,
        MPO_WIN_ID,
        MPO_LAST_TYPE_ID
    };


    /** enum used for id of MPO_Callobjects. */
    enum MPO_CALL_ID
    {
        // calls of MPI-1.2 in alphabetical order
        // (except for MPI_INIT, which is the first one)
        MPO_INIT_ID = 1,
        MPO_ABORT_ID,
        MPO_ADDRESS_ID,
        MPO_ALLGATHER_ID,
        MPO_ALLGATHERV_ID,
        MPO_ALLREDUCE_ID,
        MPO_ALLTOALL_ID,
        MPO_ALLTOALLV_ID,
        MPO_ATTR_DELETE_ID,
        MPO_ATTR_GET_ID,
        MPO_ATTR_PUT_ID,
        MPO_BARRIER_ID,
        MPO_BCAST_ID,
        MPO_BSEND_ID,
        MPO_BSEND_INIT_ID,
        MPO_BUFFER_ATTACH_ID,
        MPO_BUFFER_DETACH_ID,
        MPO_CANCEL_ID,
        MPO_CART_COORDS_ID,
        MPO_CART_CREATE_ID,
        MPO_CARTDIM_GET_ID,
        MPO_CART_GET_ID,
        MPO_CART_MAP_ID,
        MPO_CART_RANK_ID,
        MPO_CART_SHIFT_ID,
        MPO_CART_SUB_ID,
        MPO_COMM_COMPARE_ID,
        MPO_COMM_CREATE_ID,
        MPO_COMM_DUP_ID,
        MPO_COMM_FREE_ID,
        MPO_COMM_GROUP_ID,
        MPO_COMM_RANK_ID,
        MPO_COMM_REMOTE_GROUP_ID,
        MPO_COMM_REMOTE_SIZE_ID,
        MPO_COMM_SIZE_ID,
        MPO_COMM_SPLIT_ID,
        MPO_COMM_TEST_INTER_ID,
        MPO_DIMS_CREATE_ID,
        MPO_ERRHANDLER_CREATE_ID,
        MPO_ERRHANDLER_FREE_ID,
        MPO_ERRHANDLER_GET_ID,
        MPO_ERRHANDLER_SET_ID,
        MPO_ERROR_CLASS_ID,
        MPO_ERROR_STRING_ID,
        MPO_FINALIZE_ID,
        MPO_GATHER_ID,
        MPO_GATHERV_ID,
        MPO_GET_COUNT_ID,
        MPO_GET_ELEMENTS_ID,
        MPO_GET_PROCESSOR_NAME_ID,
        MPO_GET_VERSION_ID,
        MPO_GRAPH_CREATE_ID,
        MPO_GRAPHDIMS_GET_ID,
        MPO_GRAPH_GET_ID,
        MPO_GRAPH_MAP_ID,
        MPO_GRAPH_NEIGHBORS_COUNT_ID,
        MPO_GRAPH_NEIGHBORS_ID,
        MPO_GROUP_COMPARE_ID,
        MPO_GROUP_DIFFERENCE_ID,
        MPO_GROUP_EXCL_ID,
        MPO_GROUP_FREE_ID,
        MPO_GROUP_INCL_ID,
        MPO_GROUP_INTERSECTION_ID,
        MPO_GROUP_RANGE_EXCL_ID,
        MPO_GROUP_RANGE_INCL_ID,
        MPO_GROUP_RANK_ID,
        MPO_GROUP_SIZE_ID,
        MPO_GROUP_TRANSLATE_RANKS_ID,
        MPO_GROUP_UNION_ID,
        MPO_IBSEND_ID,
        MPO_INITIALIZED_ID,
        MPO_INTERCOMM_CREATE_ID,
        MPO_INTERCOMM_MERGE_ID,
        MPO_IPROBE_ID,
        MPO_IRECV_ID,
        MPO_IRSEND_ID,
        MPO_ISEND_ID,
        MPO_ISSEND_ID,
        MPO_KEYVAL_CREATE_ID,
        MPO_KEYVAL_FREE_ID,
        MPO_OP_CREATE_ID,
        MPO_OP_FREE_ID,
        MPO_PACK_ID,
        MPO_PACK_SIZE_ID,
        // MPO_PCONTROL_ID,
        MPO_PROBE_ID,
        MPO_RECV_ID,
        MPO_RECV_INIT_ID,
        MPO_REDUCE_ID,
        MPO_REDUCE_SCATTER_ID,
        MPO_REQUEST_FREE_ID,
        MPO_RSEND_ID,
        MPO_RSEND_INIT_ID,
        MPO_SCAN_ID,
        MPO_SCATTER_ID,
        MPO_SCATTERV_ID,
        MPO_SEND_ID,
        MPO_SEND_INIT_ID,
        MPO_SENDRECV_ID,
        MPO_SENDRECV_REPLACE_ID,
        MPO_SSEND_ID,
        MPO_SSEND_INIT_ID,
        MPO_STARTALL_ID,
        MPO_START_ID,
        MPO_TESTALL_ID,
        MPO_TESTANY_ID,
        MPO_TEST_CANCELLED_ID,
        MPO_TEST_ID,
        MPO_TESTSOME_ID,
        MPO_TOPO_TEST_ID,
        MPO_TYPE_COMMIT_ID,
        MPO_TYPE_CONTIGUOUS_ID,
        MPO_TYPE_EXTENT_ID,
        MPO_TYPE_FREE_ID,
        MPO_TYPE_HINDEXED_ID,
        MPO_TYPE_HVECTOR_ID,
        MPO_TYPE_INDEXED_ID,
        MPO_TYPE_LB_ID,
        MPO_TYPE_SIZE_ID,
        MPO_TYPE_STRUCT_ID,
        MPO_TYPE_UB_ID,
        MPO_TYPE_VECTOR_ID,
        MPO_UNPACK_ID,
        MPO_WAITALL_ID,
        MPO_WAITANY_ID,
        MPO_WAIT_ID,
        MPO_WAITSOME_ID,
        // MPO_WTICK_ID,
        // MPO_WTIME_ID,

        // calls of MPI-2 file I/O in alphabetical order
        // not implemented yet
        MPO_FILE_CLOSE_ID,
        MPO_FILE_DELETE_ID,
        MPO_GET_AMODE_ID,
        MPO_FILE_GET_ATOMICITY_ID,
        MPO_FILE_GET_BYTE_OFFSET_ID,
        MPO_FILE_GET_GROUP_ID,
        MPO_FILE_GET_INFO_ID,
        MPO_FILE_GET_POSITION_ID,
        MPO_FILE_GET_POSITION_SHARED_ID,
        MPO_FILE_GET_SIZE_ID,
        MPO_FILE_GET_TYPE_EXTENT_ID,
        MPO_FILE_GET_VIEW_ID,
        MPO_FILE_IREAD_ID,
        MPO_FILE_IREAD_AT_ID,
        MPO_FILE_IREAD_SHARED_ID,
        MPO_FILE_IWRITE_AT_ID,
        MPO_FILE_IWRITE_ID,
        MPO_FILE_IWRITE_SHARED_ID,
        MPO_FILE_OPEN_ID,
        MPO_FILE_PREALLOCATE_ID,
        MPO_FILE_READ_ALL_BEGIN_ID,
        MPO_FILE_READ_ALL_ID,
        MPO_FILE_READ_ALL_END_ID,
        MPO_FILE_READ_AT_ALL_BEGIN_ID,
        MPO_FILE_READ_AT_ALL_ID,
        MPO_FILE_READ_AT_ALL_END_ID,
        MPO_FILE_READ_AT_ID,
        MPO_FILE_READ_ID,
        MPO_FILE_READ_ORDERED_BEGIN_ID,
        MPO_FILE_READ_ORDERED_ID,
        MPO_FILE_READ_ORDERED_END_ID,
        MPO_FILE_READ_SHARED_ID,
        MPO_FILE_SEEK_ID,
        MPO_FILE_SEEK_SHARED_ID,
        MPO_FILE_SET_ATOMICITY_ID,
        MPO_FILE_SET_INFO_ID,
        MPO_FILE_SET_SIZE_ID,
        MPO_FILE_SET_VIEW_ID,
        MPO_FILE_WRITE_ALL_BEGIN_ID,
        MPO_FILE_WRITE_ALL_ID,
        MPO_FILE_WRITE_ALL_END_ID,
        MPO_FILE_WRITE_AT_ALL_BEGIN_ID,
        MPO_FILE_WRITE_AT_ALL_ID,
        MPO_FILE_WRITE_AT_ALL_END_ID,
        MPO_FILE_WRITE_AT_ID,
        MPO_FILE_WRITE_ID,
        MPO_FILE_WRITE_ORDERED_BEGIN_ID,
        MPO_FILE_WRITE_ORDERED_ID,
        MPO_FILE_WRITE_ORDERED_END_ID,
        MPO_FILE_WRITE_SHARED_ID,
        MPO_REGISTER_DATAREP_ID,
        MPO_FILE_CREATE_ERRHANDLER_ID,
        MPO_FILE_GET_ERRHANDLER_ID,
        MPO_FILE_SET_ERRHANDLER_ID,
        MPO_FILE_C2F_ID,
        MPO_FILE_F2C_ID,

        // calls of MPI-2 one-sided communication
        // not implemented yet
        MPO_WIN_CREATE_ID,
        MPO_WIN_FREE_ID,
        MPO_WIN_GET_GROUP_ID,
        MPO_PUT_ID,
        MPO_GET_ID,
        MPO_ACCUMULATE_ID,
        MPO_WIN_FENCE_ID,
        MPO_WIN_START_ID,
        MPO_WIN_COMPLETE_ID,
        MPO_WIN_POST_ID,
        MPO_WIN_WAIT_ID,
        MPO_WIN_TEST_ID,
        MPO_WIN_LOCK_ID,
        MPO_WIN_UNLOCK_ID,
        MPO_WIN_CREATE_ERRHANDLER_ID,
        MPO_WIN_GET_ERRHANDLER_ID,
        MPO_WIN_SET_ERRHANDLER_ID,
        MPO_WIN_CALL_ERRHANDLER_ID,
        MPO_WIN_CREATE_KEYVAL_ID,
        MPO_WIN_DELETE_KEYVAL_ID,
        MPO_GET_ATTR_ID,
        MPO_SET_ATTR_ID,
        MPO_DELETE_ATTR_ID,
        MPO_GET_NAME_ID,
        MPO_SET_NAME_ID,
        MPO_WIN_F2C_ID,
        MPO_WIN_C2F_ID,
        MPO_INIT_THREAD_ID,

        // last call id
        MPO_LAST_CALL_ID
    };                          // end enum MPO_CALL_ID

    /**
      * enum used for identifying Messageobjects.
      */
    enum MPO_MESSAGE_ID {
        MPO_STANDARDMESSAGE = 1,
	MPO_STRING,
        MPO_SOURCELOCATION
    };


    /** enum used for the states of an MPI process.
     *  @see mpo-debugserver.cc
     *  @see mpo-state.cc*/
    enum MPO_STATE_ID
    {
        MPO_STARTED = 1,
        MPO_CALLED,
        MPO_PENDING,
        MPO_FINISHED, // currently not used
        MPO_RUNNING,  // currently not used
        MPO_LAST_STATE_ID
    };


    //==========================
    // Kind of objects
    //==========================


    /** enum for the kind of count that is used.
     *  This enum replaces the magic numbers that had been used previously
     *  in the switch statements of our check functions.
     */
    enum KIND_OF_COUNT
    {
        IS_COUNT,               // used to be magic number 0
        IS_SENDCOUNT,           // used to be magic number 1
        IS_RECVCOUNT,           // used to be magic number 2
        IS_LAST_KIND_OF_COUNT
    };


    /** enum for the kind of collective call that is used.
     *  This enum replaces the magic numbers that had been used previously
     *  in the switch statements of our check functions.
     */
    enum KIND_OF_COLLECTIVE_CALL
    {
        IS_NOT_REDUCE_ALLREDUCE,  // used to be magic number 0
        IS_REDUCE_ALLREDUCE,      // used to be magic number 1
        IS_LAST_KIND_OF_COLL
    };


    /** enum for the kind of datatype that is used.
     *  This enum replaces the magic numbers that had been used previously
     *  in the switch statements of our check functions.
     *  Sometimes we have to distinguish between the C and the Fortran
     *  language binding.
     *
     *  @see  MPI-standard 1.2, pp. 211
     *  @see  mpo_Init.cc
     */
    enum KIND_OF_DATATYPE
    {
        // null handle and other invalid datatypes
        DATATYPE_NULL,
        DATATYPE_NOT_VALID,             // currently not used
        DATATYPE_CREATED_NOT_COMMITTED, // used to be magic number 0

        // all elementary and user-defined, committed datatypes
        DATATYPE_COMMITTED_C,           // used to be magic number 1
        DATATYPE_COMMITTED_C_AND_FORTRAN,  // used to be magic number 1
        DATATYPE_COMMITTED_FORTRAN,     // used to be magic number 5

        // all datatypes for reduction functions
        DATATYPE_REDUCTION_C,           // used to be magic number 3
        DATATYPE_REDUCTION_FORTRAN,     // used to be magic number 5

        // all optional datatypes
        DATATYPE_OPTIONAL_C,            // used to be magic number 4
        DATATYPE_OPTIONAL_FORTRAN,

        // special datatypes for constructing derived datatypes
        DATATYPE_CONSTRUCT_DERIVED_TYPE,

        // others
        DATATYPE_UNDEFINED,             // currently not used
        DATATYPE_LAST_KIND_OF_DATATYPE
    };


    /** enum for the kind of request that is used.
     *  This enum replaces the magic numbers that had been used previously
     *  in the constructors of MPO_Initiator and all derived classes and
     *  in the switch statements of our check functions.
     *  IMPORTANT:
     *  Whenever an enumeration value is added, make sure to also add a string
     *  to MARMOT_ERR_STRINGS.
     *  Keep the same order as in the enum, that makes life easier.
     *
     *  @see  MARMOT_ERR_STRINGS
     *  @see  mpo-debugserver.cc
     *  @see  MPI-standard 1.2, pp. 211
     *  @see  mpo_Init.cc
     */
    enum KIND_OF_REQUEST
    {
        REQUEST_NULL,
        REQUEST_NOT_PERSISTENT,          // used to be magic number 0
        REQUEST_ACTIVE_PERSISTENT,       // used to be magic number 1
        REQUEST_INACTIVE_PERSISTENT,     // used to be magic number -1
        REQUEST_UNDEFINED,               // currently not used
        REQUEST_LAST_KIND_OF_REQUEST
    };

    //==========================
    // TCheck Vars Enum
    //==========================

    /**
     * This Enum is used to select the Variables present in
     * MPO_TCheck_Races.
     */
    enum MPO_TCHECK_RACES_VARS
    {
        APP_NEEDS_MPI_THREAD_MULTIPLE = 0,
        APP_MPI_FINALIZE_TO_EARLY
    };



    //==========================
    // Error codes
    //==========================

    /** enum for MARMOT's error return codes.
     *  If you add a new error code, choose a descriptive name.
     *  Distinguish between MARMOT_ERR_FOO, MARMOT_WARN_FOO and
     *  MARMOT_NOTE_FOO.
     *
     *  Also when adding a new value refresh the MPI Standard references:
     *   * add new labels into the latex sources
     *   * go to SVN root
     *   * cd into MPI_Standards
     *   * perform an "make html"
     *   * commit the resulting changes in the DOC directory
     *   * commit the file SVNROOT/SRC/SRC/MPO_MPIStandardRef.cc
     *
     *  @see  enum KIND_OF_DATATYPE
     */
    enum MARMOT_ERR_CODES
    {
        //======================
        //======================
        // MPI-1.2
        // error types in
        // alphabetical order
        //======================
        //======================

        MARMOT_SUCCESS = 0,     // successful return code
        MARMOT_ERR_UNDEFINED,   // undefined error
        MARMOT_ERR_MARMOT,      // implementation error in MARMOT


        // blocklength
        //======================

        // errors (blocklength)
        //----------------------
        MARMOT_ERR_BLOCKLENGTH_NEG,     // blocklength is negative

        // warnings (blocklength)
        //----------------------
        MARMOT_WARN_BLOCKLENGTH_ZERO,  // blocklength is 0


        // buffer
        //======================

        // errors (buffer)
        //----------------------
        // MARMOT_ERR_BUFFER,            // invalid buffer pointer
        // MARMOT_ERR_TRUNCATE,          // message truncated on receive
        MARMOT_ERR_BUFFER_ATTACHED,      // already a buffer attached
        MARMOT_ERR_BUFFER_NOT_ATTACHED,  // no buffer attached
        MARMOT_ERR_BUFFER_IN_USE,        // a recieve or send buffer is still in use

        // color
        //======================

        // errors (color)
        //----------------------
        MARMOT_ERR_COLOR_NEG,         // color is negative

        // warnings (color)
        //----------------------
        MARMOT_WARN_COLOR_UNDEFINED,  // color is MPI_UNDEFINED


        // communicator
        //======================

        // errors (communicator)
        //----------------------
        MARMOT_ERR_COMM_NULL,          // communicator is MPI_COMM_NULL
        MARMOT_ERR_COMM_NOT_VALID,     // communicator is not valid
        MARMOT_ERR_COMM_NOT_CARTESIAN, // communicator is not cartesian
        MARMOT_ERR_COMM_NOT_GRAPH,     // communicator is not graph
        MARMOT_ERR_COMM_NOT_INTERCOMM, // communicator is not intercommunicator
        MARMOT_ERR_COMM_INTERCOMM,     // communicator is intercommunicator

        // warnings (communicator)
        //----------------------
        MARMOT_WARN_COMM_NOT_WORLD,    // communicator is not MPI_COMM_WORLD
        MARMOT_WARN_COMM_NULL,         // communicator is MPI_COMM_NULL
        MARMOT_WARN_COMM_LEFT,		   // communicator(s) not freed


        // coordinate
        //======================

        // errors (coordinate)
        //----------------------
        MARMOT_ERR_COORD_NEG,      // coordinate is negative
        MARMOT_ERR_COORD_TOO_BIG,  // coordinate is out of range


        // count
        //======================

        // errors (count)
        //----------------------
        MARMOT_ERR_COUNT_NEG,   // count is negative
        MARMOT_ERR_COUNT_ZERO,  // count is 0
        // MARMOT_ERR_COUNT_ARRAY_NEG,  //

        // warnings (count)
        //----------------------
        MARMOT_WARN_COUNT_ZERO,        // count is 0
        MARMOT_WARN_COUNTS_NOT_EQUAL,  // counts are not equal
        MARMOT_WARN_COUNT_UNDEFINED,   // count is MPI_UNDEFINED


        // datatype
        //======================

        // errors (datatype)
        //----------------------
        MARMOT_ERR_TYPE_NULL,         // datatype is MPI_DATATYPE_NULL
        MARMOT_ERR_TYPE_NOT_VALID,    // datatype is not valid
        MARMOT_ERR_TYPE_NOT_COMMITTED,// datatype is created, not committed
        MARMOT_ERR_TYPE_C,            // datatype is C datatype
        MARMOT_ERR_TYPE_F,            // datatype is Fortran datatype
        MARMOT_ERR_TYPE_REDUCTION_C,  // datatype is for reduction function (C)
        MARMOT_ERR_TYPE_REDUCTION_F,  // datatype is for reduction function (F)
        MARMOT_ERR_TYPE_OPTIONAL_C,   // datatype is optional (C)
        MARMOT_ERR_TYPE_OPTIONAL_F,   // datatype is optional (F)
        MARMOT_ERR_TYPE_CONSTR_DERIVED, // datatype is for constructing
                                      // derived datatypes
        MARMOT_ERR_TYPE_LB,           // datatype is MPI_LB
        MARMOT_ERR_TYPE_UB,           // datatype is MPI_UB

        // warnings (datatype)
        //----------------------
        MARMOT_WARN_TYPE_REDUCTION_C, //datatype is for reduction function (C)
        MARMOT_WARN_TYPE_REDUCTION_F, //datatype is for reduction function (F)
        MARMOT_WARN_TYPE_OPTIONAL_C,  // datatype is optional (C)
        MARMOT_WARN_TYPE_OPTIONAL_F,  // datatype is optional (F)
        MARMOT_WARN_TYPE_LEFT, //datatype(s) was(where) not freed

        // notes (datatype)
        //----------------------
        MARMOT_NOTE_TYPE_NULL,        // datatype is MPI_TYPE_NULL
        MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C,  // datatype is already
                                               // committed (C)
        MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C_AND_F,  // datatype is already
                                                     // committed (C and F)
        MARMOT_NOTE_TYPE_ALREADY_COMMITTED_F,  // datatype is already
                                               // committed (F)


        // dimension
        //======================

        // errors (dimension)
        //----------------------
        MARMOT_ERR_DIM_NEG,                // dimension is negative
        MARMOT_ERR_DIM_TOO_BIG,            // dimension is too big
        MARMOT_ERR_NODES_NEQ_PRODUCT_DIMS, // nodes are not a multiple of dims

        // warnings (dimension)
        //----------------------
        MARMOT_WARN_DIM_ZERO,              // dimension is zero


        // displacement
        //======================

        // errors (displacement)
        //----------------------
        MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA, // displacement overrides data

        // notes (displacement)
        //----------------------
        MARMOT_NOTE_DISPLACEMENT_READS_DATA, // displacement reads  data /
                                             // holes read from other nodes
        MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA, // displacement overrides data
        MARMOT_NOTE_DISPLACEMENT_DATA_SENT_TWICE,  // displacement sends data
                                                   // twice


        // edge
        //======================

        // errors (edge)
        //----------------------
        MARMOT_ERR_EDGE_INVALID,     // edge is invalid
        MARMOT_ERR_EDGE_NEG,         // edge is negative


        // errhandler
        //======================

        // errors (errhandler)
        //----------------------
        MARMOT_ERR_ERRHANDLER_NOT_VALID,  // errhandler is not valid

		// warnings (errhandler)
        //----------------------
        MARMOT_WARN_ERRHANDLER_LEFT, // errorhandler(s) left (not freed)

        // errorcode
        //======================

        // errors (errorcode)
        //----------------------
        MARMOT_ERR_ERRORCODE_NOT_VALID,  // errorcode is not valid


        // group
        //======================

        // errors (group)
        //----------------------
        MARMOT_ERR_GROUP_NULL,             // group is MPI_GROUP_NULL
        MARMOT_ERR_GROUP_NOT_VALID,        // group is not valid

        // warnings (group)
        //----------------------
        MARMOT_WARN_GROUP_NULL,            // group is MPI_GROUP_NULL
        MARMOT_WARN_GROUP_LEFT,			   // group(s) not freed


        // index
        //======================

        // errors (index)
        //----------------------
        MARMOT_ERR_INDEX_NEG,       // index is negative

        // warnings (index)
        //----------------------
        MARMOT_WARN_INDEX_LT_SIZE,  // index is less than size of communicator


        // leader
        //======================

        // errors (leader)
        //----------------------
        // Should we also distinguish between errors for remote
        // and for local leaders?
        MARMOT_ERR_LEADER_ANY_SOURCE,       // leader is MPI_ANY_SOURCE
        MARMOT_ERR_LEADER_NEG,              // leader is negative
        MARMOT_ERR_LEADER_OUT_OF_COMM,      // leader is out of communicator


        // neighbor
        //======================

        // errors (neighbor)
        //----------------------
        MARMOT_ERR_NEIGHBORS_NEG,           // number of neighbors is negative


        // operator
        //======================

        // errors (operator)
        //----------------------
        MARMOT_ERR_OP_NULL,                 // operator is MPO_OP_NULL
        MARMOT_ERR_OP_NOT_VALID,            // operator is not valid

        // warnings (operator)
        //----------------------
		MARMOT_WARN_OP_LEFT,				//operator(s) not freed

        // pending messages
        //======================

        // warnings (pending messages)
        //----------------------
        MARMOT_WARN_PENDING_MESSAGES_LEFT,  // pending messages are left


        // range
        //======================

        // errors (range)
        //----------------------
        MARMOT_ERR_RANK_FIRST_NEG,        // first rank is negative
        MARMOT_ERR_RANK_LAST_NEG,         // last rank is negative
        MARMOT_ERR_RANK_FIRST_TOO_BIG,    // first rank is too big
        MARMOT_ERR_RANK_LAST_TOO_BIG,     // last rank is too big
        MARMOT_ERR_RANK_COMPUTED_NEG,     // computed rank is negative
        MARMOT_ERR_RANK_COMPUTED_TOO_BIG, // computed rank is too big
        MARMOT_ERR_RANK_COMPUTED_TWICE,   // rank is computed twice
        MARMOT_ERR_STRIDE_NEG,            // stride is negative
        MARMOT_ERR_STRIDE_ZERO,           // stride is zero
        MARMOT_ERR_STRIDE_POS,            // stride is positive


        // rank
        //======================

        // errors (rank, see also range errors)
        //----------------------
        // MARMOT_ERR_RANK,               // invalid rank
        MARMOT_ERR_RANK_NEG,              // rank is negative
        MARMOT_ERR_RANK_PROC_NULL,        // rank is MPI_PROC_NULL
        MARMOT_ERR_RANK_ANY_SOURCE,       // rank is MPI_ANY_SOURCE
        MARMOT_ERR_RANK_TOO_BIG,          // rank is too big
        // MARMOT_ERR_DUP_RANK,
        // MARMOT_ERR_RANK_ARRAY,
        // MARMOT_ERR_LOCAL_RANK,
        // MARMOT_ERR_REMOTE_RANK,

        // warnings (rank)
        //----------------------
        MARMOT_WARN_RANK_ANY_SOURCE,      // rank is MPI_ANY_SOURCE

        // notes (rank)
        //----------------------
        MARMOT_NOTE_RANK_PROC_NULL,       // rank is MPI_PROC_NULL


        // request
        //======================

        // errors (request)
        //----------------------
        MARMOT_ERR_REQUEST_COUNT_NEG,          // count of requests is negative
        MARMOT_ERR_REQUEST_NOT_VALID,          // request is not valid
        MARMOT_ERR_REQUEST_STILL_USED,         // request is still in use
        MARMOT_ERR_REQUEST_NOT_PERSISTENT,     // request is not persistent
        MARMOT_ERR_REQUEST_OUTCOUNT_UNDEFINED, // outcount is MPI_UNDEFINED

        // warnings (request)
        //----------------------
        MARMOT_WARN_REQUEST_COUNT_ZERO,          // count of requests is zero
        MARMOT_WARN_REQUEST_NULL,                // request is MPI_REQUEST_NULL
        MARMOT_WARN_REQUEST_ACTIVE_NOT_PERSISTENT, // request is not persistent
                                                 // and active
        MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT,   // request is persistent and
                                                 // still active
        MARMOT_WARN_REQUEST_INACTIVE_PERSISTENT, // request is persistent
                                                 // and inactive
        MARMOT_WARN_REQUEST_OUTCOUNT_UNDEFINED,  // outcount is MPI_UNDEFINED
        MARMOT_WARN_REQUESTS_LEFT,               // requests are left


        // size
        //======================

        // errors (size)
        //----------------------
        MARMOT_ERR_SIZE_NEG,              // size is negative

        // warnings(size)
        //----------------------
        MARMOT_WARN_SIZE_ZERO,            // size is zero


        // tag
        //======================

        // errors (tag)
        //----------------------
        MARMOT_ERR_TAG_NEG,             // tag is negative
        MARMOT_ERR_TAG_TOO_BIG,         // tag is too big ( > TAG_UB)
        MARMOT_ERR_TAG_ANY_TAG,         // tag is MPI_ANY_TAG

        // warnings (tag)
        //----------------------
        MARMOT_WARN_TAG_TOO_BIG,        // tag is too big ( > guaranteed range,
                                        //  but < TAG_UB)


        // topology
        //======================

        // errors (topology, see also dimension errors, coordinates errors)
        //----------------------
        MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD, // new topology greater than
                                             // old communicator
        MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD, // new cartesian topology
                                                  // greater than
                                                  // old communicator


        // warnings (topology, see also dimension errors, coordinates errors)
        //----------------------
        MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD, // new topology smaller than
                                              // old communicator
        MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD, // new cartesian topology
                                                   // smaller than
                                                   // old communicator


        // misc
        //======================

        // misc
        // MARMOT_ERR_PENDING,           // Pending request
        MARMOT_ERR_DEADLOCK,   //Used for deadlock detection
        MARMOT_WARN_DEADLOCK,   //Used for deadlock detection (WARNING because we are actually doing deadlock detection only by timing)
        MARMOT_ERR_POINTER_NULL, //Used if a pointer should point to allocated memory but it is null instead
        MARMOT_ERR_MPI_AFTER_FINALIZE, //performed an MPI call except MPI_GET_VERSION, MPI_INITIALIZED and MPI_FINALIZED after MPI_FINALIZE !!!


        //======================
        //======================
        // MPI-2
        // file I/O
        // not implemented yet
        // error types in
        // alphabetical order
        //======================
        //======================


        // amode
        //======================

        // errors (amode)
        //----------------------
        MARMOT_ERR_AMODE_DIFF,      // file access mode is not the same
                                    // for all processes
        MARMOT_ERR_AMODE_RDONLY,    // MPI_MODE_RDONLY is specified
        MARMOT_ERR_AMODE_WRONLY,    // MPI_MODE_WRONLY is specified

        // MARMOT_ERR_AMODE_RDONLY_AND_RDWR,
        // MARMOT_ERR_AMODE_RDONLY_AND_WRONLY,
        // MARMOT_ERR_AMODE_RDWR_AND_WRONLY,
        MARMOT_ERR_AMODE_RDONLY_RDWR_WRONLY_NOT_SET_EXACTLY_ONE, // exactly one
        // of MPI_MODE_RDONLY, MPI_MODE_RDWR, MPI_MODE_WRONLY
        // must be specified
        MARMOT_ERR_AMODE_RDONLY_AND_CREATE,   // MPI_MODE_RDONLY is specified
                                              // with MPI_MODE_CREATE
        MARMOT_ERR_AMODE_RDONLY_AND_EXCL,     // MPI_MODE_RDONLY is specified
                                              // with MPI_MODE_EXCL
        MARMOT_ERR_AMODE_RDWR_AND_SEQUENTIAL, // MPI_MODE_RDWR is specified
                                              // with MPI_MODE_SEQUENTIAL
        MARMOT_ERR_AMODE_UNIQUE_OPEN,     // MPI_MODE_UNIQUE_OPEN is specified
        MARMOT_ERR_AMODE_SEQUENTIAL,      // MPI_MODE_SEQUENTIAL is specified
        MARMOT_ERR_AMODE_SEQUENTIAL_AND_NOT_DISPLACEMENT_CURRENT, //
                                     // MPI_MODE_SEQUENTIAL is specified,
                                     // but not MPI_DISPLACEMENT_CURRENT
        MARMOT_ERR_AMODE_SEQUENTIAL_NOT_SET_BUT_DISPLACEMENT_CURRENT, //
                                     // MPI_MODE_SEQUENTIAL is not specified,
                                     // but MPI_DISPLACEMENT_CURRENT

        // notes (amode)
        //----------------------
        MARMOT_NOTE_AMODE_DELETE_ON_CLOSE,      // amode is MPI_DELETE_ON_CLOSE


        // communicator - see also above
        //======================

        // errors (communicator)
        //----------------------
        // MARMOT_ERR_COMM_INTERCOMM, // communicator is intercommunicator


        // data representation
        //======================

        // errors (data representation)
        //----------------------
        MARMOT_ERR_DATAREP_DIFF, // data representation is not the same
                                 // for all processes
        MARMOT_ERR_DATAREP_GT_MAX_DATAREP_STRING,// data representation string
                                                 // is larger than
                                                 // MPI_MAX_DATAREP_STRING

        // warnings (data representation)
        //----------------------
        MARMOT_WARN_DATAREP_TOO_SMALL, // data representation is not
                                       // large enough


        // displacement
        //======================

        // errors (displacement)
        //----------------------
        MARMOT_ERR_FILETYPE_DISPLACEMENTS_NEG, // displacements in typemap of
                                               // filetype are negative
        MARMOT_ERR_FILETYPE_DISPLACEMENTS_DECREASE, // displacements in typemap
                                                    // of filetype decreasing


        // warnings (displacement)
        //----------------------
        MARMOT_WARN_DATA_IN_NEW_REGION_UNDEFINED, // values of data in new
                            // regions (those locations with displacements
                            // between old file size and size) are undefined



        // datatype - see also above datatype errors
        //======================

        // errors (datatype)
        //----------------------
        MARMOT_ERR_TYPE_OVERLAPPING,        // datatype has overlapping regions


        // etype - see also above datatype errors
        //======================

        // errors (etype)
        //----------------------
        MARMOT_ERR_ETYPE_EXTENT_DIFF, // extent of etype is not the same
                                      // on all processes
        MARMOT_ERR_ETYPE_NOT_VALID,   // etype is not valid,
                                      // see above data type errors
        MARMOT_ERR_ETYPE_OVERLAPPING, // etype has overlapping regions


        // filetype - see also above datatype errors
        //======================

        // errors (filetype)
        //----------------------
        MARMOT_ERR_FILETYPE_EXTENT_HOLE_NO_MULTIPLE_EXTENT_ETYPE, // extent of
                                     // hole in the filetype is not a multiple
                                     // of the extent of the etype
        MARMOT_ERR_FILETYPE_NOT_VALID,   // filetype is not valid,
                                         // see above data type errors
        MARMOT_ERR_FILETYPE_OVERLAPPING, // filetype has overlapping regions

        // notes (filetype)
        //----------------------
        MARMOT_NOTE_FILETYPE_HAS_HOLES,        // filetype has holes


        // filename
        //======================

        // errors (filename)
        //----------------------
        MARMOT_ERR_FILE_DIFF, // filenames do not reference the same file
        MARMOT_ERR_FILE_DOES_NOT_EXIST, // file does not exist

        // warnings (filename)
        //----------------------
        MARMOT_WARN_FILE_IN_USE,      // file still in use


        // group - see above
        //======================

        // errors (group)
        //----------------------
        // MARMOT_ERR_GROUP_NOT_VALID,        // group is not valid

        // warnings (group)
        //----------------------
        // MARMOT_WARN_GROUP_NULL,             // group is MPI_GROUP_NULL


        // info
        //======================

        // errors (info)
        //----------------------
        MARMOT_ERR_INFO_DIFF, // info entry is not the same for all processes

        // notes (info)
        //----------------------
        MARMOT_NOTE_INFO_IGNORED, // implementation is free to ignore info
	MARMOT_NOTE_PROCESSOR_NAME, //processor's name a process is running on


        // offset
        //======================

        // errors (offset)
        //----------------------
        MARMOT_ERR_OFFSET_NEG,                // offset is negative
        MARMOT_ERR_OFFSET_TO_NEG_POS_IN_VIEW, // offset reaches a negative
                                              // position in the view


        // size
        //======================

        // errors (size)
        //----------------------
        MARMOT_ERR_FILE_SIZE_DIFF, // size is not the same for all processes
        MARMOT_ERR_FILE_SIZE_NEG,  // size is negative

        // notes (size)
        //----------------------
        MARMOT_NOTE_FILE_SIZE_LT_OR_EQ_CURRENT_SIZE, // Size is less than or
                                                     // equal current size


        // misc
        //======================

        // errors (misc)
        //----------------------
        MARMOT_ERR_REQUESTS_NONBLOCKING_LEFT, // request left on file handle,
                                              // see above
        MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_LEFT, // split collective
                                       // operation left on file handle
        MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE, // more than one
                         // active split collective operation on file handle
        MARMOT_ERR_NO_MATCHING_BEGIN, // no matching begin call
        MARMOT_ERR_WHENCE_DIFF, // whence argument is not the same on all
                                // processes
        MARMOT_ERR_FLAG_DIFF, // flag argument is not the same on all processes
        MARMOT_ERR_FILEHANDLE_DIFF, // file handle argument is not the same
                                    // on all processes

        // warnings (misc)
        //----------------------
        MARMOT_WARN_FILE_OPEN_LEFT,  // open file left (after MPI_Finalize)
        MARMOT_WARN_REQUESTS_NONBLOCKING_LEFT, // request left on file handle,
                                               // see above
        MARMOT_WARN_OPERATION_SPLIT_COLLECTIVE_LEFT, // split collective
                                        // operation left on file handle
	MARMOT_WARN_DEBUGSERVER_NO_EXCLUSIVE_NODE,	//the debugserver doesn't run on an exclusive node
	MARMOT_WARN_PROCESSES_SHARE_NODE,	//two processes share one node


        //======================
        //======================
        // MPI-2
        // one-sided communication
        // not implemented yet
        // error types in
        // alphabetical order
        //======================
        //======================


        // windows
        //======================

        // errors (windows)
        //----------------------
        MARMOT_ERR_WIN_NULL,       // win is MPI_WIN_NULL
        MARMOT_ERR_WIN_NOT_VALID,  // win is not valid
        MARMOT_ERR_WIN_FREE_NOT_ALLOWED,

        // warnings ()
        //----------------------
        MARMOT_WARN_WIN_NULL,


        // size
        //======================

        // errors (size)
        //----------------------
        MARMOT_ERR_WIN_SIZE_NEG,   // win size is negative


        // displacement
        //======================

        // errors (displacement)
        //----------------------
        MARMOT_ERR_WIN_DISP_UNIT_SIZE_NONPOS,   // unit size for displacements
                                                // is not positive

        // locktype
        //======================

        // errors (locktype)
        //----------------------
        MARMOT_ERR_LOCKTYPE_NOT_VALID,   // lock type not valid


        // assert
        //======================

        // errors (assert)
        //----------------------
        MARMOT_ERR_ASSERT_NOT_VALID,   // assert type not valid



        //======================
        //======================
        // MPI-2
        // thread level
        // error types in
        // alphabetical order
        //======================
        //======================

        MARMOT_WARN_THREADLEVEL_GOT_LESS, //MPI_Init_thread returned a threadlevel lower than the one speciefied.
        MARMOT_WARN_FINALIZE_NOT_THREAD_MAIN, //MPI_Finalize was perforemed by a thread that did not perform MPI_Init.
        MARMOT_WARN_FINALIZE_OTHER_THREADS_WORKING, //MPI_Finalize was called while other threads of this process were still working.
        MARMOT_ERR_REQUEST_USED_AT_MULTIPLE_THREADS, //A request was used on to diferent calls by different threads.
        MARMOT_ERR_COMM_USED_IN_MULTIPLE_COLLECTIVES, //A comm was used in two collective calls simultaneously.
        MARMOT_WARN_PROBE_INVALIDATED, //A probe was performed and returned source and tag for a message but before this could be recieved another thread recieved it.
        MARMOT_ERR_VIOLATED_THREAD_LEVEL, //The Threadlevel provided by MPI_Init_thread/MPI_Init was violated


        // warnings ()
        //----------------------
        MARMOT_WARN_,

        // notes ()
        //----------------------
        MARMOT_NOTE_,

        // last enum value
        //======================
        MARMOT_LAST_ERR_CODE
    };                          // end enum MARMOT_ERR_CODES


}                               // end namespace MPO

#endif
