/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declarations of class Factory and utility classes 
 *  (defined here with templates).
 *
 *  This class is used to construct objects according to 
 *  an identifier. Each object has to implement the member KEY getId().
 *  returning a unique id. After the objects are registered with add(), they 
 *  can be retrieved with create(id). 
 *  The factory returns pointer to BaseType
 * 
 *  example program:
 * 
 *  enum {A_ID=0, B_ID=1, C_ID, LAST_ID};
 * 
 *  class A_base{
 *  public:
 *    virtual int getId(){return A_ID;}
 *  };
 * 
 *  class B : public A_base{
 *  public:
 *    int getId(){return B_ID;}
 *  };
 * 
 *  class C : public A_base{
 *  public:
 *    int getId(){return C_ID;}
 *  };
 * 
 * 
 *  int main(void){
 *    Factory<int,A_base> my_safe;
 *    my_safe.add<B>();
 *    my_safe.add<C>();
 * 
 *    A_base* my_A;
 *    my_A = my_safe.create(B_ID);
 *    std::cout << " created " << my_A->getId() << "\n";
 * 
 *    my_A = my_safe.create(C_ID);
 *    std::cout << " created " << my_A->getId() << "\n";
 *    return 0;
 *  }
 *
 *  @see Currently there is no implementation file, everything is defined 
 *       within this header file (templates have to be implemented here).
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller
 *
 *  $Id: mpo-factory.h 838 2008-06-13 10:44:34Z hpchimml $
 */


#ifndef MPO_FACTORY_H
#define MPO_FACTORY_H


#include <map>
#include <iostream>


namespace MPO
{


    //*************************************************************************
    //
    //                          Factory
    //
    /** This class is used to construct objects according to 
     *  an identifier. Each object has to implement the member KEY getId().
     *  returning a unique id. After the objects are registered with add(),  
     *  they can be retrieved with create(id). 
     *  The factory returns pointer to BaseType.
     * 
     *  example program:
     *  
     *  enum {A_ID=0,B_ID=1,C_ID,LAST_ID};
     * 
     *  class A_base{
     *  public:
     *    virtual int getId(){return A_ID;}
     *  };
     * 
     *  class B : public A_base{
     *  public:
     *    int getId(){return B_ID;}
     *  };
     * 
     *  class C : public A_base{
     *  public:
     *    int getId(){return C_ID;}
     *  };
     * 
     * 
     *  int main(void){
     *    Factory<int,A_base> my_safe;
     *    my_safe.add<B>();
     *    my_safe.add<C>();
     * 
     *    A_base* my_A;
     *    my_A = my_safe.create(B_ID);
     *    std::cout << " created " << my_A->getId() << "\n";
     * 
     *    my_A = my_safe.create(C_ID);
     *    std::cout << " created " << my_A->getId() << "\n";
     *    return 0;
     *  }
     */

    template < typename KEY, typename BaseType > class Factory
    {
    public:
        //==========================
        // Constructors/Destructors                            
        //==========================

        //==========================
        // Methods                            
        //==========================
        //bk// This method is not used, why?
        template < class T > void add(KEY id)
            {
                class_list[id] = new Factory_helper < T >;
            }

        template < class T > void add()
            {
                T tmp;
                class_list[tmp.getId()] = new Factory_helper < T >;
            }

        BaseType *create(int id)
            {
                return (BaseType *) class_list[id]->createObject();
            }

    private:
        //=================================================================
        //                  utility classes
        //=================================================================

        //*****************************************************************
        //
        //                  Factory_helper_base
        //
        /** Abstract base class. */
        class Factory_helper_base
        {
        public:
            //==========================
            // Constructors/Destructors                            
            //==========================

            /** The default constructor. */
            Factory_helper_base()
                {
                }


            /** The destructor is virtual because there are virtual 
             *  members here.
             */
            virtual ~Factory_helper_base()
                {
                }

            //==========================
            // Methods                            
            //==========================

            virtual void *createObject() const = 0;


        };              // end class Factory_helper_base


        //*****************************************************************
        //
        //                  Factory_helper
        //  

        template < typename T > class Factory_helper
            : public Factory_helper_base
        {
        public:
            //==========================
            // Constructors/Destructors                            
            //==========================

            /** The default constructor. */
            Factory_helper()
                {
                }

            /** The destructor is virtual because there are virtual 
             *  members here.
             */
            virtual ~Factory_helper()
                {
                }

            //==========================
            // Methods                            
            //==========================

            virtual void *createObject() const
                {
                    return new T;
                }

        };              // end class Factory_helper



        //*****************************************************************
        //
        //                  Mapping
        //  

        std::map < KEY, Factory_helper_base * >class_list;

    };                      // end class Factory


}                               // end of namespace MPO

#endif
