/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of classes Messenger and MessengerUsingMPI. 
 *  These classes define the communication between the DebugServer
 *  and the clients.
 *
 *  @see mpo-messenger.cc (implementation)
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller
 *
 *  $Id: mpo-messenger.h 838 2008-06-13 10:44:34Z hpchimml $
 */

#ifndef MPO_MESSENGER_H
#define MPO_MESSENGER_H


#include <mpi.h>
#include "mpi_undef_makros.h"


namespace MPO
{
    class MPO_Object;

    //                         Messenger
    //*************************************************************************
    /** This abstract base class defines the basic interface.
     *
     *  Class used to send messages between the DebugServer and its clients.
     * 
     *  @see Currently there is no implementation of this class
     *       in mpo-messenger.cc.
     *       Everything is defined here.
     *       Nearly all methods are pure virtual anyway. 
     */
    class Messenger
    {
    public:
        //==========================
        // Constructors/Destructors                            
        //==========================

        /** The default constructor. */
        Messenger();

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~Messenger();

        //==========================
        // Methods                            
        //==========================

        /** @return true if message is available;
         *          if true rank and tag is set, otherwise rank and tag is untouched.
         */
        virtual bool poll(int &rank,int &tag) = 0;

        /** Client/DebugServer sends id to DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  id          integer to be sent
         *  @param  rank rank of receiver
         *  @return 0
         */
        virtual int sendId(int id, int rank) = 0;
        
        /** Client/DebugServer sends id to DebugServer/Client.
         *  @param  id          integer to be sent
         *  @param  rank rank of receiver
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int sendId(int id, int rank,int tag) = 0;

        /** Client/DebugServer receives id from DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  id   integer to be received
         *  @param  rank rank of sender
         *  @return 0
         */
        virtual int receiveId(int &id, int rank) = 0;
        
        /** Client/DebugServer receives id from DebugServer/Client.
         *  @param  id   integer to be received
         *  @param  rank rank of sender
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int receiveId(int &id, int rank,int tag) = 0;

        /** Client/DebugServer sends object to DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  obj  object to be sent
         *  @param  rank rank of receiver
         *  @return 0
         */
        virtual int sendObject(MPO_Object &obj, int rank) = 0;
        
        /** Client/DebugServer sends object to DebugServer/Client.
         *  @param  obj  object to be sent
         *  @param  rank rank of receiver
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int sendObject(MPO_Object &obj, int rank,int tag) = 0;

        /** Client/DebugServer receives object from DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  obj  object to be received
         *  @param  rank rank of sender
         *  @return 0
         */
        virtual int receiveObject(MPO_Object &obj, int rank) = 0;
        
        /** Client/DebugServer receives object from DebugServer/Client.
         *  @param  obj  object to be received
         *  @param  rank rank of sender
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int receiveObject(MPO_Object &obj, int rank,int tag) = 0;

        /** Client sends error codes to DebugServer.
         *  @param  errorcodes  errorcodes to be sent
         *  @param  rank        rank of receiver
         *  @return 0
         */
        // virtual int sendErrorobject(int &errorcodes, int rank = -1) = 0;

        /** DebugServer receives error codes from Client. 
         *  @param  errorcodes  errorcodes to be received
         *  @param  rank        rank of sender
         *  @return 0
         */
        // virtual int receiveErrorobject(int &errorcodes, int rank = -1) = 0;

    };                            // end class Messenger 



    //                         MessengerUsingMPI
    //*************************************************************************
    /** This class implements Messenger functionality using MPI.
     *
     *  @see mpo-messenger.cc (implementation)
     */
    class MessengerUsingMPI : public Messenger
    {
    public:
        //==========================
        // Constructors/Destructors                            
        //==========================

        /** The default constructor initializes an invalid object. */
        MessengerUsingMPI();

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MessengerUsingMPI();

        //==========================
        // Methods                            
        //==========================
        
        /** @return true if message is available;
         *          if true rank and tag is set, otherwise rank and tag is untouched.
         */
        virtual bool poll(int &rank,int &tag);

        /** Client/DebugServer sends id to DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  id          integer to be sent
         *  @param  rank rank of receiver
         *  @return 0
         */
        virtual int sendId(int id, int rank);
                
        /** Client/DebugServer sends id to DebugServer/Client.
         *  @param  id          integer to be sent
         *  @param  rank rank of receiver
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int sendId(int id, int rank,int tag);

        /** Client/DebugServer receives id from DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  id   integer to be received
         *  @param  rank rank of sender
         *  @return 0
         */
        virtual int receiveId(int &id, int rank);
                
        /** Client/DebugServer receives id from DebugServer/Client.
         *  @param  id   integer to be received
         *  @param  rank rank of sender
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int receiveId(int &id, int rank,int tag);

        /** Client/DebugServer sends object to DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  obj  object to be sent
         *  @param  rank rank of receiver
         *  @return 0
         */
        virtual int sendObject(MPO_Object &obj, int rank);
                
        /** Client/DebugServer sends object to DebugServer/Client.
         *  @param  obj  object to be sent
         *  @param  rank rank of receiver
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int sendObject(MPO_Object &obj, int rank,int tag);

        /** Client/DebugServer receives object from DebugServer/Client.
         *  The tag used is calculated by: threadID + 2 if threads are present otherwise the standard tag is used.
         *  ThreadID is retrieved using the syncronization interface.
         *  @param  obj  object to be received
         *  @param  rank rank of sender
         *  @return 0
         */
        virtual int receiveObject(MPO_Object &obj, int rank);
                
        /** Client/DebugServer receives object from DebugServer/Client.
         *  @param  obj  object to be received
         *  @param  rank rank of sender
         *  @param  tag the tag to be used.
         *  @return 0
         */
        virtual int receiveObject(MPO_Object &obj, int rank,int tag);

        /** Client sends error codes to DebugServer. 
         *  @param  errorcodes  errorcodes to be sent
         *  @param  rank        rank of receiver
         *  @return 0
         */
        //virtual int sendErrorCodes(int &errorcodes, int rank = -1);

        /** DebugServer receives error codes from Client.
         *  @param  errorcodes  errorcodes to be received
         *  @param  rank        rank of sender
         *  @return 0
         */
        //virtual int receiveErrorCodes(int &errorcodes, int rank = -1);

        /** Sets communicator.
         */
        void setCommunicator(MPI_Comm communicator);

    private:
        //==========================
        // Data                           
        //==========================

        static const int our_tag;
        int my_server_rank;
        MPI_Comm my_communicator;

    };                          // end class MessengerUsingMPI


}                               // end of namespace MPO

#endif
