/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Templates for our check functions. 
 *  Don't spread check functions all over the code if they do exactly 
 *  the same. Define them once here and document properly where they are used. 
 *  If you declare a new check function, put it into THIS file and document
 *  exactly in the @see comment wherever it is used.
 * 
 *  HINT FOR DEVELOPERS:
 *  ==================== 
 *  How to use these template functions: 
 *
 *
 *  How to add new template functions (keep functions in alphabetical order):
 * 
 * (1) Define the template function. Its name should reflect what this method 
 *     does. 
 * (2) If the template function is similar but not exactly the same as 
 *     another, choose a different name for it. Use descriptive names.
 *     Describe exactly what's been changed if you create a new method by 
 *     copy and paste.   
 * (3) Add comments.
 * (3.1) The @param comment describes parameters, if any.
 * (3.2) The @return comment must contain all the error codes that are 
 *       returned.
 *       These error codes are in enum MARMOT_ERR_CODES. Add a new error code
 *       if necessary. The error codes distinguish between 
 *       MARMOT_ERR_FOO, MARMOT_WARN_FOO and MARMOT_NOTE_FOO.
 *       Choose the name of the error codes in a way that it is clear what the
 *       error is.
 * (3.3) The @see comment must contain all classes that use this function.
 *  
 *  @see enum MARMOT_ERR_CODES
 *
 *  
 *  TODO: ADD SOME MORE DOCUMENTATION AND ALL THE METHODS
 *  
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo-templates.h 838 2008-06-13 10:44:34Z hpchimml $
 */


#ifndef MPO_TEMPLATES_H
#define MPO_TEMPLATES_H

namespace MPO
{

// TODO
// template for mpo_CALLOBJECT_BASECLASS_FOO.cc

// template for mpo_CALLOBJECT_FOO.cc

// template for mpo_TYPEOBJECT_FOO.cc

    //************************************************************************
    //                          templates for methods 
    //                          in alphabetic order
    //************************************************************************

    //========================
    // Methods (general)
    //========================    

    /** @return  the unique family identity set by enum MPO_FAMILY_ID
     */
    virtual int getFamilyId() const;

    /** Does the MPI call after checking it. */
    virtual int callNativeMPI() = 0;

    /** @return  the name of the MPI-call (C or Fortran spelling). */
    virtual string getCallname() const = 0;

    /** Overloads << to print call-description. */
    virtual std::ostream & debugout(std::ostream & output) const = 0;


    //========================
    // Methods (insert/erase)
    //========================    

    // eraseErrhandler()
    //========================================================================
    /** @see  @see  MPO_Comm_free
     *        exactly like this template 
     */

    /** Checks whether error-handler is ready to free and erases it. 
     *
     *  @return   unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     */
    virtual int eraseErrhandler();


    // eraseRequest()
    //========================================================================
    /** @see  
     */

// TODO //

    // insertGroup()
    //========================================================================
    /** @see  
     */

// TODO //

    // insertOp()
    //========================================================================
    /** @see  
     */

// TODO //

    //========================
    // Methods (checks)
    //========================    

    // checkPrerequisites() / checkPrerequisites2()
    //========================================================================
    /** They are all alike.
     *
     *  @see  MPO_*::checkPrerequisites()
     *
     */

    /** Checks prerequisites such as correctness of parameters.
     *  Checks whether 
     *  - count  
     *  - datatype
     *  - rank
     *  - tag
     *  - communicator 
     *  are valid.
     *  
     *  @return  0
     */
    virtual int checkPrerequisites();

// TODO //



    // checkAssert()
    //========================================================================
    /** They are all alike. 
     *
     *  @see  MPO_Win_lock::checkAssert()
     *        exactly like this template
     *
     *  @see  MPO_Win_fence::checkAssert()
     *        exactly like this template
     *
     *  @see  MPO_Win_post::checkAssert()
     *        exactly like this template
     *
     *  @see  MPO_Win_start::checkAssert()
     *        exactly like this template
     */

    /** Checks whether assert is valid. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_ASSERT_NOT_VALID
     */
    virtual int checkAssert();



    // checkBlocklength() / checkBlocklengthArray()
    //========================================================================
    /** They are all alike. 
     *  The only difference between them is having a single blocklength or
     *  having an array of blocklengths.
     *
     *  @see  MPO_Type_hvector::checkBlocklength()
     *        exactly like this template
     *
     *  @see  MPO_Type_vector::checkBlocklength() 
     *        exactly like this template
     *
     *  @see  MPO_Type_struct::checkBlocklengthArray()
     *        exactly like this template
     *
     *  @see  MPO_Type_hindexed::checkBlocklengthArray()
     *        exactly like this template
     *
     *  @see  MPO_Type_indexed::checkBlocklengthArray()
     *        exactly like this template
     */

    /** Checks whether blocklength is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_BLOCKLENGTH_NEG
     *           - MARMOT_WARN_BLOCKLENGTH_ZERO
     */
    virtual int checkBlocklength();

    /** Checks whether array of blocklengths is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_BLOCKLENGTH_NEG
     *           - MARMOT_WARN_BLOCKLENGTH_ZERO
     */
    virtual int checkBlocklengthArray();


    // checkBufferAttached()
    //========================================================================
    /** @see  MPO_Buffer_detach::checkBufferAttached()
     *        exactly like this template
     *
     *  @see  MPO_Buffer_attach::checkBufferAttached()
     *        exactly like this template,
     *        exactly like MPO_Buffer_detach::checkBufferAttached(), except for
     *        - error code
     */

    /** Checks whether buffer is attached.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_BUFFER_NOT_ATTACHED
     */
    virtual int checkBufferAttached();


    /** Checks whether buffer is attached.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_BUFFER_ATTACHED
     */
    virtual int checkBufferAttached();


    // checkColor()
    //========================================================================
    /** @see  MPO_Comm_split 
     *        exactly like this template
     */

    /** Checks whether color is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COLOR_NEG
     *           - MARMOT_WARN_COLOR_UNDEFINED
     */
    virtual int checkColor();


    // checkComm() / checkComm(MPO_Comm my_comm) / checkCommSecond()
    // checkCommFreed()
    //========================================================================
    /** They are all alike.
     *
     *  @see  MPO_Blocking::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Cartesian::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Collective::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Communicatorhandler::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Graph::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Initiator::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Intercommunicators::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Probes::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Sendrecvs::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Pack_size::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Pack::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Unpack::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Attr_delete::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Attr_get::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Attr_put::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Abort::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Errhandler_get::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Errhandler_set::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Comm_group::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Comm_test_inter::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Graph_create::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Cart_create::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Cart_map::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Win_create::checkComm()
     *        exactly like this template
     *
     *  @see  MPO_Comm_test_inter::checkCommFreed()
     *        exactly like checkComm(), except for:
     *        - the name
     *        - replace rvalue = MARMOT_ERR_COMM_NULL by
     *                  rvalue = MARMOT_WARN_COMM_NULL
     *  @see  mpo_Intercomm_create::checkComm(MPO_Comm my_comm)
     *        exactly like this template, except for:
     *        -  virtual int checkComm(MPO_Comm my_comm);
     *           with parameter my_comm here
     *
     *  @see  MPO_Comm_compare::checkCommSecond()
     *        exactly like this template, 
     *        exactly like checkComm, except for:
     */

    /** Checks whether communicator is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NULL
     *           - MARMOT_ERR_COMM_NOT_VALID
     */
    virtual int checkComm();

    /** Checks whether communicator is valid.
     *
     *  @param   my_comm    communicator
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NULL
     *           - MARMOT_ERR_COMM_NOT_VALID
     */
    virtual int checkComm(MPO_Comm my_comm);

    /** Checks whether second communicator is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NULL
     *           - MARMOT_ERR_COMM_NOT_VALID
     */
    virtual int checkCommSecond();

    /** Checks whether communicator is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NOT_VALID
     *           - MARMOT_WARN_COMM_NULL
     */
    virtual int checkCommFreed();

 
    // checkCommCartesian() / checkCommGraph()
    // checkCommIntercomm() / checkCommIntracomm(MPO_Comm my_comm)
    //========================================================================
    /** They are all alike.
     *
     *  @see  MPO_Cartesian::checkCommCartesian()
     *        exactly like this template
     *
     *  @see  MPO_Graph::checkCommGraph()
     *        exactly like this template
     *
     *  @see  MPO_Intercommunicators::checkCommIntercomm()
     *        exactly like this template
     *
     *  @see  MPO_Intercomm_create::checkCommIntracomm(MPO_Comm my_comm)
     *        exactly like this template
     */

    /** Checks whether communicator has cartesian structure.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NOT_CARTESIAN
     */
    virtual int checkCommCartesian();

    /** Checks whether communicator has graph structure.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NOT_GRAPH
     */
    virtual int checkCommGraph();

    /** Checks whether communicator is inter-communicator.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_NOT_INTERCOMM
     */
    virtual int checkCommIntercomm();

    /** Checks whether communicator is intra-communicator.
     *
     *  @param   comm    communicator
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COMM_INTERCOMM
     */
    virtual int checkCommIntracomm(MPO_Comm comm);


    // checkCommWorld()
    //========================================================================
    /** @see  MPO_Abort::checkCommWorld()
     *        exactly like this template
     */

    /** Checks whether communicator is MPI_COMM_WORLD.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_WARN_COMM_NOT_WORLD
     */
    virtual int checkCommWorld();

 
    // checkCount(const string KIND_OF_COUNT) 
    // checkCountIn(const string KIND_OF_COUNT) 
    // checkCount(int my_count, const string KIND_OF_COUNT) 
    // checkCountOfReceiver(const string KIND_OF_COUNT)
    // checkCountGroup(const string KIND_OF_COUNT
    //                 const string KIND_OF_NEW_GROUP)
    // checkCountArray(int* my_counts, const string KIND_OF_COUNT)
    //========================================================================
    /** They are all alike.
     *
     *  @see  MPO_Blocking::checkCount(const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Sendrecvs::checkCount(const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Initiator::checkCount(const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Typeconstructors::checkCount(const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Pack_size::checkCountIn(const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Pack::checkCountIn(const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Unpack::checkCountout(const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Collective::checkCount(int my_count, 
     *                                   const string KIND_OF_COUNT)
     *        exactly like this template
     *
     *  @see  MPO_Sendrecv::checkCountOfReceiver(const string KIND_OF_COUNT)
     *        exactly like this template
     *        exactly like checkCount except for:
     *        - the name
     *
     *  @see  MPO_Group_excl::checkCountGroup(const string KIND_OF_COUNT,
     *                                        const string KIND_OF_NEW_GROUP)
     *        exactly like this template
     *
     *  @see  MPO_Group_incl::checkCountGroup(const string KIND_OF_COUNT,
     *                                        const string KIND_OF_NEW_GROUP)
     *        exactly like this template
     *
     *  @see  MPO_Collective::checkCountArray(int* my_counts, 
     *                                        const string KIND_OF_COUNT)
     *        exactly like this template
     *
     */

    /** Checks whether count is valid. 
     *
     *  @param   KIND_OF_COUNT  specifies the kind of count for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_WARN_COUNT_ZERO
     */
    virtual int checkCount(const string KIND_OF_COUNT);

    /** Checks whether count is valid. 
     *
     *  @param   KIND_OF_COUNT  specifies the kind of count for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_WARN_COUNT_ZERO
     */
    virtual int checkCountIn(const string KIND_OF_COUNT);

    /** Checks whether count is valid. 
     *
     *  @param   KIND_OF_COUNT  specifies the kind of count for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_WARN_COUNT_ZERO
     */
    virtual int checkCountOut(const string KIND_OF_COUNT);

    /** Checks whether count is valid. 
     *
     *  @param   my_count       send/receive count 
     *  @param   KIND_OF_COUNT  specifies the kind of count for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_WARN_COUNT_ZERO
     */
    virtual int checkCount(int my_count, const string KIND_OF_COUNT);

    /** Checks whether count is valid.
     *
     *  @param   KIND_OF_COUNT  specifies the kind of count for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_WARN_COUNT_ZERO
     */
    virtual int checkCountOfReceiver(const string KIND_OF_COUNT);

    /** Checks whether count is valid. 
     *
     *  @param   KIND_OF_COUNT  specifies the kind of count for output
     *  @param   KIND_OF_NEW_GROUP  specifies the kind of new group 
     *                              for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_WARN_COUNT_ZERO
     */
    virtual int checkCountGroup(const string KIND_OF_COUNT, 
                                const string KIND_OF_NEW_GROUP);

    /** Checks whether count is valid. 
     *
     *  @param   my_counts      array of send/receive counts 
     *  @param   KIND_OF_COUNT  specifies the kind of count for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COUNT_NEG
     *           - MARMOT_WARN_COUNT_ZERO
     */
    virtual int checkCountArray(int* my_counts, const string KIND_OF_COUNT);


    // checkCountsAreEqual()
    //========================================================================
    /** @see  MPO_Alltoall
     *        exactly like this template
     */

    /** Checks whether send and receive count are equal.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_WARN_COUNTS_NOT_EQUAL
     */
     virtual int checkCountsAreEqual();


    // checkCountUndefined()
    //========================================================================
    /** @see  MPO_Get_elements
     *        exactly like this template
     *
     *  @see  MPO_Get_count 
     *        exactly like this template
     */

    /** Checks count returned. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_WARN_COUNT_UNDEFINED
     */
     virtual int checkCountUndefined();


    // checkDatatype() / checkDatatypeOfSender() / checkDatatypeSendRecv()
    //========================================================================
    /** They are all alike.
     *
     *  @see  MPO_Blocking::checkDatatype()
     *        exactly like this template
     *
     *  @see  MPO_Initiator::checkDatatype()
     *        exactly like this template
     *
     *  @see  MPO_Typeconstructors::checkDatatype()
     *        exactly like this template
     *
     *  @see  MPO_Typehandlers::checkDatatype()
     *        exactly like this template
     *
     *  @see  MPO_Type_struct::checkDatatype()
     *        exactly like this template
     *
     *  @see  MPO_Sendrecvs::checkDatatypeOfSender()
     *        exactly like this template
     *
     *  @see  MPO_Collective::checkDatatypeSendRecv
     *        exactly like this template
     *
     *  @see  MPO_Pack_size::checkDatatype()
     *        exactly like this template, except for:
     *        - TODO
     *
     *  @see  MPO_Pack::checkDatatype()
     *        exactly like this template, except for:
     *        - TODO
     *
     *  @see  MPO_Unpack::checkDatatype()
     *        exactly like this template, except for:
     *        - TODO
     *
     *  @see  MPO_Get_elements::checkDatatype()
     *        exactly like this template, except for:
     *        - TODO
     *
     *  @see  MPO_Get_count::checkDatatype()
     *        exactly like this template, except for:
     *        - TODO
     */

// TODO // Type/sendtype/recvtype

    /** Checks whether datatype is valid. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_MARMOT
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_COMMITTED
     *           - MARMOT_ERR_TYPE_C
     *           - MARMOT_ERR_TYPE_F
     *           - MARMOT_ERR_TYPE_REDUCTION_C
     *           - MARMOT_ERR_TYPE_REDUCTION_F
     *           - MARMOT_ERR_TYPE_OPTIONAL_C
     *           - MARMOT_ERR_TYPE_OPTIONAL_F
     *           - MARMOT_ERR_TYPE_ERR_TYPE_CONSTR_DERIVED
     *           - MARMOT_WARN_TYPE_REDUCTION_C
     *           - MARMOT_WARN_TYPE_REDUCTION_F
     *           - MARMOT_WARN_TYPE_OPTIONAL_C
     *           - MARMOT_WARN_TYPE_OPTIONAL_F
     */
    virtual int checkDatatype();

// TODO // checkDatatypeArrayOfSender


    // checkDatatypeArrayTypeStruct()
    //========================================================================
    /**
     *  @see  MPO_Type_struct::checkDatatypeArrayTypeStruct()
     *        exactly like this template
     */

    /** Checks correctness of array of datatypes. 
     *  We do NOT use the checkDatatype from our base class, 
     *  but redefine it here.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_MARMOT
     *           - MARMOT_ERR_TYPE_NOT_VALID
     *           - MARMOT_ERR_TYPE_NULL
     *           - MARMOT_ERR_TYPE_NOT_COMMITTED
     *           - MARMOT_ERR_TYPE_C
     *           - MARMOT_ERR_TYPE_F
     *           - MARMOT_ERR_TYPE_REDUCTION_C
     *           - MARMOT_ERR_TYPE_REDUCTION_F
     *           - MARMOT_ERR_TYPE_OPTIONAL_C
     *           - MARMOT_ERR_TYPE_OPTIONAL_F
     *           - MARMOT_ERR_TYPE_LB
     *           - MARMOT_ERR_TYPE_UB
     *           - MARMOT_WARN_TYPE_REDUCTION_C
     *           - MARMOT_WARN_TYPE_REDUCTION_F
     *           - MARMOT_WARN_TYPE_OPTIONAL_C
     *           - MARMOT_WARN_TYPE_OPTIONAL_F
     */
    virtual int checkDatatypeArrayTypeStruct();

    // checkDatatypeCommitted()
    //========================================================================
    /**
     *  @see  MPO_Type_commit::checkDatatypeCommitted()
     *        exactly like this template
     */

    /** Checks correctness datatype. 
     *  We do NOT use the checkDatatype from our base class, 
     *  but redefine it here.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_MARMOT
     *           - MARMOT_NOTE_TYPE_NULL
     *           - MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C
     *           - MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C_AND_F
     *           - MARMOT_NOTE_TYPE_ALREADY_COMMITTED_F  
     */
    virtual int checkDatatypeCommitted();


    // checkDimsArray()
    //========================================================================
    /** @see  MPO_Dims_create
     *        exactly like this template  
     */

    /** Checks array of dimensions.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_DIM_NEG
     *           - MARMOT_ERR_DIM_TOO_BIG
     *           - MARMOT_ERR_NODES_NEQ_PRODUCT_DIMS
     *           - MARMOT_WARN_DIM_ZERO
     */
    virtual int checkDimsArray();

    // checkMaxDims()
    //========================================================================
    /** @see  MPO_Cart_get
     *        exactly like this template
     */

    /** Checks dimensions.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_DIM_NEG
     *           - MARMOT_WARN_DIM_ZERO
     */
    virtual int checkMaxDims();


    // checkDisplacementArrayOfReceiver()
    // 
    //========================================================================
    /** @see  MPO_Alltoallv::checkDisplacementArrayOfReceiver()
     *        exactly like this template
     *
     *  @see  MPO_Scatterv::checkDisplacementArrayOfSender()
     *        exactly like this template
     *
     *  @see  MPO_Allgatherv::checkDisplacementArray()
     *        exactly like this template
     *
     *  @see  MPO_Gatherv::checkDisplacementArray()
     *        exactly like this template
     */

    // MPO_Alltoallv
    /** Checks array of displacements.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA
     *           - MARMOT_NOTE_DISPLACEMENT_READS_DATA
     */
    virtual int checkDisplacementArrayOfReceiver();

    /** Checks array of displacements.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_NOTE_DISPLACEMENT_DATA_SENT_TWICE
     *           - MARMOT_NOTE_DISPLACEMENT_READS_DATA
     */
    virtual int checkDisplacementArrayOfSender();

    // MPO_Allgatherv
    /** Checks array of displacements.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA
     *           - MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA
     */
    virtual int checkDisplacementArrayOfReceiver();

    /** Checks array of displacements.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA
     *           - MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA
     */
    virtual int checkDisplacementArrayOfRoot();


    // checkEdgeArray()
    //========================================================================
    /** @see  MPO_Graph_map
     *        exactly like this template
     *  
     *  @see  MPO_Graph_create::checkEdgeArray()
     *        exactly like this template
     */

    /** Checks array of edges.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_EDGE_INVALID
     */
    virtual int checkEdgeArray();


    // checkErrhandler() / checkAndEraseErrhandler()
    //========================================================================
    /** @see  MPO_Errhandler_get::checkErrhandler()
     *        exactly like this template
     *
     *  @see  MPO_Errhandler_set::checkErrhandler()
     *        exactly like this template
     *
     *  @see  MPO_Errhandler_free::checkAndEraseErrhandler()
     *        exactly like this template
     */

    /** Checks if errhandler is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_ERRHANDLER_NOT_VALID
     */
    virtual int checkErrhandler();

    /** Checks if errhandler is valid and erases it from our map.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_ERRHANDLER_NOT_VALID
     */
    virtual int checkAndEraseErrhandler();


    // checkErrorcode()
    //========================================================================
    /** 
     *  @see  MPO_Error_class::checkErrorcode()
     *        exactly like this template
     *  
     *  @see  MPO_Error_string::checkErrorcode()
     *        exactly like this template
     *  
     */

    /** Checks if errorcode is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_ERRORCODE_NOT_VALID
     */
    virtual int checkErrorcode();


    // checkGroup() / checkGroup2()
    //========================================================================
    /** @see  
     */

// TODO //
    /** Checks whether group is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_GROUP_NULL
     *           - MARMOT_ERR_GROUP_NOT_VALID
     */
    virtual int checkGroup();

    /** Checks whether group is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_GROUP_NULL
     *           - MARMOT_ERR_GROUP_NOT_VALID
     */
    virtual int checkGroup2();


    // checkLeaderLocal() / checkLeaderRemote()
    //========================================================================
    /** @see  MPO_Intercomm_create
     *        exactly like this template
     */

    /** Checks whether local leader is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_LEADER_ANY_SOURCE
     *           - MARMOT_ERR_LEADER_NEG
     *           - MARMOT_ERR_LEADER_OUT_OF_COMM
     */
    virtual int checkLeaderLocal();

    /** Checks whether remote leader is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_LEADER_ANY_SOURCE
     *           - MARMOT_ERR_LEADER_NEG
     *           - MARMOT_ERR_LEADER_OUT_OF_COMM
     */
    virtual int checkLeaderRemote();


    // checkLockType()
    //========================================================================
    /** @see  MPO_Win_lock
     *        exactly like this template
     */

    /** Checks whether locktype is valid. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_LOCKTYPE_NOT_VALID
     */
    virtual int checkLockType();


    // checkMaxEdges()
    //========================================================================
    /** @see  MPO_Graph_get
     *        exactly like this template
     */

    /** Checks maximal number of edges.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_EDGE_NEG
     */
    virtual int checkMaxEdges();


    // checkMaxIndex()
    //========================================================================
    /** @see  MPO_Graph_get
     *        exactly like this template
     */

    /** Checks maximal index.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_INDEX_NEG
     *           - MARMOT_WARN_INDEX_LT_SIZE
     */
    virtual int checkMaxIndex();


    // checkMaxNeighbors()
    //========================================================================
    /** @see  MPO_Graph_neighbors
     *        exactly like this template
     */

    /** Checks maximal number of neighbors.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_NEIGHBORS_NEG
     */
    virtual int checkMaxNeighbors();


    // checkOp()
    //========================================================================
    /** @see  MPO_Op_free::checkOp()
     *        exactly like this template
     *
     *  @see  MPO_Collective::checkOp(MPO_Op op)
     *        exactly like this template
     */

    /** Checks whether op is valid. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OP_NOT_VALID
     */
    virtual int checkOp();

    /** Checks whether op is valid. 
     *
     *  @param   op   Operation
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_OP_NULL
     *           - MARMOT_ERR_OP_NOT_VALID
     */
    virtual int checkOp(MPO_Op op);


    // checkPendingMessagesLeft()
    //========================================================================
    /** @see  MPO_Finalize::checkPendingMessagesLeft()
     *        exactly like this template
     */

    /** Checks whether pending messages are left. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_WARN_PENDING_MESSAGES_LEFT
     */
    virtual int checkPendingMessagesLeft();



    // checkCoordinates()
    //========================================================================
    /** @see  MPO_Cart_rank
     *        exactly like this template
     */

    /** Checks whether coordinates are valid. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_COORD_NEG
     *           - MARMOT_ERR_COORD_TOO_BIG
     */
    virtual int checkCoordinates();


    // checkRangeArray()
    //========================================================================
    /** @see  MPO_Group_range_excl::checkRangeArray
     *        exactly like this template
     *
     *  @see  MPO_Group_range_incl::checkRangeArray
     *        exactly like this template
     */

    /** Checks whether array of ranges is valid. 
     *  Checks correctness of
     *  - first rank
     *  - last rank
     *  - stride
     *  - computed rank
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES 
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_RANK_FIRST_NEG
     *           - MARMOT_ERR_RANK_LAST_NEG
     *           - MARMOT_ERR_RANK_FIRST_TOO_BIG
     *           - MARMOT_ERR_RANK_LAST_TOO_BIG
     *           - MARMOT_ERR_STRIDE_NEG
     *           - MARMOT_ERR_STRIDE_ZERO
     *           - MARMOT_ERR_STRIDE_POS
     *           - MARMOT_ERR_RANK_COMPUTED_NEG
     *           - MARMOT_ERR_RANK_COMPUTED_TOO_BIG
     *           - MARMOT_ERR_RANK_COMPUTED_TWICE
     */
    virtual int checkRangeArray();




    // checkRank() / checkRankTooBig()
    // checkRankOfDestination() / checkRankOfDestTooBig()
    // checkRankOfSource() / checkRankOfSourceTooBig() 
    //========================================================================
    /** @see  
     */

// TODO //

    /** Checks whether rank is valid. 
     *
     *  @return  0
     */
    virtual int checkRank();

    /** Checks whether rank of source is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_RANK_NEG
     *           - MARMOT_WARN_RANK_ANY_SOURCE
     *           - MARMOT_NOTE_RANK_PROC_NULL
     */
    virtual int checkRankOfSource();

    /** Checks whether rank of destination is valid. 
     *  
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_RANK_NEG
     *           - MARMOT_ERR_RANK_ANY_SOURCE
     *           - MARMOT_NOTE_RANK_PROC_NULL
     */
    virtual int checkRankOfDestination();

    /** Checks whether rank is within size. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_RANK_TOO_BIG
     */
    virtual int checkRankTooBig();


    // checkRequest() / checkRequestArray() / checkRequestCountArray()
    //========================================================================
    /** @see  
     */

// TODO //


    /** Checks whether request is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_REQUEST_STILL_USED
     */
    virtual int checkRequest();

    /** Checks whether count is valid. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_REQUEST_COUNT_NEG
     *           - MARMOT_WARN_REQUEST_COUNT_ZERO
     */
    virtual int checkRequestCountArray();

    /** Checks whether requests are valid. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_MARMOT
     *           - MARMOT_ERR_REQUEST_NOT_VALID
     *           - MARMOT_ERR_REQUEST_NOT_PERSISTENT
     *           - MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT
     *           - MARMOT_WARN_REQUEST_NULL
     */
    virtual int checkRequestArray();


    // checkRequest() / checkRequestArray() / checkRequestCountArray()
    //========================================================================
    /** @see  
     */

// TODO //


    // checkRequestsLeft()
    //========================================================================
    /** @see  MPO_Finalize::checkRequestsLeft()
     *        exactly like this template
     */

    /** Checks whether requests are left. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_WARN_REQUESTS_LEFT
     */
    virtual int checkRequestsLeft();


    // checkRoot()
    //========================================================================
    /** @see  MPO_Collective
     *        exactly like this template
     */

    /** Checks whether roots is valid. 
     *
     *  @param   root  rank of root
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_
     *           - MARMOT_ERR_
     *           - MARMOT_WARN_
     *           - MARMOT_WARN_
     */
    virtual int checkRoot(int root);

    // checkSizeIn()
    //========================================================================
    /** @see  MPO_Unpack::checkSizeIn(const string KIND_OF_SIZE)
     *        exactly like this template
     *
     *  @see  MPO_Pack::checkSizeOut(const string KIND_OF_SIZE)
     *        exactly like this template, except for:
     *        - the name 
     */

    /** Checks whether size is valid. 
     *
     *  @param   KIND_OF_SIZE  specifies the kind of size for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_SIZE_NEG
     *           - MARMOT_WARN_SIZE_ZERO
     */
    virtual int checkSizeIn(const string KIND_OF_SIZE);


    // checkTag(const string KIND_OF_TAG)
    // checkTagOfSender() / checkTagOfReceiver()
    //========================================================================
    /** @see  MPO_Blocking::checkTag(const string KIND_OF_TAG)
     *        exactly like this template
     *
     *  @see  MPO_Probes::checkTag(const string KIND_OF_TAG)
     *        exactly like this template
     *
     *  @see MPO_Initiator::checkTag(const string KIND_OF_TAG)
     *        exactly like this template
     *
     *  @see MPO_Intercomm_create::checkTag(const string KIND_OF_TAG)
     *        exactly like this template, except for:
     *        - add MARMOT_ERR_TAG_ANY_TAG
     *
     *  @see  MPO_Sendrecvs::checkTagOfSender(const string KIND_OF_TAG)
     *        exactly like this template, except for:
     *        - the name
     *
     *  @see  MPO_Sendrecvs::checkTagOfReceiver(const string KIND_OF_TAG)
     *        exactly like this template, except for:
     *        - the name
     */

    /** Checks whether tag is valid.
     *
     *  @param   KIND_OF_TAG  specifies the kind of count for output
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_TAG_NEG
     *           - MARMOT_ERR_TAG_TOO_BIG
     *           - MARMOT_WARN_TAG_TOO_BIG
     */
    virtual int checkTag(const string KIND_OF_TAG);


    // checkToplogyNew() / checkTopologyCartNew()
    //========================================================================
    /** @see  MPO_Graph_map::checkToplogyNew()
     *        exactly like this template
     *
     *  @see  MPO_Graph_map::checkTopologyNew()
     *        exactly like this template
     *
     *  @see  MPO_Cart_create::checkTopologyCartNew()
     *        exactly like this template, 
     *
     *  @see  MPO_Cart_map::checkTopologyCartNew()
     *        exactly like this template
     */

    /** Checks whether new topology is correct. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD
     *           - MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD
     */
    virtual int checkTopologyNew();

    /** Checks whether new topology is correct. 
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD
     *           - MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD
     */
    virtual int checkTopologyCartNew();




    // checkWin()
    //========================================================================
    /** @see  MPO_Winhandler::checkWin()
     *        exactly like this template
     *
     *  @see  MPO_Win_free:checkWin()
     *        exactly like this template
     */

    /** Checks whether window is valid.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_WIN_NULL
     *           - MARMOT_ERR_WIN_NOT_VALID
     */
    virtual int checkWin();

}                               // end of namespace MPO

#endif


