/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  The namespace MPO contains the whole project.
 *
 *  Declarations of all classes that serve as base classes
 *  from which all the classes representing MPI-calls
 *  and MPI-types are derived.
 *  Utility classes like Messenger and Factory are also wrapped in this
 *  namespace.
 *
 *  @see MPO_Space.cc (implementation of MPO_Space)
 *  @see MPO_*.cc (implementation of base classes)
 *  @see mpo_Comm.cc, mpo_Group.cc, mpo_Datatype.cc (implementation)
 *  @see mpo-debugserver.cc (environment variables)
 *  @see mpo-objects.h (derived classes representing MPI-calls)
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo.h 979 2009-07-17 12:54:04Z hpcdiche $
 */

#ifndef MPO_H
#define MPO_H

#include <string>
#include <map>
#include <list>
#include <set>
#include <iostream>

#include <mpi.h>
#include "mpi_undef_makros.h"

#ifdef MARMOT_VAMPI
#include "replacepmpi.h"
#endif

#include "mpo_SourceLocation.h"
#include "mpo-messenger.h"
#include "mpo-factory.h"
#include "mpo-debugserver.h"
#include "mpo-enums.h"
#include "mpo-envvar.h"
#include "config.h"
#include "MPO_IThreadSynchronisation.h"
#include "ParseErrCodesEnv.h"

using std::string;


namespace MPO
{

    //************************************************************************
    //
    //                          Declarations
    //
    /* We use these classes in the set and map templates, that's why we
     * pre-declare them here, proper declarations follow later below.
     */
    class MPO_Callobject;
    class MPO_Comm;
    class MPO_Datatype;
    class MPO_Group;
    class MPO_Op;
    class MPO_Request;
#ifdef MARMOT_1SIDED
    class MPO_Win;

    enum MPI_WIN_MARMOT_STATES{
    	MPI_WIN_FREE_ALLOWED,
    	MPI_WIN_FREE_NOT_ALLOWED
    } ;
    typedef enum MPI_WIN_MARMOT_STATES mpi_win_marmot_states_t;


#endif
    class MPO_LogFile;
    template <class T> class MPO_ResourceGatherer;
    class MPO_IThreadSynchronisation;
// not implemented
//     class MPO_Errhandler;
// not needed here
//     class MPO_Timestamp;

    //************************************************************************
    //
    //                          MPO_Space
    //
    /** MPO_Space represents the MPI library with all resources.
     *
     *  For every resource that is shared between the object we use
     *  a static member variable in MPO_Space.
     *
     */
    class MPO_Space
    {
        friend class DebugServer;
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Space();

        /** The destructor is virtual because this is a base class.*/
        virtual ~MPO_Space();

        //========================
        // Methods
        //========================

		/**
         * Initializes and returns the syncronisation object.
         * @return pointer to the synchronisation object.
         */
        static MPO_IThreadSynchronisation* getSyncObject (void);

        /**
         * Returns a pointer to the logfile of this process.
         * @return pointer to the logfile of this process.
         */
        static MPO_LogFile* getLogFile (void);

        /**
         * Returns the thread level provided by MPI_Init_thread.
         * (MPI_Init sets the provided thread level to MPI_THREAD_SINGLE)
         * @return the threadlevel provided by this mpi implementation.
         */
		#ifdef MARMOT_THREADED
        	static int getProvidedThreadLevel (void);
		#endif

        /**
         * Returns whether MPI_Finalize was called or not.
         * @return true if MPI_Finalize was called, false otherwise.
         */
        static bool wasFinalizeCalled (void);

        /**
         * Returns whether MPI_Finalize was called or not.
         * @return true if MPI_Finalize was called, false otherwise.
         */
        static bool wasInitCalled (void);

        /**
         * Returns a string for the given thread level.
         * @param the thread level to be named.
         * @return the name for the given thread level.
         */
        static const char* getThreadLevelName(int level);

        /**
         * Returns the value of MPI_STATUS_SIZE in Fortran
         * @return the value of MPI_STATUS_SIZE in Fortran
         */
        static int getStatusSizeFortran (void);

        /**
         * Disables the DebugServer, must be called before
         * MPO_Init's constructor is called (the one with
         * arguments), otherwise the call is erroneous.
         */
        static void disableDebugServer (void);

        /**
         * Returns true if the DebugServer is enabled and
         * false otherwise.
         * @return true if the DebugServer is enabled and
         * false otherwise.
         */
        static bool isDebugServerEnabled (void);

        /**
         * Forces Marmot to use the specified logfile type.
         * Only valid if valled before MPO_Init's real
         * constructor is called.
         * @param type the desired logfile type.
         */
        static void forceLogType (MPO_LOGFILE_TYPE type);

        /**
         * Returns whether a logfile type was forced.
         * @return true if a lofgile type is forced.
         */
        static bool isLogTypeForced (void);

        /**
         * Returns the forced logfile type.
         * @return the forced type if a type is forced,
         *         LAST_LOGFILE_TYPE otherwise.
         */
        static MPO_LOGFILE_TYPE getForcedLogType (void);

	protected:
        /** Initializes factories.
         */
        static void initFactories();

        //========================
        // Requests
        //========================

        /** Map to handle requests.
         *
         *  Herewith we can check if a used request is valid or not:
         *  -  All requests are automatically inserted in
         *     our_requests.
         *  -  All newly created requests are automatically inserted in
         *     our_requests, and removed when freed.
         *  The argument KIND_OF_REQUEST marks whether request is persistent or
         *  not, etc.
         */
        static std::map < MPO_Request , enum KIND_OF_REQUEST > our_requests;

        /**
		 * Pointer to a manager object which cares about all the resource
		 * traces of requests.
		 * */
        static MPO_ResourceGatherer<MPO_Request > *our_request_traces;

        //========================
        // Datatypes
        //========================

        /** Map to handle datatypes.
         *
         *  Herewith we can check if a used datatype is valid or not:
         *  -  All predefined datatypes are automatically inserted in
         *     our_datatypes.
         *  -  All newly created datatypes are automatically inserted in
         *     our_datatypes, and removed when freed.
         *  The argument KIND_OF_DATATYPE marks whether datatype is valid or
         *  not, etc.
         */
        static std::map < MPO_Datatype, enum KIND_OF_DATATYPE > our_datatypes;

        /** Map to get datatype-names for output. */
        static std::map < MPO_Datatype, std::string > our_datatype_names;

		/**
		 * Pointer to a manager object which cares about all the resource
		 * traces of datatypes.
		 * */
        static MPO_ResourceGatherer<MPO_Datatype> *our_datatype_traces;

        //========================
        // Errhandlers
        //========================

        /** Map to handle errhandlers.
         *
         *  The int array is to contain the following information:
         *  -  In the 1st int we count the communicators the errhandler is
         *     associated with.
         *  -  In the 2nd int we mark for deallocation (1) when
         *     MPI_Errhandler_free is called.
         */
        static std::map < MPI_Errhandler, int * > our_errhandlers;

        /**
		 * Pointer to a manager object which cares about all the resource
		 * traces of errorhandlers.
		 * */
        static MPO_ResourceGatherer<MPI_Errhandler> *our_errhandler_traces;

        //========================
        // Communicators
        //========================

        /** Set to handle (known) communicators.
         *
         *  Herewith we can check if a used communicator is valid or not:
         *  -  All predefined communicators are automatically inserted
         *     in our_comms.
         *  -  All newly created communicators are inserted when created,
         *     and removed when freed.
         */
        static std::set < MPO_Comm > our_comms;

        /**
		 * Pointer to a manager object which cares about all the resource
		 * traces of communicators.
		 * */
        static MPO_ResourceGatherer<MPO_Comm> *our_comm_traces;

        //========================
        // Groups
        //========================

        /** Set to handle (known) groups.
         *
         *  Herewith we can check if a used group is valid or not:
         *  -  All predefined groups are automatically inserted in
         *     our_groups.
         *  -  All newly created groups are automatically inserted in
         *     our_groups, and removed when freed.
         */
        static std::set < MPO_Group > our_groups;

        /**
		 * Pointer to a manager object which cares about all the resource
		 * traces of groups.
		 * */
        static MPO_ResourceGatherer<MPO_Group> *our_group_traces;

        //========================
        // Operations
        //========================

        /** Set to handle (known) operators.
         *
         *  Herewith we can check if a used operator is valid or not:
         *  -  All predefined operators are automatically inserted in
         *     our_ops.
         *  -  All newly created operators are automatically inserted in
         *     our_ops, and removed when freed.
         */
        static std::set < MPO_Op > our_ops;

        /** Map to get operator-names for output. */
        static std::map < MPO_Op, std::string > our_op_names;

        /**
		 * Pointer to a manager object which cares about all the resource
		 * traces of operators.
		 * */
        static MPO_ResourceGatherer<MPO_Op> *our_op_traces;

        //========================
        // Windows
        //========================

#ifdef MARMOT_1SIDED
        /** Set to handle (known) windows.
         *
         *  Herewith we can check if a used window is valid or not:
         *  -  All newly created windows are automatically inserted in
         *     our_wins, and removed when freed.
         */
        static std::map < MPI_Win, mpi_win_marmot_states_t> our_wins;
#endif

        //========================
        // Others
        //========================

        /** To check whether there's already a buffer attached. */
        static int buffer_attached;

        /** Our MPI_COMM_WORLD (excluding the debug server) */
        static MPI_Comm our_mpi_comm_world;

        /** Our communication channel. */
        static MessengerUsingMPI our_channel;

        /** Factory for all our MPI-call-objects. */
        static Factory < int, MPO_Callobject > our_call_factory;

        /** The Loging interface. */
        static MPO_LogFile *our_log;

        /** Flag that describes whether MPI_Finalize was called. */
        static bool our_finalize_was_called;

        /** Flag that describes whether MPI_Init was called. */
        static bool our_init_was_called;

#ifdef MARMOT_THREADED
		/** The threadlevel that was required by MPI_Init. */
		static int our_threadlevel_required;
		/** The threadlevel that was provided by MPI_Init. */
		static int our_threadlevel_provided;
#endif

		/** The MPI_STATUS_SIZE value defined in FORTRAN */
        static int our_mpi_status_size;

	/**vector for the errcodes we're interested in*/
	static std::vector<int> our_errcodes;

	private:
		/** The Thread Synchronisation object we are using. */
		static MPO_IThreadSynchronisation *our_thread_sync;

		/** Selects whether the debugserver is used or not, default: used. */
		static bool our_no_debugserver;
		/** Used to force Marmot to use a certain logfile. */
		static MPO_LOGFILE_TYPE our_forced_logtype;
    };                          // end class MPO_Space

    //************************************************************************
    //
    //                          MPO_Object
    //

    /** MPO_Object represents calls, types, messages.
     *
     *  All MPO-Objects should be transferrable to and from a stream.
     *  The member functions getId(), getFamilyId(), save() and restore() are
     *  intended to provide this facility.
     *
     */
    class MPO_Object : public MPO_Space
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor. */
        MPO_Object();

        /** The destructor is virtual because this is an (abstract)
         *  base class.
         */
        virtual ~MPO_Object();

        //==========================
        // Methods
        //==========================

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const = 0;

        /** @return  the unique identity set by
         *           - enum MPO_CALL_ID
         *           - enum MPO_TYPE_ID
         *           - enum MPO_STATE_ID
         */
        virtual int getId() const = 0;

        /** Overloads output-operator << to print call-description.
         *  We distinguish between C and Fortran interface.
         *
         *  @return  output    The call description
         */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

        /** Serializes an object to ostream.
         *  We don't save and restore pointers.
         *
         *  @return  out
         */
        virtual std::ostream & save(std::ostream & out) const = 0;

        /** Deserializes an object from istream.
         *  We don't save and restore pointers.
         *
         *  @return  in
         */
        virtual std::istream & restore(std::istream & in) = 0;

    };                          // end class MPO_Object

    //==========================
    // Operators
    //==========================

    /** Overloads output-operator << for our own objects.
     *
     *  @return  rhs.debugout(output)
     */
    std::ostream & operator <<(std::ostream & output, const MPO_Object & rhs);



    //========================================================================
    // The following classes represent MPI-Types.
    //========================================================================


    //************************************************************************
    //
    //                          MPO_Typeobject
    //

    /** MPO_Typeobject represents MPI-Types.
     *
     *  For each MPI_Type MPI_foo there is one object MPO_foo.
     *  The constructor takes an argument of type  MPI_foo. These arguments are
     *  stored in private variables. Necessary conversions take place in the
     *  constructor. For ease of use a conversion function to MPI_foo exists.
     *  Therefore an object of type MPO_foo can be used, wherever an object
     *  of type MPI_foo is required. Vice versa, an argument of type MPI_foo
     *  can be passed, wherever type MPO_foo is expected.
     *
     */
    class MPO_Typeobject : public MPO_Object
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor. */
        MPO_Typeobject();

        /** The destructor is virtual because this is a base class.*/
        virtual ~MPO_Typeobject();

        //==========================
        // Methods
        //==========================

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

    };                          // end class MPO_Typeobject



    //************************************************************************
    //
    //                          MPO_Comm
    //
    /** This class represents the type MPI_Comm.
     */
    class MPO_Comm : public MPO_Typeobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm();

        /** Constructor taking proper arguments.
         *
         *  @param  communicator    The MPI communicator
         */
        MPO_Comm(MPI_Comm communicator);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Comm();

        //==========================
        // Methods
        //==========================

        /** @return  the unique type identity set by enum MPO_TYPE_ID
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

        //==========================
        // Operators
        //==========================

        /** @return  my_comm    Our communicator */
        operator  MPI_Comm() const;

    private:
        //==========================
        // Data
        //==========================

        /** Our communicator */
        MPI_Comm my_comm;

    };                          // end class MPO_Comm



    //************************************************************************
    //
    //                          MPO_Group
    //
    /** This class represents the type MPI_Group.
     */
    class MPO_Group : public MPO_Typeobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group();

        /** Constructor taking proper arguments.
         *
         *  @param  group    The MPI group
         */
        MPO_Group(MPI_Group group);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Group();

        //==========================
        // Methods
        //==========================

        /** @return  the unique type identity set by enum MPO_TYPE_ID
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

        //==========================
        // Operators
        //==========================

        /** @return my_group    Our group */
        operator  MPI_Group() const;

    private:
        //==========================
        // Data
        //==========================

        /** Our group */
        MPI_Group my_group;

    };                          // end class MPO_Group



    //************************************************************************
    //
    //                          MPO_Datatype
    //
    /** This class represents the type MPI_Datatype.
     */
    class MPO_Datatype : public MPO_Typeobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Datatype();

        /** Constructor taking proper arguments.
         *
         *  @param  datatype    The MPI datatype
         */
        MPO_Datatype(MPI_Datatype datatype);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Datatype();

        //==========================
        // Methods
        //==========================

        /** @return  the unique family identity set by enum MPO_TYPE_ID
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

        //==========================
        // Operators
        //==========================

        /** @return  my_datatype    Our datatype */
        operator  MPI_Datatype() const;

    private:
        //==========================
        // Data
        //==========================

        /** Our datatype */
        MPI_Datatype my_datatype;

    };                          // end class MPO_Datatype



    //************************************************************************
    //
    //                          MPO_Op
    //
    /** This class represents the type MPI_Op - the Operators for Reduce etc.
     */
    class MPO_Op : public MPO_Typeobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Op();

        /** Constructor taking proper arguments.
         *
         *  @param  op    The MPI operation
         */
        MPO_Op(MPI_Op op);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Op();

        //==========================
        // Methods
        //==========================

        /** @return  the unique type identity set by enum MPO_TYPE_ID
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

        //==========================
        // Operators
        //==========================

        /** @return  my_op    Our operation. */
        operator  MPI_Op() const;

    private:
        //==========================
        // Data
        //==========================

        /** Our operation */
        MPI_Op my_op;

    };                          // end class MPO_Op



    //************************************************************************
    //
    //                          MPO_Win
    //
#ifdef MARMOT_1SIDED
    /** This class represents the type MPI_Win.
     */
    class MPO_Win : public MPO_Typeobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win();

        /** Constructor taking proper arguments.
         *
         *  @param  win    The MPI window
         */
        MPO_Win(MPI_Win win);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Win();

        //==========================
        // Methods
        //==========================

        /** @return  the unique type identity set by enum MPO_TYPE_ID
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

        //==========================
        // Operators
        //==========================

        /** @return my_win    Our window */
        operator  MPI_Win() const;

    private:
        //==========================
        // Data
        //==========================

        /** Our window */
        MPI_Win my_win;

    };                          // end class MPO_Win
#endif


    //************************************************************************
    //
    //                          MPO_Request
    //
    /**
     * This class helps representing the type MPI_Request.
     * It is not used within the MARMOT callobjects as MPI uses Requests only
     * as OUT or IN-OUT arguments thus Requests will always appear as Array or
     * pointer argument.
     *
     * This Object is used for comparing Requests and in order to store
     * lists of knownn Requests in MARMOT.
     *
     * What it stores (Value or Adress) is selected with the Macro
     *   -> MARMOT_REQUESTS_AS_VALUES
     * This also influences the implementation of the compare operator.
     */
    class MPO_Request : public MPO_Typeobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Request();

        /** Constructor taking proper arguments.
         *
         *  @param  op    The MPI operation
         */
        MPO_Request(MPI_Request* request);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Request();

        //==========================
        // Methods
        //==========================

        /** @return  the unique type identity set by enum MPO_TYPE_ID
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

        //==========================
        // Operators
        //==========================

        /** Used to compare Requests.
         * @param other request to compare with.
         * @return  true if both requests are equal false otherwise.
         */
        bool operator == (MPO_Request other) const;

        /** Used to compare Requests.
         * @param other request to compare with.
         * @return true if this request != other.
         */
        bool operator < (MPO_Request other) const;

        /** Used to retrieve the encapsulated Request.
         * You should not use this method outside the appropriate Macros !!!
         * @return encapsulated request.
         */
#ifdef MARMOT_REQUESTS_AS_VALUES
        operator MPI_Request () const;
#else
        operator MPI_Request* () const;
#endif
    private:
        //==========================
        // Data
        //==========================

        /** Our Request */
#ifdef MARMOT_REQUESTS_AS_VALUES
        MPI_Request my_request;
#else
        MPI_Request *my_request;
#endif

    };                          // end class MPO_Op

    //=========================================================================
    // The following classes serve as base classes from which all the classes
    // representing MPI-calls are derived.
    //=========================================================================

    //************************************************************************
    //
    //                          MPO_Callobject
    //
    /** This abstract base class represents MPI-calls.
     *
     *  For each MPI-call MPI_Foo there is one object MPO_Foo.
     *  -  The constructor takes all arguments of the MPI Call.
     *     These arguments are stored in private variables.
     *  -  The call to the MPI_Foo is made when the member function
     *     callNativeMPI() is called.
     */
    class MPO_Callobject : public MPO_Object
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor. */
        MPO_Callobject();

        /** The destructor is virtual because this is an (abstract)
         *  base class.
         */
        virtual ~MPO_Callobject();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters etc.
         * Executed before the actual MPI call is issued.*/
        virtual int checkPrerequisites() = 0;

        /** Checks correctness of parameters etc.
         * Executed after the actual MPI call.*/
        virtual int checkPost() = 0;

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Executes pre and post checks/monitoring and
         * calls the actuall MPI call (PMPI).*/
        int callMPIAndCheck(void);

        /** Handles communication with the debug process.
         *
         *  @return  callNativeMPI()
         */
        virtual int callDebugServer();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

        /** Serializes an object to ostream.
         *  We don't save and restore pointers.
         *  @param out the ostream to store the data in.
         *  @return  out
         */
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.
         *  We don't save and restore pointers.
         *  @param in the ostream to store the data in.
         *  @return  in
         */
        virtual std::istream & restore(std::istream & in);

        /** sets the data for my_SourceLine and my_SourceFile.
         * @param line the code line of the MPI_Call.
         * @param fileName the name of the sourcefile of the callee.
         */
        void setSourceLocation (int line,char *fileName);

        /** sets the data for my_SourceLine and my_SourceFile.
         * @param line the code line of the MPI_Call.
         * @param fileName the name of the sourcefile of the callee.
         * @param extra_id an ID associated with the MPI call of this location.
         */
        void setSourceLocation (int line,char *fileName, unsigned long long extra_id);

        /** sets the data for my_SourceLine and my_SourceFile.
         * @param line the code line of the MPI_Call.
         * @param fileName the name of the sourcefile of the callee.
         */
        void setSourceLocation (int line,const char *fileName);

        /** sets the data for my_SourceLine and my_SourceFile.
         * @param line the code line of the MPI_Call.
         * @param fileName the name of the sourcefile of the callee.
         * @param extra_id an ID associated with the MPI call of this location.
         */
        void setSourceLocation (int line,const char *fileName, unsigned long long extra_id);

        /** returns a MPO_SourceLocation object representing the location data of this call.
         * @return the object representing the sourcelocation of the call.
         */
        MPO_SourceLocation getSourceLocation (void);

		/**
		 * Returns the rank of the thread associated with this callobject.
		 * @return rank of thread (id) associated with this callobject.
		 */
		int getThreadRank (void);

		/**
		 * This method must be called by all MPI calls that use a send or recieve
		 * buffer in an asynchronous MPI call that uses a request.
		 * @param buf the buffer used.
		 * @param count the amount of elements of type "type" used for the communication.
		 * @param type the MPI_Datatype used for the communication.
		 * @param the request that was generated by the asynchronous call.
		 */
		void registerBuffer (void *buf, int count, MPO_Datatype type, MPO_Request request);

		/**
		 * This call is used to remove a buffer from the registration. That must be done
		 * after the communication has finished or was aborted.
		 * @param the request associated with the finished or aborted communication.
		 */
		void unregisterBuffer (MPO_Request request);

		/**
		 * This call is used to verify whether a buffer or a part of it is already in use.
		 * This method should be called by all communication calls that use one or multiple
		 * send/recieve buffers.
		 * @param buf the buffer used.
         * @param count the amount of elements of type "type" used for the communication.
         * @param type the MPI_Datatype used for the communication.
         * @return MARMOT_SUCCESS or a code for an error/warning raised.
		 */
		int checkBufferInUse (void *buf, int count, MPO_Datatype type);

		/**
		 * Sets whether pre MPI call data is restored,
		 * use case is for UniMCI support, were the pre and post
		 * checks are executed by different instances of the
		 * MPO_Callobject associated with the MPI call.
		 * Data that is stored in the pre checks and will be
		 * needed in the post checks has to be stored in that
		 * case.
		 * @param true/false.
		 */
		static void SetPreDataRestore (bool on);

		/**
		 * If the MPO_Callobject implementation needs to
		 * store/restore pre MPI call data, it has to call
		 * this function in its constructor.
		 * @return true if data was restored, false otherwise.
		 */
		bool handlePreDataConstructor (void);

		/**
		 * If the MPO_Callobject implementation needs to
		 * store/restore pre MPI call data, it has to call
		 * this function in its destructor.
		 * @return true if data was saved, false otherwise.
		 */
		bool handlePreDataDestructor (void);

		/**
		 * Restores pre MPI call data from a MPO_Callobject
		 * created with clonePreData of the same call.
		 * Warning:
		 *  if you copy pointers in obj into the current
		 *  MPO_Callobject, be sure to not deallocate these
		 *  pointers in the destructor of obj !
		 * @param obj the object containing the pre data.
		 */
		virtual void setPreData (MPO_Callobject *obj);

		/**
		 * Used to create an MPO_Callobject that stores
		 * pre call data for a post call.
		 * @return an MPO_Callobject holding the data.
		 */
		virtual MPO_Callobject* clonePreData (void);

    protected:

        //==========================
        // Data
        //==========================
        int my_SourceLine; /**< Speciefies the code line of the call in Application space. */
        unsigned long long my_extra_id;
        char *my_SourceFile; /**< Speciefies the source file of the call in Application space. */
        int my_thread_rank; /**< Speciefies the rank of the thread (used to pass thread information to the debugserver). */

        /**
         * Associated with SetPreDataRestore.
         */
        static bool our_store_pre_object_data;

        /**
         * This map is used for UniMCI support to store
         * data of MPO objects for the post check.
         * Details:
         *   For UniMCI an MPO object is created for the
         *   pre and for the post checks, data from the
         *   pre object will be lost when the post object
         *   is created. In some cases it will be necessary
         *   to store data, e.g., for MPI_Waitsome the request
         *   array has to be stored before the MPI call is issued
         *   from this array the completed requests have to be
         *   removed from the Marmot registries.
         */
        static std::map<int, MPO_Callobject*> our_pre_object_data;

        /**
         * Stores whether data of a pre MPI call object was
         * restored in the constructor. If so, this is a post
         * MPI call object. Must only be used if
         * our_store_pre_object_data is set to true.
         */
        bool my_persitent_restored;

        /**
         * Types and map for information about the used buffers.
         */
        typedef struct {
            MPI_Datatype type;
            MPI_Aint start_add;
            MPI_Aint end_add;
            int count;
            MPO_SourceLocation location;
            string callname;
        } MPO_BufferInfoType;
        static std::map<MPO_Request,MPO_BufferInfoType> our_used_buffers;


#ifdef MARMOT_USE_THREADSYNC
        /*
         * Datatype necessary for storing succesful probes.
         */
        typedef struct
        {
        	int 		source;
        	int 		tag;
        	MPI_Comm 	comm;
        	int 		sourceLine;
        	std::string sourceName;
        	int 		threadNum;
        	int 		callId;
        	std::string callName;

#ifdef MARMOT_TCHECK
        	//for artifical races we add an index into an array of conflict variables
        	int         index;
#endif /*MARMOT_TCHECK*/
        } MPO_ProbeInfo;

        /**
         * List that stores source,tag,communicator for messages that were
         * probed but not yet recieved.
         */
        static std::list<MPO_ProbeInfo> our_succesful_probes;

#ifdef MARMOT_TCHECK
        static int our_next_probe_array_index;
#endif /*MARMOT_TCHECK*/

#endif /*MARMOT_USE_THREADSYNC*/

    private:
        char *my_SourceFileMem; /**< Used after restore operations to create Memory for storing a passed source filename. */

    };                          // end class MPO_Callobject



    //************************************************************************
    //
    //                          MPO_Environment
    //
    /** This abstract base class represents calls concerning MPI itself.
     *
     */
    class MPO_Environment : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor. */
        MPO_Environment();

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Environment();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters etc. */
        virtual int checkPrerequisites() = 0;

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    };                          // end class MPO_Environment



    //************************************************************************
    //
    //                          MPO_Pt2Pt
    //
    /** This abstract base class represents pt2pt-calls.
     */
    class MPO_Pt2Pt : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor. */
        MPO_Pt2Pt();

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Pt2Pt();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters etc. */
        virtual int checkPrerequisites() = 0;

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:

     	#ifdef MARMOT_USE_THREADSYNC

	    	/**
	    	 * This call is used to detect race conditions for revieve calls and probe calls.
	    	 * Itl tests whether the given source, tag and comm matches a successfully
	    	 * performed probe. If this Thread is distinct from the probing thread a warning is issued.
	    	 * All probed messages that match this recv call are unregistered.
	    	 * @param source the source rank of this recieve call.
			 * @param tag the tag used in this recieve call.
			 * @param comm the communicator used in this recieve call.
			 * @param sourceLine the source file line of this recieve call.
			 * @param sourceName the source file name of this recieve call.
			 * @param threadNum the thread that perfoms this recieve call.
			 * @param callId the id of the call that issued the recieve (used to distinc MPI_Recv,MPI_Irecv,...).
			 * @param callName the name of the call that issued this recieve (used to distinc MPI_Recv,MPI_Irecv,...).
	    	 */
	        int testIfProbed (int source, int tag, MPI_Comm comm, int sourceLine, std::string sourceName, int threadNum, int callId, std::string callName);

	    #endif /*MARMOT_USE_THREADSYNC*/

    };                          // end class MPO_Pt2Pt



    //************************************************************************
    //
    //                          MPO_Blocking
    //
    /** This abstract base class represents blocking calls.
     *
     *  All blocking calls are derived from this one. Checking
     *  common prerequisites (correctness of tags and ranks, datatypes
     *  and communicators) for these calls is implemented as member
     *  functions here.
     *
     *  @see MPO_Blocking.cc (implementation)
     */
    class MPO_Blocking : public MPO_Pt2Pt
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Blocking();

        /** Constructor taking proper arguments.
         *
         *  @param  count      maximum number of elements in send/receive
         *                     buffer
         *  @param  datatype   datatype of each receive buffer element
         *  @param  rank       rank of source/destination
         *  @param  tag        message tag
         *  @param  comm       communicator
         */
        MPO_Blocking(int count, MPO_Datatype datatype, int rank, int tag,
                     MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Blocking();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - count
         *  - datatype
         *  - rank
         *  - tag
         *  - communicator
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCount(const string KIND_OF_COUNT);

        /** Checks whether datatype is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_TYPE_NOT_VALID
         *           - MARMOT_ERR_TYPE_NULL
         *           - MARMOT_ERR_TYPE_NOT_COMMITTED
         *           - MARMOT_ERR_TYPE_C
         *           - MARMOT_ERR_TYPE_F
         *           - MARMOT_ERR_TYPE_REDUCTION_C
         *           - MARMOT_ERR_TYPE_REDUCTION_F
         *           - MARMOT_ERR_TYPE_OPTIONAL_C
         *           - MARMOT_ERR_TYPE_OPTIONAL_F
         *           - MARMOT_ERR_TYPE_ERR_TYPE_CONSTR_DERIVED
         *           - MARMOT_WARN_TYPE_REDUCTION_C
         *           - MARMOT_WARN_TYPE_REDUCTION_F
         *           - MARMOT_WARN_TYPE_OPTIONAL_C
         *           - MARMOT_WARN_TYPE_OPTIONAL_F
         */
        virtual int checkDatatype();

        /** Checks whether rank is valid.
         *
         *  @return  0
         */
        virtual int checkRank();

        /** Checks whether rank of source is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_WARN_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRankOfSource();

        /** Checks whether rank of destination is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRankOfDestination();

        /** Checks whether rank is within size.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRankTooBig();

        /** Checks whether tag is valid.
         *
         *  @param   KIND_OF_TAG  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TAG_NEG
         *           - MARMOT_ERR_TAG_TOO_BIG
         *           - MARMOT_WARN_TAG_TOO_BIG
         */
        virtual int checkTag(const string KIND_OF_TAG);

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const = 0;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in) = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        int my_count;
        MPO_Datatype my_datatype;
        int my_rank;
        int my_tag;
        MPO_Comm my_comm;

    };                          // end class MPO_Blocking



    //************************************************************************
    //
    //                          MPO_Sendrecvs
    //
    /** This abstract base class represents Sendrecv and Sendrecv_replace.
     *
     *  Sendrecv and Sendrecv_replace are derived from this. Checking
     *  common pre-requisites (correctness of tags and ranks, datatypes, count
     *  and communicators) for these two calls is implemented as a member
     *  function here.
     */
    class MPO_Sendrecvs : public MPO_Pt2Pt
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Sendrecvs();

        /** Constructor taking proper arguments.
         *
         *  @param  count      number of elements in send/receive buffer
         *  @param  datatype   type of elements in send/receive buffer
         *  @param  dest       rank of destination
         *  @param  sendtag    send tag
         *  @param  source     rank of source
         *  @param  recvtag    receive tag
         *  @param  comm       communicator
         */
        MPO_Sendrecvs(int count, MPO_Datatype datatype, int dest,
                      int sendtag,
                      int source, int recvtag, MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Sendrecvs();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - count,
         *  - datatype,
         *  - communicator,
         *  - tag,
         *  - ranks
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCount(const string KIND_OF_COUNT);

        /** Checks whether send datatype is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_TYPE_NOT_VALID
         *           - MARMOT_ERR_TYPE_NULL
         *           - MARMOT_ERR_TYPE_NOT_COMMITTED
         *           - MARMOT_ERR_TYPE_C
         *           - MARMOT_ERR_TYPE_F
         *           - MARMOT_ERR_TYPE_REDUCTION_C
         *           - MARMOT_ERR_TYPE_REDUCTION_F
         *           - MARMOT_ERR_TYPE_OPTIONAL_C
         *           - MARMOT_ERR_TYPE_OPTIONAL_F
         *           - MARMOT_ERR_TYPE_ERR_TYPE_CONSTR_DERIVED
         *           - MARMOT_WARN_TYPE_REDUCTION_C
         *           - MARMOT_WARN_TYPE_REDUCTION_F
         *           - MARMOT_WARN_TYPE_OPTIONAL_C
         *           - MARMOT_WARN_TYPE_OPTIONAL_F
         */
        virtual int checkDatatypeOfSender();

        /** Checks whether ranks of source and destination are valid.
         *
         *  @return  0
         */
        virtual int checkRankSendRecv();

        /** Checks whether rank of source is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_WARN_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRankOfSource();

        /** Checks whether rank of destination is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRankOfDestination();

        /** Checks whether rank of source is within size.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRankOfSourceTooBig();

        /** Checks whether rank of destination is within size.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRankOfDestTooBig();

        /** Checks whether send tag is valid.
         *
         *  @param   KIND_OF_TAG  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TAG_NEG
         *           - MARMOT_ERR_TAG_TOO_BIG
         *           - MARMOT_WARN_TAG_TOO_BIG
         */
        virtual int checkTagOfSender(const string KIND_OF_TAG);

        /** Checks whether receive tag is valid.
         *
         *  @param   KIND_OF_TAG  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TAG_NEG
         *           - MARMOT_ERR_TAG_TOO_BIG
         *           - MARMOT_WARN_TAG_TOO_BIG
         */
        virtual int checkTagOfReceiver(const string KIND_OF_TAG);

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** our arguments */
        int my_count;
        MPO_Datatype my_datatype;
        int my_dest;
        int my_sendtag;
        int my_source;
        int my_recvtag;
        MPO_Comm my_comm;

    };                          // end class MPO_Sendrecvs



    //************************************************************************
    //
    //                          MPO_Requestcall
    //
    /** This abstract base class represents calls using requests.
     *
     *  All calls using requests are derived from this class.
     */
    class MPO_Requestcall : public MPO_Pt2Pt
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor. */
        MPO_Requestcall();

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Requestcall();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters etc. */
        virtual int checkPrerequisites() = 0;

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    };                          // end class MPO_Requestcall



    //************************************************************************
    //
    //                          MPO_Initiator
    //
    /** This abstract base class represents calls that initiate
     *  communication using requests.
     *
     *  All calls initiating communication with requests are derived from
     *  this. Call to check prerequisites (correctness of rank, communicator,
     *  count, datatype, tag and request) is implemented as a member function
     *  here.
     */
    class MPO_Initiator : public MPO_Requestcall
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Initiator();

        /** Constructor taking proper arguments.
         *
         *  @param  count      number of elements in send/receive buffer
         *  @param  datatype   type of elements in send/receive buffer
         *  @param  rank       rank of source/destination
         *  @param  tag        message tag
         *  @param  comm       communicator
         *  @param  request    communication request
         *  @param  persistent unique enumeration value to describe
         *                     kind of request
         */
        MPO_Initiator(int count,
                      MPO_Datatype datatype, int rank, int tag,
                      MPO_Comm comm, MPI_Request * request,
                      enum KIND_OF_REQUEST persistent);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Initiator();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Currently no checks implemented.
         *  Implementation is done in checkPrerequisites(enum MPO_CALL_ID id)
         *
         *  @return 0
         */
        virtual int checkPrerequisites();

        /** Checks prerequisites such as correctness of parameters.
         *  Checks correctness of
         *  - count,
         *  - datatype,
         *  - rank,
         *  - tag,
         *  - comm,
         *  - request
         *
         *  @param  id    unique call identity
         *  @return 0
         */
        virtual int checkPrerequisites(enum MPO_CALL_ID id);

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCount(const string KIND_OF_COUNT);

        /** Checks whether datatype is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_TYPE_NOT_VALID
         *           - MARMOT_ERR_TYPE_NULL
         *           - MARMOT_ERR_TYPE_NOT_COMMITTED
         *           - MARMOT_ERR_TYPE_C
         *           - MARMOT_ERR_TYPE_F
         *           - MARMOT_ERR_TYPE_REDUCTION_C
         *           - MARMOT_ERR_TYPE_REDUCTION_F
         *           - MARMOT_ERR_TYPE_OPTIONAL_C
         *           - MARMOT_ERR_TYPE_OPTIONAL_F
         *           - MARMOT_ERR_TYPE_ERR_TYPE_CONSTR_DERIVED
         *           - MARMOT_WARN_TYPE_REDUCTION_C
         *           - MARMOT_WARN_TYPE_REDUCTION_F
         *           - MARMOT_WARN_TYPE_OPTIONAL_C
         *           - MARMOT_WARN_TYPE_OPTIONAL_F
         */
        virtual int checkDatatype();

        /** Checks whether ranks of source and destination are valid.
         *
         *  @param   id    unique call identity
         *  @return  0
         */
        virtual int checkRank(enum MPO_CALL_ID id);

        /** Checks whether rank of source is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_WARN_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRankOfSource();

        /** Checks whether rank of destination is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRankOfDestination();

        /** Checks whether rank of source is within size.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRankOfSourceTooBig();

        /** Checks whether rank of destination is within size.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRankOfDestTooBig();

        /** Checks whether tag is valid.
         *
         *  @param   KIND_OF_TAG  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TAG_NEG
         *           - MARMOT_ERR_TAG_TOO_BIG
         *           - MARMOT_WARN_TAG_TOO_BIG
         */
        virtual int checkTag(const string KIND_OF_TAG);

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether request is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_REQUEST_STILL_USED
         */
        virtual int checkRequest();

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        int my_count;
        MPO_Datatype my_datatype;
        int my_rank;
        int my_tag;
        MPO_Comm my_comm;
        MPI_Request * my_request_p;
        enum KIND_OF_REQUEST my_persistent;

    };                          // end class MPO_Initiator



    //************************************************************************
    //
    //                          MPO_Starts
    //
    /** This abstract base class represents Startall and Start.
     *
     *  Startall and Start are derived from this class. The call to check
     *  prerequisites (correct use of request) is implemented here.
     */
    class MPO_Starts : public MPO_Requestcall
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Starts();

        /** Constructor taking proper arguments.
         *
         *  @param  count    list length
         *  @param  array    array of communication requests
         */
        MPO_Starts(int count, MPI_Request * array);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Starts();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - count of requests
         *  - requests
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_REQUEST_COUNT_NEG
         *           - MARMOT_WARN_REQUEST_COUNT_ZERO
         */
        virtual int checkRequestCountArray();

        /** Checks whether requests are valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_REQUEST_NOT_VALID
         *           - MARMOT_ERR_REQUEST_NOT_PERSISTENT
         *           - MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT
         *           - MARMOT_WARN_REQUEST_NULL
         */
        virtual int checkRequestArray();

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        int my_count;
        MPI_Request *my_request_array;

    };                          // end class MPO_Starts



    //************************************************************************
    //
    //                          MPO_Completion
    //
    /** This abstract base class represents calls that complete
     *  communication using requests.
     *
     *  Calls completing communication with requests are derived from this
     *  class.
     *  Checking pre-requisites (correctness of count and request-use) is
     *  implemented as member function here. To free the request for further
     *  use, the request will be removed from the map our_requests after doing
     *  the MPI-call.
     */
    class MPO_Completion : public MPO_Requestcall
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Completion();

        /** Constructor taking proper arguments.
         *
         *  @param  count    list length
         *  @param  array    array of communication requests
         */
        MPO_Completion(int count, MPI_Request * array);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Completion();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - number of requests
         *  - requests
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_REQUEST_COUNT_NEG
         *           - MARMOT_WARN_REQUEST_COUNT_ZERO
         */
        virtual int checkRequestCountArray();

        /** Checks whether requests are valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_REQUEST_NOT_VALID
         *           - MARMOT_WARN_REQUEST_INACTIVE_PERSISTENT
         *           - MARMOT_WARN_REQUEST_NULL
         */
        virtual int checkRequestArray();

        /** Checks whether request is to remove from our_requests.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_WARN_REQUEST_OUTCOUNT_UNDEFINED
         */
        virtual int eraseRequest(int *my_outcount, int *my_index_array, MPI_Request *request_array);

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        int my_count;
        MPI_Request *my_request_array;

    };                          // end class MPO_Completion



    //************************************************************************
    //
    //                          MPO_Tests
    //
    /** This abstract base class represents calls that test communication
     *  using requests.
     *
     *  All calls testing communication with requests are derived from this.
     *  Checking correctness of count and request(s) is implemented as
     *  member function here. If communication is completed, request will be
     *  removed from our_requests.
     *
     *  @see  MPO_Tests.cc (implementation)
     */
    class MPO_Tests : public MPO_Requestcall
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Tests();

        /** Constructor taking proper arguments.
         *  @param  count    list length
         *  @param  array    array of communication requests
         *  @param  flags    array of flags (true if operation completed)
         */
        MPO_Tests(int count, MPI_Request * array, int *flags);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Tests();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - count of requests
         *  - requests
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_REQUEST_COUNT_NEG
         *           - MARMOT_WARN_REQUEST_COUNT_ZERO
         */
        virtual int checkRequestCountArray();

        /** Checks whether requests are valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_REQUEST_NOT_VALID
         *           - MARMOT_WARN_REQUEST_INACTIVE_PERSISTENT
         *           - MARMOT_WARN_REQUEST_NULL
         */
        virtual int checkRequestArray();

        /** Checks whether request is to remove from our_requests.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_REQUEST_OUTCOUNT_UNDEFINED
         */
        virtual int eraseRequest(int *my_outcount, int *my_index_array, MPI_Request *request_array);

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        int my_count;
        MPI_Request *my_request_array;
        int *my_flags;

    };                          // end class MPO_Tests



    //************************************************************************
    //
    //                          MPO_Collective
    //
    /** This abstract base class represents collective calls.
     *
     *  All calls concerning collective communication are derived from this.
     *  Checking the parameters ist implemented as several member functions.
     *
     *  @see  MPO_Collective.cc (implementation)
     */
    class MPO_Collective : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Collective();

        /** Constructor taking proper arguments.
         *
         *  @param  comm    communicator
         */
        MPO_Collective(MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Collective();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - communicator
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether count is valid.
         *
         *  @param   my_count       send/receive count
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCount(int my_count, const string KIND_OF_COUNT);

        /** Checks whether count is valid.
         *
         *  @param   my_counts      array of send/receive counts
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCountArray(int* my_counts,
                                    const string KIND_OF_COUNT);

        /** Checks whether datatype is valid.
         *
         *  @param   my_count         number of elements in send/receive buffer
         *  @param   my_datatype      data type of elements of send/receive
         *                            buffer
         *  @param   kind_of_count    kind of count argument (count,
         *                            send count or receive count)
         *  @param   kind_of_coll     kind of collective call
         *                            (Reduce/Allreduce or not)
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TYPE_NOT_VALID
         *           - MARMOT_ERR_TYPE_NULL
         *           - MARMOT_ERR_TYPE_NOT_COMMITTED
         *           - MARMOT_ERR_TYPE_C
         *           - MARMOT_ERR_TYPE_F
         *           - MARMOT_ERR_TYPE_REDUCTION_C
         *           - MARMOT_ERR_TYPE_REDUCTION_F
         *           - MARMOT_ERR_TYPE_OPTIONAL_C
         *           - MARMOT_ERR_TYPE_OPTIONAL_F
         *           - MARMOT_ERR_TYPE_ERR_TYPE_CONSTR_DERIVED
         *           - MARMOT_WARN_TYPE_REDUCTION_C
         *           - MARMOT_WARN_TYPE_REDUCTION_F
         *           - MARMOT_WARN_TYPE_OPTIONAL_C
         *           - MARMOT_WARN_TYPE_OPTIONAL_F
         */
        virtual int checkDatatypeSendRecv(int my_count, MPO_Datatype my_datatype,
                                      enum KIND_OF_COUNT kind_of_count,
                                      enum KIND_OF_COLLECTIVE_CALL
                                        kind_of_coll);

        /** Checks whether root rank is valid.
         *
         *  @param   root    rank of root
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRoot(int root);

        /** Checks whether operation is valid.
         *
         *  @param   op   reduce operation
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_OP_NULL
         *           - MARMOT_ERR_OP_NOT_VALID
         */
        virtual int checkOp(MPO_Op op);

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        MPO_Comm my_comm;

    };                          // end class MPO_Collective



    //************************************************************************
    //
    //                          MPO_Communicatorhandler
    //
    /**  This abstract base class represents calls handling communicators.
     *
     *  All calls for communicator-handling are derived from this.
     *  Correctnes of used communicator is checked in implemented
     *  member function.
     */
    class MPO_Communicatorhandler : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Communicatorhandler();

        /** Constructor taking proper arguments.
         *
         *  @param  comm    communicator
         */
        MPO_Communicatorhandler(MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Communicatorhandler();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - comm
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;

    };                          // end class MPO_Communicatorhandler



    //************************************************************************
    //
    //                          MPO_Cartesian
    //
    /** This abstract base class represents calls concerning
     *  cartesian coordiantes.
     *
     *  All calls concerning cartesian coordinates are derived from this.
     *  Correctness of communicator is checked in implemented member function.
     *  In MPO_Cart_create we have to check different things
     *  in checkPrerequisites than in the other calls...
     *
     *  @see  MPO_Cartesian.cc (implementation)
     */
    class MPO_Cartesian : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cartesian();

        /** Constructor taking proper arguments.
         *
         *  @param  comm    communicator
         */
        MPO_Cartesian(MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Cartesian();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - comm is valid
         *  - comm has cartesian structure.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether communicator has cartesian structure.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NOT_CARTESIAN
         */
        virtual int checkCommCartesian();

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our Arguments */
        MPO_Comm my_comm;

    };                          // end class MPO_Cartesian



    //************************************************************************
    //
    //                          MPO_Graph
    //
    /** This abstract base class represents calls concerning graph coordinates.
     *
     *  All calls concerning graph coordinates are derived from this.
     *  Correctness of communicator is checked in implemented member function.
     *  In MPO_Graph_create we have to check different things
     *  in checkPrerequisites than in the other calls...
     */
    class MPO_Graph : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Graph();

        /** Constructor taking proper arguments.
         *
         *  @param  comm    communicator
         */
        MPO_Graph(MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Graph();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - comm is valid
         *  - comm has graph structure.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether communicator has graph structure.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NOT_GRAPH
         */
        virtual int checkCommGraph();

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        MPO_Comm my_comm;

    };                          // end class MPO_Graph



    //************************************************************************
    //
    //                          MPO_Intercommunicators
    //
    /** This abstract base class represents inter-communicator-handlers.
     *
     *  All calls handling inter-communicators are derived from this.
     *  Correctness of communicator is checked in implemented member function.
     *
     *  @see  MPO_Intercommunicators.cc (implementation)
     */
    class MPO_Intercommunicators : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Intercommunicators();

        /** Constructor taking proper arguments.
         *
         *  @param  comm    communicator
         */
        MPO_Intercommunicators(MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Intercommunicators();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks if comm is valid inter-communicator.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether communicator has graph structure.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NOT_INTERCOMM
         */
        virtual int checkCommIntercomm();

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        MPO_Comm my_comm;

    };                          // end class MPO_Intercommunicators



    //************************************************************************
    //
    //                          MPO_Probes
    //
    /** This abstract base class represents MPO_Probe and MPO_Iprobe.
     *
     *  MPO_Probe and MPO_Iprobe are derived from this class. Checking
     *  correctness of ranks, tag and communicator is implemented as
     *  member function.
     */
    class MPO_Probes : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Probes();

        /** Constructor taking proper arguments.
         *
         *  @param   source    source rank, or 'MPI_ANY_SOURCE'
         *  @param   tag       tag value or 'MPI_ANY_TAG'
         *  @param   comm      communicator
         *
         *  @return 0
         */
        MPO_Probes(int source, int tag, MPO_Comm comm);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Probes();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - rank of source
         *  - tag
         *  - communicator
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether rank of source is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_WARN_RANK_ANY_SOURCE
         *           - MARMOT_NOTE_RANK_PROC_NULL
         */
        virtual int checkRankOfSource();

        /** Checks whether rank of source is within size.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRankOfSourceTooBig();

        /** Checks whether tag is valid.
         *
         *  @param   KIND_OF_TAG  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TAG_NEG
         *           - MARMOT_ERR_TAG_TOO_BIG
         *           - MARMOT_WARN_TAG_TOO_BIG
         */
        virtual int checkTag(const string KIND_OF_TAG);

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

        #ifdef MARMOT_USE_THREADSYNC
	        /*
	         * This call registers a message that was probed for but is not yet recieved.
	         * The message will be unregistered when a matching recv call is performed.
	         * @param source the source process of the message.
	         * @param tag the tag of the massage.
	         * @param comm the communicator used for the message.
	         * @param sourceLine the source line of the probe call.
	         * @param sourceName the source file name of the probe call.
	         * @param threadNum the thread id that performed the probe call.
	         * @param callId the id of the probe call (used to destinct between MPI_Probe and MPI_Iprobe).
	         * @param callName the name of the probe call (used to destinct between MPI_Probe and MPI_Iprobe).
	         */
	        void registerProbe (int source, int tag, MPI_Comm Communicator, int sourceLine, std::string sourceName, int threadNum, int callId, std::string callName);

	    #endif /*MARMOT_USE_THREADSYNC*/

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        int my_source;
        int my_tag;
        MPO_Comm my_comm;

    };                          // end class MPO_Probes



    //************************************************************************
    //
    //                          MPO_Groupconstructor
    //
    /** This abstract base class represents calls that construct new groups.
     *
     *  All calls constructing new groups are derived from this class.
     *  Checking prerequisites is implemented as member function, and
     *  new groups are added to our_groups.
     */
    class MPO_Groupconstructor : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Groupconstructor();

        /** Constructor taking proper arguments.
         *
         *  @param   group      first group
         *  @param   group2     second group
         *  @param   newgroup   new group
         *  @param   groupcount number of groups
         */
        MPO_Groupconstructor(MPO_Group group, MPO_Group group2,
                             MPI_Group * newgroup, int groupcount);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Groupconstructor();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks correctness of
         *  - group 1
         *  - group 2
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether group is valid.
         *
         *  @param   valid
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup(int valid);

        /** Checks whether group2 is valid.
         *
         *  @param   valid
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup2(int valid);

// TODO // valid???

        /** Adds new group to our set.
         *
         *  @return 0
         */
        virtual int insertGroup();

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        MPO_Group my_group;
        MPO_Group my_group2;
        MPI_Group *my_newgroup;
        int my_groupcount;

    };                          // end class MPO_Groupconstructor



    //************************************************************************
    //
    //                          MPO_Grouphandler
    //
    /** This abstract base class represents calls that access groups.
     *
     *  All calls handling groups are derived from this class. Checking
     *  correctness of pre-requisites are implemented as member function.
     */
    class MPO_Grouphandler : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Grouphandler();

        /** Constructor taking proper arguments.
         *
         *  @param   group      first group
         *  @param   group2     second group
         *  @param   groupcount number of groups
         */
        MPO_Grouphandler(MPO_Group group, MPO_Group group2, int groupcount);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Grouphandler();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks correctness of
         *  - group 1
         *  - group 2
         *
         *  return 0
         */
        virtual int checkPrerequisites();

        /** Checks whether group is valid.
         *
         *  @param   valid
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup(int valid);

        /** Checks whether group2 is valid.
         *
         *  @param   valid
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup2(int valid);

// TODO // valid???

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        MPO_Group my_group;
        MPO_Group my_group2;
        int my_groupcount;

    };                          // end class MPO_Grouphandler



    //************************************************************************
    //
    //                          MPO_Typeconstructors
    //
    /** This abstract base class represents type constructing calls.
     *
     *  All calls constructing new datatypes are derived from this class.
     *  Checking correctness of prerequisites are implemented as member
     *  function.
     */
    class MPO_Typeconstructors : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Typeconstructors();

        /** Constructor taking proper arguments.
         *
         *  @param  count    number of blocks
         *                   also number of entries in arrays array_of_types
         *  @param  oldtype  type of elements in each block
         *  @param  id       unique call identity
         */
        MPO_Typeconstructors(int count, MPO_Datatype oldtype, int id);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Typeconstructors();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - count,
         *  - datatype
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCount(const string KIND_OF_COUNT);

        /** Checks whether datatype is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_TYPE_NOT_VALID
         *           - MARMOT_ERR_TYPE_NULL
         *           - MARMOT_ERR_TYPE_NOT_COMMITTED
         *           - MARMOT_ERR_TYPE_C
         *           - MARMOT_ERR_TYPE_F
         *           - MARMOT_ERR_TYPE_REDUCTION_C
         *           - MARMOT_ERR_TYPE_REDUCTION_F
         *           - MARMOT_ERR_TYPE_OPTIONAL_C
         *           - MARMOT_ERR_TYPE_OPTIONAL_F
         *           - MARMOT_WARN_TYPE_REDUCTION_C
         *           - MARMOT_WARN_TYPE_REDUCTION_F
         *           - MARMOT_WARN_TYPE_OPTIONAL_C
         *           - MARMOT_WARN_TYPE_OPTIONAL_F
         */
        virtual int checkDatatype();

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        int my_count;
        MPO_Datatype my_oldtype;
        int my_id;

    };                          // end class MPO_Typeconstructors



    //************************************************************************
    //
    //                          MPO_Typehandlers
    //
    /** This abstract base class represents type handling calls.
     *
     *  All calls handling datatypes are derived from this class.
     *  Checking correctness of handelled datatype is implemented as
     *  member function.
     */
    class MPO_Typehandlers : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Typehandlers();

        /** Constructor taking proper arguments.
         *
         *  @param  datatype    datatype
         */
        MPO_Typehandlers(MPO_Datatype datatype);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Typehandlers();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether datatypes are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether datatype is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_TYPE_NOT_VALID
         *           - MARMOT_ERR_TYPE_NULL
         *           - MARMOT_ERR_TYPE_NOT_COMMITTED
         *           - MARMOT_ERR_TYPE_C
         *           - MARMOT_ERR_TYPE_F
         *           - MARMOT_ERR_TYPE_REDUCTION_C
         *           - MARMOT_ERR_TYPE_REDUCTION_F
         *           - MARMOT_ERR_TYPE_OPTIONAL_C
         *           - MARMOT_ERR_TYPE_OPTIONAL_F
         *           - MARMOT_ERR_TYPE_ERR_TYPE_CONSTR_DERIVED
         *           - MARMOT_WARN_TYPE_REDUCTION_C
         *           - MARMOT_WARN_TYPE_REDUCTION_F
         *           - MARMOT_WARN_TYPE_OPTIONAL_C
         *           - MARMOT_WARN_TYPE_OPTIONAL_F
         */
        virtual int checkDatatype();

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        /** Our arguments */
        MPO_Datatype my_datatype;

    };                          // end class MPO_Typehandlers



    //************************************************************************
    //
    //                          MPO_Errhandlers
    //
    /** This abstract base class represents error-handling calls.
     *
     *  All calls concerning error-handling are derived from this.
     */
    class MPO_Errhandlers : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor. */
        MPO_Errhandlers();

        /** The default constructor initializes an invalid object. */
        //MPO_Errhandlers();

        /** Constructor taking proper arguments. */
        //MPO_Errhandlers(MPO_Errhandler errhandler);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Errhandlers();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters etc. */
        virtual int checkPrerequisites() = 0;

        /** @return  the unique family identity set by enum MPO_FAMILY_ID
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

        //protected:

    };                      // end class MPO_Errhandlers


}                               // end namespace MPO

#endif

