/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This Headerfile describes the MPO_LogCube object.
 * This object recieves MPO_MsgObject objects and creates a logfile from their data.
 * The logfile will be in a CUBE readable format.
 * @author Tobias Hilbrich
 * @date 13.11.2005
 * $Id: mpo_LogCube.h 694 2007-10-08 07:33:05Z tobias $ 
 */

#ifndef MPO_LOGCUBE_H
#define MPO_LOGCUBE_H

#include <string>
#include <iostream>
#include <vector>
#include <fstream>

#include "config.h"

/** This ifdef disables Cube when no Cubelogging was enabled during configure.
 */
#ifdef CUBE

#include <Cube.h> //The Cube Header
   
#include "mpo_LogHTML.h"
#include "mpo_TemplateTree.h"

using std::string;
using std::ofstream;
using namespace cube;

namespace MPO 
{
    /**
     * This Objects recieves MPO_MsgObject objects and creates an ASCII logfile from them.
     * The ASCII Log file object is the most simple logfile and should be useable on every 
     * System.
     */
    class MPO_LogCube : public MPO_LogFile
    {
    private:
		
        //==========================
        // Data
        //==========================
        bool                            m_bInitialized; /**< Flag speciefies wheter the Log is initialized or not. */
        Cube                            myCube; /**< The Cube logging object. */
        MPO_TemplateTree<Metric*> *mypMetricTree; /**< Stores the MetricTree. */
        MPO_TemplateTree<MPO_LogHTML*>  *mypMetricTreeHTML; /**< Stores the ofstreams of the Descriptive HTML. */
        //const Module*                   myModule; /**< Root of the Region Tree. */
        MPO_TemplateTree<const Region*> *mypRegionTree; /**< Stores all the Regions. */
        MPO_TemplateTree<Cnode*>  *mypCNodeTree; /**< Stores all the Call Nodes almost equivalent to myRegionTree. */
        MPO_TemplateTree<MPO_LogHTML*>  *mypCallTreeHTML; /**< Stores all the Call Nodes almost equivalent to myRegionTree. */
        Machine*	                myMachine; /**< The root of the System Tree. */
        Node*                     myNode; /**< We will only use one Node for the whole system. */
        std::vector<Process*>     myProcess; /**< List of all the process we have (MPI-Processes).*/
        //std::vector<const Thread*>      myThread; /**< List of all the threads we have (just one per process).*/
        std::vector<std::map<int,Thread*> > myThread; /**< List with one entry per process containing all the known threads.*/
        string                          myLogFileName; /**< Stores the name of the output file. */
        char                            *myCurrentDir; /**< The absolute path to this directory. */
        unsigned int                    myCurrentDirSize; /**< The length of the directory path. */
        
        MPO_TemplateTree<char*>         myCharIdMap;
        unsigned long                   myNextCharId;

    public:
        //==========================
        // Constructors/Destructors
        //==========================
		
        MPO_LogCube (string logFileName); /**< Constructor. */
        ~MPO_LogCube (); /**< Standard Destructor. */
		
        //==========================
        // Methods own
        //==========================
    public:	
        bool getMsg (MPO_MsgObject &pMsg); /**< Gets a message object, looks for MPI-Standard reference text and calls logMsg. */
        bool closeLog (void); /**< Closes the Log. */
        bool flushOfstream (void); /**< Flushs the Cube logfile. */

    protected:
        bool init(string logFileName);
        bool initAttributes(void); /**< Sets informative attributes like date for the logfile. */
        bool initPerformanceMetrics (void); /**< Initializes the Metrics for all the messages. */
        bool initCallTree (void); /**< Initalizes the call locations tree. */
        bool initSystemTree (void); /**< Initializes the system tree. */
        bool initSeverities (void); /**< Initializes Severities. */
        bool initMirrors (void); /**< Initializes the mirrors. */
                
        bool logInfoMsg(MPO_MsgObject &pMsg); /**< Logs a message of the MSG_TYPE_INFO type. */
        bool logWarningMsg(MPO_MsgObject &pMsg); /**< Logs a message of the MSG_TYPE_WARNING type. */
        bool logErrorMsg(MPO_MsgObject &pMsg); /**< Logs a message of the MSG_TYPE_ERROR type. */
        bool logNoteMsg(MPO_MsgObject &pMsg); /**< Logs a message of the MSG_TYPE_NOTE type. */
        
        unsigned long calculateCallTreeId(MPO_CALL_ID CallID,const MPO_SourceLocation &Location) const; /**< Calculates a unique CallTree Id from the given values. */
        unsigned long calculatePerformanceTreeId(MARMOT_ERR_CODES ErrorCode,MPO_MSG_TYPE MsgType); /**< Calculates a unique PerformanceTree(MessagesTree) Id from the given values. */
        unsigned long calculateIdFromString (const char *str); /**< Calculates an Id from the given char string.*/
        Cnode* logMessageToCallTree (unsigned long id, MPO_MsgObject &pMsg,const MPO_SourceLocation& SourceLocation); /**< Logs the message to the Cnode with the given id or creates the new Cnode and logs to it. */
        bool logMsgToExsitingCallTreeNode (MPO_MsgObject &pMsg,unsigned int nodeId,unsigned int subtreeId); /**< Logs the message to the node with the given id. */
        bool createPerformanceMetric (MARMOT_ERR_CODES err,unsigned int metricsId,unsigned int idFather); /**< Creates a child of type err for the node with idFather. */
        std::string getFileNameFromPath(const char* str);
        
        void checkForThreadInMetric (MPO_MsgObject &pMsg);
                        		
    };//end of class MPO_LogCube

}//end of namespace MPO

#endif //endif for cube logging
#endif //endif for HeaderProtector







