/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This File descibes the MPO_MsgObject class.
 * Message Objects are used for logging. They get Information about
 * the data to be logged and can be passed to a implementation of the
 * MPO_LogFile interface.
 * @author Tobias Hilbrich
 * @date 11.10.2005
 * $Id: mpo_MsgObject.h 892 2008-12-05 08:57:12Z tobias $
 */

#ifndef MPO_MSGOBJECTS_H
#define MPO_MSGOBJECTS_H

#include "mpo.h"
#include "mpo-enums.h"
#include "mpo-timer.h"
#include "mpo_SourceLocation.h"
#include "mpo_String.h"

#include <string>
#include <iostream>

class CTcpSocket;

using std::string;

namespace MPO
{
    /**
      * This class represents a output message.
      * The data of the message is abstract and independent from the logtype used.
      * When initializing these Messages all available data should be passed to this
      * class. The MPO_LogFile implementation will decide which data should be logged.
      * So there is no need for all those "if" statments within the code sections that
      * detect errors or create other kinds of messages.
      */
    class MPO_MsgObject : public MPO_Object
    {
    protected:

        //==========================
        // Data
        //==========================

        MPO_MSG_TYPE          m_MsgType; /**< Describes the message type. */
        long                  m_Timestamp; /**< The timestamp of the message. */
        int                   m_rank; /**< Rank of the node which created this message. */
        MPO_OUTPUT_TYPE       m_OutputType; /**< Describes the conntent of this message. */
        MPO_SourceLocation    m_CodeReference; /**< References Codefile + Codeline must provide a MPO_LOCATION_UNKNOWN test. */

        MPO_String            m_Callinfo; /**< Describes the call with its name and parameters. */

        MPO_String            m_Text; /**< Holds text part of the message (only for OUTPUT_TYPE_PURE_TEXT or OUTPUT_TYPE_TEXT_AND_CALLINFO). */
        MPO_CALL_ID           m_CallID; /**< Used to find MPI-Standard text information when an error occured (or a Note, Warning). */
        MARMOT_ERR_CODES      m_ErrorCode; /**< Used to find MPI-Standard text information when an error occured (or a Note, Warning). */

        static int            m_GlobalStaticRank; /**< The rank of the Message creating process. */
        int                   m_ThreadRank; /**< The numerical id of the message creating thread. */

    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /**
         * We provide the creation of 3 different Msg Types depending on the given data).
         * For each of this 3 types we offer a version with and withoout of a SourceLocation and a Rank.
         * So we get 4 different versions per type leading to 12 different Constructors.
         */

        MPO_MsgObject (); /**< Create invalid object constructor. */

        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode); /**< Constructor for OUTPUT_TYPE_PURE_TEXT Messages. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Constructor for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location. */
        //Constructors not nedding a rank
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode); /**< Constructor for OUTPUT_TYPE_PURE_TEXT Messages - rank is determined by the rank of the message creating process. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Constructor for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location - rank is determined by the rank of the message creating process. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages - rank is determined by the rank of the message creating process. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location - rank is determined by the rank of the message creating process. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages - rank is determined by the rank of the message creating process. */
        MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location - rank is determined by the rank of the message creating process. */

        ~MPO_MsgObject (); /**< Standard destructor. */

        //==========================
        // Methods own
        //==========================
    private:
        void initMsgObject (MPO_MSG_TYPE MsgType, MPO_OUTPUT_TYPE OutputType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Init Method called by all the different constructors. */
        void initGlobalStaticRank(); /** Initializes the static rank of the message creating process by using a MPI-Call. */

    public:

        /** In order to enable a recycling of MPO_MsgObject`s we provide all the Constructors as
         *  normal public mehtods.
         */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode); /**< Init for OUTPUT_TYPE_PURE_TEXT Messages. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Init for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Init for OUTPUT_TYPE_PURE_CALLINFO Messages. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Init for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location. */
        //init methods not nedding a rank
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode); /**< Init for OUTPUT_TYPE_PURE_TEXT Messages - rank is determined by the rank of the message creating process. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Init for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location - rank is determined by the rank of the message creating process. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Init for OUTPUT_TYPE_PURE_CALLINFO Messages - rank is determined by the rank of the message creating process. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Init for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location - rank is determined by the rank of the message creating process. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode); /**< Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages - rank is determined by the rank of the message creating process. */
        void initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference); /**< Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location - rank is determined by the rank of the message creating process. */

        MPO_MSG_TYPE getMsgType (void) const; /**< Returns the type of this message. */
        long getTimestamp (void) const; /**< Returns the timestamp. */
        int getRank (void) const; /**< Returns the rank. */
        MPO_OUTPUT_TYPE getOutputType (void) const; /**< Returns the output type. */
        const MPO_SourceLocation& getSourceLocation (void) const; /**< Returns the source location when present. */
        const string& getText (void) const; /**< Returns the associated text. */
        const string& getCallinfo (void) const; /**< Returns the associated callinfo. */
        bool hasMPIStandardReferenceInformation (void) const; /**< Returns wheter information for creating a reference to the MPI-Standard is present or not. */
        bool getMPIStandardReferenceInformation (MPO_CALL_ID *pCallID, MARMOT_ERR_CODES *pErrorCode) const; /**< Returns the refernce data.*/
        MPO_CALL_ID getCallId (void); /**< Returns the ID of the Call assigned to this Object. */
        MARMOT_ERR_CODES getErrorCode (void); /**< Returns the Errorcode assigned to this Object. */
        string getMessageTypeString (void) const;/**< Returns a string that descibres the Messagetype (like "Error" or "Information"). */
        string getOutputTypeString (void) const; /**< Returns a string that describes the Outputtype (like "Texutal information" or "Call information"). */
        bool setTimestamp (long time); /**< Sets the timestamp to the given value might be used to modify the timestamp after sending it with MPO_LogToServer. */
        int getThreadRank(void); /**< Returns the rank of the thread that created this message. */
        void setThreadRank(int threadRank); /**< Sets the rank for the thread that created this message. */
        void appendToText (std::string append);

        //==========================
        // Methods inherited
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        int getId() const;

        /** @return  Family ID. */
        int getFamilyId() const;

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        string getCallname() const;

        /** Overloads << to print call-description. */
        std::ostream & debugout(std::ostream & output) const;

        /**
         * @brief  Sends contents of the message to a remote host as a "degree-sign" delimited and <CR><LF> terminated string
         * @param  socketptr  : Pointer to socket (may be NULL)
         * @return Bytes sent if successful, -1 otherwise
         */
        int writeToSocket(CTcpSocket *socketptr, const char *mpi_reference_path) const;

        /** Serializes an object to ostream.*/
        std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        std::istream & restore(std::istream & in);

    };//end of class MPO_MsgObject

}//end of namespace MPO

#endif







