/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
  * \file mpo_ThreadSynchronisationDummy.h
  * This Headerfile describes the Dummy implementation of the interface
  * MPO_IThreadSynchronisation.
  * @author Tobias Hilbrich
  * @date 01.06.2007
  * $ID$
  */

#ifndef MPO_THREADSYNCHRONISATIONDUMMY_H
#define MPO_THREADSYNCHRONISATIONDUMMY_H

#include "MPO_IThreadSynchronisation.h"
#include "config.h"

namespace MPO
{
    /**
      * This class is a dummy implementation of the MPO_IThreadSynchronisation interface.
      * It provides an implmentation that is suficent for non threaded environments.
      */
   class MPO_ThreadSynchronisationDummy : public MPO_IThreadSynchronisation
    {
    public:

        /**
         * This method is used when a thread enters the marmot wrappers and after a
         * thread made its PMPI call in order to guarantie that only one thread
         * is using marmots global data at any given time.
         * This method also has to check for violations to the used level of thread
         * support.
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @return returns true if this thread is not already in MARMOT and false otherwise.
         *         Note: this might happen if the Implementation of an MPI call uses other MPI calls.
         */
        bool enterMARMOT (int line,const char * file);

        /**
         * This method is used when a thread leaves the marmot wrappers and before a
         * thread makes the PMPI call to guarantie that only one thread is using
         * marmots global data at any given time
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param enter_succeeded value returned by enterMARMOT.
         */
        void leaveMARMOT (int line,const char * file, bool enter_succeeded);

        /**
         * This method is used before a thread makes its PMPI call in order to
         * guarantie that only one thread is using marmots global data at any
         * given time.
         */
        void enterPMPI ();

        /**
         * This method is used after a thread made its PMPI call to
         * guarantie that only one thread is using marmots global data
         * at any given time.
         */
        void leavePMPI ();

        /**
         * This method is usefull for determining whether thread parallelisation is
         * activated or not.
         * i.e. whether we are in a omp parallel region or not.
         * @return true if execution is in a parallel region, false otherwise.
         */
        bool isInParallelRegion (void);

        /**
         * This method is used to return the number/id of the thread actually executing.
         * @return the id of the thread in execution or 0 if no parallel region is present.
         */
        int getThreadRank (void);

        /**
         * Swichtes wrapping of MPI calls on for this
         * thread.
         * Must not be called from while inside Marmot.
         */
        void wrappingOn (void);

        /**
         * Swichtes wrapping of MPI calls off for this
         * thread.
         * Must not be called from while inside Marmot.
         */
        void wrappingOff (void);

        /**
         * Returns whether the enterMARMOT succeeded.
         * Must only be called between Enter- and Leave-
         * MARMOT.
         */
        bool didEnterSucceed (void);

        /**
         * This method returns the name of the used implementation of this interface.
         * Used for logging data about the used thread synchronisation.
         * @return the name of the implementation used.
         */
        std::string getImplementationName (void);

        /**
         * This method returns true if another thread is waiting at enterMARMOT.
         * @return true if another thread is waiting at enterMARMOT, false otherwise.
         */
        bool otherThreadsWaiting (void);

        /**
         * Does nothing as this implementation assumes that there are no threads.
         * @param request the request to be registered.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @param callId the call id of the associated call.
         * @param callName the name of the associated call.
         * @return an id used to unregister the requests after usage.
         */
        int registerRequest (MPI_Request *request, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * Does nothing as this implementation assumes that there are no threads.
         * @param requests the request to be registered.
         * @param count the amount of requests in the given array.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @param callId the call id of the associated call.
         * @param callName the name of the associated call.
         * @return an id used to unregister the requests after usage.
         */
        int registerRequests (MPI_Request *requests, int count, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * Does nothing as this implementation assumes that there are no threads.
         * @param id the id of the requests to be unregistered.
         */
        void unregisterRequests (int id);

#ifdef MARMOT_TCHECK
        /**
         * Used to remove a request registered in the request to int conversion for
         * TCheck conflict Variables.
         * Before calling make sure you are not within MARMOT i.e. by calling enterPMPI
         * and leavePMPI after the call !
         */
        void unregisterRequestToIntEntry(MPI_Request *request);

        /**
         * Used to remove a comm registered in the request to int conversion for
         * TCheck conflict Variables.
         * Before calling make sure you are not within MARMOT i.e. by calling enterPMPI
         * and leavePMPI after the call !
         */
        void unregisterCommToIntEntry(MPI_Comm comm);
#endif

        /**
         * Does nothing as this implementation assumes that there are no threads.
         * @param comm the communicator to be registered.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @param threadNum the id of the calling thread.
         * @param callId the call id of the associated call.
         * @param callName the name of the associated call.
         * @return an id used to unregister the communicator after usage.
         */
        int registerCollComm (MPI_Comm comm, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName);

        /**
         * Does nothing as this implementation assumes that there are no threads.
         * list of currently used communicators.
         * @param id the id of the communicator to be unregistered.
         */
        void unregisterCollComm (int id);

        /**
         * Always returns MPI_THREAD_SINGLE as this implementation assumes that there are no threads.
         * @return the minimal level of thread support neded for this application and run.
         */
        int getMinThreadLvl (void);

        /**
         * Always returns an empty violation info as this implementation assumes
         * that there are no threads.
         * @param not of interest
         * @return an invalid source location.
         */
        MPO_ThreadViolationInfo getViolationInfoForThreadLvl (int threadLevel);

    };//end of class IThreadSynchronisation

}//end of namespace MPO

#endif







