/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
  * \file mpo_ThreadSynchronisationOpenMP.h
  * This Headerfile describes the OpenMP implementation of the interface
  * MPO_IThreadSynchronisation.
  * @author Tobias Hilbrich
  * @date 01.06.2007
  * $ID$
  */

#ifndef MPO_THREADSYNCHRONISATIONOPENMP_H
#define MPO_THREADSYNCHRONISATIONOPENMP_H

#include "config.h"

#ifdef MARMOT_OPENMP

#include <omp.h>
#include "MPO_ThreadSynchronisationSyncBase.h"
#include <mpo.h>

namespace MPO
{
    /**
      * This class is an OpenMP implementation of the MPO_IThreadSynchronisation interface.
      * As it is a synchronized implementation it used MPO_ThreadSynchronisationSyncBase as
      * its base class. Thus it also has to implement the beginCritical and endCrititcal
      * methods.
      * It provides an implmentation that supports threadsynchonisation for Threads created
      * with OpenMP.
      */
   class MPO_ThreadSynchronisationOpenMP : public MPO_ThreadSynchronisationSyncBase
    {
    protected:

        /**
         * This variable is used for the OpenMP locks.
         */
        omp_lock_t my_lock;

        /**
         * This variable is used short critical sections.
         */
        omp_lock_t my_criticalLock;

        /**
         * Holds the MPO_Callobject associated with the thread currently in MARMOT.
         */
        std::pair<int,MPO_SourceLocation> my_infoOfThreadInMARMOT;

        /**
         * This variable counts the number of threads actually within MARMOT.
         */
        int my_NumInMarmot;

        /**
         * Hook method used in registerRequest, registerRequests, unregisterRequests.
         * This method has to start a protection for a critical section.
         * If it uses a lock (probably the implementation will do) it should use another
         * lock than in enterMARMOT !
         */
        void beginCritical ();

        /**
         * Hook method used in registerRequest, registerRequests, unregisterRequests.
         * This method has to end the protection for a critical section.
         * If it uses a lock (probably the implementation will do) it should use another
         * lock than in enterMARMOT !
         */
        void endCritical ();

        /**
         * Returns the thread id and source location of the mpi call associated with
         * the thread that passed enterMARMOT.
         * This methos should only be called by one thread at a time.
         * @return thread id and source location of the mpi call associated with
         * the thread that passed enterMARMOT.
         */
        std::pair<int,MPO_SourceLocation> getInfoForThreadInMARMOT(void);

        /**
         * Returns true if another thread is somewhere enterMARMOTSync and
         * leaveMARMOTSync.
         * @return true if other thread(s) are in MARMOT otherwise false.
         */
        bool otherThreadsInMARMOT (void);

        /**
         * This method is used when a thread enters the marmot wrappers and after a
         * thread made its PMPI call in order to guarantie that only one thread
         * is using marmots global data at any given time.
         * This method also has to check for violations to the used level of thread
         * support.
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         */
        void enterMARMOTSync (int line,const char * file);

        /**
         * This method is used when a thread leaves the marmot wrappers and before a
         * thread makes the PMPI call to guarantie that only one thread is using
         * marmots global data at any given time
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         */
        void leaveMARMOTSync (int line,const char * file);

        /**
         * This method returns the amount of threads currently used.
         * @return the amount of threads currently used.
         *
         * Its implementation depends upon the concrete threading implementation used.
         */
        int getNumThreads(void);

    public:

        /**
         * Constructor responsible for necessary initialisations.
         */
        MPO_ThreadSynchronisationOpenMP(void);

        /**
         * Destructor responsible releasing memory and objects.
         */
        ~MPO_ThreadSynchronisationOpenMP(void);

        /**
         * This method is usefull for determining whether thread parallelisation is
         * activated or not.
         * i.e. whether we are in a omp parallel region or not.
         * @return true if execution is in a parallel region, false otherwise.
         */
        bool isInParallelRegion (void);

        /**
         * This method is used to return the number/id of the thread actually executing.
         * @return the id of the thread in execution or 0 if no parallel region is present.
         */
        int getThreadRank (void);

        /**
         * This method returns the name of the used implementation of this interface.
         * Used for logging data about the used thread synchronisation.
         * @return the name of the implementation used.
         */
        std::string getImplementationName (void);

        /**
         * This method returns true if another thread is waiting at enterMARMOT.
         * @return true if another thread is waiting at enterMARMOT, false otherwise.
         */
        bool otherThreadsWaiting (void);

        /**
         * This method is used before a thread makes its PMPI call in order to
         * guarantie that only one thread is using marmots global data at any
         * given time.
         */
        void enterPMPI ();

        /**
         * This method is used after a thread made its PMPI call to
         * guarantie that only one thread is using marmots global data
         * at any given time.
         */
        void leavePMPI ();

    };//end of class IThreadSynchronisation

}//end of namespace MPO

#endif //#ifdef MARMOT_OPENMP

#endif







