/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Accumulate class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Accumulate.cc 838 2008-06-13 10:44:34Z hpchimml $
 */


#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Accumulate::MPO_Accumulate()
        : MPO_WinCommunicator(0,
                              -1,
                              MPI_DATATYPE_NULL,
                              MPI_PROC_NULL,
                              -1,
                              -1,
                              MPI_DATATYPE_NULL,
                              MPI_WIN_NULL),
          my_op(MPI_OP_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Accumulate::MPO_Accumulate(void *origin_addr, 
                                  int origin_count, 
                                  MPO_Datatype origin_datatype, 
                                  int target_rank, 
                                  MPI_Aint target_disp,
                                  int target_count, 
                                  MPO_Datatype target_datatype, 
                                  MPO_Op op,
                                  MPO_Win win)
        : MPO_WinCommunicator(origin_addr,
                              origin_count,
                              origin_datatype,
                              target_rank,
                              target_disp,
                              target_count,
                              target_datatype,
                              win),
          my_op(op)
    {
    }


    // destructor
    //========================================================================
    MPO_Accumulate::~MPO_Accumulate()
    {
    }


    // checkOp 
    //========================================================================
    int MPO_Accumulate::checkOp() 
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;


        return rvalue;
    }



    // getId
    //========================================================================
    int MPO_Accumulate::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ACCUMULATE_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Accumulate::callNativeMPI()
    {
        int rvalue = 1;

        checkPrerequisites();
        checkPrerequisites2();
        checkOp();
        rvalue = PMPI_Accumulate(my_origin_addr, my_origin_count,
                                 my_origin_datatype, my_target_rank,
                                 my_target_disp, my_target_count,
                                 my_target_datatype, my_op, my_win);

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Accumulate::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Accumulate";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ACCUMULATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Accumulate::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(*origin_addr"
            << ", origin_count = "
            << my_origin_count
            << ", origin_datatype = "
            << my_origin_datatype
            << ", target_rank = "
            << my_target_rank
            << ", target_disp = "
            << my_target_disp
            << ", target_count = "
            << my_target_count
            << ", target_datatype = "
            << my_target_datatype
            << ", op = "
            << my_op
            << ", win = "
            << my_win;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }

    // save
    //========================================================================
    std::ostream & MPO_Accumulate::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_origin_addr;
        out < my_origin_count;
        out < my_origin_datatype;
        out < my_target_rank;
        out < my_target_disp;
        out < my_target_count;
        out < my_target_datatype;
        out < my_op;
        out < my_win;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Accumulate::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_origin_addr;
        in > my_origin_count;
        in > my_origin_datatype;
        in > my_target_rank;
        in > my_target_disp;
        in > my_target_count;
        in > my_target_datatype;
        in > my_op;
        in > my_win;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/accumulate.c

/*@
   MPI_Accumulate - Accumulate data into the target process using remote 
   memory access 

   Input Parameters:
+ origin_addr - initial address of buffer (choice) 
. origin_count - number of entries in buffer (nonnegative integer) 
. origin_datatype - datatype of each buffer entry (handle) 
. target_rank - rank of target (nonnegative integer) 
. target_disp - displacement from start of window to beginning of target 
  buffer (nonnegative integer)  
. target_count - number of entries in target buffer (nonnegative integer) 
. target_datatype - datatype of each entry in target buffer (handle) 
. op - predefined reduce operation (handle) 
- win - window object (handle) 

   Notes:
The basic components of both the origin and target datatype must be the same 
predefined datatype (e.g., all 'MPI_INT' or all 'MPI_DOUBLE_PRECISION').

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COUNT
.N MPI_ERR_RANK
.N MPI_ERR_OP

int MPI_Accumulate(void *origin_addr, int origin_count, MPI_Datatype
                   origin_datatype, int target_rank, MPI_Aint
                   target_disp, int target_count, MPI_Datatype
                   target_datatype, MPI_Op op, MPI_Win win) 
@*/
