/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Win_free class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Win_free.cc 979 2009-07-17 12:54:04Z hpcdiche $
 */


#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Win_free::MPO_Win_free()
        : mpo_win_free_Win(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Win_free::MPO_Win_free(MPI_Win *win)

    {
    	mpo_win_free_Win=win;
    }


    // destructor
    //========================================================================
    MPO_Win_free::~MPO_Win_free()
    {
    }

    int MPO_Win_free::checkPost(){
    	return MARMOT_SUCCESS;

    }

    // checkPrerequisites
    //========================================================================
    int MPO_Win_free::checkPrerequisites()
    {
        // check correctness of parameters
        checkWin();

        return 0;
    }


    // checkWin
    //========================================================================
    int MPO_Win_free::checkWin()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if window is null handle.
        if ((MPI_WIN_NULL == *mpo_win_free_Win) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": window to free is already MPI_WIN_NULL!"
                << std::endl;
            rvalue = MARMOT_WARN_WIN_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if window is valid, i.e. registered in our_wins.
        /*
         * Kiril: comment the next block because I am currently not able to set the window attribute in a proper
         * way before this call (e.g. in MPO_Space ?)
         *
         **/
        else if ((our_wins.end() == our_wins.find(*mpo_win_free_Win)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Window is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_WIN_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

		//The synchronization call of the one-sided communication is missing, but we still try to free the window.
        //This is not correct and must be logged as an error
        else{
       	}

        return rvalue;
    }                           // end checkWin


    // getId
    //========================================================================
    int MPO_Win_free::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WIN_FREE_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Win_free::callNativeMPI(){
        int rvalue = 1;

        checkPrerequisites();
        //our_wins.erase(*my_win);



    	std::map < MPI_Win, mpi_win_marmot_states_t>::iterator it =	our_wins.find(*mpo_win_free_Win);
    	if (it == our_wins.end())
    	{
    		std::cout << "Window not found anymore in mpo_Win_free " << std::endl;
    	}
		mpi_win_marmot_states_t mpi_win_marmot_state=it->second;
		if (mpi_win_marmot_state == MPI_WIN_FREE_NOT_ALLOWED){
			std::stringstream stream;
			stream   << "ERROR: " << getCallname()<< " not allowed because a synchronization call of the one-sided communication is missing" << std::endl;
			rvalue = MARMOT_ERR_WIN_FREE_NOT_ALLOWED;
			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

        rvalue = PMPI_Win_free(mpo_win_free_Win);

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Win_free::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Win_free";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WIN_FREE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Win_free::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*win";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Win_free::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
//        out < my_win;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Win_free::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in < my_win;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/win_free.c

/*@
   MPI_Win_free - Free an MPI RMA window

   Input Parameter:
. win - window object (handle)

   Notes:
   If successfully freed, 'win' is set to 'MPI_WIN_NULL'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN

int MPI_Win_free(MPI_Win *win)
@*/


