/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/** 
 *  @file
 *
 *  Definition of the base class MPO_Blocking.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: MPO_Blocking.cc 849 2008-07-16 16:06:49Z hpchimml $
 */


#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Blocking::MPO_Blocking()
        : my_count(-1),
          my_datatype(MPI_DATATYPE_NULL),
          my_rank(MPI_PROC_NULL), 
          my_tag(MPI_ANY_TAG), 
          my_comm(MPI_COMM_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Blocking::MPO_Blocking(int count,
                               MPO_Datatype datatype,
                               int rank,
                               int tag,
                               MPO_Comm comm)
        : my_count(count),
          my_datatype(datatype), 
          my_rank(rank), 
          my_tag(tag), 
          my_comm(comm)
    {
    }


    // destructor
    //========================================================================
    MPO_Blocking::~MPO_Blocking()
    {
    }


    /* There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Blocking::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Blocking::checkPrerequisites()
    {
        // check correctness of parameters
        if (getId() == MPO_RECV_ID)
        {
            checkCount(MSG_RECVCOUNT);
        }
        else 
        {
            checkCount(MSG_SENDCOUNT);
        }

        checkDatatype();
        checkRank();
        checkTag(MSG_TAG);
        checkComm();

        return 0;
    }


    // checkCount
    //========================================================================
    int MPO_Blocking::checkCount(const string KIND_OF_COUNT)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if count is negative.
	if (our_errcodes.at(MARMOT_ERR_COUNT_NEG))
	{
        if ((my_count < 0) && 
            (MSG_ERR <= DEBUG_MODE) )
        {
            std::stringstream stream;
            stream 
                << MSG_ERROR
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << " = "
                << my_count
                << " less than 0 !" 
                << std::endl;
            
            rvalue = MARMOT_ERR_COUNT_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                                   
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);            
        }
	}

        // (2) Check if count is 0.
	if (our_errcodes.at(MARMOT_WARN_COUNT_ZERO))
	{
        if ((my_count == 0) && 
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << MSG_WARNING
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << " = 0!" 
                << std::endl;
            rvalue = MARMOT_WARN_COUNT_ZERO;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);            
        }
	}
        return rvalue;
    }


    // checkDatatype
    //========================================================================
    int MPO_Blocking::checkDatatype()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Datatype, enum KIND_OF_DATATYPE >::iterator pos;
        pos = our_datatypes.find(my_datatype);

        // (1) Datatype is not valid, i.e. not registered in
        // our MPO datatypes.
	
        if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_VALID)) && (pos == our_datatypes.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": datatype is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_TYPE_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);            
        }
	
        // (2) Datatype is registered in our MPO datatypes.
        // In this case, we have to check different violations 
        // in the switch statement.
        else if ((pos != our_datatypes.end()) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            //used to store errors
            std::stringstream stream; 
            MPO_MSG_TYPE type = MSG_TYPE_ERROR;
            
            // all the cases from enum KIND_OF_DATATYPE
            switch (pos->second)
            {
                case (DATATYPE_NULL):
			if(our_errcodes.at(MARMOT_ERR_TYPE_NULL))
		    	{
                    		stream
                        	<< "ERROR: "
                        	<< getCallname() 
                        	<< ": datatype is MPI_DATATYPE_NULL!" 
                        	<< std::endl;
                    		rvalue = MARMOT_ERR_TYPE_NULL;
			}
                    break;
                case (DATATYPE_CREATED_NOT_COMMITTED):
		    if(our_errcodes.at(MARMOT_ERR_TYPE_NOT_COMMITTED))
		    {
                    	stream
                        << "ERROR: "
                        << getCallname() 
                        << ": datatype is not committed!" 
                        << std::endl;
                    	rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;
			}
                    break;
                case (DATATYPE_COMMITTED_C):
		    if(our_errcodes.at(MARMOT_ERR_TYPE_C))
		    {
                    	if (F_INTERFACE == INTERFACE_MODE)
                    	{
                        	stream
                            	<< "ERROR: "
                            	<< getCallname() 
                            	<< ": datatype is C-Type!" 
                            	<< std::endl;
                        	rvalue = MARMOT_ERR_TYPE_C;
			}
                    }
                    break;
                case (DATATYPE_COMMITTED_C_AND_FORTRAN):
                    // Do nothing, because everything is fine.
                    break;
                case (DATATYPE_COMMITTED_FORTRAN):
		    if(our_errcodes.at(MARMOT_ERR_TYPE_F))
		    {
                    	if (C_INTERFACE == INTERFACE_MODE)
                    	{
                        	stream
                            << "ERROR: "
                            << getCallname() 
                            << ": datatype is Fortran-Type!" 
                            << std::endl;
                        	rvalue = MARMOT_ERR_TYPE_F;
			}
                    }
                    break;
                case (DATATYPE_REDUCTION_C):
                    if (our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_C) && (MSG_ERR_WARN <= DEBUG_MODE) &&
                        (C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname()
                            << ": datatype is for reduction functions (C)!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_REDUCTION_C;
                        type = MSG_TYPE_WARNING;
                    }
                    if (our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_C) && F_INTERFACE == INTERFACE_MODE)
                    {
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": datatype is for reduction functions (C)!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_REDUCTION_C;
                    }
                    break;
                case (DATATYPE_REDUCTION_FORTRAN):
                    if (our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_F) && C_INTERFACE == INTERFACE_MODE)
                    {
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": datatype is for reduction functions "
                            << "(Fortran)!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_REDUCTION_F;
                    }
                    if (our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_F) && (MSG_ERR_WARN <= DEBUG_MODE) &&
                        (F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname()
                            << ": datatype is for reduction functions "
                            << "(Fortran)!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_REDUCTION_F;
                        type = MSG_TYPE_WARNING;
                    }
                    break;
                case (DATATYPE_OPTIONAL_C):
                    if (our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_C) && (MSG_ERR_WARN <= DEBUG_MODE) &&
                        (C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname() 
                            << ": datatype is only optional!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_OPTIONAL_C;
                        type = MSG_TYPE_WARNING;
                    }
                    if (our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_C) && F_INTERFACE == INTERFACE_MODE)
                    {
                        stream
                            << "ERROR: "
                            << getCallname() 
                            << ": datatype is optional C-Type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_OPTIONAL_C;
                    }
                    break;
                case (DATATYPE_OPTIONAL_FORTRAN):
                    if (our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_F) && C_INTERFACE == INTERFACE_MODE)
                    {
                        stream
                            << "ERROR: "
                            << getCallname() 
                            << ": datatype is optional Fortran-type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_OPTIONAL_F;
                    }
                    if (our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_F) && (MSG_ERR_WARN <= DEBUG_MODE) && 
                        (F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname() 
                            << ": datatype is only optional!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_OPTIONAL_F;
                        type = MSG_TYPE_WARNING;
                    }
                    break;
                case (DATATYPE_CONSTRUCT_DERIVED_TYPE):
		    if (our_errcodes.at(MARMOT_ERR_TYPE_CONSTR_DERIVED))
		    {
                    	stream
                        << "ERROR: "
                        << getCallname() 
                        << ": datatype is for constructing derived datatypes!"
                        << std::endl;
                    	rvalue = MARMOT_ERR_TYPE_CONSTR_DERIVED;
		    }
                    break;
                default:
		    if(our_errcodes.at(MARMOT_ERR_MARMOT))
		    {
                    	stream
                        << "Check mapping of datatypes."
                        << std::endl
                        << "Check implementation of switch statement."
                        << std::endl;
                    	rvalue = MARMOT_ERR_MARMOT;
		    }
                    break;
            }                   // end switch
            
            if (rvalue != MARMOT_SUCCESS)
            {
            	our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            	our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            	
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);            
            }//if error occured
        }                       // end else if (2)

        return rvalue;
    }                           // end checkDatatype


    // checkRank
    //========================================================================
    int MPO_Blocking::checkRank()
    {
        // We distinct between MPO_Recv and other blocking calls, 
        // because my_rank may be source or destination.

        if (getId() == MPO_RECV_ID)
        {
            checkRankOfSource();
        }
        else 
        {
            checkRankOfDestination();
        }

        checkRankTooBig();

        return 0;
    }


    // checkRankOfSource
    //========================================================================
    int MPO_Blocking::checkRankOfSource()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if rank of source < 0.
        if ((our_errcodes.at(MARMOT_ERR_RANK_NEG)) && (my_rank != MPI_ANY_SOURCE) &&
            (my_rank != MPI_PROC_NULL) && 
            (my_rank < 0) && 
            (MSG_ERR <= DEBUG_MODE))
    	    {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank of source = "
                << my_rank
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_RANK_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        // (2) Check if rank of source is MPI_PROC_NULL.
        else if ((our_errcodes.at(MARMOT_NOTE_RANK_PROC_NULL))&&(my_rank == MPI_PROC_NULL) && 
                 (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << MSG_NOTE
                << getCallname() 
                << ": Source = MPI_PROC_NULL !" 
                << std::endl;
            rvalue = MARMOT_NOTE_RANK_PROC_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        // (3) Warn of race conditions if MPI_ANY_SOURCE is used.
        else if ((our_errcodes.at(MARMOT_WARN_RANK_ANY_SOURCE))&&(my_rank == MPI_ANY_SOURCE) && 
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": Use of MPI_ANY_SOURCE may cause race conditions!"
                << std::endl;
            rvalue = MARMOT_WARN_RANK_ANY_SOURCE;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }                      

        return rvalue;
    }                           // end checkRankOfSource


    // checkRankOfDestination
    //========================================================================
    int MPO_Blocking::checkRankOfDestination()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if rank of destination < 0.
        if ((our_errcodes.at(MARMOT_ERR_RANK_NEG))&&(my_rank < 0) && 
            (my_rank != MPI_PROC_NULL) && 
            (my_rank != MPI_ANY_SOURCE) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Rank of destination = "
                << my_rank
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_RANK_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        // (2) Check if rank of destination is MPI_PROC_NULL.
        else if ((our_errcodes.at(MARMOT_NOTE_RANK_PROC_NULL))&&(my_rank == MPI_PROC_NULL) && 
                 (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << getCallname()
                << ": Rank of destination = MPI_PROC_NULL !" 
                << std::endl;
            rvalue = MARMOT_NOTE_RANK_PROC_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        // (3) Check if rank of destination is MPI_ANY_SOURCE.
        else if ((our_errcodes.at(MARMOT_ERR_RANK_ANY_SOURCE))&&(my_rank == MPI_ANY_SOURCE) && 
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR "
                << getCallname()
                << ": Rank of destination = MPI_ANY_SOURCE !" 
                << std::endl;
            rvalue = MARMOT_ERR_RANK_ANY_SOURCE;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        return rvalue;
    }                           // end checkRankOfDestination


    // checkRankTooBig
    //========================================================================
    int MPO_Blocking::checkRankTooBig()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;        

        // Check if rank is out of communicator.     
        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        if ((our_errcodes.at(MARMOT_ERR_RANK_TOO_BIG))&&(my_rank >= size) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank is out of communicator: rank = "
                << my_rank 
                << ", size of communicator = " 
                << size 
                << std::endl;
            rvalue = MARMOT_ERR_RANK_TOO_BIG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        return rvalue;
    }


    // checkTag
    //========================================================================
    int MPO_Blocking::checkTag(const string KIND_OF_TAG)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if tag is < 0.
        if ((our_errcodes.at(MARMOT_ERR_TAG_NEG))&&(my_tag != MPI_ANY_TAG) && 
            (my_tag < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream  << "ERROR: "
                    << KIND_OF_TAG
                    << " = "
                    << my_tag
                    << " less than 0!" 
                    << std::endl;
            rvalue = MARMOT_ERR_TAG_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);           
        }

        // (2) Check if tag is within guaranteed range.
        else if ((our_errcodes.at(MARMOT_ERR_TAG_TOO_BIG)||our_errcodes.at(MARMOT_WARN_TAG_TOO_BIG))&&(my_tag != MPI_ANY_TAG) && 
                 (my_tag > 32767) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            int ub = 0;
            int flag = 1;
            PMPI_Attr_get(MPI_COMM_WORLD, MPI_TAG_UB, &ub, &flag);

            if ((my_tag > ub) && 
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream  << "ERROR: "
                        << KIND_OF_TAG
                        << " = "
                        << my_tag 
                        << " is larger than allowed!" 
                        << std::endl
                        << "THIS implementation allows tags up to "
                        << ub 
                        << "."
                        << std::endl;
                rvalue = MARMOT_ERR_TAG_TOO_BIG;
            
            	our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            	our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                
                MPO_String text;
                text.assign(&stream);     
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);
            }

            else if ((my_tag <= ub) && 
                     (MSG_ERR_WARN <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream  << "WARNING: "
                        << KIND_OF_TAG
                        << " = "
                        << my_tag
                        << " greater then 32767 ! MPI only guarantees tags up to this." 
                        << std::endl
                        << "THIS implementation allows tags up to "
                        << ub 
                        << "."
                        << std::endl;
                rvalue = MARMOT_WARN_TAG_TOO_BIG;
            
            	our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            	our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                
                MPO_String text;
                text.assign(&stream);            
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);
            }
        }                       // end else if (2)

        return rvalue;
    }                           // end checkTag


    // checkComm
    //========================================================================
    int MPO_Blocking::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");            
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " :Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        return rvalue;
    }                           // end checkComm



}                               // end of namespace MPO
