/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the abstract base class MPO_Callobject.
 *
 *  This abstract base class represents MPI calls.
 *
 *  @author Bettina Krammer, Tobias Hilbrich
 *
 *  $Id: MPO_Callobject.cc 964 2009-06-02 06:38:44Z tobias $
 */


#include <string.h>
#include <assert.h>
#include "mpo.h"
#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo-messenger.h"
#include "mpo-autogenerated.h"
#include "mpo_SourceLocation.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

namespace MPO
{
#ifdef MARMOT_USE_THREADSYNC
		std::list<MPO_Callobject::MPO_ProbeInfo> MPO_Callobject::our_succesful_probes;

#ifdef MARMOT_TCHECK
        int MPO_Callobject::our_next_probe_array_index = 0;
#endif /*MARMOT_TCHECK*/

#endif /*MARMOT_USE_THREADSYNC*/

	std::map<MPO_Request,MPO_Callobject::MPO_BufferInfoType> MPO_Callobject::our_used_buffers;
	std::map<int, MPO_Callobject*> MPO_Callobject::our_pre_object_data;
	bool MPO_Callobject::our_store_pre_object_data = false;

    // default constructor
    //========================================================================
    MPO_Callobject::MPO_Callobject() :
        my_SourceLine(0),my_SourceFile(NULL),my_SourceFileMem(NULL)
    {
    	my_thread_rank = MPO_Space::getSyncObject()->getThreadRank();
    	my_persitent_restored = true;
    }


    // destructor
    //========================================================================
    MPO_Callobject::~MPO_Callobject()
    {
        if (my_SourceFileMem)
            delete (my_SourceFileMem);
        my_SourceFileMem = NULL;

        my_SourceFile = NULL;
    }


    /** There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Callobject::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }


    // callDebugServer
    //========================================================================
    int MPO_Callobject::callDebugServer()
    {
        //test whether
	// - MPI_Finalize was already called
	// - MPI_Init was not called yet
        if ((our_finalize_was_called == true)||(our_init_was_called == false))
        {
            //only MPI_Get_version, MPI_Initialized and MPI_Finalized is now allowed,
            //for all other calls we have to log an error.
            //additionally for the above mentioned calls we MUST NOT call
            //the DebugServer !
            if ((getId() == MPO_GET_VERSION_ID) ||
                (getId() == MPO_INITIALIZED_ID) )//||
                //(MPO_FINALIZED_ID))  //why does this one not exist ???
            {
                return callNativeMPI();
            }
            else
            {
                //Log Error !
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": MPI call performed after MPI_Finalize ! ";

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_MPI_AFTER_FINALIZE,getSourceLocation());
                our_log->getMsg(msg);

                return callNativeMPI();
            }
        }//if finalize was called

        //Do Ping Pong with DebugServer, if it is enabled
        if (MPO_Space::isDebugServerEnabled())
        {
#ifdef MARMOT_USE_THREADSYNC
            /**
             * To speedup synchronization in the MPI_THREAD_MULTIPLE case we leave synchronization here.
             */
            if (MPO_Space::getProvidedThreadLevel() >= MPI_THREAD_MULTIPLE)
                MPO_Space::getSyncObject()->enterPMPI();
#endif

            // first we notify debug server
            our_channel.sendId(this->getFamilyId(),-1);
            our_channel.sendId(this->getId(),-1);
            our_channel.sendObject(*this,-1);

            // wait for reply from server
            int server_token = -1;
            our_channel.receiveId(server_token,-1);

#ifdef MARMOT_USE_THREADSYNC
            if (MPO_Space::getProvidedThreadLevel() >= MPI_THREAD_MULTIPLE)
                MPO_Space::getSyncObject()->leavePMPI();
#endif
        }
// TEST errorcodes
//       our_channel.sendId(this->checkPrerequisites());
// end TEST errorcodes

        //now we call the MPI call and perform the checks ...
        int rvalue = callMPIAndCheck();

        // return return value from MPI call
        return rvalue;
    }

    // callMPIAndCheck
    //========================================================================
    int MPO_Callobject::callMPIAndCheck(void)
    {
        int rvalue;

        /*What about the return code ....*/
        checkPrerequisites ();
        rvalue = callNativeMPI ();
        checkPost ();

        return rvalue;
    }

    // save
    //========================================================================
    std::ostream & MPO_Callobject::save(std::ostream & out) const
    {
        /** \todo @Tobias from Tobias: add a Enviromental variable that controls whether we save&restore this data or not and define global maxlength for filenames.*/
        if (my_SourceLine != 0)
        {
            out < my_SourceLine;
            for (int i = 0;(my_SourceFile[i] != '\0')&&(i<FILENAME_LEN-1);i++) //<64 forces max length
                out < my_SourceFile[i];
            out < '\0';
        }
        else
            out < 0;

        out < my_extra_id;

        out < my_thread_rank;

        return out;
    }

    // restore
    //========================================================================
    std::istream & MPO_Callobject::restore(std::istream & in)
    {
        /** \todo @Tobias from Tobias: add a Enviromental variable that controls whether we save&restore this data or not */
        in > my_SourceLine;

        if (my_SourceLine != 0)
        {
            int i = 0;
            if (my_SourceFileMem)
                delete (my_SourceFileMem);
            my_SourceFileMem = new char[FILENAME_LEN]; //using fixed length for performance purposes
            do
            {
                in > my_SourceFileMem[i];
            }
            while (my_SourceFileMem[i++] != '\0');
            my_SourceFile = my_SourceFileMem;
        }

        in  > my_extra_id;

        in  > my_thread_rank;

        return in;
    }

    // setSourceLocation
    //========================================================================
    void MPO_Callobject::setSourceLocation (int line,char *fileName)
    {
        my_SourceLine = line;
        my_SourceFile = fileName;

        //that should not happen but I will handle it here, just in case
        if (my_SourceFileMem)
            delete (my_SourceFileMem);
        my_SourceFileMem = NULL;
    }

    // setSourceLocation
    //========================================================================
    void MPO_Callobject::setSourceLocation (int line,char *fileName, unsigned long long extra_id)
    {
        my_SourceLine = line;
        my_SourceFile = fileName;
        my_extra_id = extra_id;

        //that should not happen but I will handle it here, just in case
        if (my_SourceFileMem)
            delete (my_SourceFileMem);
        my_SourceFileMem = NULL;
    }

    // setSourceLocation
    //========================================================================
    void MPO_Callobject::setSourceLocation (int line,const char *fileName)
    {
        my_SourceLine = line;
        int len = strlen (fileName);

        //that should not happen but I will handle it here, just in case
        if (my_SourceFileMem)
            delete (my_SourceFileMem);

        my_SourceFileMem = new char[len+1];
        strcpy (my_SourceFileMem,fileName);
        my_SourceFile = my_SourceFileMem;
    }

    // setSourceLocation
    //========================================================================
    void MPO_Callobject::setSourceLocation (int line,const char *fileName, unsigned long long extra_id)
    {
        my_SourceLine = line;
        my_extra_id = extra_id;
        int len = strlen (fileName);

        //that should not happen but I will handle it here, just in case
        if (my_SourceFileMem)
            delete (my_SourceFileMem);

        my_SourceFileMem = new char[len+1];
        strcpy (my_SourceFileMem,fileName);
        my_SourceFile = my_SourceFileMem;
    }

    // getSourceLocation
    //========================================================================
    MPO_SourceLocation MPO_Callobject::getSourceLocation (void)
    {
        return MPO_SourceLocation (my_SourceFile,my_SourceLine,my_extra_id);
    }

    // getThreadRank
    //========================================================================
    int MPO_Callobject::getThreadRank (void)
    {
    	return my_thread_rank;
    }

    // registerBuffer
    //========================================================================
    void MPO_Callobject::registerBuffer (void *buf, int count, MPO_Datatype type, MPO_Request request)
    {
        MPO_BufferInfoType info;
        MPI_Aint extent;
        int last_index;

        //create the info structure
        info.type = type;
        info.count = count;
        PMPI_Address(buf, &info.start_add);

        PMPI_Type_extent(type, &extent);
        last_index=count*extent; //TODO: is this correct ????
        PMPI_Address(((unsigned char*) buf)+last_index, &info.end_add );

        info.location = getSourceLocation();
        info.callname = getCallname();

        //store the info structure
        our_used_buffers.insert(std::make_pair(request,info));
    }

    // unregisterBuffer
    //========================================================================
    void MPO_Callobject::unregisterBuffer (MPO_Request request)
    {
        our_used_buffers.erase(request);
    }

    // checkBufferInUse
    //========================================================================
    int MPO_Callobject::checkBufferInUse (void *buf, int count, MPO_Datatype type)
    {
        if(our_errcodes.at(MARMOT_ERR_BUFFER_IN_USE))
        {
            //calculate address ranges
            MPI_Aint extent,extent_other;
            int last_index;
            MPI_Aint end_add,start_add;

            if ((buf != NULL) && (buf != MPI_BOTTOM))
            {
                PMPI_Address(buf, &start_add);
                PMPI_Type_extent(type, &extent);
                last_index=count*extent; //TODO: is this correct ????
                PMPI_Address(((unsigned char*) buf)+last_index, &end_add );

                //compare the used add to all registered bufs
                std::map<MPO_Request,MPO_Callobject::MPO_BufferInfoType>::iterator i = our_used_buffers.begin();

                for (;i != our_used_buffers.end();i++)
                {
                    if (    (end_add > i->second.start_add) &&
                            (start_add < i->second.end_add)    )
                    {
                        PMPI_Type_extent(i->second.type, &extent_other);

                        std::stringstream stream;
                        stream
                        << "ERROR: "
                        << getCallname()
                        << ": At least a part of the speciefied buffer is still in use!"
                        << " The buffer is still used by a call to "
                        << i->second.callname;

                        if (i->second.location.isValidLocation())
                        {
                            stream << " in ";
                            i->second.location.debugout(stream);
                        }

                        stream
                        << ". "
                        << std::endl
                        << std::endl
                        << "=== Detailed information ==="
                        << std::endl
                        << "This buffer is speciefied by:"
                        << std::endl
                        << "Starting address: "
                        << start_add
                        << std::endl
                        << "Count: "
                        << count
                        << std::endl
                        << "Extent of used datatype: "
                        << extent
                        << std::endl
                        << "Resulting end address (first non used byte): "
                        << end_add
                        << std::endl
                        << "Other buffer is speciefied by:"
                        << std::endl
                        << "Starting address: "
                        << i->second.start_add
                        << std::endl
                        << "Count: "
                        << i->second.count
                        << std::endl
                        << "Extent of used datatype: "
                        << extent_other
                        << std::endl
                        << "Resulting end address (first non used byte): "
                        << i->second.end_add
                        << std::endl
                        << std::endl;


                        our_request_traces->findAndPrintTraceNamed(i->first,stream,"request associated with the call that is already using the buffer");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_BUFFER_IN_USE,getSourceLocation());
                        our_log->getMsg(msg);
                        return MARMOT_ERR_BUFFER_IN_USE;
                    }
                }
            }
        }
        return MARMOT_SUCCESS;
    }

    // setPreData
    //========================================================================
    void MPO_Callobject::setPreData (MPO_Callobject *obj)
    {
        /*Default implementation does nothing*/
    }

    // clonePreData
    //========================================================================
    MPO_Callobject* MPO_Callobject::clonePreData (void)
    {
        /*Default returns NULL -> no need for cloned pre data*/
        return NULL;
    }

    // handlePreDataConstructor
    //========================================================================
    bool MPO_Callobject::handlePreDataConstructor (void)
    {
        //restore pre data if necessary
        if (our_store_pre_object_data)
        {
            my_persitent_restored = false;

            std::map<int, MPO_Callobject*>::iterator i =
                our_pre_object_data.find (my_thread_rank);
            if (i != our_pre_object_data.end())
            {
                if (i->second)
                {
                    //DEBUG
                      //std::cout << "Reading PreData for: " << getCallname()<<std::endl;
                    setPreData (i->second);
                    delete i->second;
                }

                our_pre_object_data.erase(i);
                my_persitent_restored = true;
                return true;
            }
        }

        return false;
    }

    // handlePreDataDestructor
    //========================================================================
    bool MPO_Callobject::handlePreDataDestructor (void)
    {
        //Store pre data if necessary
        if (our_store_pre_object_data &&
            !my_persitent_restored)
        {
            MPO_Callobject *p = clonePreData ();
            if (p)
            {
                //DEBUG
                  //std::cout << "Stored PreData for: " << getCallname()<<std::endl;
                std::pair<std::map<int, MPO_Callobject*>::iterator,bool> ret
                      = our_pre_object_data.insert(
                              std::make_pair (my_thread_rank,p));
                assert (ret.second); /*Maybe remove that ???*/
            }
            return true;
        }

        return false;
    }

    // SetPreDataRestore
    //========================================================================
    void MPO_Callobject::SetPreDataRestore (bool on)
    {
        our_store_pre_object_data = on;
    }
}                               // end of namespace MPO
