/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the base class MPO_Cartesian.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller
 *
 *  $Id: MPO_Cartesian.cc 857 2008-07-17 13:02:48Z hpchimml $
 */

#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Cartesian::MPO_Cartesian()
        : my_comm(MPI_COMM_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Cartesian::MPO_Cartesian(MPO_Comm comm)
        : my_comm(comm)
    {
    }


    // destructor
    //========================================================================
    MPO_Cartesian::~MPO_Cartesian()
    {
    }


    /* There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Cartesian::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Cartesian::checkPrerequisites()
    {
        // check correctness of parameters
        checkComm();
        checkCommCartesian();

        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Cartesian::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " :Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        return rvalue;
    }                           // end checkComm


    // checkCommCartesian
    //========================================================================
    int MPO_Cartesian::checkCommCartesian()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;
	if(!our_errcodes.at(MARMOT_ERR_COMM_NOT_CARTESIAN))
		return rvalue;

        int status = MPI_UNDEFINED;

        PMPI_Topo_test(my_comm, &status);

        // Check if communicator has cartesian structure.
        if ((status != MPI_CART) &&
            (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream 
                    << "ERROR: "
                    << getCallname()
                    << ": Communicator has no cartesian structure ! " 
                    << std::endl;
                rvalue = MARMOT_ERR_COMM_NOT_CARTESIAN;
                
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg); 
            }

        return rvalue;
    }                           // end checkCommCartesian


}                               // end of namespace MPO
