/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the base class MPO_Collective.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: MPO_Collective.cc 856 2008-07-17 11:59:35Z hpchimml $
 */


#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Collective::MPO_Collective()
        : my_comm(MPI_COMM_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Collective::MPO_Collective(MPO_Comm comm)
        : my_comm(comm)
    {
    }


    // destructor
    //========================================================================
    MPO_Collective::~MPO_Collective()
    {
    }


    /** There are some pure virtual methods, see mpo.h.
     */

 
    // getFamilyId
    //========================================================================
    int MPO_Collective::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }



    // checkPrerequisites
    //========================================================================
    int MPO_Collective::checkPrerequisites()
    {
        // check correctness of parameters
        checkComm();

        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Collective::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " :Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


    // checkCount
    //========================================================================
    int MPO_Collective::checkCount(int my_count,
                                   const string KIND_OF_COUNT)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if count is negative.
        if ((our_errcodes.at(MARMOT_ERR_COUNT_NEG))&&(my_count < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << " = "
                << my_count
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_COUNT_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if count is 0.
        if ((our_errcodes.at(MARMOT_WARN_COUNT_ZERO))&&(my_count == 0) && 
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: " 
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << "= 0!" 
                << std::endl;
            rvalue = MARMOT_WARN_COUNT_ZERO;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkCount
                             

    // checkCountArray
    //========================================================================
    int MPO_Collective::checkCountArray(int* my_counts, 
                                        const string KIND_OF_COUNT)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        int testrank = -1;
        PMPI_Comm_rank(my_comm, &testrank);

        for (int i = 0; i < size; i++)
        {
            // (1) Check if count is negative.
            if ((our_errcodes.at(MARMOT_ERR_COUNT_NEG))&&(my_counts[i] < 0) && 
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname() 
                    << ": On rank "
                    << testrank 
                    << ": "
                    << KIND_OF_COUNT
                    << "[" << i << "]"
                    << " = "
                    << my_counts[i]
                    << " less than 0 !" 
                    << std::endl;
                rvalue = MARMOT_ERR_COUNT_NEG;
                
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }

            // (2) Check if count is 0.
            if ((our_errcodes.at(MARMOT_WARN_COUNT_ZERO))&&(my_counts[i] == 0) && 
                (MSG_ERR_WARN <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "WARNING: " 
                    << getCallname() 
                    << ": On rank "
                    << testrank 
                    << ": "
                    << KIND_OF_COUNT
                    << "[" << i << "]"
                    << "= 0!" 
                    << std::endl;
                rvalue = MARMOT_WARN_COUNT_ZERO;
                
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }

        return rvalue;
    }                           // end checkCountArray


    // checkDatatypeSendRecv
    //========================================================================
    int MPO_Collective::checkDatatypeSendRecv(int my_count,
                                          MPO_Datatype my_datatype,
                                          enum KIND_OF_COUNT kind_of_count, 
                                          enum KIND_OF_COLLECTIVE_CALL
                                          kind_of_coll)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Datatype, enum KIND_OF_DATATYPE >::iterator pos;
        pos = our_datatypes.find(my_datatype);

        // (1) Datatype is not valid, i.e. not registered in
        // our MPO datatypes.
        if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_VALID))&&(pos == our_datatypes.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: " 
                << getCallname();
            switch (kind_of_count)
            {
                case (IS_COUNT):
                    stream
                        << ": Datatype";
                    break;
                case (IS_SENDCOUNT):
                    stream
                        << ": Sendtype";
                    break;
                case (IS_RECVCOUNT):
                    stream
                        << ": Recvtype";
                    break;
                default:
                    // do nothing
                    break;
            }
            stream 
                << " is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_TYPE_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }                       // end if (1)

        // (2) Datatype is registered in our MPO datatypes.
        // In this case, we have to check different violations 
        // in the switch statement.
        else if ((pos != our_datatypes.end()) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            
            std::stringstream stream;
            MPO_MSG_TYPE type = MSG_TYPE_ERROR;
            
                        
            // all the cases from enum KIND_OF_DATATYPE
            switch (pos->second)
            {
                case (DATATYPE_NULL):
			if(our_errcodes.at(MARMOT_ERR_TYPE_NOT_COMMITTED))
		    {
                    stream
                        << "ERROR: "
                        << getCallname();
                    switch (kind_of_count)
                    {
                        case (IS_COUNT):
                            stream
                                << ": Datatype";
                            break;
                        case (IS_SENDCOUNT):
                            stream
                                << ": Sendtype";
                            break;
                        case (IS_RECVCOUNT):
                            stream
                                << ": Recvtype";
                            break;
                        default:
                            // do nothing
                            break;
                    }
                    stream
                        << " is MPI_DATATYPE_NULL!" 
                        << std::endl;
                    rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;
		    }
                    break;
                case (DATATYPE_CREATED_NOT_COMMITTED):
		    if(our_errcodes.at(MARMOT_ERR_TYPE_NOT_COMMITTED))
		    {
                    stream
                        << "ERROR: "
                        << getCallname(); 
                    switch (kind_of_count)
                    {
                        case (IS_COUNT):
                            stream
                                << ": Datatype";
                            break;
                        case (IS_SENDCOUNT):
                            stream
                                << ": Sendtype";
                            break;
                        case (IS_RECVCOUNT):
                            stream
                                << ": Recvtype";
                            break;
                        default:
                            // do nothing
                            break;
                    }
                    stream
                        << " is not committed!" 
                        << std::endl;
                    rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;
		    }
                    break;
                case (DATATYPE_COMMITTED_C):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_C))&&(F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname();
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is C-Type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_C;
                    }
                    break;
                case (DATATYPE_COMMITTED_C_AND_FORTRAN):
                    // Do nothing, because everything is fine.
                    break;
                case (DATATYPE_COMMITTED_FORTRAN):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_F))&&(C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname();
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is Fortran-Type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_F;
                    }
                    break;
                case (DATATYPE_REDUCTION_C):
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_C))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                        (C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname();
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is for reduction functions (C)!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_REDUCTION_C;
                        type = MSG_TYPE_WARNING;
                    }
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_C))&&(F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname();
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is for reduction functions (C)!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_REDUCTION_C;
                    }
                    break;
                case (DATATYPE_REDUCTION_FORTRAN):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_F))&&(C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname();
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is for reduction functions "
                            << "(Fortran)!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_REDUCTION_F;
                    }
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_F))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                        (F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname();
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is for reduction functions "
                            << "(Fortran)!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_REDUCTION_F;
                        type = MSG_TYPE_WARNING;
                    }
                    break;
                case (DATATYPE_OPTIONAL_C):
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_C))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                        (C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname(); 
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is only optional!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_OPTIONAL_C;
                        type = MSG_TYPE_WARNING;
                    }
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_C))&&(F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname(); 
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is optional C-Type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_OPTIONAL_C;
                    }
                    break;
                case (DATATYPE_OPTIONAL_FORTRAN):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_F))&&(C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname();
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is optional Fortran-type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_OPTIONAL_F;
                    }
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_F))&&(MSG_ERR_WARN <= DEBUG_MODE) && 
                        (F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname(); 
                        switch (kind_of_count)
                        {
                            case (IS_COUNT):
                                stream
                                    << ": Datatype";
                                break;
                            case (IS_SENDCOUNT):
                                stream
                                    << ": Sendtype";
                                break;
                            case (IS_RECVCOUNT):
                                stream
                                    << ": Recvtype";
                                break;
                            default:
                                // do nothing
                                break;
                        }
                        stream
                            << " is only optional!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_OPTIONAL_F;
                        type = MSG_TYPE_WARNING;
                    }
                    break;
                case (DATATYPE_CONSTRUCT_DERIVED_TYPE):
		    if(our_errcodes.at(MARMOT_ERR_TYPE_CONSTR_DERIVED))
		    {
                    stream
                        << "ERROR: "
                        << getCallname(); 
                    switch (kind_of_count)
                    {
                        case (IS_COUNT):
                            stream
                                << ": Datatype";
                            break;
                        case (IS_SENDCOUNT):
                            stream
                                << ": Sendtype";
                            break;
                        case (IS_RECVCOUNT):
                            stream
                                << ": Recvtype";
                            break;
                        default:
                            // do nothing
                            break;
                    }
                    stream
                        << " is for constructing derived datatypes!"
                        << std::endl;
                    rvalue = MARMOT_ERR_TYPE_CONSTR_DERIVED;
		    }
                    break;
                default:
		    if(our_errcodes.at(MARMOT_ERR_MARMOT))
		    {
                    	stream
                        << "Check mapping of datatypes."
                        << std::endl
                        << "Check implementation of switch statement."
                        << std::endl;
                    	rvalue = MARMOT_ERR_MARMOT;
		    }
                    break;
            }                   // end switch
            
            if (rvalue != MARMOT_SUCCESS)
            {
            	our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            	
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }                       // end else if (2)

        return rvalue;
    }                           // end checkSendRecvDatatype


    // checkRoot
    //========================================================================
    int MPO_Collective::checkRoot(int root)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if rank of root < 0.
        if ((our_errcodes.at(MARMOT_ERR_RANK_NEG))&&(root != MPI_ANY_SOURCE) &&
            (root != MPI_PROC_NULL) && 
            (root < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank of source = "
                << root
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_RANK_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if rank of root is MPI_PROC_NULL.
        else if ((our_errcodes.at(MARMOT_NOTE_RANK_PROC_NULL))&&(root == MPI_PROC_NULL) && 
                 (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << getCallname() 
                << ": Source = MPI_PROC_NULL !" 
                << std::endl;
            rvalue = MARMOT_NOTE_RANK_PROC_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (3) Warn of race conditions if MPI_ANY_SOURCE is used.
        else if ((our_errcodes.at(MARMOT_WARN_RANK_ANY_SOURCE))&&(root == MPI_ANY_SOURCE) && 
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": Use of MPI_ANY_SOURCE may cause race conditions!"
                << std::endl;
            rvalue = MARMOT_WARN_RANK_ANY_SOURCE;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }                      

        return 0;
    }                           // end checkRoot


    // checkOp
    //========================================================================
    int MPO_Collective::checkOp(MPO_Op op)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        // (1) Check if op is null handle.
        if ((our_errcodes.at(MARMOT_ERR_OP_NULL))&&(op == MPI_OP_NULL) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Operation is MPI_OP_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_OP_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if op is valid, i.e. registered in our_ops.
        if ((our_errcodes.at(MARMOT_ERR_OP_NOT_VALID))&&(our_ops.find(op) == our_ops.end()) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Operation not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_OP_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkOp

}                               // end of namespace MPO
