/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the base class MPO_Grouphandler.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: MPO_GroupHandler.cc 860 2008-07-18 08:33:01Z hpchimml $
 */


#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Grouphandler::MPO_Grouphandler()
        : my_group(MPI_GROUP_NULL),
          my_group2(MPI_GROUP_NULL), 
          my_groupcount(-1)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Grouphandler::MPO_Grouphandler(MPO_Group group,
                                       MPO_Group group2,
                                       int groupcount)
        : my_group(group),
          my_group2(group2), 
          my_groupcount(groupcount)
    {
    }


    // destructor
    //========================================================================
    MPO_Grouphandler::~MPO_Grouphandler()
    {
    }


    /** There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Grouphandler::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Grouphandler::checkPrerequisites()
    {
        /** \todo @creator of this method: what about return values ? */
        
        // check correctness of parameters
        static int valid = my_groupcount;

        if (my_groupcount > 0)
        {
            checkGroup(valid);
        } 
        
        if (my_groupcount > 1)
        {
            checkGroup2(valid);
        }

        if ((valid == 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": no valid group!" 
                << std::endl;
                
            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group1");
            our_group_traces->findAndPrintTraceNamed(my_group2,stream,"group2");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_GROUP_NOT_VALID,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return 0;
    }


    // checkGroup
    //========================================================================
    int MPO_Grouphandler::checkGroup(int valid)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        // (1) Check if group is null handle.
        if ((our_errcodes.at(MARMOT_WARN_GROUP_NULL))&&(my_group == MPI_GROUP_NULL) && 
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname() 
                << ": Group1 is MPI_GROUP_NULL!" 
                << std::endl;
            rvalue = MARMOT_WARN_GROUP_NULL;
            
            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group1");
            our_group_traces->findAndPrintTraceNamed(my_group2,stream,"group2");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if group is valid, i.e. registered in our_groups.
        if ((our_errcodes.at(MARMOT_ERR_GROUP_NOT_VALID))&&(our_groups.find(my_group) == our_groups.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Group1 not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_GROUP_NOT_VALID;
            
            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group1");
            our_group_traces->findAndPrintTraceNamed(my_group2,stream,"group2");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }
        valid--;

        return rvalue;
    }                           // end checkGroup


    // checkGroup2
    //========================================================================
    int MPO_Grouphandler::checkGroup2(int valid)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        // (1) Check if group is null handle.
        if ((our_errcodes.at(MARMOT_WARN_GROUP_NULL))&&(my_group2 == MPI_GROUP_NULL) && 
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname() 
                << ": Group2 is MPI_GROUP_NULL!" 
                << std::endl;
            rvalue = MARMOT_WARN_GROUP_NULL;
            
            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group1");
            our_group_traces->findAndPrintTraceNamed(my_group2,stream,"group2");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if group is valid, i.e. registered in our_groups.
        if ((our_errcodes.at(MARMOT_ERR_GROUP_NOT_VALID))&&(our_groups.find(my_group2) == our_groups.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Group2 not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_GROUP_NOT_VALID;
            
            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group1");
            our_group_traces->findAndPrintTraceNamed(my_group2,stream,"group2");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }
        valid--;

        return rvalue;
    }                           // end checkGroup2

}                               // end of namespace MPO
