/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the base class MPO_Probes. 
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: MPO_Probes.cc 860 2008-07-18 08:33:01Z hpchimml $
 */


#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include "mpo_TCheck_Races.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Probes::MPO_Probes()
        : my_source(MPI_PROC_NULL),
          my_tag(MPI_ANY_TAG),
          my_comm(MPI_COMM_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Probes::MPO_Probes(int source,
                           int tag,
                           MPO_Comm comm)
        : my_source(source),
          my_tag(tag),
          my_comm(comm)
    {
    }


    // destructor
    //========================================================================
    MPO_Probes::~MPO_Probes()
    {
    }


    /** There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Probes::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Probes::checkPrerequisites()
    {
        // check correctness of parameters
        checkRankOfSource();
        checkRankOfSourceTooBig();
        checkTag(MSG_TAG);
        checkComm();

        return 0;
    }


    // checkRankOfSource
    //========================================================================
    int MPO_Probes::checkRankOfSource()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if rank of source < 0.
        if ((our_errcodes.at(MARMOT_ERR_RANK_NEG))&&(my_source != MPI_ANY_SOURCE) &&
            (my_source != MPI_PROC_NULL) && 
            (my_source < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank of source = "
                << my_source
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_RANK_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if rank of source is MPI_PROC_NULL.
        else if ((our_errcodes.at(MARMOT_NOTE_RANK_PROC_NULL))&&(my_source == MPI_PROC_NULL) && 
                 (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << getCallname() 
                << ": Source = MPI_PROC_NULL !" 
                << std::endl;
            rvalue = MARMOT_NOTE_RANK_PROC_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (3) Warn of race conditions if MPI_ANY_SOURCE is used.
        else if ((our_errcodes.at(MARMOT_WARN_RANK_ANY_SOURCE))&&(my_source == MPI_ANY_SOURCE) && 
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": Use of MPI_ANY_SOURCE may cause race conditions!"
                << std::endl;
            rvalue = MARMOT_WARN_RANK_ANY_SOURCE;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }                      

        return rvalue;
    }                           // end checkRankOfSource


    // checkRankOfSourceTooBig
    //========================================================================
     int MPO_Probes::checkRankOfSourceTooBig()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;        
	if(!our_errcodes.at(MARMOT_ERR_RANK_TOO_BIG))
		return rvalue;

        // Check if rank of source is out of communicator.     
        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        if ((my_source >= size) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank of source is out of communicator: rank = "
                << my_source 
                << ", size of communicator = " 
                << size 
                << std::endl;
            rvalue = MARMOT_ERR_RANK_TOO_BIG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }


    // checkTag
    //========================================================================
    int MPO_Probes::checkTag(const string KIND_OF_TAG)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if tag is < 0.
        if ((our_errcodes.at(MARMOT_ERR_TAG_NEG))&&(my_tag != MPI_ANY_TAG) && 
            (my_tag < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": "
                << KIND_OF_TAG
                << " = "
                << my_tag
                << " less than 0!" 
                << std::endl;
            rvalue = MARMOT_ERR_TAG_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if tag is within guaranteed range.
        else if ((our_errcodes.at(MARMOT_ERR_TAG_TOO_BIG))&&(my_tag != MPI_ANY_TAG) && 
                 (my_tag > 32767) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            int ub = 0;
            int flag = 1;
            PMPI_Attr_get(MPI_COMM_WORLD, MPI_TAG_UB, &ub, &flag);

            if ((my_tag > ub) && 
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname() 
                    << ": "
                    << KIND_OF_TAG
                    << " = "
                    << my_tag 
                    << " is larger than allowed!" 
                    << std::endl
                    << "THIS implementation allows tags up to "
                    << ub 
                    << "."
                    << std::endl;
                rvalue = MARMOT_ERR_TAG_TOO_BIG;
                
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }

            else if ((our_errcodes.at(MARMOT_WARN_TAG_TOO_BIG))&&(my_tag <= ub) && 
                     (MSG_ERR_WARN <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "WARNING: "
                    << getCallname()
                    << ": "
                    << KIND_OF_TAG
                    << " = "
                    << my_tag
                    << " greater than 32767 ! MPI only guarantees tags up to this." 
                    << std::endl
                    << "THIS implementation allows tags up to "
                    << ub 
                    << "."
                    << std::endl;
                rvalue = MARMOT_WARN_TAG_TOO_BIG;
                
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }                       // end else if (2)

        return rvalue;
    }                           // end checkTag


    // checkComm
    //========================================================================
    int MPO_Probes::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " :Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm
    
    
#ifdef MARMOT_USE_THREADSYNC
	    
    // registerProbe
    //========================================================================
    void MPO_Probes::registerProbe (int source, int tag, MPI_Comm comm, int sourceLine, std::string sourceName, int threadNum, int callId, std::string callName)
    {
    	//create the info
		MPO_ProbeInfo info;
		info.source 	= source;
		info.tag    	= tag;
        info.comm       = comm;  
        info.sourceLine = sourceLine;
        info.sourceName = sourceName;
        info.threadNum  = threadNum;
        info.callId		= callId;
        info.callName   = callName;

#ifdef MARMOT_TCHECK
        //enter values for the inidizes
        info.index= our_next_probe_array_index;
        our_next_probe_array_index=(our_next_probe_array_index+1)%TCHECK_NUM_PROBE_SLOTS;
#endif

        //add the info to the list
        our_succesful_probes.push_back(info);
        
#ifdef MARMOT_TCHECK
        //Stop Synchronisation and make a read of a conflict Variable
        MPO_Space::getSyncObject()->enterPMPI();
        MPO_TCheck_Races::getInstance()->writeIntForProbe(info.index);
        MPO_Space::getSyncObject()->leavePMPI();
#endif
    }
	    
#endif /*MARMOT_USE_THREADSYNC*/

}                               // end of namespace MPO
