/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the abstract base class MPO_Pt2Pt.
 *
 *  This abstract base class represents calls for point-to-point
 *  communication.
 * 
 *  @author Bettina Krammer
 *
 *  $Id: MPO_Pt2Pt.cc 849 2008-07-16 16:06:49Z hpchimml $
 */


#include "mpo.h"
#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "mpo_SourceLocation.h"
#include "MPO_LogFile.h"
#include "mpo_TCheck_Races.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Pt2Pt::MPO_Pt2Pt()
    {
    }


    // destructor
    //========================================================================
    MPO_Pt2Pt::~MPO_Pt2Pt()
    {
    }


    /** There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Pt2Pt::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }

#ifdef MARMOT_USE_THREADSYNC

	// testIfProbed
    //========================================================================
	int MPO_Pt2Pt::testIfProbed (int source, int tag, MPI_Comm comm, int sourceLine, std::string sourceName, int threadNum, int callId, std::string callName)
	{
		int rvalue = MARMOT_SUCCESS;
			
		if(our_errcodes.at(MARMOT_WARN_PROBE_INVALIDATED))
		{
		
		for (std::list<MPO_ProbeInfo>::iterator i = our_succesful_probes.begin(); i != our_succesful_probes.end();i++)
		{
			if ( (i->tag != tag) &&
				 (tag != MPI_ANY_TAG) &&
				 (i->tag != MPI_ANY_TAG) )
				continue;
				
			if ( (i->source != source) &&
				 (i->source != MPI_ANY_SOURCE) &&
				 (source != MPI_ANY_SOURCE) )
				continue;
				
			if (i->comm != comm)
				continue;  
			
#ifdef MARMOT_TCHECK
			//Stop Synchronisation and make a write of a conflict Variable
	        MPO_Space::getSyncObject()->enterPMPI();
	        MPO_TCheck_Races::getInstance()->writeIntForProbe(i->index);
	        MPO_Space::getSyncObject()->leavePMPI();
#endif
			
			//recieve of a msg already probed by another thread -> Warn
			if (i->threadNum != threadNum)
			{					 
				//a) log first direction
				std::stringstream stream;
				stream
			    	<< "WARNING: "
			        << "call: "
			        << callName
			        << " invalidated a probe (tag="
			        << tag
			        << ", source="
			        << source
			        << ")."
			        << std::endl
			        << "Probe made by thread "
			        << i->threadNum
			        << " call "
			        << i->callName.c_str()
			        << " in line "
			        << i->sourceLine
			        << " of file "
			        << i->sourceName;
	   
				MPO_String text;
				text.assign(&stream);      
				MPO_MsgObject msg1 (MSG_TYPE_WARNING,0,text, callName ,(MPO_CALL_ID) callId,MARMOT_WARN_PROBE_INVALIDATED,MPO_SourceLocation(sourceName,sourceLine)); 
				
				//b) log second direction
				stream.str("");
				stream.clear();
				stream
			    	<< "WARNING: "
			        << "callname: "
			        << i->callName
			        << " was invalidated by a recieve of another thread (tag="
			        << tag
			        << ", source="
			        << source
			        << ")."
			        << std::endl
			        << "Recieve made by thread "
			        << threadNum
			        << " call "
			        << callName.c_str()
			        << " in line "
			        << sourceLine
			        << " of file "
			        << sourceName;
	   
				text.assign(&stream);      
				MPO_MsgObject msg2 (MSG_TYPE_WARNING,0,text, i->callName ,(MPO_CALL_ID) i->callId,MARMOT_WARN_PROBE_INVALIDATED,MPO_SourceLocation(i->sourceName,i->sourceLine));
				msg2.setThreadRank(i->threadNum);
				
				our_log->getMsg(msg1);
				our_log->getMsg(msg2);
				
				rvalue = MARMOT_WARN_PROBE_INVALIDATED;
			}
			
			//remove the registration
			our_succesful_probes.erase(i);
				
			//restart the iterator
			i = our_succesful_probes.begin();				
		} /*for*/
		}
		return rvalue;	
	} /*testIfProbed*/ 
	
#endif /*MARMOT_USE_THREADSYNC*/

}                               // end of namespace MPO
