/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This file implements the MPO_ResourceTrace class.
 * @author Tobias Hilbrich
 * @date 28.11.2006
 * $Id: $
 */

#include "MPO_ResourceTrace.h"

namespace MPO
{

    //Method MPO_ResourceTrace
    //========================================================================
    MPO_ResourceTrace::MPO_ResourceTrace (std::string resource_name,bool has_usage,bool has_free,bool has_commit,bool has_creation,bool has_dependencies) :
        my_free_time(0.0),
        my_usage_flag(has_usage),
        my_free_flag(has_free),
        my_commit_flag(has_commit),
        my_creation_flag(has_creation),
        my_dependencies_flag(has_dependencies),
        my_resource_name(resource_name)
    {
        //store the creation time
        my_creation_time = PMPI_Wtime();
    } //end of method MPO_ResourceTrace

    //Method ~MPO_ResourceTrace
    //========================================================================
    MPO_ResourceTrace::~MPO_ResourceTrace (void)
    {

    } //end of method ~MPO_ResourceTrace

    //Method addUsage
    //========================================================================
    void MPO_ResourceTrace::addUsage (MPO_SourceLocation location)
    {
        while (my_usage_locations.size() >=  NUM_USAGE_CALLS_TRACED)
        {
            my_usage_locations.pop_back();
        }

        my_usage_locations.push_front(location);

    } //end of method addUsage

    //Method setFree
    //========================================================================
    void MPO_ResourceTrace::setFree (MPO_SourceLocation location)
    {
        my_free_location = location;
        my_free_time = PMPI_Wtime();
    } //end of method setFree

    //Method isFreed
    //========================================================================
    bool MPO_ResourceTrace::isFreed (void)
    {
        return my_free_location.isValidLocation();
    }//end of method isFreed

    //Method setCommit
    //========================================================================
    void MPO_ResourceTrace::setCommit (MPO_SourceLocation location)
    {
        my_commit_location = location;
    } //end of method setCommit

    //Method setCreation
    //========================================================================
    void MPO_ResourceTrace::setCreation (MPO_SourceLocation location)
    {
        my_creation_location = location;
    } //end of method setCreation

    //Method getCreation
    //========================================================================
    MPO_SourceLocation MPO_ResourceTrace::getCreation ()
    {
        return my_creation_location;
    }

    //Method addDependency
    //========================================================================
    void MPO_ResourceTrace::addDependency (MPO_SourceLocation location)
    {
        my_dependency_locations.push_back (location);
    } //end of method addDependency

    //Method printTrace
    //========================================================================
    std::ostream& MPO_ResourceTrace::printTrace (std::ostream & out)
    {

        //Intro
        out << std::endl
            << "Information for Resource of type "
            << my_resource_name
            << ":"
            << std::endl;

        //Creation
        if (my_creation_flag)
        {
            out	<< "created at ";
            my_creation_location.debugout(out);
            out << std::endl;
        }

        //Commit
        if (my_commit_flag)
        {
            if (my_commit_location.isValidLocation())
            {
                out	<< "commited at ";
                my_commit_location.debugout(out);
            }
            else
            {
                out << "not yet commited.";
            }

            out << std::endl;
        }

        //Dependencies
        if ( my_dependencies_flag &&
             (my_dependency_locations.size() > 0))
        {
            out	<< "depends upon the following resources: "
                << std::endl;

            std::deque <MPO_SourceLocation >::iterator i;

            for (i = my_dependency_locations.begin(); i != my_dependency_locations.end(); i++)
            {
                i->debugout(out);
                out << std::endl;
            }
        }

        //Usage
        if ( my_usage_flag &&
             (my_usage_locations.size() > 0))
        {
            out	<< "Last "
                << NUM_USAGE_CALLS_TRACED
                << " calls that used this resource (most recent first):"
                << std::endl;

            std::deque <MPO_SourceLocation >::iterator i;

            for (i = my_usage_locations.begin(); i != my_usage_locations.end(); i++)
            {
                i->debugout(out);
                out << std::endl;
            }
        }

        //Free
        if (my_free_flag)
        {
            if (my_free_location.isValidLocation())
            {
                out	<< "freed at ";
                my_free_location.debugout(out);
            }
            else
            {
                out << "not yet freed.";
            }

            out << std::endl;
        }

        return out;
    } //end of method printTrace

    //Method getCreationTime
    //========================================================================
    double MPO_ResourceTrace::getCreationTime (void)
    {
        return my_creation_time;
    }

    //Method getFreeTime
    //========================================================================
    double MPO_ResourceTrace::getFreeTime (void)
    {
        return my_free_time;
    }

    //Method getCreationTime
    //========================================================================
    bool MPO_ResourceTrace::wasUsedAtTime (double timeToCheck)
    {
        if (timeToCheck >= my_creation_time)
        {
            if (!my_free_flag)
                return true;

            if (!my_free_location.isValidLocation())
                return true;

            if (timeToCheck < my_free_time)
                return true;
        }

        return false;
    }

} //end of namespace MPO
