/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
#include <cstring>
#include <cstdlib>
#include "NetUtil.h"
#include <assert.h>
#include <stdarg.h>
#include <iostream>
#include <sstream>

using namespace std;

// Link to "Ws2_32.lib" on windows if not forbidden by NET_DONT_LINK_WINSOCK
#ifdef _WIN32
#ifndef NET_DONT_LINK_WINSOCK
#pragma comment(lib, "Ws2_32.lib")
#endif
#endif

// -------- helper for debug macro ----------
std::string NetUtil::DBG_FILE="";
int NetUtil::DBG_LINE=0;

void NetUtil::myHelperDebug(const char *msg, ...)
{
  static char linestr[10];
  static char usermes[2001];
  memset(usermes, 0, sizeof(usermes));
  std::string finalmes, hstr;
  // get rid of path in DBG_FILE
  string::size_type pos = DBG_FILE.find_last_of("/\\");
  if (pos != string::npos)
  {
    hstr = DBG_FILE.substr(pos+1);
    DBG_FILE = hstr;
  }

  va_list ap;
  va_start(ap, msg);
  if (vsnprintf(usermes, sizeof(usermes)-1, msg, ap) == -1)
  {
    // compose a message without the user message
    finalmes = DBG_FILE + ":";
    NetUtil::itoa(DBG_LINE, linestr, 10);
    finalmes.append(linestr);
    finalmes.append(" - Undefined Error!");
  } else {
    // compose with usermessage
    finalmes = DBG_FILE + ":";
    NetUtil::itoa(DBG_LINE, linestr, 10);
    finalmes.append(linestr);
    finalmes.append(" - ");
    finalmes.append(usermes);
    //qDebug("%s:%i - %s", (fi.fileName()).latin1(), DBG_LINE, usermes);
  }
  // output
#ifdef _WIN32
  OutputDebugString(finalmes.c_str());
#else
  std::cerr << finalmes << std::endl;
#endif
  va_end(ap);
}
// ------------------------------------------

NetUtil *NetUtil::m_instance = NULL;

NetUtil::NetUtil()
{
  // Initialize winsock
#ifdef _WIN32
  std::string errstr;
  int err = WSAStartup(MAKEWORD(2, 2), &m_WSAData);
	if (err != 0)
  {
    errstr = "NetUtil - WSAStartup() failed - ";
    switch (err)
    {
      case WSASYSNOTREADY:
        errstr.append("Network subsystem is not ready for network communication.");
        break;
      case WSAVERNOTSUPPORTED:
        errstr.append("The version of Windows Sockets support requested is not provided by this particular Windows Sockets implementation.");
        break;
      case WSAEINPROGRESS:
        errstr.append("A blocking Windows Sockets 1.1 operation is in progress.");
        break;
      case WSAEPROCLIM:
        errstr.append("Limit on the number of tasks supported by the Windows Sockets implementation has been reached.");
        break;
    }
    cerr << errstr << endl;
    DBGOUT(errstr.c_str());
  }
#endif
}

NetUtil *NetUtil::instance(void)
{
  if (m_instance == NULL) m_instance = new NetUtil;
  assert(m_instance != NULL);
  return m_instance;
}

void NetUtil::initialize()
{
  NetUtil::instance();
}

bool NetUtil::isipv4(const std::string &str)
{
  int dots = 0;
  for (size_t i = 0; i < str.size(); i++)
  {
    if (str[i] == '.')
    {
			dots++;
    } else {
		  if (!isdigit(str[i])) return false;
    }
	}
	if (dots != 3)
		return false;
	return true;
}

bool NetUtil::resolve(const char *hostname, std::vector<std::string> &output)
{
  std::string hname;
  struct in_addr addr;
  struct hostent *host;
  int i;

  NetUtil::initialize();

  output.clear();
  if (hostname == NULL)
  {
    char tmpbuf[80];
    if (gethostname(tmpbuf, sizeof(tmpbuf)) == 0)
      hname = tmpbuf;
  } else {
    hname = hostname;
  }

  // test whether IP-Address in dotted notation
  addr.s_addr = inet_addr(hname.c_str());
  if (addr.s_addr == INADDR_NONE)
  {
    // it's a hostname (passing NULL here is the same as calling gethostname())
    host = gethostbyname(hname.c_str());
    if (!host)
    {
      //herror("gethostbyname() failed");
      return false;
    }
    // add all to a list
    for (i=0; host->h_addr_list[i]!=NULL; i++)
    {
      addr = *(struct in_addr *)host->h_addr_list[i];
      output.push_back(inet_ntoa(addr));
      //printf("%s -> %s\n", hname.c_str(), inet_ntoa(addr));
    }
  } else {
    // so it's an IP-address
    host = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
    if (!host)
    {
      //herror("gethostbyaddr() failed");
      return false;
    }
    // output the "real" name first
    output.push_back(host->h_name);
    //printf("%s -> %s\n", inet_ntoa(addr), host->h_name);
    // then all the alias
    for (i=0; host->h_aliases[i]!=NULL; i++)
    {
      output.push_back(host->h_aliases[i]);
      //printf("%s -> %s\n", inet_ntoa(addr), host->h_aliases[i]);
    }
  }
  return true;
}

net_hostent_ptr NetUtil::getHostEnt(const std::string &hostname)
{
  struct in_addr addr;

  // test whether IP-Address in dotted notation
  addr.s_addr = inet_addr(hostname.c_str());
  if (addr.s_addr == INADDR_NONE)
  {
    // it's a hostname
    return (gethostbyname(hostname.c_str()));
  }
  // so it's an IP-address
  return (gethostbyaddr((char *)&addr, sizeof(addr), AF_INET));
}

void NetUtil::getTime(struct timeval *p)
{
  assert(p != NULL);

#ifdef _WIN32
	FILETIME ft; // Contains a 64-bit value representing the number of 100-nanosecond intervals since January 1, 1601 (UTC).
	GetSystemTimeAsFileTime(&ft);
  unsigned __int64 tt;
	memcpy(&tt, &ft, sizeof(tt));
	tt /= 10; // make it usecs
	p->tv_sec  = (long)tt / 1000000;
	p->tv_usec = (long)tt % 1000000;
#else
	gettimeofday(p, NULL);
#endif
}


const char * NetUtil::getStrError(int code)
{
#ifdef _WIN32
  // do some mapping on windows
  static char hbuf[64];
	switch (code)
	{
    case 10004: return "Interrupted function call.";
    case 10013: return "Permission denied.";
    case 10014: return "Bad address.";
    case 10022: return "Invalid argument.";
    case 10024: return "Too many open files.";
    case 10035: return "Resource temporarily unavailable.";
    case 10036: return "Operation now in progress.";
    case 10037: return "Operation already in progress.";
    case 10038: return "Socket operation on nonsocket.";
    case 10039: return "Destination address required.";
    case 10040: return "Message too long.";
    case 10041: return "Protocol wrong type for socket.";
    case 10042: return "Bad protocol option.";
    case 10043: return "Protocol not supported.";
    case 10044: return "Socket type not supported.";
    case 10045: return "Operation not supported.";
    case 10046: return "Protocol family not supported.";
    case 10047: return "Address family not supported by protocol family.";
    case 10048: return "Address already in use.";
    case 10049: return "Cannot assign requested address.";
    case 10050: return "Network is down.";
    case 10051: return "Network is unreachable.";
    case 10052: return "Network dropped connection on reset.";
    case 10053: return "Software caused connection abort.";
    case 10054: return "Connection reset by peer.";
    case 10055: return "No buffer space available.";
    case 10056: return "Socket is already connected.";
    case 10057: return "Socket is not connected.";
    case 10058: return "Cannot send after socket shutdown.";
    case 10060: return "Connection timed out.";
    case 10061: return "Connection refused.";
    case 10064: return "Host is down.";
    case 10065: return "No route to host.";
    case 10067: return "Too many processes.";
    case 10091: return "Network subsystem is unavailable.";
    case 10092: return "Winsock.dll version out of range.";
    case 10093: return "Successful WSAStartup not yet performed.";
    case 10101: return "Graceful shutdown in progress.";
    case 10109: return "Class type not found.";
    case 11001: return "Host not found.";
    case 11002: return "Nonauthoritative host not found.";
    case 11003: return "This is a nonrecoverable error.";
    case 11004: return "Valid name, no data record of requested type.";
	default:
		break;
	}
	sprintf(hbuf, "Winsock error code: %d", code);
	return hbuf;
#else
  // not windows
  return strerror(code);
#endif
}

string NetUtil::composeErrorStr(int err_code, const char *usertxt)
{
  string mestxt;
  stringstream sstr;
  string codestr;
  sstr << err_code;
  sstr >> codestr;
  mestxt = "NetError: ";
  mestxt.append(usertxt);
  mestxt.append("  (Code: ");
  mestxt.append(codestr);
  mestxt.append(" - ");
  mestxt.append(NetUtil::getStrError(err_code));
  mestxt.append(")\n");
  return mestxt;
}

int NetUtil::getLastError(void)
{
#if defined(_WIN32)
	return WSAGetLastError();
#else
	return errno;
#endif
}

void NetUtil::strreverse(char *begin, char *end)
{
	char aux;
	while (end>begin)
		aux=*end, *end--=*begin, *begin++=aux;
}

void NetUtil::itoa(int value, char* str, int base)
{
	static char num[]="0123456789abcdefghijklmnopqrstuvwxyz";
	char *wstr=str;
	int   sign;
  div_t res;

  // Validate base
	if (base<2 || base>35) { *wstr='\0'; return; }
	// Take care of sign
	if ((sign=value) < 0) value = -value;
	// Conversion. Number is reversed.
	do
  {
    res = div(value, base);
    *wstr++ = num[res.rem];
	} while ((value=res.quot));
	if (sign<0) *wstr++='-';
	*wstr='\0';
	// Reverse string
	strreverse(str, wstr-1);
}

#ifdef _SX
unsigned long NetUtil::sx_workaround_inetaddr(const char *cp)
{
  unsigned long retval;
  unsigned int len = strlen(cp);
  char *buf = new char[len+1];
  strcpy(buf, cp);
  retval = inet_addr(buf);
  delete [] buf;
  return retval;
}
#endif
