/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#include <cstring>
#include <iostream>
#include <assert.h>
#include "NetUtil.h"
#include "SocketLib.h"

using namespace std;

// disable depracation warning caused by sprintf (we know it's safe because no user defined strings are used)
#ifdef _MSC_VER
#pragma warning(disable:4996)
#endif


CSocket::CSocket() : m_socket(NET_INVALID_SOCKET), m_error_code(0), m_ok(false)
{
  memset(&m_sockaddr, 0, sizeof(m_sockaddr));
  // do WSAStartup if not already done
  NetUtil::initialize();
}

void CSocket::handleError(const char *usermsg)
{
  m_error_code = NetUtil::getLastError();
  m_error = NetUtil::composeErrorStr(m_error_code, usermsg);
  DBGOUT(m_error.c_str());
}

/**
 * Close the socket
 */
void CSocket::close()
{
	if (m_socket != NET_INVALID_SOCKET)
  {
/* -----------
#ifdef _WIN32
    if (shutdown(m_socket, SD_BOTH) == NET_SOCKET_ERROR)
#else
    if (shutdown(m_socket, SHUT_RDWR) == NET_SOCKET_ERROR)
#endif
    {
      m_ok = false; m_error = NetUtil::composeErrorStr("CSocket - shutdown() failed");
      DBGOUT(m_error.c_str());
    }
----------- */
    if (net_closesocket(m_socket) == NET_SOCKET_ERROR)
    {
      handleError("CSocket - closesocket() failed");
    } else {
      m_socket = NET_INVALID_SOCKET;
    }
    m_ok = false;
  }
}

int CSocket::available(void)
{
  int retval;
#if defined(_WIN32)
	int rc = ioctlsocket(m_socket, FIONREAD, reinterpret_cast<unsigned long *>(&retval));
#else
	int rc = ::ioctl(m_socket, FIONREAD, &retval);
#endif
  if (rc != 0)
  {
    m_ok = false;  handleError("CSocket - available() failed");
    return -1;
  }
  return retval;
}

bool CSocket::setBlocking(bool state)
{
  std::string err_str("NetError: CSocket::setBlocking() failed");
  bool retval = false;
  int retcode;
#ifdef _WIN32
  unsigned long arg = 1;
  if (!state)
  {
    retcode = ioctlsocket(m_socket, FIONBIO, &arg);
  } else {
    arg = 0;
    retcode = ioctlsocket(m_socket, FIONBIO, &arg);
  }
  if (retcode == NET_SOCKET_ERROR)
  {
    m_ok = false; handleError(err_str.c_str());
    return false;
  }
#else
	int oldflags = fcntl(m_socket, F_GETFL);
	if (oldflags == -1)
  {
    m_ok = false; handleError(err_str.c_str());
    return false;
  }
  if (!state) { oldflags |= O_NONBLOCK; }
  else        { oldflags &= ~O_NONBLOCK; }

  retcode = fcntl(m_socket, F_SETFL, oldflags);
	if (retcode == -1)
  {
    m_ok = false; handleError(err_str.c_str());
    return false;
  }
#endif
  // here everything ok
  retval = true;
  return retval;
}

bool CSocket::setLinger(bool do_linger, int timeout)
{
  struct linger sockl;
  if (do_linger)
  {
    sockl.l_onoff  = 1;
    sockl.l_linger = timeout;
  } else {
    sockl.l_onoff  = 0;
    sockl.l_linger = timeout; // will be ignored anyway
  }
  if (setsockopt(m_socket, SOL_SOCKET, SO_LINGER, (char *)&sockl, sizeof(sockl)) == NET_SOCKET_ERROR)
  {
    char buf[128];
    sprintf(buf, "CSocket - setsockopt() failed in setLinger(%d, %d)", (int)do_linger, timeout);
    m_ok = false; handleError(buf);
    return false;
  }
  return true;
}

bool CSocket::getLinger(bool &is_lingering, int &set_timeout)
{
  struct linger sockl;
  net_socklen_t opt_len = sizeof(sockl);
  memset(&sockl, 0, opt_len);
  if (getsockopt(m_socket, SOL_SOCKET, SO_LINGER, (char *)&sockl, &opt_len) == NET_SOCKET_ERROR)
  {
    m_ok = false; handleError("CSocket - getsockopt() failed in getLinger()");
    return false;
  }
  if (sockl.l_onoff)
    is_lingering = true;
  else
    is_lingering = false;
  set_timeout  = sockl.l_linger;
  return true;
}

bool CSocket::checkSocketReadable(unsigned int timeout_sec, unsigned int timeout_usec)
{
  int rc;
  fd_set readfds; //writefds, exceptfds;
  struct timeval timeout;

  FD_ZERO(&readfds);
  FD_SET(m_socket, &readfds);
  timeout.tv_sec  = timeout_sec;
  timeout.tv_usec = timeout_usec;
  rc = select(m_socket+1, &readfds, NULL, NULL, &timeout);
  if (rc > 0)
  {
    // socket read will not block
    return true;
  } else if (rc == 0) {
    // time limit expired
    return false;
  }
  // (rc == NET_SOCKET_ERROR)
  char buf[128];
  sprintf(buf, "CSocket - select() in checkSocketReadable(%d, %d) failed", timeout_sec, timeout_usec);
  m_ok = false; handleError(buf);
  return false;
}

bool CSocket::checkSocketWritable(unsigned int timeout_sec, unsigned int timeout_usec)
{
  int rc;
  fd_set writefds;
  struct timeval timeout;

  FD_ZERO(&writefds);
  FD_SET(m_socket, &writefds);
  timeout.tv_sec  = timeout_sec;
  timeout.tv_usec = timeout_usec;
  rc = select(m_socket+1, NULL, &writefds, NULL, &timeout);
  if (rc > 0)
  {
    // socket write will not block
    return true;
  } else if (rc == 0) {
    // timeout
    return false;
  }
  // (rc == NET_SOCKET_ERROR)
  char buf[128];
  sprintf(buf, "CSocket - select() in checkSocketWritable(%d, %d) failed", timeout_sec, timeout_usec);
  m_ok = false; handleError(buf);
  return false;
}

bool CSocket::checkSocketException(unsigned int timeout_sec, unsigned int timeout_usec)
{
  int rc;
  fd_set exceptfds;
  struct timeval timeout;

  FD_ZERO(&exceptfds);
  FD_SET(m_socket, &exceptfds);
  timeout.tv_sec  = timeout_sec;
  timeout.tv_usec = timeout_usec;
  rc = select(m_socket+1, NULL, NULL, &exceptfds, &timeout);
  if (rc > 0)
  {
    // socket exception
    return true;
  } else if (rc == 0) {
    // timeout
    return false;
  }
  // (rc == NET_SOCKET_ERROR)
  char buf[128];
  sprintf(buf, "CSocket - select() in checkSocketException(%d, %d) failed", timeout_sec, timeout_usec);
  m_ok = false; handleError(buf);
  return false;
}

bool CSocket::blockTillReadable(void)
{
  int rc;
  fd_set readfds; //writefds, exceptfds;

  FD_ZERO(&readfds);
  FD_SET(m_socket, &readfds);
  rc = select(m_socket+1, &readfds, NULL, NULL, NULL);
  if (rc > 0)
  {
    // socket read will not block
    return true;
  } else if (rc == 0) {
    // time limit expired (should not happen here)
    return false;
  }
  // (rc == NET_SOCKET_ERROR)
  char buf[128];
  sprintf(buf, "CSocket - select() in blockTillReadable() failed");
  m_ok = false; handleError(buf);
  return false;
}



// ------------------------------------------------------------
// CTcpSocket Impl.
// ------------------------------------------------------------

CTcpSocket::CTcpSocket(int iPort) : CSocket()
{
  //DBGOUT("CTcpSocket(%d) called!\n", iPort);
  // common init to all ctor.
	m_sockaddr.sin_family = AF_INET;
	m_sockaddr.sin_port = htons(iPort);
	m_sockaddr.sin_addr.s_addr = INADDR_ANY;
  // create the socket
  m_socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (m_socket == NET_INVALID_SOCKET)
  {
    handleError("CTcpSocket - socket() failed (IVALID_SOCKET)");
    return;
  }
  // bind to socket
	if (bind(m_socket, (SOCKADDR *)&m_sockaddr, sizeof(sockaddr)) != 0)
  {
    handleError("CTcpSocket - bind() failed");
    return;
  }
  // listen
	if (listen(m_socket, 0) != 0)
  {
    handleError("CTcpSocket - listen() failed");
    return;
  }
  // if we got that far everything is ok
  m_ok = true;
}

CTcpSocket::CTcpSocket(const char *szRemoteAddr, int iPort) : CSocket()
{
  //DBGOUT("CTcpSocket(%s:%d) called !\n");
  // simple param check
  if (!szRemoteAddr)
  {
    m_error_code = NET_EINVAL;
    m_error = "NetError: CTcpSocket - Invalid params";
    DBGOUT(m_error.c_str());
    return;
  }
  // common init to all ctor.
	m_sockaddr.sin_family = AF_INET;
	m_sockaddr.sin_port = htons(iPort);
	m_sockaddr.sin_addr.s_addr = INADDR_ANY;
  // create the socket
  m_socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (m_socket == NET_INVALID_SOCKET)
  {
    handleError("CTcpSocket - socket() failed (IVALID_SOCKET)");
    return;
  }

  // first guess => try to resolve it as IP@
	m_sockaddr.sin_addr.s_addr = inet_addr(szRemoteAddr);
	if (m_sockaddr.sin_addr.s_addr == INADDR_NONE)
	{
    // screwed => try to resolve it as name
    LPHOSTENT lpHost = gethostbyname(szRemoteAddr);
		if (!lpHost)
    {
      handleError("CTcpSocket - Unable to resolve address");
      return;
    }
		m_sockaddr.sin_addr.s_addr = **(int**)(lpHost->h_addr_list);
	}

  // actually performs connection
	if (connect(m_socket, (SOCKADDR*)&m_sockaddr, sizeof(sockaddr)) != 0)
  {
    handleError("CTcpSocket - connect() failed");
    return;
  }
  // everything ok here
  m_ok = true;
}

CTcpSocket::CTcpSocket(net_socket_t sfd,  SOCKADDR_IN *saptr) : CSocket()
{
	m_socket = sfd;
  if (saptr)
  {
    memcpy(&m_sockaddr, saptr, sizeof(m_sockaddr));
  }
  m_ok = true; // assume it's ok
}


/**
 * Destructor
 */
CTcpSocket::~CTcpSocket()
{
	close();
}

CTcpSocket *CTcpSocket::accept(void)
{
  SOCKADDR_IN sa;
  net_socklen_t nlen = sizeof(sa);
  net_socket_t sfd = ::accept(m_socket, reinterpret_cast<SOCKADDR *>(&sa), reinterpret_cast<net_socklen_t *>(&nlen));
  if (sfd == NET_INVALID_SOCKET) return(NULL);
  // socket ok
  nlen = sizeof(sa);
  if (getsockname(sfd, reinterpret_cast<struct sockaddr*>(&sa), reinterpret_cast<net_socklen_t *>(&nlen)) == NET_SOCKET_ERROR)
  {
    handleError("CTcpSocket - getsockname() failed");
    return NULL;
  }
	return (new CTcpSocket(sfd, &sa));
}

int CTcpSocket::read(void * pData, unsigned int iLen, int flags)
{
	//if(!pData || !iLen) return( -1);
  assert(pData != NULL);

	int rc = recv(m_socket, (char *)pData, iLen, flags);
  if (rc == NET_SOCKET_ERROR)
  {
    char buf[128];
    sprintf(buf, "CTcpSocket - recv(%d, 0x%p, %d) failed", (int)m_socket, pData, iLen);
    m_ok = false;  handleError(buf);
    return -1;
  }
  return rc;
}

std::string CTcpSocket::readln(bool block)
{
  string retval("");

  // anything in the queue, if "yes" return front-element
  if (m_readln_queue.size() > 0)
  {
    retval = m_readln_queue.front();
    m_readln_queue.pop();
    return retval;
  }

  // nothing in queue, so read from socket and do the dirty work
  bool line_read=false, done=false;
  const int maxbytes = sizeof(m_readln_buf) - 1;
  int availbytes = 0;

  while (!done)
  {
    if (block)
    {
      if (!blockTillReadable())  return retval;
    } else {
      if (!checkSocketReadable(0, 0)) return retval;
    }
    // how much data is available for reading?
    availbytes = available();
    if (!m_ok)
      return retval;
    if (availbytes==0)
    {
      // closed be remote peer?
      if (CTcpSocket::read(m_readln_buf, 0) == -1)
      {
        // hard reset
        DBGOUT("Connection reset by peer!");
        return retval;
      } else {
        DBGOUT("Connection closed by peer!");
        m_ok = false;  // let the caller close the socket
        return retval;
      }
    } else if (availbytes > 0) {
      if (availbytes > maxbytes)
      {
        // there is more data than we can handle with our fixed size buffer, so handle what we can
        availbytes = maxbytes;
      }
      if (read_and_breakdown(availbytes, line_read))
      {
        if (line_read)
        {
          // successfully read at least one line
          retval = m_readln_queue.front();
          m_readln_queue.pop();
          done = true;
        } else {
          // successful read but no terminated line
          // so select/wait and try to read next chunk of data
          continue;
        }
      } else {
        return retval;
      }
    } else {
      // socket error occured in available()
      // return empty string
      return retval;
    }
  }
  return retval;
}

int CTcpSocket::write(void * pData, unsigned int iSize, int flags)
{
	//if(!pData || !iSize) return(-1);
  assert(pData != NULL);
#ifdef _WIN32
	int rc = send(m_socket, (LPCSTR)pData, iSize, flags);
#else
	int rc = send(m_socket, (char *)pData, iSize, flags);
#endif
  if (rc == NET_SOCKET_ERROR)
  {
    char buf[128];
    sprintf(buf, "CTcpSocket - send(%d, %p, %d) failed", (int)m_socket, pData, iSize);
    m_ok = false;  handleError(buf);
    return -1;
  }
  return rc;
}

int CTcpSocket::writeln(const std::string &str)
{
  string hstr(str);
  hstr.append("\r\n");
  return CTcpSocket::write((void *)hstr.c_str(), hstr.length());
}

bool CTcpSocket::read_and_breakdown(int available_bytes, bool &line_read)
{
  assert((unsigned int) available_bytes < sizeof(m_readln_buf));
  // copy what is in the socket buffer (not removing it yet)
  int readbytes = CTcpSocket::read(m_readln_buf, available_bytes, MSG_PEEK);
  if (readbytes != available_bytes)
  {
    m_ok = false;  handleError("CTcpSocket - readln() failed!");
    return false;
  }
  m_readln_buf[readbytes] = 0; // terminate with zero
  int i, current_pos=0;
  string hstr(m_unfinished_line);
  line_read = false;
  for (i=0; i<readbytes; ++i)
  {
    if ((m_readln_buf[i]==13) || (m_readln_buf[i]==10)) // '\r' or '\n'
    {
      if ((i<readbytes-1) && (m_readln_buf[i+1]==10)) // it was a "\r\n" <CR><LF>, so skip the <CR>
        i++;
      //i++;
      // discard line from socket buffer
      if (CTcpSocket::read(m_readln_buf, (i+1-current_pos)) == -1)
      {
        DBGOUT("CTcpSocket - readln() buffer discard (1) failed!");
        return false;
      }
      current_pos = i+1;
      // push the line to the queue (if there is something in it)
      if (!hstr.empty())
      {
        m_readln_queue.push(hstr);
        line_read = true;
      }
      hstr.clear();
      m_unfinished_line.clear();
    } else {
      hstr.push_back(m_readln_buf[i]);
    }
  }
  // discard what's left in socket (it's already saved in hstr)
  int leftover = readbytes - current_pos;
  if (leftover > 0)
  {
    if (CTcpSocket::read(m_readln_buf, leftover) == -1)
    {
      DBGOUT("CTcpSocket - readln() buffer discard (2) failed!");
      return false;
    }
  }
  // keep unterminated string
  m_unfinished_line.append(hstr);
  return true;
}

CSocketAddr CTcpSocket::getPeerAddress()
{
  assert(m_socket != NET_INVALID_SOCKET);

  struct sockaddr_in sa;
	net_socklen_t salen = sizeof(sa);
	int rc = ::getpeername(m_socket, reinterpret_cast<struct sockaddr *>(&sa), &salen);
	if (rc == 0)
  {
    // everything ok
		return CSocketAddr(&sa);
  }
  // failed! this does not change socket status but sets error-string (and error-code) and returns an invalid address
  handleError("CTcpSocket - getPeerAddress() failed!");
  return CSocketAddr();
}

CSocketAddr CTcpSocket::getAddress()
{
  if (m_socket != NET_INVALID_SOCKET)
    return CSocketAddr(reinterpret_cast<struct sockaddr_in *>(&m_sockaddr));
  return CSocketAddr();
}
