/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definitions of DebugServer member functions.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo-debugserver.cc 955 2009-05-15 14:02:17Z hpczink $
 */

//bk// TODO:
// - think about setting client states
// - and timeouts
// - and execution flow (may be options for users)
//   - serialization is very bad for performance
//   - round robin
//   - random
//   - first in first out
// - According to MPI-1.2, it is allowed to call MPI_Get_version and
//   MPI_Initialized before MPI_Init/after MPI_Finalize.
//   According to MPI-2.0, it is additionally allowed to call MPI_Finalized
//   before MPI_Init/after MPI_Finalize.
//   Currently we can't check calls that are issued
//   before MPI_Init/after MPI_Finalize. In fact, calls after MPI_Finalize lead
//   to an abortion of the program if run with MARMOT.
// - file name as commandline argument, logging to std::cerr or to file


#include <cstring>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <string>
#include <map>
#include <set>

#include "config.h"
#include "mpo-objects.h"
#include "mpo-debugserver.h"
#include "mpo-enums.h"
#include "mpo-timer.h"
#include "mpo-state.h"
#include "mpo.h"
#include "mpo_String.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "mpo_LogASCII.h"
#include "mpo_LogHTML.h"
#ifdef CUBE2
	#include "mpo_LogCube.h"
#else
	#include "mpo_LogCube3.h"
#endif

namespace MPO
{
	//static attributes
	//=======================================================================
	bool DebugServer::myErrorsSet = false;

    // proper constructor
    //========================================================================
    /** Constructor sets up server with provided communication_channel.
     */
    DebugServer::DebugServer(Messenger & comm_channel, char *ExeName)
        : my_channel(comm_channel)
    {
        std::string hstr;
        std::string exepath;
        if (ExeName == NULL)
          myExeName = "LogFile";
        else
          myExeName = ExeName;
        // get rid of path
        string::size_type pos = myExeName.find_last_of("/\\");
        if (pos != string::npos)
        {
          hstr = myExeName.substr(pos+1);
          exepath = myExeName.substr(0, pos+1);
          myExeName = hstr;
        }
#ifdef _WIN32
        if (myExeName.length() > 5)
        {
          // get rid of ".exe" (on windows)
          hstr = myExeName.substr(myExeName.length()-4);
          transform(hstr.begin(), hstr.end(), hstr.begin(), (int(*)(int))tolower);
          if (hstr == ".exe")
            myExeName.erase(myExeName.length()-4);
        }
        // preset log-file-path (works around problems creating a log-file on mounted shares)
        MARMOT_LOGFILE_PATH = exepath;
#endif
        initLogFile ();
        setAndPrintEnvironment();
        setErrorMessages();
        logThreadSyncInfo();
    }


    // destructor
    //========================================================================
    DebugServer::~DebugServer()
    {
        if (my_log)
        {
            delete (my_log);
        }
        my_log = NULL;

        CSocketLogger::destroyInstance();
    }


    // runMainLoop
    //========================================================================
    void DebugServer::runMainLoop()
    {
        int comm_size = 0;
        char **pn = NULL;   //processor names
        int pn_length;      //processor name length
        MPO_Timer timer;
        MPO_Timestamp recv_time;
        PMPI_Comm_size(MPI_COMM_WORLD, &comm_size);
        //printf("DEBUG --- debugserver - comm_size=%i\n",comm_size);
        //pn == processor names
        pn=new char* [comm_size];
        for(int i=0;i<comm_size;i++)
        {
            pn[i]=NULL;
        }
        pn[comm_size-1]=new char [MPI_MAX_PROCESSOR_NAME];
        //get the processor's name the sedubserver is running on
        PMPI_Get_processor_name(pn[comm_size-1],&pn_length);
        comm_size = comm_size - 1;
        int running = comm_size;
        int not_initialized = running;
        int pend_count = 0;

        //std::vector < MPO_Clientstate > client_status(comm_size);
        /**
         * Client status now collected for each thread of each process.
         */
        std::vector < std::vector<MPO_Clientstate> > client_status(comm_size);
        for (int i = 0; i < comm_size; i++)
        {
            client_status[i].clear();
            client_status[i].resize(1);
            client_status[i][0].setState(MPO_PENDING);
        }

        CallTimestampPair my_call_pair;

        MPO_MsgObject logmsg;

        /**
         * In this queue we store the last 10 calls, to be able to print
         * them out in case of deadlock or so...
         */
        std::vector < std::deque < CallTimestampPair > > Call_Queue(comm_size);

        int next_mpi_rank = 0;

        // In this while-loop we receive messages from the clients,
        // trace their calls (if tracing is set by the user via TRACE_CALLS).
        // and send them the permission to execute (if non-serialization is
        // set by the user via SERIALIZE.)
        // In case of a deadlock a warning is issued, calls may be traced
        // back if this is set by the user via TRACE_CALLS and MAX_PEND_COUNT.
        while (running > 0)
        {
            int sender_rank = -1;
            int sender_tag  = -1;

            // timeout to wait for new messages (timeout set to 0)
           timer.sleep(MAX_TIMEOUT_ONE);

            // We loop as long as we have messages
            while (my_channel.poll(sender_rank,sender_tag) ||
                   not_initialized)
            {
                if (sender_rank != -1)
                {
                    // (1) receive message
                    int family = -1;
                    my_channel.receiveId(family, sender_rank, sender_tag);

                    int object_type = -1;
                    my_channel.receiveId(object_type, sender_rank, sender_tag);

                    // only increment time when a new call appeared
                    // (we do not want to count Messages)
                    if (family == MPO_CALLOBJECT_ID)
                        ++recv_time;

                    if (family == MPO_CALLOBJECT_ID)
                    {
                        MPO_Callobject *obj_p = 0;
                        obj_p = MPO_Space::our_call_factory.create(object_type);
                        my_channel.receiveObject(*obj_p, sender_rank,sender_tag);

                        // (2) set client state to MPO_STARTED
                        //do we have to add a new entry for a thread in the client states
                        int old_size = client_status[sender_rank].size();
                        if (old_size <= sender_tag-MESSENGER_TAG_OFFSET)
                            client_status[sender_rank].resize(sender_tag-MESSENGER_TAG_OFFSET+1);

                        //we might have created multiple new entries - so we set a state for each
                        for (unsigned int i_new = old_size;i_new < client_status[sender_rank].size();i_new++)
                            client_status[sender_rank][i_new].setState(MPO_PENDING);

                        client_status[sender_rank][sender_tag-MESSENGER_TAG_OFFSET].setState(MPO_STARTED);

                        my_call_pair.my_call_p = obj_p;
                        my_call_pair.my_timestamp = recv_time;
                        my_call_pair.my_location = obj_p->getSourceLocation();
                        client_status[sender_rank][sender_tag-MESSENGER_TAG_OFFSET].setCallTimestampPair(my_call_pair);
                        // (3) collect calls for traceback in case of deadlock
                        // add call to queue if user sets tracing via
                        // TRACE_CALLS
                        if (TRACE_CALLS >= 0)
                        {
                            if (Call_Queue[sender_rank].size() >= (unsigned int) MAX_PEND_COUNT)
                            {
                                Call_Queue[sender_rank].pop_front();
                            }
                            Call_Queue[sender_rank].push_back(my_call_pair);
                        }

                        // (4) trace calls: level of verbosity
                        // show message content if user sets this option via
                        // TRACE_CALLS
                        if (TRACE_CALLS > 0)
                        {
                            MPO_String  strText("performing"),
                                strCallinfo;
                            std::stringstream stream;

                            // show complete call with arguments
                            if (TRACE_CALLS >= 2)
                            {
                                obj_p->debugout(stream);
                            }
                            // show just callname
                            else
                            {
                                stream  << obj_p->getCallname()
                                        << std::endl;
                            }

                            strCallinfo.assign(&stream);

                            logmsg.initMsgObject (MSG_TYPE_NOTE,recv_time,sender_rank,strText,strCallinfo,(MPO_CALL_ID)obj_p->getId(),MARMOT_SUCCESS,obj_p->getSourceLocation());
                            logmsg.setThreadRank(obj_p->getThreadRank());
                            my_log->getMsg(logmsg);
                        }
                        else if (TRACE_CALLS == 0)
                        {
                            //for  cube we create a message used to add the source locations and severitys
                            if (LOGFILE_TYPE == LOGFILE_TYPE_CUBE)
                            {
                                MPO_String  strCallinfo;
                                std::stringstream stream;

                                stream  << obj_p->getCallname()
                                        << std::endl;

                                strCallinfo.assign(&stream);

                                logmsg.initMsgObject (MSG_TYPE_NOTE,recv_time,sender_rank,"",strCallinfo,(MPO_CALL_ID)obj_p->getId(),MARMOT_SUCCESS,obj_p->getSourceLocation());
                                logmsg.setThreadRank(obj_p->getThreadRank());
                                my_log->getMsg(logmsg);
                            }
                        } //endif (TRACE_CALLS...)


                        // (5) count how many ranks have performed MPO_Init
                        if ((family == MPO_CALLOBJECT_ID) &&
                            ((object_type == MPO_INIT_ID) ||
                             (object_type == MPO_INIT_THREAD_ID) ))
                        {
                            not_initialized--;
                        }

                    }                   // end  if(family==MPO_CALLOBJECT_ID)

                    else if (family == MPO_MESSAGEOBJECT_ID)
                    {
                        //is it a Message Object ?
                        if (object_type == MPO_STANDARDMESSAGE)
                        {
                            //Receive and Log it !
                            MPO_MsgObject obj;
                            my_channel.receiveObject(obj,sender_rank,sender_tag);
                            obj.setTimestamp(recv_time);
                            //my_log->getMsg(obj);
                //VALENTIN TEST
                if (MARMOT_NOTE_PROCESSOR_NAME == (MARMOT_ERR_CODES)obj.getErrorCode())
                {
                    std::string procname(obj.getText());
                    pn[sender_rank]=new char [MPI_MAX_PROCESSOR_NAME];
                    strcpy(pn[sender_rank],procname.c_str());
                    //printf("DEBUG --- MARMOT_NOTE_PROCESSOR_NAME (sender_rank=%i, processor=%s)\n",sender_rank,procname.c_str());
                   //Here we check if all processes have already sent their processor's name
                   //if one hasn't yet, pn[i]==NULL
                   int test=1;
                   for(int i=0;i<comm_size+1;i++)
                {
                    if(!pn[i])  //at least one processor name is still unknown
                    {
                        test=0;
                        break;
                    }
                }


                   if(test) //If all processes have sent therir processor's name
                {
                    //check whether the debugserver runs on the same node as some process
                    for(int i=0;i<comm_size;i++)
                    {
                      if(strcmp(pn[comm_size],pn[i])==0)  //debugserver runs on same node as process i
                      {
                            //printf("DEBUG --- debugserver runs on same node as process %i (%s)\n",i,pn[i]);
                            MPO_String  strText;
                            std::stringstream stream;
                            stream << "Debugserver runs on same node as process ";
                            stream << i << " (" << pn[i] << ")" << std::endl;

                                        strText.assign(&stream);

                                    logmsg.initMsgObject (MSG_TYPE_WARNING,recv_time,strText,MARMOT_WARN_DEBUGSERVER_NO_EXCLUSIVE_NODE);
                                    my_log->getMsg(logmsg);
                    }
                    }
                        //check whether some of the processes share one node
                    for(int i=0;i<comm_size-1;i++)
                    {
                        for(int j=i+1;j<comm_size;j++)
                        {
                            if(strcmp(pn[i],pn[j])==0) //two processes running on the same node
                            {
                                //printf("DEBUG --- processes %i and %i both run on %s\n",i,j,pn[i]);
                                MPO_String strText;
                                std::stringstream stream;
                                stream << "Processes " << i << " and " << j;
                                stream << " both run on " << pn[i] << std::endl;
                                strText.assign(&stream);
                                logmsg.initMsgObject (MSG_TYPE_WARNING,recv_time,strText,MARMOT_WARN_PROCESSES_SHARE_NODE);
                                               my_log->getMsg(logmsg);
                        }
                    }
                }

                }
                }
                else
                {
                    my_log->getMsg(obj);
                }
                    //END VALENTIN TEST
                            // If error/warning/note was found
                            // indicate it to debugger by calling one of
                            // these functions
                            // (when using CUBE_DDT the processes automatically call these functions)
                            if (LOGFILE_TYPE != LOGFILE_TYPE_CUBE_DDT)
                            {
                                if (MSG_TYPE_ERROR == obj.getMsgType() )
                                {
                                    //insertBreakpointMarmotError("MarmotError");
                                    insertBreakpointMarmotError((char*)obj.getText().c_str());
                                }
                                else if (MSG_TYPE_WARNING == obj.getMsgType() )
                                {
                                    //insertBreakpointMarmotWarning("MarmotWarning");
                                    insertBreakpointMarmotWarning((char*)obj.getText().c_str());
                                }
                                else if (MSG_TYPE_NOTE == obj.getMsgType() )
                                {
                                    //insertBreakpointMarmotNote("MarmotNote");
                                    insertBreakpointMarmotNote((char*)obj.getText().c_str());
                                }
                                else
                                {
                                    // do nothing
                                }
                            }

                            // if we received an error message we have to
                            // synchronize in order to log correctly !
                            // otherwise we might lose data
                            if (MSG_TYPE_ERROR == obj.getMsgType())
                            {
                                my_channel.sendId(1, sender_rank,sender_tag);
                            }
                        }

                    }                   // end else if(family==MPO_MESSAGEOBJECT_ID)

                    else if (family == MPO_FINALIZE_READY)
                    {
                        //send the OK to perform MPI_Finalize
                        my_channel.sendId(1, sender_rank,-1);

                        //decrease running
                        running--;
                    }
                }                       // end if(sender_rank!=-1)


                // (6) if there is no message in the queue we wait
                // for a new message
                sender_rank = -1;

            }  // end while(my_channel.poll(sender_rank) || not_initialized )

// TODO where do we set the client state MPO_RUNNING?

            // (7) pending clients:
            // after the MAX_TIMEOUT we know that every client that was
            // in MPO_CALLED is now in MPO_PENDING
            for (int i = 0; i < comm_size; i++)
            {
                for (unsigned int j = 0; j < client_status[i].size();j++)
                {
                    assert(client_status[i][j].getState() != MPO_LAST_STATE_ID);
                    if ((MAX_PEND_COUNT >= 0) &&
                        (client_status[i][j].getState() == MPO_CALLED))
                    {
                        client_status[i][j].setState(MPO_PENDING);
                    }
                }
            }

            //=====================
            // Check for deadlocks:
            //=====================

            // To check for deadlocks, we count the number of pending clients.
            // If all clients are pending, we issue a deadlock warning.
            // To avoid unjustified deadlock warnings, we use
            // additional if-loops to check my_channel.poll(sender_rank)
            // and we use timer.sleep(MAX_TIMEOUT..)

            /**
             * @todo:
             * in a multithreaded case we would have to check whether all threads are pending.
             * But as the amount of threads used can vary over time this would require to know
             * the amount of threads actually used by each client. To get this information we would
             * have to get the current amount of threads with each call information.
             * This is currently not implemented and might never be implemented as the current
             * deadlock detection is probably going to be replaced.
             *
             * Currently we use the old scheme and only for thread 0.
             */

            int nr_of_pending_clients = 0;

            // (8) get number of pending clients
            for (int i = 0; i < comm_size; i++)
            {
                if (client_status[i][0].getState() == MPO_PENDING)
                {
                    nr_of_pending_clients++;
                }
            }

            if (nr_of_pending_clients == comm_size)
            {
                pend_count++;
            }
            else
            {
                pend_count = 0;
            }

            if ((TRACE_CALLS >= 0) &&
                (MAX_PEND_COUNT >= 0) &&
                (pend_count == MAX_PEND_COUNT))
            {
                if (!my_channel.poll(sender_rank,sender_tag))
                {
                    //bk// we poll once a millisecond
                    //bk// if there are messages before we don't have
                    //bk// to wait MAX_TIMEOUT_DEADLOCK
                   for (int i = 0; i < MAX_TIMEOUT_DEADLOCK; i += 1000)
                   {
                       if (!my_channel.poll(sender_rank,sender_tag))
                       {
                           timer.sleep(1000);
                       }
                       else
                       {
                           break;
                       }
                   }
                }


                // (9) now issue the deadlock warning (with traceback...)
                if (!my_channel.poll(sender_rank,sender_rank))
                {

                    MPO_String  strText;
                    std::stringstream stream;

                    stream  << "WARNING: "
                            << "all clients are pending!"
                            <<std::endl;

                    for (int i = 0; i < comm_size; i++)
                    {
                        stream
                            << "Last calls (max. "
                            << MAX_PEND_COUNT
                            << ") on node "
                            << i
                            << ":"
                            << std::endl;

                        int queue_size = Call_Queue[i].size();
                        for (int j = 0; j < queue_size; j++)
                        {
                            stream
                                << " timestamp "
                                << (Call_Queue[i])[j].my_timestamp
                                << ": ";
                            (Call_Queue[i])[j].my_call_p->debugout(stream);
                            if ((Call_Queue[i])[j].my_location.isValidLocation())
                                (Call_Queue[i])[j].my_location.debugout(stream);

                            stream << std::endl;
                        }
                        stream
                            << std::endl;
                    }

                    strText.assign(&stream);

                    //logmsg.initMsgObject (MSG_TYPE_WARNING,recv_time,strText,MARMOT_WARN_DEADLOCK);
                    logmsg.initMsgObject (MSG_TYPE_ERROR,recv_time,strText,MARMOT_WARN_DEADLOCK);
                    my_log->getMsg(logmsg);
                    //print a warning to the user !
                    std::cerr << "WARNING: all clients are pending! (Details see the LogFile)\n";

                    // indicate it to debugger by calling this function
                    insertBreakpointMarmotError("MarmotWarning: all clients are pending");

                }                       // end if( !my_channel.poll(sender_rank) )

            }                         // end if ( (TRACE_CALLS >= 0) && (MAX_PEND_COUNT >= 0) && ...)

            //=====================
            // Serialized:
            //=====================

            // if user wants to have code serialized
            // start clients now instead of before in the if-loop
            if (SERIALIZE == 1)
            {
                // the next client that is MPO_STARTED can now execute:
                int tries = 0;
                bool started_a_call = false;

                for (unsigned int i = 0; i < client_status[next_mpi_rank].size(); i++)
                {
                    if (client_status[next_mpi_rank][i].getState() == MPO_STARTED)
                    {
                        started_a_call = true;

                        // send the client the permission to execute:
                        my_channel.sendId(1, next_mpi_rank, i + MESSENGER_TAG_OFFSET);

                        if (MAX_TIMEOUT_SERIALIZE > 0)
                            timer.sleep(MAX_TIMEOUT_SERIALIZE);

                        client_status[next_mpi_rank][i].setState(MPO_CALLED);
                    }
                }//end for threads of client

                if ( (!started_a_call) && (tries < comm_size))
                {
                    tries++;
                }

                next_mpi_rank = (next_mpi_rank + 1) % comm_size;
            }                         // end if( SERIALIZE == 1 )

            //=====================
            // Not serialized:
            //=====================

            // send the client the permission to execute
            // if user does not want code serialized
            if (SERIALIZE == 0)
            {
                for (next_mpi_rank = 0; next_mpi_rank < comm_size;
                     next_mpi_rank++)
                {
                    for (unsigned int next_thread = 0; next_thread < client_status[next_mpi_rank].size(); next_thread++)
                    {
                        if (client_status[next_mpi_rank][next_thread].getState() == MPO_STARTED)
                        {
                            // send the client the permission to execute:
                            my_channel.sendId(1, next_mpi_rank,next_thread + MESSENGER_TAG_OFFSET);

                            timer.sleep(MAX_TIMEOUT_TWO);
                            client_status[next_mpi_rank][next_thread].setState(MPO_CALLED);
                        } // end if
                    } //end for threads in client
                } // end for
            } // end if (SERIALIZE == 0)
        } // end while(running>0)


        // Accorcing to MPI-1.2, it is allowed to call MPI_Get_version and
        // MPI_Initialized after MPI-Finalize.
        // According to MPI-2.0, it is additionally allowed to call
        // MPI-Finalized after MPI-Finalize.

        //close the LogFile in case the Destructor of the DebugServer won`t be called
        my_log->closeLog ();

        PMPI_Finalize();
        // exit, because we don't want to give the control to the program
        // at the position where MPI_Init was called.
        exit(0);

    }                           // end main loop




    // setAndPrintEnvironment
    //========================================================================
    /** Gets/sets and prints all our environment variables:
     *  - DEBUG_MODE, INTERFACE_MODE are set in the constructor of MPO_Init.
     *  - MAX_TIMEOUT, TRACE_CALLS, MAX_PEND_COUNT, SERIALIZE
     *    (and maybe other test variables such as MAX_TIME_OUT_ONE etc.)
     *    are set here.
     */
    int DebugServer::setAndPrintEnvironment() const
    {
        MPO_MsgObject logmsg;
        MPO_String msg;
        std::stringstream stream;

        //==========================
        // DEBUG_MODE
        //==========================

        // see mpo_Init.cc
        //         if (getenv("MARMOT_DEBUG_MODE") == NULL)
        //         {
        //             DEBUG_MODE = DEBUG_MODE_DEFAULT;
        //         }
        //         else
        //         {
        //             DEBUG_MODE = atoi(getenv("MARMOT_DEBUG_MODE"));
        //         }

        stream << "The following entries describe all the Environmental variables.";

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        stream  << "MARMOT_DEBUG_MODE = "
                << DEBUG_MODE
                << std::endl
                << "    (negative int: nothing,"
                << std::endl
                << "     0: errors,"
                << std::endl
                << "     1: errors and warnings,"
                << std::endl
                << "     2: errors, warnings and remarks are reported,"
                << std::endl
                << "     default: "
                << DEBUG_MODE_DEFAULT
                << ")"
                << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MARMOT_LOGFILE_TYPE
        //==========================
        stream  << "MARMOT_LOGFILE_TYPE = "
                << LOGFILE_TYPE
                << std::endl
                << "    (0: ASCII Logging,"
                << std::endl
                << "     1: HTML Logging,"
                << std::endl
                << "     2: CUBE Logging (when enabled via configure),"
                << std::endl
                << "     3: Cube Logging for DDT (with runtime output),"
                << std::endl
                << "     default: "
                << LOGFILE_TYPE_DEFAULT
                << ")"
                << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MARMOT_LOGFILE_NAME
        //==========================
        stream  << "MARMOT_LOGFILE_NAME"
                << std::endl
                << "     You can set a name for the log file here"
                << std::endl
                << "     If left undefined, the log file will have the following name:"
                << std::endl
                << "     Marmot_[Executable Name]_[YYYYMMDD_HHMMSS].[txt|html|cube]"
                << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

    //==========================
        // MARMOT_LOG_FLUSH_TYPE
        //==========================
        stream  << "MARMOT_LOG_FLUSH_TYPE = "
                << LOGFLUSH_TYPE
                << std::endl
                << "    (0: Flush on Error,"
                << std::endl
                << "     1: Immediate Flush,"
                << std::endl
                << "     default: "
                << LOGFLUSH_TYPE_DEFAULT
                << ")"
                << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MARMOT_LOG_FILTER_COUNT
        //==========================
        stream  << "MARMOT_LOG_FILTER_COUNT = "
                << LOGFILTER_COUNT
                << std::endl
                << "    (sets the maximal amount of messages of the same type that are logged for each source location,"
                << std::endl
                << "     default: "
                << LOGFILTER_COUNT_DEFAULT
                << ")"
                << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MARMOT_RESOURCE_TRACE_TYPE
        //==========================
        stream  << "MARMOT_RESOURCE_TRACE_SELECTION = "
                << RESOURCE_TRACE_TYPE
                << std::endl
                << "    (0: On (Full resource tracing if source-locations are present and quantitativ tracing otherwise),"
                << std::endl
                << "     1: Off (no additional resource tracing, some tracing is still done in order to detect certain errors),"
                << std::endl
                << "     default: "
                << RESOURCE_TRACE_TYPE_DEFAULT
                << ")"
                << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // INTERFACE_MODE
        //==========================

        // see mpo_Init.cc
        //         if (getenv("MARMOT_INTERFACE_MODE") == NULL)
        //         {
        //             INTERFACE_MODE = INTERFACE_MODE_DEFAULT;
        //         }
        //         else
        //         {
        //             INTERFACE_MODE = atoi(getenv("MARMOT_INTERFACE_MODE"));
        //         }
        stream
            << "MARMOT_INTERFACE_MODE = "
            << INTERFACE_MODE
            << std::endl
            << "    (0: C interface,"
            << std::endl
            << "     1: Fortran interface,"
            << std::endl
            << "     interface mode is set automatically)"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // SERIALIZE
        //==========================

        if (getenv("MARMOT_SERIALIZE") == NULL)
        {
            SERIALIZE = SERIALIZE_DEFAULT;
        }
        else
        {
            SERIALIZE = atoi(getenv("MARMOT_SERIALIZE"));
        }
        stream
            << "MARMOT_SERIALIZE = "
            << SERIALIZE
            << std::endl
            << "    (0: code is not serialized,"
            << std::endl
            << "     1: code is serialized,"
            << std::endl
            << "     default: "
            << SERIALIZE_DEFAULT
            << ")"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // TRACE_CALLS
        //==========================

        if (getenv("MARMOT_TRACE_CALLS") == NULL)
        {
            TRACE_CALLS = TRACE_CALLS_DEFAULT;
        }
        else
        {
            TRACE_CALLS = atol(getenv("MARMOT_TRACE_CALLS"));
        }
        stream
            << "MARMOT_TRACE_CALLS = "
            << TRACE_CALLS
            << std::endl
            << "    (1: calls are traced with output to stderr,"
            << std::endl
            << "        traceback in case of a deadlock is possible,"
            << std::endl
            << "     0: calls are traced without output to stderr,"
            << std::endl
            << "        traceback in case of a deadlock is possible,"
            << std::endl
            << "    -1: calls are not traced,"
            << std::endl
            << "        traceback in case of a deadlock is NOT possible."
            << std::endl
            << "     default: "
            << TRACE_CALLS_DEFAULT
            << std::endl
            << "     Number of calls to be traced back in case of deadlock"
            << std::endl
            << "     can be set via MAX_PEND_COUNT.)"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MAX_PEND_COUNT
        //==========================

        if (getenv("MARMOT_MAX_PEND_COUNT") == NULL)
        {
            MAX_PEND_COUNT = MAX_PEND_COUNT_DEFAULT;
        }
        else
        {
            MAX_PEND_COUNT = atol(getenv("MARMOT_MAX_PEND_COUNT"));
        }
        stream
            << "MARMOT_MAX_PEND_COUNT = "
            << MAX_PEND_COUNT
            << std::endl
            << "    (maximum number of calls that are traced back,"
            << std::endl
            << "     default: "
            << MAX_PEND_COUNT_DEFAULT
            << ")"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MAX_TIMEOUT_DEADLOCK
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_DEADLOCK") == NULL)
        {
            MAX_TIMEOUT_DEADLOCK = MAX_TIMEOUT_DEADLOCK_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_DEADLOCK = atol(getenv("MARMOT_MAX_TIMEOUT_DEADLOCK"));
        }
        stream
            << "MARMOT_MAX_TIMEOUT_DEADLOCK = "
            << MAX_TIMEOUT_DEADLOCK
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_DEADLOCK_DEFAULT
            << " microseconds)"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MAX_TIMEOUT_SERIALIZE
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_SERIALIZE") == NULL)
        {
            MAX_TIMEOUT_SERIALIZE = MAX_TIMEOUT_SERIALIZE_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_SERIALIZE =
                atol(getenv("MARMOT_MAX_TIMEOUT_SERIALIZE"));
        }
        stream
            << "MARMOT_MAX_TIMEOUT_SERIALIZE = "
            << MAX_TIMEOUT_SERIALIZE
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_SERIALIZE_DEFAULT
            << " microseconds)"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //================================================================
        // additional variables for MAX_TIMEOUT
        // to test performance and deadlock detection
        //================================================================

        //==========================
        // MAX_TIMEOUT_ONE
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_ONE") == NULL)
        {
            MAX_TIMEOUT_ONE = MAX_TIMEOUT_ONE_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_ONE = atol(getenv("MARMOT_MAX_TIMEOUT_ONE"));
        }
        stream
            << "MARMOT_MAX_TIMEOUT_ONE = "
            << MAX_TIMEOUT_ONE
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_ONE_DEFAULT
            << " microseconds)"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        //==========================
        // MAX_TIMEOUT_TWO
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_TWO") == NULL)
        {
            MAX_TIMEOUT_TWO = MAX_TIMEOUT_TWO_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_TWO = atol(getenv("MARMOT_MAX_TIMEOUT_TWO"));
        }

        stream
            << "MARMOT_MAX_TIMEOUT_TWO = "
            << MAX_TIMEOUT_TWO
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_ONE_DEFAULT
            << " microseconds)"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();


        stream
            << "MARMOT_LOGFILE_PATH = "
            << MARMOT_LOGFILE_PATH
            << std::endl
            << "    (path of Marmot log file output,"
        << std::endl
        << "     default: "
            << MARMOT_LOGFILE_PATH_DEFAULT
        << ")"
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();

        std::string errcodes_set_str("(not set)");
        const char *tmp_ptr = getenv("MARMOT_ERRCODES_SET");
        if (tmp_ptr)
          errcodes_set_str = tmp_ptr;

        stream
            << "MARMOT_ERRCODES_SET = "
            << errcodes_set_str
            << std::endl
	    << "      (not functional yet)"
	    << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
        stream.clear();



        //==========================
        // End of info
        //==========================

        stream
            << "End of the environmental variables info."
            << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);


        return 0;
    }                           // end setAndPrintEnvironment



    // setAndPrintEnvironment
    //========================================================================
    /** Gets/sets and prints all our environment variables:
     *  - DEBUG_MODE, INTERFACE_MODE are set in the constructor of MPO_Init.
     *  - MAX_TIMEOUT, TRACE_CALLS, MAX_PEND_COUNT, SERIALIZE
     *    (and maybe other test variables such as MAX_TIME_OUT_ONE etc.)
     *    are set here.
     *  Same as above, just defined as ostream.
     */
    std::ostream & DebugServer::setAndPrintEnvironment(std::ostream & output)
        const
    {

        output
            << "=============== environmental variables ===================="
            << std::endl;

        //==========================
        // DEBUG_MODE
        //==========================
        output
            << "MARMOT_DEBUG_MODE = "
            << DEBUG_MODE
            << std::endl
            << "    (negative int: nothing,"
            << std::endl
            << "     0: errors,"
            << std::endl
            << "     1: errors and warnings,"
            << std::endl
            << "     2: errors, warnings and remarks are reported,"
            << std::endl
            << "     default: "
            << DEBUG_MODE_DEFAULT
            << ")"
            << std::endl;

        //==========================
        // MARMOT_LOGFILE_TYPE
        //==========================
        output  << "MARMOT_LOGFILE_TYPE = "
                << LOGFILE_TYPE
                << std::endl
                << "    (0: ASCII Logging,"
                << std::endl
                << "     1: HTML Logging,"
                << std::endl
                << "     2: CUBE Logging (when enabled via configure),"
                << std::endl
                << "     3: Cube Logging for DDT (with runtime output),"
                << std::endl
                << "     default: "
                << LOGFILE_TYPE_DEFAULT
                << ")"
                << std::endl;

        //==========================
        // MARMOT_LOG_FLUSH_TYPE
        //==========================
        output  << "MARMOT_LOG_FLUSH_TYPE = "
                << LOGFLUSH_TYPE
                << std::endl
                << "    (0: Flush on Error,"
                << std::endl
                << "     1: Immediate Flush,"
                << std::endl
                << "     default: "
                << LOGFLUSH_TYPE_DEFAULT
                << ")"
                << std::endl;

        //==========================
        // MARMOT_LOG_FILTER_COUNT
        //==========================
        output  << "MARMOT_LOG_FILTER_COUNT = "
                << LOGFILTER_COUNT
                << std::endl
                << "    ( sets the maximal amount of messages of the same type that are logged for each source location,"
                << std::endl
                << "     default: "
                << LOGFILTER_COUNT_DEFAULT
                << ")"
                << std::endl;

        //==========================
        // MARMOT_RESOURCE_TRACE_TYPE
        //==========================
        output  << "MARMOT_RESOURCE_TRACE_SELECTION = "
                << RESOURCE_TRACE_TYPE
                << std::endl
                << "    (0: On (Full resource tracing if source-locations are present and quantitativ tracing otherwise),"
                << std::endl
                << "     1: Off (no additional resource tracing, some tracing is still done in order to detect certain errors),"
                << std::endl
                << "     default: "
                << RESOURCE_TRACE_TYPE_DEFAULT
                << ")"
                << std::endl;

        //==========================
        // INTERFACE_MODE
        //==========================

        // see mpo_Init.cc
        //         if (getenv("MARMOT_INTERFACE_MODE") == NULL)
        //         {
        //             INTERFACE_MODE = INTERFACE_MODE_DEFAULT;
        //         }
        //         else
        //         {
        //             INTERFACE_MODE = atoi(getenv("MARMOT_INTERFACE_MODE"));
        //         }
        output
            << "MARMOT_INTERFACE_MODE = "
            << INTERFACE_MODE
            << std::endl
            << "    (0: C interface,"
            << std::endl
            << "     1: Fortran interface,"
            << std::endl
            << "     interface mode is set automatically)"
            << std::endl;

        //==========================
        // SERIALIZE
        //==========================

        if (getenv("MARMOT_SERIALIZE") == NULL)
        {
            SERIALIZE = SERIALIZE_DEFAULT;
        }
        else
        {
            SERIALIZE = atoi(getenv("MARMOT_SERIALIZE"));
        }
        output
            << "MARMOT_SERIALIZE = "
            << SERIALIZE
            << std::endl
            << "    (0: code is not serialized,"
            << std::endl
            << "     1: code is serialized,"
            << std::endl
            << "     default: "
            << SERIALIZE_DEFAULT
            << ")"
            << std::endl;

        //==========================
        // TRACE_CALLS
        //==========================

        if (getenv("MARMOT_TRACE_CALLS") == NULL)
        {
            TRACE_CALLS = TRACE_CALLS_DEFAULT;
        }
        else
        {
            TRACE_CALLS = atol(getenv("MARMOT_TRACE_CALLS"));
        }
        output
            << "MARMOT_TRACE_CALLS = "
            << TRACE_CALLS
            << std::endl
            << "    (1: calls are traced with output to stderr,"
            << std::endl
            << "        traceback in case of a deadlock is possible,"
            << std::endl
            << "     0: calls are traced without output to stderr,"
            << std::endl
            << "        traceback in case of a deadlock is possible,"
            << std::endl
            << "    -1: calls are not traced,"
            << std::endl
            << "        traceback in case of a deadlock is NOT possible."
            << std::endl
            << "     default: "
            << TRACE_CALLS_DEFAULT
            << std::endl
            << "     Number of calls to be traced back in case of deadlock"
            << std::endl
            << "     can be set via MAX_PEND_COUNT.)"
            << std::endl;

        //==========================
        // MAX_PEND_COUNT
        //==========================

        if (getenv("MARMOT_MAX_PEND_COUNT") == NULL)
        {
            MAX_PEND_COUNT = MAX_PEND_COUNT_DEFAULT;
        }
        else
        {
            MAX_PEND_COUNT = atol(getenv("MARMOT_MAX_PEND_COUNT"));
        }
        output
            << "MARMOT_MAX_PEND_COUNT = "
            << MAX_PEND_COUNT
            << std::endl
            << "    (maximum number of calls that are traced back,"
            << std::endl
            << "     default: "
            << MAX_PEND_COUNT_DEFAULT
            << ")"
            << std::endl;

        //==========================
        // MAX_TIMEOUT_DEADLOCK
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_DEADLOCK") == NULL)
        {
            MAX_TIMEOUT_DEADLOCK = MAX_TIMEOUT_DEADLOCK_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_DEADLOCK = atol(getenv("MARMOT_MAX_TIMEOUT_DEADLOCK"));
        }
        output
            << "MARMOT_MAX_TIMEOUT_DEADLOCK = "
            << MAX_TIMEOUT_DEADLOCK
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_DEADLOCK_DEFAULT
            << " microseconds)"
            << std::endl;

        //==========================
        // MAX_TIMEOUT_SERIALIZE
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_SERIALIZE") == NULL)
        {
            MAX_TIMEOUT_SERIALIZE = MAX_TIMEOUT_SERIALIZE_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_SERIALIZE =
                atol(getenv("MARMOT_MAX_TIMEOUT_SERIALIZE"));
        }
        output
            << "MARMOT_MAX_TIMEOUT_SERIALIZE = "
            << MAX_TIMEOUT_SERIALIZE
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_SERIALIZE_DEFAULT
            << " microseconds)"
            << std::endl;

        //================================================================
        // additional variables for MAX_TIMEOUT
        // to test performance and deadlock detection
        //================================================================

        //==========================
        // MAX_TIMEOUT_ONE
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_ONE") == NULL)
        {
            MAX_TIMEOUT_ONE = MAX_TIMEOUT_ONE_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_ONE = atol(getenv("MARMOT_MAX_TIMEOUT_ONE"));
        }
        output
            << "MARMOT_MAX_TIMEOUT_ONE = "
            << MAX_TIMEOUT_ONE
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_ONE_DEFAULT
            << " microseconds)"
            << std::endl;

        //==========================
        // MAX_TIMEOUT_TWO
        //==========================

        if (getenv("MARMOT_MAX_TIMEOUT_TWO") == NULL)
        {
            MAX_TIMEOUT_TWO = MAX_TIMEOUT_TWO_DEFAULT;
        }
        else
        {
            MAX_TIMEOUT_TWO = atol(getenv("MARMOT_MAX_TIMEOUT_TWO"));
        }
        output
            << "MARMOT_MAX_TIMEOUT_TWO = "
            << MAX_TIMEOUT_TWO
            << std::endl
            << "    (maximum message time,"
            << std::endl
            << "     default: "
            << MAX_TIMEOUT_ONE_DEFAULT
            << " microseconds)"
            << std::endl;

        //==========================
        // End of info
        //==========================

        output
            << "============================================================"
            << std::endl;

        return output;
    }                           // end setAndPrintEnvironment



    // setErrorMessages
    //========================================================================

    std::map < enum MARMOT_ERR_CODES, std::string >
    DebugServer::MARMOT_ERR_STRINGS;

    /** For every enumeration value of MARMOT_ERR_CODES,
     *  we set an error string here. This is to keep MARMOT's logging
     *  in one place here instead of splattering the logging all over
     *  the source code.
     *  IMPORTANT:
     *  Whenever an enumeration value is added, make sure to also add a string
     *  to MARMOT_ERR_STRINGS.
     *  Keep the same order as in the enum, that makes life easier.
     *
     *  @see enum MARMOT_ERR_CODES
     */
    int DebugServer::setErrorMessages()
    {

		if (myErrorsSet)
			return 0; //already set !

#ifdef NEC_SX_COMPILER
// problems with compiler on SX-8
#else
        //======================
        //======================
        // MPI-1.2
        // error types in
        // alphabetical order
        //======================
        //======================

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_SUCCESS,
                                   std::string("SUCCESS!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_UNDEFINED,
                                   std::string("ERROR - Undefined Error!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_MARMOT,
                                   std::string("ERROR - Implementation Error in MARMOT!")));


        // blocklength
        //======================

        // errors (blocklength)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_BLOCKLENGTH_NEG,
                                   std::string("ERROR - Blocklength is negative!")));

        // warnings (blocklength)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_BLOCKLENGTH_ZERO,
                                   std::string("WARNING - Blocklength is 0!")));


        // buffer
        //======================

        // errors (buffer)
        //----------------------
        // MARMOT_ERR_BUFFER,            // invalid buffer pointer
        // MARMOT_ERR_TRUNCATE,          // message truncated on receive
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_BUFFER_ATTACHED,
                                   std::string("ERROR - There is already a buffer attached!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_BUFFER_NOT_ATTACHED,
                                   std::string("ERROR - No buffer attached!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_BUFFER_IN_USE,
                                   std::string("ERROR - Speciefied buffer still in use!")));

        // color
        //======================

        // errors (color)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COLOR_NEG,
                                   std::string("ERROR - Color is negative!")));

        // warnings (color)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_COLOR_UNDEFINED,
                                   std::string("WARNING - Color is MPI_UNDEFINED!")));


        // communicator
        //======================

        // errors (communicator)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COMM_NULL,
                                   std::string("ERROR - Communicator is MPI_COMM_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COMM_NOT_VALID,
                                   std::string("ERROR - Communicator is not valid!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COMM_NOT_CARTESIAN,
                                   std::string("ERROR - Communicator is not cartesian!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COMM_NOT_GRAPH,
                                   std::string("ERROR - Communicator is not graph!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COMM_NOT_INTERCOMM,
                                   std::string("ERROR - Communicator is not intercommunicator!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COMM_INTERCOMM,
                                   std::string("ERROR - Communicator is intercommunicator!")));

        // warnings (communicator)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_COMM_NOT_WORLD,
                                   std::string("WARNING - Communicator is not MPI_COMM_WORLD!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_COMM_NULL,
                                   std::string("WARNING - Communicator is MPI_COMM_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_COMM_LEFT,
                                   std::string("WARNING - Communicator(s) not freed!")));


        // coordinate
        //======================

        // errors (coordinate)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COORD_NEG,
                                   std::string("ERROR - Coordinate is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COORD_TOO_BIG,
                                   std::string("ERROR - Coordinate is out of range!")));


        // count
        //======================

        // errors (count)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COUNT_NEG,
                                   std::string("ERROR - Count is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COUNT_ZERO,
                                   std::string("ERROR - Count is 0!")));
        // MARMOT_ERR_COUNT_ARRAY_NEG,

        // warnings (count)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_COUNT_ZERO,
                                   std::string("WARNING - Count is 0!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_COUNTS_NOT_EQUAL,
                                   std::string("WARNING - Counts are not equal!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_COUNT_UNDEFINED,
                                   std::string("WARNING - Count is MPI_UNDEFINED!")));


        // datatype
        //======================

        // errors (datatype)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_NULL,
                                   std::string("ERROR - Datatype is MPI_DATATYPE_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_NOT_VALID,
                                   std::string("ERROR - Datatype is not valid!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_NOT_COMMITTED,
                                   std::string("ERROR - Datatype is created, not committed!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_C,
                                   std::string("ERROR - Datatype is C datatype!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_F,
                                   std::string("ERROR - Datatype is Fortran datatype!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_REDUCTION_C,
                                   std::string("ERROR - Datatype is for reduction function (C)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_REDUCTION_F,
                                   std::string("ERROR - Datatype is for reduction function (F)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_OPTIONAL_C,
                                   std::string("ERROR - Datatype is optional (C)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_OPTIONAL_F,
                                   std::string("ERROR - Datatype is optional (F)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_CONSTR_DERIVED,
                                   std::string("ERROR - Datatype is for constructing derived datatypes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_LB,
                                   std::string("ERROR - Datatype is MPI_LB!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_UB,
                                   std::string("ERROR - Datatype is MPI_UB!")));

        // warnings (datatype)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TYPE_REDUCTION_C,
                                   std::string("WARNING - Datatype is for reduction function (C)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TYPE_REDUCTION_F,
                                   std::string("WARNING - Datatype is for reduction function (F)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TYPE_OPTIONAL_C,
                                   std::string("WARNING - Datatype is optional (C)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TYPE_OPTIONAL_F,
                                   std::string("WARNING - Datatype is optional (F)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TYPE_LEFT,
                                   std::string("WARNING - Datatype(s) not freed!")));


        // notes (datatype)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_TYPE_NULL,
                                   std::string("NOTE - Datatype is MPI_TYPE_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C,
                                   std::string("NOTE - Datatype is already committed (C)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C_AND_F,
                                   std::string("NOTE - Datatype is already committed (C and F)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_TYPE_ALREADY_COMMITTED_F,
                                   std::string("NOTE - Datatype is already committed (F)!")));


        // dimension
        //======================

        // errors (dimension)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_DIM_NEG,
                                   std::string("ERROR - Dimension is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_DIM_TOO_BIG,
                                   std::string("ERROR - Dimension is too big!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_NODES_NEQ_PRODUCT_DIMS,
                                   std::string("ERROR - Nodes are not a multiple of dims!")));

        // warnings (dimension)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_DIM_ZERO,
                                   std::string("WARNING - Dimension is zero!")));


        // displacement
        //======================

        // errors (displacement)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA,
                                   std::string("ERROR - Displacement overrides data!")));

        // notes (displacement)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_DISPLACEMENT_READS_DATA,
                                   std::string("NOTE - Displacement reads  data / holes read from other nodes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA,
                                   std::string("NOTE - Displacement overrides data!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_DISPLACEMENT_DATA_SENT_TWICE,
                                   std::string("NOTE - Displacement sends data twice!")));


        // edge
        //======================

        // errors (edge)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_EDGE_INVALID,
                                   std::string("ERROR - Edge is invalid!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_EDGE_NEG,
                                   std::string("ERROR - Edge is negative!")));

        // errhandler
        //======================

        // errors (errhandler)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_ERRHANDLER_NOT_VALID,
                                   std::string("ERROR - Errhandler is not valid!")));

        // warnings (errhandler)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_ERRHANDLER_LEFT,
                                   std::string("WARNING - Errhandler(s) not freed!")));


        // errorcode
        //======================

        // errors (errorcode)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_ERRORCODE_NOT_VALID,
                                   std::string("ERROR - Errorcode is not valid!")));



        // group
        //======================

        // errors (group)
        //----------------------
        // MARMOT_ERR_GROUP_NULL,             // group is MPI_GROUP_NULL
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_GROUP_NOT_VALID,
                                   std::string("ERROR - Group is not valid!")));

        // warnings (group)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_GROUP_NULL,
                                   std::string("WARNING - Group is MPI_GROUP_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_GROUP_LEFT,
                                   std::string("WARNING - Group(s) not freed!")));


        // index
        //======================

        // errors (index)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_INDEX_NEG,
                                   std::string("ERROR - Index is negative!")));

        // warnings (index)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_INDEX_LT_SIZE,
                                   std::string("WARNING - Index is less than size of communicator!")));


        // leader
        //======================

        // errors (leader)
        //----------------------
        // Should we also distinguish between errors for remote
        // and for local leaders?
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_LEADER_ANY_SOURCE,
                                   std::string("ERROR - Leader is MPI_ANY_SOURCE!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_LEADER_NEG,
                                   std::string("ERROR - Leader is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_LEADER_OUT_OF_COMM,
                                   std::string("ERROR - Leader is out of communicator!")));


        // neighbor
        //======================

        // errors (neighbor)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_NEIGHBORS_NEG,
                                   std::string("ERROR - Number of neighbors is negative!")));


        // operator
        //======================

        // errors (operator)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_OP_NULL,
                                   std::string("ERROR - Operator is MPO_OP_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_OP_NOT_VALID,
                                   std::string("ERROR - Operator is not valid!")));

        // warnings (operator)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_OP_LEFT,
                                   std::string("WARNING - Operator(s) not freed!")));



        // pending messages
        //======================

        // warnings (pending messages)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_PENDING_MESSAGES_LEFT,
                                   std::string("WARNING - Pending messages are left!")));


        // range
        //======================

        // errors (range)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_FIRST_NEG,
                                   std::string("ERROR - First rank is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_LAST_NEG,
                                   std::string("ERROR - Last rank is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_FIRST_TOO_BIG,
                                   std::string("ERROR - First rank is too big!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_LAST_TOO_BIG,
                                   std::string("ERROR - Last rank is too big!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_COMPUTED_NEG,
                                   std::string("ERROR - Computed rank is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_COMPUTED_TOO_BIG,
                                   std::string("ERROR - Computed rank is too big!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_COMPUTED_TWICE,
                                   std::string("ERROR - Rank is computed twice!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_STRIDE_NEG,
                                   std::string("ERROR - Stride is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_STRIDE_ZERO,
                                   std::string("ERROR - Stride is zero!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_STRIDE_POS,
                                   std::string("ERROR - Stride is positive!")));


        // rank
        //======================

        // errors (rank, see also range errors)
        //----------------------
        // MARMOT_ERR_RANK,              // invalid rank
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_NEG,
                                   std::string("ERROR - Rank is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_PROC_NULL,
                                   std::string("ERROR - Rank is MPI_PROC_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_ANY_SOURCE,
                                   std::string("ERROR - Rank is MPI_ANY_SOURCE!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_RANK_TOO_BIG,
                                   std::string("ERROR - Rank is too big!")));
        // MARMOT_ERR_DUP_RANK,
        // MARMOT_ERR_RANK_ARRAY,
        // MARMOT_ERR_LOCAL_RANK,
        // MARMOT_ERR_REMOTE_RANK,

        // warnings (rank)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_RANK_ANY_SOURCE,
                                   std::string("WARNING - Rank is MPI_ANY_SOURCE!")));

        // notes (rank)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_RANK_PROC_NULL,
                                   std::string("NOTE - Rank is MPI_PROC_NULL!")));


        // request
        //======================

        // errors (request)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_REQUEST_COUNT_NEG,
                                   std::string("ERROR - Count of requests is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_REQUEST_NOT_VALID,
                                   std::string("ERROR - Request is not valid!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_REQUEST_STILL_USED,
                                   std::string("ERROR - Request is still in use!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_REQUEST_NOT_PERSISTENT,
                                   std::string("ERROR - Request is not persistent!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_REQUEST_OUTCOUNT_UNDEFINED,
                                   std::string("ERROR - Outcount is MPI_UNDEFINED!")));

        // warnings (request)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUEST_COUNT_ZERO,
                                   std::string("WARNING - Count of requests is zero!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUEST_NULL,
                                   std::string("WARNING - Request is MPI_REQUEST_NULL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUEST_ACTIVE_NOT_PERSISTENT,
                                   std::string("WARNING - Request is not persistent and active !")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT,
                                   std::string("WARNING - Request is persistent and still active!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUEST_INACTIVE_PERSISTENT,
                                   std::string("WARNING - Request is persistent and inactive!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUEST_OUTCOUNT_UNDEFINED,
                                   std::string("WARNING - Outcount is MPI_UNDEFINED!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUESTS_LEFT,
                                   std::string("WARNING - Requests are left!")));


        // size
        //======================

        // errors (size)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_SIZE_NEG,
                                   std::string("ERROR - Size is negative!")));

        // warnings (size)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_SIZE_ZERO,
                                   std::string("WARNING - Size is zero!")));


        // tag
        //======================

        // errors (tag)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TAG_NEG,
                                   std::string("ERROR - Tag is negative !")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TAG_TOO_BIG,
                                   std::string("ERROR - Tag is too big ( greather than TAG_UB)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TAG_ANY_TAG,
                                   std::string("ERROR - Tag is MPI_ANY_TAG!")));

        // warnings (tag)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TAG_TOO_BIG,
                                   std::string("WARNING - Tag is too big ( greather than guaranteed range, but less than TAG_UB)!")));


        // topology
        //======================

        // errors (topology, see also dimension errors, coordinates errors)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD,
                                   std::string("ERROR - New topology is greater than old communicator!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD,
                                   std::string("ERROR - New cartesian topology is greater than old communicator!")));


        // warnings (topology, see also dimension errors, coordinates errors)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD,
                                   std::string("WARNING - New topology is smaller than old communicator!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD,
                                   std::string("WARNING - New cartesian topology is smaller than old communicator!")));


        // misc
        //======================

        // misc
        // MARMOT_ERR_PENDING,           // Pending request
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_DEADLOCK,
                                   std::string("ERROR - A Deadlock has appeared or is possible!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_DEADLOCK,
                                   std::string("WARNING - A Deadlock might have appeared !")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_POINTER_NULL,
                                   std::string("ERROR - A pointer was null !")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_MPI_AFTER_FINALIZE,
                                   std::string("ERROR - called MPI after MPI_Finalize !")));

        //======================
        //======================
        // MPI-2
        // file I/O
        // not implemented yet
        // error types in
        // alphabetical order
        //======================
        //======================


        // amode
        //======================

        // errors (amode)
        //----------------------

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_DIFF,
                                   std::string("ERROR - File access mode is not the same for all processes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_RDONLY,
                                   std::string("ERROR - MPI_MODE_RDONLY is specified!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_WRONLY,
                                   std::string("ERROR - MPI_MODE_WRONLY is specified!")));

        // MARMOT_ERR_AMODE_RDONLY_AND_RDWR,
        // MARMOT_ERR_AMODE_RDONLY_AND_WRONLY,
        // MARMOT_ERR_AMODE_RDWR_AND_WRONLY,
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_RDONLY_RDWR_WRONLY_NOT_SET_EXACTLY_ONE,
                                   std::string("ERROR - Exactly one of MPI_MODE_RDONLY, MPI_MODE_RDWR, MPI_MODE_WRONLY must be specified!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_RDONLY_AND_CREATE,
                                   std::string("ERROR - MPI_MODE_RDONLY is specified with MPI_MODE_CREATE!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_RDONLY_AND_EXCL,
                                   std::string("ERROR - MPI_MODE_RDONLY is specified with MPI_MODE_EXCL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_RDWR_AND_SEQUENTIAL,
                                   std::string("ERROR - MPI_MODE_RDWR is specified with MPI_MODE_SEQUENTIAL!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_UNIQUE_OPEN,
                                   std::string("ERROR - MPI_MODE_UNIQUE_OPEN is specified!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_SEQUENTIAL,
                                   std::string("ERROR - MPI_MODE_SEQUENTIAL is specified!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_SEQUENTIAL_AND_NOT_DISPLACEMENT_CURRENT,
                                   std::string("ERROR - MPI_MODE_SEQUENTIAL is specified, but not MPI_DISPLACEMENT_CURRENT!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_AMODE_SEQUENTIAL_NOT_SET_BUT_DISPLACEMENT_CURRENT,
                                   std::string("ERROR - MPI_MODE_SEQUENTIAL is not specified, but MPI_DISPLACEMENT_CURRENT!")));

        // notes (amode)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_AMODE_DELETE_ON_CLOSE,
                                   std::string("NOTE - Amode is MPI_DELETE_ON_CLOSE!")));


        // communicator - see also above
        //======================

        // errors (communicator)
        //----------------------

        // communicator is intercommunicator, see above
        // MARMOT_ERR_COMM_INTERCOMM,


        // data representation
        //======================

        // errors (data representation)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_DATAREP_DIFF,
                                   std::string("ERROR - Data representation is not the same for all processes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_DATAREP_GT_MAX_DATAREP_STRING,
                                   std::string("ERROR - Data representation string is larger than MPI_MAX_DATAREP_STRING!")));

        // warnings (data representation)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_DATAREP_TOO_SMALL,
                                   std::string("WARNING - Data representation is not large enough!")));


        // displacement
        //======================

        // errors (displacement)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILETYPE_DISPLACEMENTS_NEG,
                                   std::string("ERROR - Displacements in typemap of filetype are negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILETYPE_DISPLACEMENTS_DECREASE,
                                   std::string("ERROR - Displacements in typemap of filetype are decreasing!")));

        // warnings (displacement)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_DATA_IN_NEW_REGION_UNDEFINED,
                                   std::string("WARNING - Values of data in new regions (those locations with displacements between old file size and size) are undefined!")));



        // datatype - see also above datatype errors
        //======================

        // errors (datatype)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_TYPE_OVERLAPPING,
                                   std::string("ERROR - Datatype has overlapping regions!")));


        // etype - see also above datatype errors
        //======================

        // errors (etype)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (       MARMOT_ERR_ETYPE_EXTENT_DIFF,
                                          std::string("ERROR - Extent of etype is not the same on all processes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_ETYPE_NOT_VALID,
                                   std::string("ERROR - Etype is not valid, see above data type errors!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_ETYPE_OVERLAPPING,
                                   std::string("ERROR - Etype has overlapping regions!")));


        // filetype - see also above datatype errors
        //======================

        // errors (filetype)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILETYPE_EXTENT_HOLE_NO_MULTIPLE_EXTENT_ETYPE,
                                   std::string("ERROR - Extent of hole in the filetype is not a multiple of the extent of the etype!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILETYPE_NOT_VALID,
                                   std::string("ERROR - Filetype is not valid, see above data type errors!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILETYPE_OVERLAPPING,
                                   std::string("ERROR - Filetype has overlapping regions!")));

        // notes (filetype)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_FILETYPE_HAS_HOLES,
                                   std::string("NOTE - Filetype has holes!")));


        // filename
        //======================

        // errors (filename)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILE_DIFF,
                                   std::string("ERROR - Filenames do not reference the same file!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILE_DOES_NOT_EXIST,
                                   std::string("ERROR - File does not exist!")));

        // warnings (filename)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_FILE_IN_USE,
                                   std::string("WARNING - File is still in use!")));


        // group - see above
        //======================

        // errors (group)
        //----------------------
        // MARMOT_ERR_GROUP_NOT_VALID,        // group is not valid

        // warnings (group)
        //----------------------
        // MARMOT_WARN_GROUP_NULL,             // group is MPI_GROUP_NULL


        // info
        //======================

        // errors (info)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_INFO_DIFF,
                                   std::string("ERROR - Info entry is not the same for all processes !")));

        // notes (info)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_INFO_IGNORED,
                                   std::string("NOTE - Implementation is free to ignore info!")));


        // offset
        //======================

        // errors (offset)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_OFFSET_NEG,
                                   std::string("ERROR - Offset is negative!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_OFFSET_TO_NEG_POS_IN_VIEW,
                                   std::string("ERROR - Offset reaches a negative position in the view!")));


        // size
        //======================

        // errors (size)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILE_SIZE_DIFF,
                                   std::string("ERROR - Size is not the same for all processes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILE_SIZE_NEG,
                                   std::string("ERROR - Size is negative!")));

        // notes (size)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_NOTE_FILE_SIZE_LT_OR_EQ_CURRENT_SIZE,
                                   std::string("NOTE - Size is less than or equal current size!")));


        // misc
        //======================

        // errors (misc)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_REQUESTS_NONBLOCKING_LEFT,
                                   std::string("ERROR - Request left on file handle, see above!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_LEFT,
                                   std::string("ERROR - Split collective operation left on file handle!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_OPERATION_SPLIT_COLLECTIVE_COUNT_GT_ONE,
                                   std::string("ERROR - There is more than one active split collective operation on file handle!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_NO_MATCHING_BEGIN,
                                   std::string("ERROR - There is no matching begin call!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_WHENCE_DIFF,
                                   std::string("ERROR - Whence argument is not the same on all processes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FLAG_DIFF,
                                   std::string("ERROR - Flag argument is not the same on all processes!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_FILEHANDLE_DIFF,
                                   std::string("ERROR - File handle argument is not the same on all processes!")));

        // warnings (misc)
        //----------------------
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_FILE_OPEN_LEFT,
                                   std::string("WARNING - Open file left (after MPI_Finalize)!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_REQUESTS_NONBLOCKING_LEFT,
                                   std::string("WARNING - Request left on file handle!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_OPERATION_SPLIT_COLLECTIVE_LEFT,
                                   std::string("WARNING - Split collective operation left on file handle!")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_DEBUGSERVER_NO_EXCLUSIVE_NODE,
                                   std::string("WARNING - The debugserver does not run on an exclusive node")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_PROCESSES_SHARE_NODE,
                                   std::string("WARNING - Two processes share the same node")));
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_LAST_ERR_CODE,
                                   std::string("LAST_ERR_CODE!")));

        //======================
        //======================
        // MPI-2
        // thread level
        // error types in
        // alphabetical order
        //======================
        //======================
        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_THREADLEVEL_GOT_LESS,
                                   std::string("Required Threadlevel not achieved")));

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_FINALIZE_NOT_THREAD_MAIN,
                                   std::string("MPI_Finalize was not performed by the thread that performed MPI_Init")));

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_FINALIZE_OTHER_THREADS_WORKING,
                                   std::string("MPI_Finalize was issued while other Threads of the process were still working")));

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_REQUEST_USED_AT_MULTIPLE_THREADS,
                                   std::string("Multiple threads made MPI calls that used the same Request")));

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_COMM_USED_IN_MULTIPLE_COLLECTIVES,
                                   std::string("A communicator was used in two collective calls of a process simultaneously")));

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_WARN_PROBE_INVALIDATED,
                                   std::string("A call to MPI_Probe or MPI_Iprobe was invalidated by another thread.")));

        MARMOT_ERR_STRINGS.insert(std::make_pair
                                  (MARMOT_ERR_VIOLATED_THREAD_LEVEL,
                                   std::string("The thread level provided by MPI_Init[_thread] was violated.")));



#endif

		myErrorsSet = true;

        return 0;
    }                           // end setErrorMessages



    // getErrorMessages
    //========================================================================
    /** For every enumeration value of MARMOT_ERR_CODES,
     *  we output an error string here. This is to keep MARMOT's logging
     *  in one place here instead of splattering the logging all over
     *  the source code.
     *
     *  @see enum MARMOT_ERR_CODES
     *  @see MARMOT_ERR_STRINGS
     */
    int DebugServer::getErrorMessages(int errorcodes)
    {
		setErrorMessages();

        if (MARMOT_SUCCESS != errorcodes)
        {
            // Currently the check functions return ints, not enums
            // (if we want to be very correct, we'll have to return an
            // enum MARMOT_ERR_CODES, not an int...)
            // Here we have to cast the int back to an enum,
            // or else it does not compile... Alternatively, we could change
            // the types in the map MARMOT_ERR_STRINGS...
//             std::cerr << "errrocode is "
//                       << errorcode
//                       << std::endl;
            std::cerr << MARMOT_ERR_STRINGS.find(static_cast
                                                 <enum MARMOT_ERR_CODES>
                                                 (errorcodes))->second
                      << std::endl;
        }


        return 0;
    }                           // end getErrorMessages



    // outputErrorMessages
    //========================================================================
    /** For any enumeration value of MARMOT_ERR_CODES,
     *  we output an error string here. This is to keep MARMOT's logging
     *  in one place here instead of splattering the logging all over
     *  the source code.
     *  Same as getErrorMessages, just defined as ostream.
     *
     *  @see enum MARMOT_ERR_CODES
     *  @see MARMOT_ERR_STRINGS
     */
    std::ostream
    & DebugServer::outputErrorMessages(int errorcodes,
                                       std::ostream & output)
    {
        // Currently the check functions return ints, not enums
        // (if we want to be very correct, we'll have to return an
        // enum MARMOT_ERR_CODES, not an int...)
        // Here we have to cast the int back to an enum,
        // or else it does not compile... Alternatively, we could change
        // the types in the map MARMOT_ERR_STRINGS...
//            output << "errrocode is "
//                   << errorcode
//                   << std::endl;

		setErrorMessages();

        output << MARMOT_ERR_STRINGS.find(static_cast
                                          <enum MARMOT_ERR_CODES>
                                          (errorcodes))->second
               << std::endl;

        return output;
    }                           // end outputErrorMessages







    // initLogFile
    //========================================================================
    int DebugServer::initLogFile (void)
    {
      std::string LogFileName;  // for the name of the log file
      char TimeStr[16];         // for the timestamp
      time_t now;
      struct tm *snow = NULL;
      if (getenv("MARMOT_LOGFILE_NAME") == NULL)    // is this environemnt varibale defined?
      {
        time(&now);
        snow = localtime(&now);
        sprintf(TimeStr, "%i%02i%02i_%02i%02i%02i", snow->tm_year+1900, snow->tm_mon+1, snow->tm_mday, snow->tm_hour, snow->tm_min, snow->tm_sec);
        LogFileName = "Marmot_";
        LogFileName.append(myExeName);
        LogFileName.append("_");
        LogFileName.append(TimeStr);
      }
      else                      //user provides the log file name
      {
          LogFileName=getenv("MARMOT_LOGFILE_NAME");
      }
      if ( getenv("MARMOT_LOGFILE_PATH") == NULL ) {
#ifndef _WIN32
        MARMOT_LOGFILE_PATH = MARMOT_LOGFILE_PATH_DEFAULT;
#endif
      } else {
#ifdef _WIN32
        MARMOT_LOGFILE_PATH = getenv("MARMOT_LOGFILE_PATH") + string("\\");
#else
        MARMOT_LOGFILE_PATH = getenv("MARMOT_LOGFILE_PATH") + string("/");
#endif
      }
      switch (LOGFILE_TYPE)
      {
            case LOGFILE_TYPE_ASCII:
                my_log = new MPO_LogASCII (MARMOT_LOGFILE_PATH+LogFileName+".txt");
                break;
            case LOGFILE_TYPE_HTML:
                my_log = new MPO_LogHTML (MARMOT_LOGFILE_PATH+LogFileName+".html",true);
                break;
            case LOGFILE_TYPE_CUBE:
            case LOGFILE_TYPE_CUBE_DDT:
            {
#ifdef CUBE
                my_log = new MPO_LogCube (MARMOT_LOGFILE_PATH+LogFileName+".cube");
#endif

#ifndef CUBE
                my_log = new MPO_LogASCII (MARMOT_LOGFILE_PATH + LogFileName+"MarmotLog.txt");
                MPO_MsgObject msg;
                msg.initMsgObject (MSG_TYPE_ERROR,0,-1,"You have set CUBE logging via environmental variable but you haven`t set \"--enable-cube\" as you run \"sh configure\" for MARMOT.",MARMOT_ERR_MARMOT);
                my_log->getMsg(msg);
#endif

                break;
            }
      }
      return 0;
    }//end of initLogFile

    // initLogFile
    //========================================================================
    void DebugServer::logThreadSyncInfo(void) const
    {
        MPO_MsgObject logmsg;
        MPO_String msg;
        std::stringstream stream;
        stream
                << "Thread Syncronisation is "
#ifdef MARMOT_USE_THREADSYNC
                << "enabled."
                << std::endl
                << "Implementation used for Synchronisation: "
                << MPO_Space::getSyncObject()->getImplementationName()
#else
                << "disabled."
                << "If you are using multiple threads errors might occur"
#endif
                << std::endl;

        msg.assign(&stream);
        logmsg.initMsgObject (MSG_TYPE_INFO,0,-1,msg,MARMOT_SUCCESS);
        my_log->getMsg(logmsg);
    }
}                               // end of namespace MPO
