/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MessengerUsingMPI.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo-messenger.cc 838 2008-06-13 10:44:34Z hpchimml $
 */

#include <string>
#include "config.h"
#include "mpo-messenger.h"
#include "mpo.h"

#include <sstream>
#include <iostream>


namespace MPO
{
    //                         Messenger
    //***********************************************************************
 
    // default constructor
    //========================================================================
    Messenger::Messenger()
    {
    }

    // destructor
    //========================================================================
    Messenger::~ Messenger()
    {
    }

 

    //                         MessengerUsingMPI
    //***********************************************************************

    const int MessengerUsingMPI::our_tag = 1; //our_tag mst be less than MESSENGER_TAG_OFFSET !!!

//bk// old version, in X# testbed I got sometimes a message such as
//bk// MPI_RECV: Invalid rank 1073792080
//bk// right after the table of environment variables, befor the MPI_Init.


//     // default constructor
    //========================================================================
//     MessengerUsingMPI::MessengerUsingMPI()
//     {

//         // server rank is the highest rank in MPI_COMM_WORLD;
//         // MPI_Comm_size(MPI_COMM_WORLD,&my_server_rank);
//         // my_server_rank--;
//         my_server_rank = MPI_PROC_NULL;

//         // we create an own communicator to prevent any interference with 
//         // messages sent by the application
//         // MPI_Comm_dup(MPI_COMM_WORLD,&my_communicator);
//         my_communicator = MPI_COMM_NULL;
//     }


    // default constructor
    //========================================================================
    MessengerUsingMPI::MessengerUsingMPI()
        : my_server_rank(MPI_PROC_NULL),
          my_communicator(MPI_COMM_NULL)
    {
    }

/** proper definition of constructor? **/


    // destructor
    //========================================================================
    MessengerUsingMPI::~MessengerUsingMPI()
    {
    }


    // poll
    //========================================================================
    bool MessengerUsingMPI::poll(int &rank,int &tag)
    {
        MPI_Status status;
        int flag = 0;
        bool rvalue = true;

        PMPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, my_communicator, &flag, 
                    &status);

        if (!flag)
        {
            rvalue = false;
        }
		else
		{
			//Tobias: put this command in the else -> as status is undefined if !flag holds true !!!!
			//		  we should improve our MPI ;)
        	rank = status.MPI_SOURCE;
        	tag = status.MPI_TAG;
		}

        // default true
        return rvalue;
    }


    // sendId
    //========================================================================
    int MessengerUsingMPI::sendId(int id, int rank)
    {
        int tag = MESSENGER_TAG_OFFSET;
        
#ifdef MARMOT_USE_THREADSYNC
        //the +2 is used to have a tag distinct from the standard tag 
        tag = MPO_Space::getSyncObject()->getThreadRank() + MESSENGER_TAG_OFFSET;
#endif        

        // client sends id to debug server / debug server sends id to client
        if (rank == -1)
        {
            rank = my_server_rank;
        }

        PMPI_Send(&id, 1, MPI_INT, rank, tag, my_communicator);

        return 0;
    }
    
    // sendId
    //========================================================================
    int MessengerUsingMPI::sendId(int id, int rank, int tag)
    {
        
        // client sends id to debug server / debug server sends id to client
        if (rank == -1)
        {
            rank = my_server_rank;
        }
        
        if (tag < 0)
            tag = our_tag;

        PMPI_Send(&id, 1, MPI_INT, rank, tag, my_communicator);

        return 0;
    }


    // receiveId
    //========================================================================
    int MessengerUsingMPI::receiveId(int &id, int rank)
    {
        
        int tag = MESSENGER_TAG_OFFSET;

#ifdef MARMOT_USE_THREADSYNC
        //the +2 is used to have a tag distinct from the standard tag 
        tag = MPO_Space::getSyncObject()->getThreadRank() + MESSENGER_TAG_OFFSET;
#endif
        
        // client receives id from debug server / debug server receives id
        // from client
        if (rank == -1)
        {
            rank = my_server_rank;
        }        

        MPI_Status status;
        PMPI_Recv(&id, 1, MPI_INT, rank, tag, my_communicator, &status);

        return 0;
    }
    
    // receiveId
    //========================================================================
    int MessengerUsingMPI::receiveId(int &id, int rank, int tag)
    {

        // client receives id from debug server / debug server receives id
        // from client
        if (rank == -1)
        {
            rank = my_server_rank;
        }
        
        if (tag < 0)
            tag = our_tag;

        MPI_Status status;
        PMPI_Recv(&id, 1, MPI_INT, rank, tag, my_communicator, &status);

        return 0;
    }


    // sendObject
    //========================================================================
    int MessengerUsingMPI::sendObject(MPO_Object &obj, int rank)
    {
        int tag = MESSENGER_TAG_OFFSET;

        #ifdef MARMOT_USE_THREADSYNC
                //the +2 is used to have a tag distinct from the standard tag 
                tag = MPO_Space::getSyncObject()->getThreadRank() + MESSENGER_TAG_OFFSET;
        #endif
        
        // client sends object to debug server / debug server sends 
        // object to client
        if (rank == -1)
        {
            rank = my_server_rank;
        }
        
        std::ostringstream s_buffer;
        obj.save(s_buffer);
        
        // message does not contain the final null character
        int msg_size = s_buffer.str().size();
        PMPI_Send((void *) s_buffer.str().c_str(), msg_size, MPI_BYTE,
                  rank, tag, my_communicator);

        return 0;
    }
    
    // sendObject
    //========================================================================
    int MessengerUsingMPI::sendObject(MPO_Object &obj, int rank, int tag)
    {
        // client sends object to debug server / debug server sends 
        // object to client
        if (rank == -1)
        {
            rank = my_server_rank;
        }
        
        if (tag < 0)
            tag = our_tag;
        
        std::ostringstream s_buffer;
        obj.save(s_buffer);
        
        // message does not contain the final null character
        int msg_size = s_buffer.str().size();
        PMPI_Send((void *) s_buffer.str().c_str(), msg_size, MPI_BYTE,
                  rank, tag, my_communicator);

        return 0;
    }


    // receiveObject
    //========================================================================
    int MessengerUsingMPI::receiveObject(MPO_Object &obj, int rank)
    {
        int tag = MESSENGER_TAG_OFFSET;

#ifdef MARMOT_USE_THREADSYNC
        //the +2 is used to have a tag distinct from the standard tag 
        tag = MPO_Space::getSyncObject()->getThreadRank() + MESSENGER_TAG_OFFSET;
#endif
                        
        // client receives object from debug server / debug server receives 
        // object from client
        if (rank == -1)
        {
            rank = my_server_rank;
        }

        // first we ask for the message size without receiving the message:
        MPI_Status status;
        PMPI_Probe(rank, tag, my_communicator, &status);

        int minsize = 0;
        PMPI_Get_count(&status, MPI_BYTE, &minsize);

        char *buffer = new char[minsize];
        PMPI_Recv(buffer, minsize, MPI_BYTE, rank, tag,
                  my_communicator, &status);
        std::istringstream i_stream(std::string(buffer, minsize));
        obj.restore(i_stream);

        // as far as I know istringstream now takes care of the buffer
        return 0;
    }
    
    // receiveObject
    //========================================================================
    int MessengerUsingMPI::receiveObject(MPO_Object &obj, int rank, int tag)
    {

        // client receives object from debug server / debug server receives 
        // object from client
        if (rank == -1)
        {
            rank = my_server_rank;
        }
        
        if (tag < 0)
            tag = our_tag;

        // first we ask for the message size without receiving the message:
        MPI_Status status;
        PMPI_Probe(rank, tag, my_communicator, &status);

        int minsize = 0;
        PMPI_Get_count(&status, MPI_BYTE, &minsize);

        char *buffer = new char[minsize];
        PMPI_Recv(buffer, minsize, MPI_BYTE, rank, tag,
                  my_communicator, &status);
        std::istringstream i_stream(std::string(buffer, minsize));
        obj.restore(i_stream);

        // as far as I know istringstream now takes care of the buffer
        return 0;
    }


    // sendErrorCodes
    //========================================================================
//     int MessengerUsingMPI::sendErrorobject(int &errorcodes, int rank)
//     {

//         // client sends id to debug server / debug server sends id to client
//         if (rank == -1)
//         {
//             rank = my_server_rank;
//         }

//         PMPI_Send(&errorcodes, 1, MPI_INT, rank, our_tag, my_communicator);

//         return 0;
//     }


    // receiveErrorCodes
    //========================================================================
//     int MessengerUsingMPI::receiveErrorobject(int &errorcodes, int rank)
//     {

//         // client receives id from debug server / debug server receives id
//         // from client
//         if (rank == -1)
//         {
//             rank = my_server_rank;
//         }

//         MPI_Status status;
//         PMPI_Recv(&errorcodes, 1, MPI_INT, rank, our_tag, my_communicator,
//                   &status);

//         return 0;
//     }


    // setCommunicator
    //========================================================================
    void MessengerUsingMPI::setCommunicator(MPI_Comm comm)
    {

        // we duplicate the communicator to avoid interference with the 
        // application
        PMPI_Comm_dup(comm, &my_communicator);

        // server rank is the highest rank in the comm
        // TODO: the server rank could be implemented as an option
        PMPI_Comm_size(my_communicator, &my_server_rank);
        my_server_rank--;
    }


}                               // end of namespace MPO
