/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Allreduce.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Allreduce.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Allreduce::MPO_Allreduce()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_recvbuf(0),
          my_count(-1),
          my_datatype(MPI_DATATYPE_NULL),
          my_op(MPI_OP_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Allreduce::MPO_Allreduce(void *sendbuf,
                                 void *recvbuf,
                                 int count,
                                 MPO_Datatype datatype,
                                 MPO_Op op,
                                 MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_recvbuf(recvbuf),
          my_count(count),
          my_datatype(datatype),
          my_op(op)
    {
    }


    // destructor
    //========================================================================
    MPO_Allreduce::~MPO_Allreduce()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Allreduce::checkPrerequisites2()
    {
        // check correctness of parameters
        checkCount(my_count, MSG_COUNT);
        checkDatatypeSendRecv(my_count, my_datatype,
                              IS_COUNT, IS_REDUCE_ALLREDUCE);
        checkOp(my_op);

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Allreduce::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ALLREDUCE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Allreduce::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Allreduce::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkBufferInUse(my_sendbuf,my_count,my_datatype);
        checkBufferInUse(my_recvbuf,my_count,my_datatype);

        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_op_traces->findAndAddUsage(my_op,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Allreduce::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Allreduce(my_sendbuf, my_recvbuf, my_count, my_datatype,
                              my_op, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Allreduce::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Allreduce";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ALLREDUCE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Allreduce::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "*recvbuf, "
            << "count = "
            << my_count
            << ", datatype = "
            << my_datatype
            << ", op = "
            << my_op
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Allreduce::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        // out < my_recvbuf;
        out < my_count;
        out < my_datatype;
        out < my_op;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Allreduce::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        // in > my_recvbuf;
        in > my_count;
        in > my_datatype;
        in > my_op;
        in > my_comm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/allreduce.c
/*@

MPI_Allreduce - Combines values from all processes and distribute the
result back to all processes

Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. count - number of elements in send buffer (integer)
. datatype - data type of elements of send buffer (handle)
. op - operation (handle)
- comm - communicator (handle)

Output Parameter:
. recvbuf - starting address of receive buffer (choice)

.N fortran

.N collops

.N Errors
.N MPI_ERR_BUFFER
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_OP
.N MPI_ERR_COMM

int MPI_Allreduce (
void *sendbuf,
void *recvbuf,
int count,
MPI_Datatype datatype,
MPI_Op op,
MPI_Comm comm )
@ */

