/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Alltoall class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Alltoall.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Alltoall::MPO_Alltoall()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcount(-1),
          my_sendtype(MPI_DATATYPE_NULL),
          my_recvbuf(0),
          my_recvcount(-1),
          my_recvtype(MPI_DATATYPE_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Alltoall::MPO_Alltoall(void *sendbuf,
                               int sendcount,
                               MPO_Datatype sendtype,
                               void *recvbuf,
                               int recvcount,
                               MPO_Datatype recvtype,
                               MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcount(sendcount),
          my_sendtype(sendtype),
          my_recvbuf(recvbuf),
          my_recvcount(recvcount),
          my_recvtype(recvtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Alltoall::~MPO_Alltoall()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Alltoall::checkPrerequisites2()
    {
        // check correctness of parameters
        checkCount(my_sendcount, MSG_SENDCOUNT);
        checkCount(my_recvcount, MSG_RECVCOUNT);
        checkDatatypeSendRecv(my_sendcount, my_sendtype,
                              IS_SENDCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        checkDatatypeSendRecv(my_recvcount, my_recvtype,
                              IS_RECVCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        checkCountsAreEqual();

        //check buffers
        checkBufferInUse(my_sendbuf,my_sendcount,my_sendtype);
        checkBufferInUse(my_recvbuf,my_recvcount,my_recvtype);

        return 0;
    }


    // checkCountsAreEqual
    //========================================================================
    int MPO_Alltoall::checkCountsAreEqual()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_WARN_COUNTS_NOT_EQUAL))
        return rvalue;

        if ((my_recvcount != my_sendcount) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": Send- and recvcount are NOT equal!"
                << std::endl;
            rvalue = MARMOT_WARN_COUNTS_NOT_EQUAL;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
            our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Alltoall::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ALLTOALL_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Alltoall::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Alltoall::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkPrerequisites2();

        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_recvtype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Alltoall::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Alltoall(my_sendbuf, my_sendcount, my_sendtype, my_recvbuf,
                             my_recvcount, my_recvtype, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Alltoall::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Alltoall";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ALLTOALL";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Alltoall::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "sendcount = "
            << my_sendcount
            << ", sendtype = "
            << my_sendtype
            << ", *recvbuf, "
            << "recvcount = "
            << my_recvcount
            << ", recvtype = "
            << my_recvtype
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Alltoall::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        out < my_sendcount;
        out < my_sendtype;
        // out < my_recvbuf;
        out < my_recvcount;
        out < my_recvtype;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Alltoall::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        in > my_sendcount;
        in > my_sendtype;
        // in > my_recvbuf
        in > my_recvcount;
        in > my_recvtype;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/alltoall.c.
/*@

MPI_Alltoall - Sends data from all to all processes

Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. sendcount - number of elements to send to each process (integer)
. sendtype - data type of send buffer elements (handle)
. recvcount - number of elements received from any process (integer)
. recvtype - data type of receive buffer elements (handle)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice)

.N fortran

.N Errors
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER

int MPI_Alltoall(
void *sendbuf,
int sendcount,
MPI_Datatype sendtype,
void *recvbuf,
int recvcnt,
MPI_Datatype recvtype,
MPI_Comm comm )
@ */


