/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Alltoallv class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Alltoallv.cc 909 2008-12-17 09:48:19Z tobias $
 */


#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Alltoallv::MPO_Alltoallv()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcounts(0),
          my_sdispls(0),
          my_sendtype(MPI_DATATYPE_NULL),
          my_recvbuf(0),
          my_recvcounts(0),
          my_rdispls(0),
          my_recvtype(MPI_DATATYPE_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Alltoallv::MPO_Alltoallv(void *sendbuf,
                                 int *sendcounts,
                                 int *sdispls,
                                 MPO_Datatype sendtype,
                                 void *recvbuf,
                                 int *recvcounts,
                                 int *rdispls,
                                 MPO_Datatype recvtype,
                                 MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcounts(sendcounts),
          my_sdispls(sdispls),
          my_sendtype(sendtype),
          my_recvbuf(recvbuf),
          my_recvcounts(recvcounts),
          my_rdispls(rdispls),
          my_recvtype(recvtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Alltoallv::~MPO_Alltoallv()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Alltoallv::checkPrerequisites2(int *counts,
                                           MPO_Datatype datatype,
                                           int type)
    {
        // check correctness of parameters
        checkDisplacementArrayOfReceiver();
        checkDatatypeOfSender(counts,
                              datatype,
                              type);
        checkCountArray(my_sendcounts, MSG_SENDCOUNT);
        checkCountArray(my_recvcounts, MSG_RECVCOUNT);

        //check buffers
        int size;
        PMPI_Comm_size(my_comm,&size);
        for (int i = 0;i < size;i++)
        {
            checkBufferInUse(my_sendbuf,my_sdispls[i]+my_sendcounts[i],my_sendtype);
            checkBufferInUse(my_recvbuf,my_rdispls[i]+my_recvcounts[i],my_sendtype);
        }

        return 0;
    }


    // checkDisplacementArrayOfReceiver
    //========================================================================
    int MPO_Alltoallv::checkDisplacementArrayOfReceiver()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        int rank = -1;
        PMPI_Comm_rank(my_comm, &rank);

        // compilation error:  MPI_Aint typeextent may be int or long, so
        // use instead of
        //     int used = 0;
        // an MPI_Aint and cast typesize to MPI_Aint in the if condition
        // below.
        MPI_Aint used = 0;

        MPI_Aint typeextent = 0;
        PMPI_Type_extent(my_recvtype, &typeextent);

        int typesize = -1;
        PMPI_Type_size(my_recvtype, &typesize);

        // (1) Here we have a compact datatype without holes
        // so overriding are errors.
        if (typeextent == (MPI_Aint) typesize)
        {
            //test all pairs of intervals
            for (int i = 0; i < size;i++)
            {
                for (int j = i+1; j < size;j++)
                {
                    //some variables to describe the intervals
                    // (s_i - start of interval i, e_i - end of interval i, ...)
                    int s_i = my_rdispls[i],
                        s_j = my_rdispls[j],
                        e_i = s_i +  my_recvcounts[i],
                        e_j = s_j +  my_recvcounts[j];

                    //are the two intervals intersecting ?
                    if ( (e_i > s_j)             &&
                         (s_i < e_j)             &&
                         (MSG_ERR <= DEBUG_MODE)    )
                    {
                        //yes they intersect each other - give an error
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": On rank "
                            << rank
                            << ": "
                            << "recieve buffer location writen more than once."
                            << "(Conflict: rdispls["
                            << i
                            << "] = "
                            << my_rdispls[i]
                            << ", recvcounts["
                            << i
                            << "] = "
                            << my_recvcounts[i]
                            << ", rdispls["
                            << j
                            << "] = "
                            << my_rdispls[j]
                            << ", recvcounts["
                            << j
                            << "] = "
                            << my_recvcounts[j]
                            << ")"
                            << std::endl;

                        rvalue = MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA;

                        our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                        our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                        our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
                    }//if intersection
                }//for j
            }//for i
        }
        // (2) Here we have a datatype with holes
        // so we don't check whether it is correct.
        else
        {
            // Check array of displacements.
            for (int i = 0; i < size; i++)
            {
                if ((my_rdispls[i] < used) &&
                    (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                {
                    std::stringstream stream;
                    stream
                        << "REMARK: "
                        << getCallname()
                        << ": Datatype has holes, "
                        << "so we don't check whether it's correct,"
                        << "but on rank "
                        << rank
                        << ": displs["
                        << i
                        << "] reads data/ holes read from other nodes!"
                        << std::endl;
                    rvalue = MARMOT_NOTE_DISPLACEMENT_READS_DATA;

                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                    our_log->getMsg(msg);
                }
                used = my_rdispls[i] + (typeextent * my_recvcounts[i]);
            }
        }

        return rvalue;
    }                           // end checkDisplacementArray


    // checkDatatypeOfSender
    //========================================================================
    int MPO_Alltoallv::checkDatatypeOfSender(int *counts,
                                             MPO_Datatype datatype,
                                             int type)
    {
        /** \todo @creator of this method: what about return values, I think there are casses missing when using FORTRAN like see MARMOT_ERR_TYPE_C,...! */

        if (datatype == MPI_DATATYPE_NULL)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                if (type == 1)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname()
                        << ": Sendtype is MPI_DATATYPE_NULL!"
                        << std::endl;

                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                    our_log->getMsg(msg);
                }
                else
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname()
                        << ": Recvtype is MPI_DATATYPE_NULL!"
                            << std::endl;

                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
        }
        else
        {
            if (our_datatypes.find(datatype) == our_datatypes.end())
            {
                if (MSG_ERR <= DEBUG_MODE)
                {
                    if (type == 1)
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": Sendtype is not valid!"
                            << std::endl;

                        our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                        our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                        our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                        our_log->getMsg(msg);
                    }
                    else
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": Recvtype is not valid!"
                            << std::endl;

                        our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                        our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                        our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                        our_log->getMsg(msg);
                    }
                }
            }
            else
            {
                switch (our_datatypes.find(datatype)->second)
                {
                    case (DATATYPE_CREATED_NOT_COMMITTED):
                        if (MSG_ERR <= DEBUG_MODE)
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Sendtype is not committed!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_COMMITTED,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Recvtype is not committed!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_COMMITTED,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;
                    case (DATATYPE_REDUCTION_C):
                        if (MSG_ERR_WARN <= DEBUG_MODE)
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "WARNING: "
                                    << getCallname()
                                    << ": Sendtype is for reduction functions!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_REDUCTION_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Recvtype is for reduction functions!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_REDUCTION_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;
                    case (DATATYPE_OPTIONAL_C):
                        if (MSG_ERR_WARN <= DEBUG_MODE)
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "WARNING: "
                                    << getCallname()
                                    << ": Sendtype is optional!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_OPTIONAL_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "WARNING: "
                                    << getCallname()
                                    << ": Recvtype is optional!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_OPTIONAL_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;
                    case (DATATYPE_COMMITTED_FORTRAN):
                        if ((MSG_ERR <= DEBUG_MODE) &&
                            (C_INTERFACE == INTERFACE_MODE))
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Sendtype is Fortran-type!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_F,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Recvtype is Fortran-type!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_F,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;

                    default:
                        // do nothing
                        break;

                }
            }
        }
        return 1;
    }


    // getId
    //========================================================================
    int MPO_Alltoallv::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ALLTOALLV_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Alltoallv::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Alltoallv::checkPrerequisites(void)
    {


        // ??? what about that ???  MPO_Collective::checkPrerequisites();
        checkPrerequisites2(my_recvcounts, my_recvtype, 2);

        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_recvtype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Alltoallv::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Alltoallv(my_sendbuf, my_sendcounts, my_sdispls,
                              my_sendtype,
                              my_recvbuf, my_recvcounts, my_rdispls,
                              my_recvtype,
                              my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Alltoallv::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Alltoallv";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ALLTOALLV";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Alltoallv::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << " *sendcounts, "
            << "*sdispls, "
            << "sendtype = "
            << my_sendtype
            << ", *recvbuf, "
            << "*recvcounts, "
            << "*rdispls, "
            << "recvtype = "
            << my_recvtype
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Alltoallv::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf
        // out < my_sendcounts;
        // out < my_displs;
        out < my_sendtype;
        // out < my_recvbuf
        // out < my_recvcounts;
        // out < my_rdispls;
        out < my_recvtype;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Alltoallv::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf
        // in > my_sendcounts;
        // in > my_displs;
        in > my_sendtype;
        // in > my_recvbuf
        // in > my_recvcounts;
        // in > my_rdispls;
        in > my_recvtype;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/alltoallv.c.
/*@

MPI_Alltoallv - Sends data from all to all processes, with a displacement

Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. sendcounts - integer array equal to the group size
specifying the number of elements to send to each processor
. sdispls - integer array (of length group size). Entry
'j'  specifies the displacement (relative to sendbuf  from
which to take the outgoing data destined for process  'j'
. sendtype - data type of send buffer elements (handle)
. recvcounts - integer array equal to the group size
specifying the maximum number of elements that can be received from
each processor
. rdispls - integer array (of length group size). Entry
'i'  specifies the displacement (relative to recvbuf  at
which to place the incoming data from process  'i'
. recvtype - data type of receive buffer elements (handle)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice)

.N fortran

.N Errors
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER

int MPI_Alltoallv (
void *sendbuf,
int *sendcnts,
int *sdispls,
MPI_Datatype sendtype,
void *recvbuf,
int *recvcnts,
int *rdispls,
MPI_Datatype recvtype,
MPI_Comm comm )
@ */
