/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Attr_get class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Attr_get.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Attr_get::MPO_Attr_get()
        : my_comm(MPI_COMM_NULL),
          my_keyval(MPI_KEYVAL_INVALID),
          my_attribute_val(0), 
          my_flag(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Attr_get::MPO_Attr_get(MPO_Comm comm,
                               int keyval,
                               void *attribute_val,
                               int *flag)
        : my_comm(comm),
          my_keyval(keyval), 
          my_attribute_val(attribute_val), 
          my_flag(flag)
    {
    }


    // destructor
    //========================================================================
    MPO_Attr_get::~MPO_Attr_get()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Attr_get::checkPrerequisites()
    {


        // check correctness of parameters
        checkComm();
    
        
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        
        
        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Attr_get::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


    // getId
    //========================================================================
    int MPO_Attr_get::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ATTR_GET_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Attr_get::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Attr_get::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Attr_get(my_comm, my_keyval, my_attribute_val, my_flag);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Attr_get::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Attr_get";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ATTR_GET";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Attr_get::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(comm = "
            << my_comm
            << ", keyval = " 
            << my_keyval 
            << ", *attribute_val, "
            << "*flag";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Attr_get::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_keyval;
        // out < my_attribute_val;
        // out < my_flag;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Attr_get::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_keyval;
        // in > my_attribute_val;
        // in > my_flag;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from 
// mpich-1.2.4/src/context/attr_getval.c.
/*@C

MPI_Attr_get - Retrieves attribute value by key

Input Parameters:
+ comm - communicator to which attribute is attached (handle) 
- keyval - key value (integer) 

Output Parameters:
+ attr_value - attribute value, unless 'flag' = false 
- flag -  true if an attribute value was extracted;  false if no 
attribute is associated with the key 

Notes:
Attributes must be extracted from the same language as they were inserted 
in with 'MPI_ATTR_PUT'.  The notes for C and Fortran below explain why.

Notes for C:
Even though the 'attr_value' arguement is declared as 'void *', it is
really the address of a void pointer.  See the rationale in the 
standard for more details. 

.N fortran

The 'attr_value' in Fortran is a pointer to a Fortran integer, not
a pointer to a 'void *'.  

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_KEYVAL

int MPI_Attr_get ( 
MPI_Comm comm, 
int keyval, 
void *attr_value, 
int *flag )
@ */

