/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Attr_put class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller
 *
 *  $Id: mpo_Attr_put.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Attr_put::MPO_Attr_put()
        : my_comm(MPI_COMM_NULL),
          my_keyval(MPI_KEYVAL_INVALID), 
          my_attribute_val(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Attr_put::MPO_Attr_put(MPO_Comm comm,
                               int keyval,
                               void *attribute_val)
        : my_comm(comm),
          my_keyval(keyval), 
          my_attribute_val(attribute_val)
    {
    }


    // destructor
    //========================================================================
    MPO_Attr_put::~MPO_Attr_put()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Attr_put::checkPrerequisites()
    {


        // check correctness of parameters
        checkComm();
    
        
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        
        
        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Attr_put::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


    // getId
    //========================================================================
    int MPO_Attr_put::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ATTR_PUT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Attr_put::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Attr_put::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Attr_put(my_comm, my_keyval, my_attribute_val);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Attr_put::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Attr_put";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ATTR_PUT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Attr_put::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(comm = "
            << my_comm
            << ", keyval = " 
            << my_keyval 
            << ", *attribute_val" ;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")" 
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Attr_put::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_keyval;
        // out < my_attribute_val;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Attr_put::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_keyval;
        // in > my_attribute_val;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from 
// mpich-1.2.4/src/context/attr_putval.c.  
/*@

MPI_Attr_put - Stores attribute value associated with a key

Input Parameters:
+ comm - communicator to which attribute will be attached (handle) 
. keyval - key value, as returned by  'MPI_KEYVAL_CREATE' (integer) 
- attribute_val - attribute value 

Notes:
Values of the permanent attributes 'MPI_TAG_UB', 'MPI_HOST', 'MPI_IO', and
'MPI_WTIME_IS_GLOBAL' may not be changed. 

The type of the attribute value depends on whether C or Fortran is being 
used.
In C, an attribute value is a pointer ('void *'); in Fortran, it is a 
single integer (`not` a pointer, since Fortran has no pointers and there 
are systems for which a pointer does not fit in an integer 
(e.g., any > 32 bit address 
system that uses 64 bits for Fortran 'DOUBLE PRECISION').

If an attribute is already present, the delete function (specified when the
corresponding keyval was created) will be called.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_KEYVAL
.N MPI_ERR_PERM_KEY

.seealso MPI_Attr_get, MPI_Keyval_create, MPI_Attr_delete

int MPI_Attr_put ( 
MPI_Comm comm, 
int keyval, 
void *attr_value )
@ */

    
    
