/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Buffer_attach.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Buffer_attach.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Buffer_attach::MPO_Buffer_attach()
        : my_buf(0),
          my_size(-1)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Buffer_attach::MPO_Buffer_attach(void *buf,
                                         int size)
        : my_buf(buf),
          my_size(size)
    {
    }


    // destructor
    //========================================================================
    MPO_Buffer_attach::~MPO_Buffer_attach()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Buffer_attach::checkPrerequisites()
    {


        // check correctness of parameters
        checkBufferAttached();



        return 0;
    }


    // checkBufferAttached
    //========================================================================
    int MPO_Buffer_attach::checkBufferAttached()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_ERR_BUFFER_ATTACHED))
        return rvalue;

        // Check if buffer is attached.
        if ((buffer_attached) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": There's already a buffer attached!"
                << std::endl;
            rvalue = MARMOT_ERR_BUFFER_ATTACHED;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return 1;
    }                           // end checkBufferAttached


    // getId
    //========================================================================
    int MPO_Buffer_attach::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_BUFFER_ATTACH_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Buffer_attach::checkPost(void)
    {

        buffer_attached = 1;

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Buffer_attach::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Buffer_attach(my_buf, my_size);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Buffer_attach::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Buffer_attach";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_BUFFER_ATTACH";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Buffer_attach::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*buffer, "
            << "size="
            << my_size;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Buffer_attach::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_buf;
        out < my_size;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Buffer_attach::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_buf;
        in > my_size;

        return in;
    }


}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/bufattach.c.
/*@
  MPI_Buffer_attach - Attaches a user-defined buffer for sending

  Input Parameters:
  + buffer - initial buffer address (choice)
  - size - buffer size, in bytes (integer)

  Notes:
  The size given should be the sum of the sizes of all outstanding Bsends
  that you intend to have, plus a few hundred bytes for each Bsend that
  you do.
  For the purposes of calculating size, you should use 'MPI_Pack_size'.
  In other words, in the code
  .vb
  MPI_Buffer_attach( buffer, size );
  MPI_Bsend( ..., count=20, datatype=type1,  ... );
  ...
  MPI_Bsend( ..., count=40, datatype=type2, ... );
  .ve
  the value of 'size' in the MPI_Buffer_attach call should be greater than
  the value computed by
  .vb
  MPI_Pack_size( 20, type1, comm, &s1 );
  MPI_Pack_size( 40, type2, comm, &s2 );
  size = s1 + s2 + 2 * MPI_BSEND_OVERHEAD;
  .ve
  The 'MPI_BSEND_OVERHEAD' gives the maximum amount of space that may be
  used in the buffer for use by the BSEND routines in using the buffer.
  This value is in 'mpi.h' (for C) and 'mpif.h' (for Fortran).

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_BUFFER
  .N MPI_ERR_INTERN

  .seealso: MPI_Buffer_detach, MPI_Bsend

  int MPI_Buffer_attach(
  void *buffer,
  int size )
  @ */



