/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Cancel.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Cancel.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Cancel::MPO_Cancel()
        : MPO_Completion(0,
                         0)
    {
    }


    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether request is in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove request from our_requests
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     *  (it is a opaque type and thus  may contain some internal status etc. )
     */
    // proper constructor
    //========================================================================
    MPO_Cancel::MPO_Cancel(MPI_Request * request)
        : MPO_Completion(1,
                         request)
    {
    }


    // destructor
    //========================================================================
    MPO_Cancel::~MPO_Cancel()
    {
    }


    // getId
    //========================================================================
    int MPO_Cancel::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_CANCEL_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Cancel::checkPost(void)
    {
        // We still need to call Test or Wait or so to remove request!

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Cancel::checkPrerequisites(void)
    {


        MPO_Completion::checkPrerequisites();

        our_request_traces->findAndAddUsage(my_request_array,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Cancel::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Cancel(my_request_array);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Cancel::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Cancel";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_CANCEL";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Cancel::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*request";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Cancel::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_request_p;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Cancel::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_request_p;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/cancel.c.
/*@
  MPI_Cancel - Cancels a communication request

  Input Parameter:
  . request - communication request (handle)

  Note:
  Cancel has only been implemented for receive requests; it is a no-op for
  send requests.  The primary expected use of MPI_Cancel is in multi-
  buffering schemes, where speculative MPI_Irecvs are made.  When the
  computation completes, some of these receive requests may remain;
  using MPI_Cancel allows the user to cancel these unsatisfied requests.

  Cancelling a send operation is much more difficult, in large part because
  the send will usually be at least partially complete (the information on
  the tag, size, and source are usually sent immediately to the
  destination).
  As of version 1.2.0, MPICH supports cancelling of sends.  Users are
  advised that cancelling a send, while a local operation (as defined by
  the MPI standard), is likely to be expensive (usually generating one or
  more internal messages).

  .N fortran

  .N NULL

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_REQUEST
  .N MPI_ERR_ARG

  int MPI_Cancel( MPI_Request *request )
  @ */


