/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Cart_coords.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Cart_coords.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Cart_coords::MPO_Cart_coords()
        : MPO_Cartesian(MPI_COMM_NULL),
          my_rank(MPI_PROC_NULL),
          my_maxdims(-1),
          my_coords(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Cart_coords::MPO_Cart_coords(MPO_Comm communicator,
                                     int rank,
                                     int maxdims,
                                     int *coords)
        : MPO_Cartesian(communicator),
          my_rank(rank),
          my_maxdims(maxdims),
          my_coords(coords)
    {
    }


    // destructor
    //========================================================================
    MPO_Cart_coords::~MPO_Cart_coords()
    {
    }


    // checkRank
    //========================================================================
    int MPO_Cart_coords::checkRank()
    {

        /** \todo @creator of this method: what about return types ? */

        if (my_rank < 0)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Rank less than 0 !"
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_NEG,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            int commsize = 0;
            PMPI_Comm_size(my_comm, &commsize);
            if ((my_rank >= commsize) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Rank out of communicator!"
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_TOO_BIG,getSourceLocation());
                our_log->getMsg(msg);
            }
        }

        return 1;
    }


    // getId
    //========================================================================
    int MPO_Cart_coords::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_CART_COORDS_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Cart_coords::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Cart_coords::checkPrerequisites(void)
    {


        MPO_Cartesian::checkPrerequisites();
        checkRank();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Cart_coords::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Cart_coords(my_comm, my_rank, my_maxdims, my_coords);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Cart_coords::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Cart_coords";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_CART_COORDS";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Cart_coords::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", rank = "
            << my_rank
            << ", maxdims = "
            << my_maxdims
            << ", *coords";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Cart_coords::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_rank;
        out < my_maxdims;
        // out < my_coords;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Cart_coords::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_rank;
        in > my_maxdims;
        // in > my_coords;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied
// from mpich-1.2.4/src/topol/cart_coords.c.
/*@

MPI_Cart_coords - Determines process coords in cartesian topology given
rank in group

Input Parameters:
+ comm - communicator with cartesian structure (handle)
. rank - rank of a process within group of 'comm' (integer)
- maxdims - length of vector 'coords' in the calling program (integer)

Output Parameter:
. coords - integer array (of size 'ndims') containing the Cartesian
coordinates of specified process (integer)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_RANK
.N MPI_ERR_DIMS
.N MPI_ERR_ARG

int MPI_Cart_coords (
MPI_Comm comm,
int rank,
int maxdims,
int *coords )
@ */

