/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Cart_create.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Cart_create.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Cart_create::MPO_Cart_create()
        : my_comm_old(MPI_COMM_NULL),
          my_ndims(-1), 
          my_dims(0), 
          my_periods(0), 
          my_reorder(0), 
          my_comm_cart(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Cart_create::MPO_Cart_create(MPO_Comm comm_old,
                                     int ndims,
                                     int *dims,
                                     int *periods,
                                     int reorder,
                                     MPI_Comm *
                                     comm_cart)
        : my_comm_old(comm_old),
          my_ndims(ndims), 
          my_dims(dims), 
          my_periods(periods), 
          my_reorder(reorder),
          my_comm_cart(comm_cart)
    {
    }


    // destructor
    //========================================================================
    MPO_Cart_create::~MPO_Cart_create()
    {
    }


    /** We overload MPO_Cartesian::checkPrerequisites to 
     *  check other pre_requisites.
     */
    // checkPrerequisites
    //========================================================================
    int MPO_Cart_create::checkPrerequisites()
    {


        // check correctness of parameters
        checkComm();
        checkTopologyCartNew();
    
        
        
        return 0;
    }

 
    // checkComm
    //========================================================================
    int MPO_Cart_create::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm_old) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm_old,stream,"comm_old");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm_old)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm_old,stream,"comm_old");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


    // checkTopologyCartNew
    //========================================================================
    int MPO_Cart_create::checkTopologyCartNew()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;
    if(!(our_errcodes.at(MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD)||our_errcodes.at(MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD)))
        return rvalue;

        int comm_size = -1;
        PMPI_Comm_size(my_comm_old, &comm_size);

        // Compute total size of cartesian grid
        int i = 0;
        int total_size = 1;

        for (i = 0; i <= my_ndims - 1; i++)
        {
            total_size = total_size * my_dims[i];
        }

        // (1) Check if total size > size of old communicator
        if ((our_errcodes.at(MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD))&&(comm_size < total_size) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream    << "ERROR: "
                      << getCallname()
                      << ": New topology is larger than old communicator:"
                      << std::endl
                      << " comm size = "
                      << comm_size 
                      << ", total size = " 
                      << total_size 
                      << std::endl;
            rvalue = MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm_old,stream,"comm_old");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if total_size < size of old communicator
        else if ((our_errcodes.at(MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD))&&(comm_size > total_size) &&
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": New topology is smaller than old communicator:"
                << " Some nodes will get MPI_COMM_NULL." 
                << std::endl
                << " comm size = "
                << comm_size 
                << ", total size = " 
                << total_size 
                << std::endl;
            rvalue = MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm_old,stream,"comm_old");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkTopologyCartNew


    // getId
    //========================================================================
    int MPO_Cart_create::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_CART_CREATE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Cart_create::checkPost(void)
    {
        
        our_comms.insert(*my_comm_cart);
        
        our_comm_traces->findAndAddUsage(my_comm_old,getSourceLocation());
        our_comm_traces->addAndSetCreation(*my_comm_cart,getSourceLocation());
        our_comm_traces->findAndTryAddDependendency(*my_comm_cart,my_comm_old);
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Cart_create::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Cart_create(my_comm_old, my_ndims, my_dims, my_periods,
                                  my_reorder, my_comm_cart);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Cart_create::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Cart_create";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_CART_CREATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Cart_create::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(comm_old = "
            << my_comm_old
            << ", ndims = "
            << my_ndims
            << ", *dims, "
            << "*periods, "
            << "reorder = "
            << my_reorder 
            << ", *comm_cart";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Cart_create::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm_old;
        out < my_ndims;
        // out < my_dims;
        // out < my_periods;
        out < my_reorder;
        // out < my_comm_cart;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Cart_create::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm_old;
        in > my_ndims;
        // in > my_dims;
        // in > my_periods;
        in > my_reorder;
        // in > my_comm_cart;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied 
// from mpich-1.2.4/src/topol/cart_create.c.
/*@

MPI_Cart_create - Makes a new communicator to which topology information
has been attached

Input Parameters:
+ comm_old - input communicator (handle) 
. ndims - number of dimensions of cartesian grid (integer) 
. dims - integer array of size ndims specifying the number of processes in 
each dimension 
. periods - logical array of size ndims specifying whether the grid is 
periodic (true) or not (false) in each dimension 
- reorder - ranking may be reordered (true) or not (false) (logical) 

Output Parameter:
. comm_cart - communicator with new cartesian topology (handle) 

Algorithm:
We ignore 'reorder' info currently.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_DIMS
.N MPI_ERR_ARG

int MPI_Cart_create 
( 
MPI_Comm comm_old, 
int ndims, 
int *dims, 
int *periods, 
int reorder, 
MPI_Comm *comm_cart 
)
@ */

