/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Cart_get.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Cart_get.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Cart_get::MPO_Cart_get()
        : MPO_Cartesian(MPI_COMM_NULL),
          my_maxdims(-1),
          my_dims(0),
          my_periods(0),
          my_coords(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Cart_get::MPO_Cart_get(MPO_Comm communicator,
                               int maxdims,
                               int *dims,
                               int *periods,
                               int *coords)
        : MPO_Cartesian(communicator),
          my_maxdims(maxdims),
          my_dims(dims),
          my_periods(periods),
          my_coords(coords)
    {
    }


    // destructor
    //========================================================================
    MPO_Cart_get::~MPO_Cart_get()
    {
    }


    // checkMaxDims
    //========================================================================
    int MPO_Cart_get::checkMaxDims()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if maxdims is negative.
        if ((our_errcodes.at(MARMOT_ERR_DIM_NEG))&&(my_maxdims < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Maxdims less than 0 !"
                << std::endl;
            rvalue = MARMOT_ERR_DIM_NEG;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if maxdims is 0.
        if ((our_errcodes.at(MARMOT_WARN_DIM_ZERO))&&(my_maxdims == 0) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << "Maxdims = 0 !"
                << std::endl;
            rvalue = MARMOT_WARN_DIM_ZERO;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Cart_get::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_CART_GET_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Cart_get::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Cart_get::checkPrerequisites(void)
    {


        MPO_Cartesian::checkPrerequisites();
        checkMaxDims();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Cart_get::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Cart_get(my_comm, my_maxdims, my_dims, my_periods,
                             my_coords);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Cart_get::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Cart_get";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_CART_GET";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Cart_get::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", maxdims = "
            << my_maxdims
            << ", *dims, "
            << "*periods, "
            << "*coords";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Cart_get::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_maxdims;
        // out < my_dims;
        // out < my_periods;
        // out < my_coords;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Cart_get::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_maxdims;
        // in > my_dims
        // in > my_periods;
        // in > my_coords;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/topol/cart_get.c.
/*@

MPI_Cart_get - Retrieves Cartesian topology information associated with a
communicator

Input Parameters:
+ comm - communicator with cartesian structure (handle)
- maxdims - length of vectors  'dims', 'periods', and 'coords'
in the calling program (integer)

Output Parameters:
+ dims - number of processes for each cartesian dimension
(array of integer)
. periods - periodicity (true/false) for each cartesian dimension
(array of logical)
- coords - coordinates of calling process in cartesian structure
(array of integer)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG

int MPI_Cart_get (
MPI_Comm comm,
int maxdims,
int *dims,
int *periods,
int *coords )
@ */



