/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Cart_rank.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Cart_rank.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Cart_rank::MPO_Cart_rank()
        : MPO_Cartesian(MPI_COMM_NULL),
          my_coords(0),
          my_rank(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Cart_rank::MPO_Cart_rank(MPO_Comm communicator,
                                 int *coords,
                                 int *rank)
        : MPO_Cartesian(communicator),
          my_coords(coords),
          my_rank(rank)
    {
    }


    // destructor
    //========================================================================
    MPO_Cart_rank::~MPO_Cart_rank()
    {
    }


    // checkCoordinates
    //========================================================================
    int MPO_Cart_rank::checkCoordinates()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!(our_errcodes.at(MARMOT_ERR_COORD_NEG)||our_errcodes.at(MARMOT_ERR_COORD_TOO_BIG)))
        return rvalue;

        int ndims = -1;
        PMPI_Cartdim_get(my_comm, &ndims);

        int *dims = 0;
        int *periods = 0;
        int *coords = 0;

        dims = new int[ndims];
        periods = new int[ndims];
        coords = new int[ndims];

        PMPI_Cart_get(my_comm, ndims, dims, periods, coords);

        // Check array of coordinates
        for (int i = 0; i < ndims; i++)
        {
            // (1) Check if coordinate is negative.
            if ((our_errcodes.at(MARMOT_ERR_COORD_NEG))&&(!periods[i]) &&
                (coords[i] < 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": coordinate("
                    << i
                    << ") less than 0 in non-periodic dimension!"
                    << std::endl;
                rvalue = MARMOT_ERR_COORD_NEG;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (2) Check if coordinate is out of range.
            else if ((our_errcodes.at(MARMOT_ERR_COORD_TOO_BIG))&&(!periods[i]) &&
                     (coords[i] >= dims[i]) &&
                     (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": coordinate("
                    << i
                    << ") out of range in non-periodic dimension!"
                    << std::endl;
                rvalue = MARMOT_ERR_COORD_TOO_BIG;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }
        }                       // end for

        delete[] dims;
        delete[] periods;
        delete[] coords;

        return rvalue;
    }                           // end checkCoordinates


    // getId
    //========================================================================
    int MPO_Cart_rank::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_CART_RANK_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Cart_rank::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Cart_rank::checkPrerequisites(void)
    {


        MPO_Cartesian::checkPrerequisites();
        checkCoordinates();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Cart_rank::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Cart_rank(my_comm, my_coords, my_rank);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Cart_rank::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Cart_rank";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_CART_RANK";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Cart_rank::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", *coords, "
            << "*rank";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Cart_rank::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        // out < my_coords;
        // out << my_rank;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Cart_rank::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        // in > my_coords;
        // in >> my_rank;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/topol/cart_rank.c.
/*@

MPI_Cart_rank - Determines process rank in communicator given Cartesian
location

Input Parameters:
+ comm - communicator with cartesian structure (handle)
- coords - integer array (of size  'ndims') specifying the cartesian
coordinates of a process

Output Parameter:
. rank - rank of specified process (integer)

Notes:
Out-of-range coordinates are erroneous for non-periodic dimensions.
Versions of MPICH before 1.2.2 returned 'MPI_PROC_NULL' for the rank in
this case.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_RANK
.N MPI_ERR_ARG

int MPI_Cart_rank (
MPI_Comm comm,
int *coords,
int *rank )
@ */

