/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Comm_compare.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Comm_compare.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Comm_compare::MPO_Comm_compare()
        : MPO_Communicatorhandler(MPI_COMM_NULL),
          my_comm2(MPI_COMM_NULL),
          my_result(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Comm_compare::MPO_Comm_compare(MPO_Comm comm,
                                       MPO_Comm comm2,
                                       int
                                       *result)
        : MPO_Communicatorhandler(comm),
          my_comm2(comm2),
          my_result(result)
    {
    }


    // destructor
    //========================================================================
    MPO_Comm_compare::~MPO_Comm_compare()
    {
    }


    // checkCommSecond
    //========================================================================
    int MPO_Comm_compare::checkCommSecond()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Communicator is MPI_COMM_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_comm_traces->findAndPrintTraceNamed(my_comm2,stream,"comm2");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Communicator is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_comm_traces->findAndPrintTraceNamed(my_comm2,stream,"comm2");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkCommSecond


    // getId
    //========================================================================
    int MPO_Comm_compare::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_COMM_COMPARE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Comm_compare::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Comm_compare::checkPrerequisites(void)
    {


        MPO_Communicatorhandler::checkPrerequisites();
        checkCommSecond();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm2,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Comm_compare::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Comm_compare(my_comm, my_comm2, my_result);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Comm_compare::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Comm_compare";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_COMM_COMPARE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Comm_compare::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm1 = "
            << my_comm
            << ", comm2 = "
            << my_comm2
            << ", *result";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Comm_compare::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_comm2;
        // out < my_result;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Comm_compare::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_comm2;
        // in > my_result;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/commcompare.c.
/*@

MPI_Comm_compare - Compares two communicators

Input Parameters:
+ comm1 - comm1 (handle)
- comm2 - comm2 (handle)

Output Parameter:
. result - integer which is 'MPI_IDENT' if the contexts and groups are the
same, 'MPI_CONGRUENT' if different contexts but identical groups,
'MPI_SIMILAR' if different contexts but similar groups, and 'MPI_UNEQUAL'
otherwise

Using 'MPI_COMM_NULL' with 'MPI_Comm_compare':

It is an error to use 'MPI_COMM_NULL' as one of the arguments to
'MPI_Comm_compare'.  The relevant sections of the MPI standard are

$(2.4.1 Opaque Objects)
A null handle argument is an erroneous 'IN' argument in MPI calls, unless
an exception is explicitly stated in the text that defines the function.

$(5.4.1. Communicator Accessors)
<no text in 'MPI_COMM_COMPARE' allowing a null handle>

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG

int MPI_Comm_compare (
MPI_Comm  comm1,
MPI_Comm  comm2,
int *result)
@ */

