/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Comm_create.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Comm_create.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Comm_create::MPO_Comm_create()
        : MPO_Communicatorhandler(MPI_COMM_NULL),
          my_group(MPI_GROUP_NULL),
          my_newcomm(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Comm_create::MPO_Comm_create(MPO_Comm comm,
                                     MPO_Group group,
                                     MPI_Comm *
                                     newcomm)
        : MPO_Communicatorhandler(comm),
          my_group(group),
          my_newcomm(newcomm)
    {
    }


    // destructor
    //========================================================================
    MPO_Comm_create::~MPO_Comm_create()
    {
    }


    // checkGroup
    //========================================================================
    int MPO_Comm_create::checkGroup()
    {
        /** \todo @creator of this method: what about return types ? */

        int rvalue = 1;
        MPI_Group testgroup = MPI_GROUP_NULL;
        int count = -1;
        int count1 = -1;
        int count2 = -1;
        int notfound = 0;
        int i = 0;
        int *ranks1 = 0;
        int *ranks2 = 0;

        if ((our_groups.find(my_group) == our_groups.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Group is not valid!"
                << std::endl;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_GROUP_NOT_VALID,getSourceLocation());
            our_log->getMsg(msg);
        }

        PMPI_Comm_group(my_comm, &testgroup);
        PMPI_Group_size(testgroup, &count2);
        PMPI_Group_size(my_group, &count1);

        if (count2 < count1)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Group is NOT a subset of communicator:"
                    << std::endl
                    << " groupsize="
                    << count1
                    << ",  commsize="
                    << count2
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                /** \todo create a new ERROR_CODE for this case, also see todo below ???? */
                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_GROUP_NOT_VALID,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            count = count1;
            ranks1 = new int[count];
            ranks2 = new int[count];
            for (i = 0; i <= count - 1; i++)
            {
                ranks1[i] = i;
            }
            PMPI_Group_translate_ranks(my_group, count, ranks1, testgroup,
                                       ranks2);
            for (i = 0; i <= count - 1; i++)
            {
                if (ranks2[i] == MPI_UNDEFINED)
                {
                    notfound++;
                }
            }
            if ((notfound > 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Group is NOT a subset of communicator: "
                    << notfound
                    << " rank(s) not found in communicator! "
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                /** \todo see todo above ! */
                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_GROUP_NOT_VALID,getSourceLocation());
                our_log->getMsg(msg);
            }

            delete[]ranks1;
            delete[]ranks2;
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Comm_create::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_COMM_CREATE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Comm_create::checkPost(void)
    {

        our_comms.insert(*my_newcomm);

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_group_traces->findAndAddUsage(my_group,getSourceLocation());
        our_comm_traces->addAndSetCreation(*my_newcomm,getSourceLocation());
        our_comm_traces->findAndTryAddDependendency(*my_newcomm,my_comm);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Comm_create::checkPrerequisites(void)
    {


        MPO_Communicatorhandler::checkPrerequisites();
        checkGroup();


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Comm_create::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Comm_create(my_comm, my_group, my_newcomm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Comm_create::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Comm_create";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_COMM_CREATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Comm_create::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", group = "
            << my_group
            << ", *newcomm";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Comm_create::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_group;
        // out < my_newcomm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Comm_create::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_group;
        // in > my_newcomm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/comm_create.c.
/*@

MPI_Comm_create - Creates a new communicator

Input Parameters:
+ comm - communicator (handle)
- group - group, which is a subset of the group of 'comm'  (handle)

Output Parameter:
. comm_out - new communicator (handle)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Comm_free

int MPI_Comm_create (
MPI_Comm comm,
MPI_Group group,
MPI_Comm *comm_out )
@ */

