/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Comm_dup.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Comm_dup.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Comm_dup::MPO_Comm_dup()
        : MPO_Communicatorhandler(MPI_COMM_NULL),
          my_newcomm(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Comm_dup::MPO_Comm_dup(MPO_Comm comm,
                               MPI_Comm *
                               newcomm)
        : MPO_Communicatorhandler(comm),
          my_newcomm(newcomm)
    {
    }


    // destructor
    //========================================================================
    MPO_Comm_dup::~MPO_Comm_dup()
    {
    }


    // getId
    //========================================================================
    int MPO_Comm_dup::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_COMM_DUP_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Comm_dup::checkPost(void)
    {

        our_comms.insert(*my_newcomm);

        our_comm_traces->addAndSetCreation(*my_newcomm,getSourceLocation());
        our_comm_traces->findAndTryAddDependendency(*my_newcomm,my_comm);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Comm_dup::checkPrerequisites(void)
    {


        MPO_Communicatorhandler::checkPrerequisites();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Comm_dup::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Comm_dup(my_comm, my_newcomm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Comm_dup::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Comm_dup";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_COMM_DUP";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Comm_dup::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", *newcomm";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Comm_dup::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        // out < my_newcomm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Comm_dup::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        // in > my_newcomm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/context/comm_dup.c.
/*@

MPI_Comm_dup - Duplicates an existing communicator with all its cached
information

Input Parameter:
. comm - communicator (handle)

Output Parameter:
. newcomm - A new communicator over the same group as 'comm' but with a new
context. See notes.  (handle)

Notes:
This routine is used to create a new communicator that has a new
communication context but contains the same group of processes as
the input communicator.  Since all MPI communication is performed
within a communicator (specifies as the group of processes `plus`
the context), this routine provides an effective way to create a
private communicator for use by a software module or library.  In
particular, no library routine should use 'MPI_COMM_WORLD' as the
communicator; instead, a duplicate of a user-specified communicator
should always be used.  For more information, see Using MPI, 2nd
edition.

Because this routine essentially produces a copy of a communicator,
it also copies any attributes that have been defined on the input
communicator, using the attribute copy function specified by the
'copy_function' argument to 'MPI_Keyval_create'.  This is
particularly useful for (a) attributes that describe some property
of the group associated with the communicator, such as its
interconnection topology and (b) communicators that are given back
to the user; the attibutes in this case can track subsequent
'MPI_Comm_dup' operations on this communicator.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Comm_free, MPI_Keyval_create, MPI_Attr_set, MPI_Attr_delete

int MPI_Comm_dup (
MPI_Comm comm,
MPI_Comm *comm_out )
@ */

