/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Comm_free.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Comm_free.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Comm_free::MPO_Comm_free()
        : my_comm(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Comm_free::MPO_Comm_free(MPI_Comm * comm)
        : my_comm(comm)
    {
    }


    // destructor
    //========================================================================
    MPO_Comm_free::~MPO_Comm_free()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Comm_free::checkPrerequisites()
    {


        // check correctness of parameters
        checkCommFreed();
        our_comms.erase(*my_comm);
        eraseErrhandler();
    
        
        our_comm_traces->findAndSetFree(*my_comm,getSourceLocation());
        

        return 0;
    }


    // checkCommFreed
    //========================================================================
    int MPO_Comm_free::checkCommFreed()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_WARN_COMM_NULL))&&(MPI_COMM_NULL == *my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator to be freed is already MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_WARN_COMM_NULL;
            
            if (my_comm)
                our_comm_traces->findAndPrintTraceNamed(*my_comm,stream,"comm");
                        
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(*my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            if (my_comm)
                our_comm_traces->findAndPrintTraceNamed(*my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkCommFreed


    // eraseErrhandler
    //========================================================================
    int MPO_Comm_free::eraseErrhandler()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        MPI_Errhandler err;
        PMPI_Errhandler_get(*my_comm, &err);

        if ((err != MPI_ERRORS_ARE_FATAL) && 
            (err != MPI_ERRORS_RETURN))
        {
            (our_errhandlers.find(err)->second[0])--;
        }

        if ((our_errhandlers.find(err)->second[0] == 0) &&
            (our_errhandlers.find(err)->second[1] == 1))
        {
            our_errhandlers.erase(err);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Comm_free::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_COMM_FREE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Comm_free::checkPost(void)
    {
            
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Comm_free::callNativeMPI()
    {
        int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
#endif
        rvalue = PMPI_Comm_free(my_comm);
        
#ifdef MARMOT_TCHECK
        MPO_Space::getSyncObject()->unregisterCommToIntEntry(*my_comm);
#endif
        
#ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
#endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Comm_free::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Comm_free";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_COMM_FREE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Comm_free::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname() 
            << "(*comm";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Comm_free::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Comm_free::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_comm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied                               
// from mpich-1.2.4/src/context/comm_free.c.
/*@

MPI_Comm_free - Marks the communicator object for deallocation

Input Parameter:
. comm - communicator to be destroyed (handle) 

Null Handles:
The MPI 1.1 specification, in the section on opaque objects, explicitly
disallows freeing a null communicator.  The text from the standard is:
.vb
A null handle argument is an erroneous IN argument in MPI calls, unless an
exception is explicitly stated in the text that defines the function. Such
exception is allowed for handles to request objects in Wait and Test calls
(sections Communication Completion and Multiple Completions ). Otherwise, a
null handle can only be passed to a function that allocates a new object 
and returns a reference to it in the handle.
.ve

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG

int MPI_Comm_free ( MPI_Comm *commp )
@ */
