/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Comm_split.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Comm_split.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Comm_split::MPO_Comm_split()
        : MPO_Communicatorhandler(MPI_COMM_NULL),
          my_color(-1),
          my_key(-1),
          my_newcomm(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Comm_split::MPO_Comm_split(MPO_Comm comm,
                                   int color,
                                   int key,
                                   MPI_Comm *
                                   newcomm)
        : MPO_Communicatorhandler(comm),
          my_color(color),
          my_key(key),
          my_newcomm(newcomm)
    {
    }


    // destructor
    //========================================================================
    MPO_Comm_split::~MPO_Comm_split()
    {
    }


    // checkColor
    //========================================================================
    int MPO_Comm_split::checkColor()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if color is MPI_UNDEFINED.
        if ((our_errcodes.at(MARMOT_WARN_COLOR_UNDEFINED))&&(my_color == MPI_UNDEFINED) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": Color = MPI_UNDEFINED ! "
                << "New communicator will be MPI_COMM_NULL."
                << std::endl;
            rvalue = MARMOT_WARN_COLOR_UNDEFINED;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if color is < 0.
        else if ((our_errcodes.at(MARMOT_ERR_COLOR_NEG))&&(my_color != MPI_UNDEFINED) &&
                 (my_color < 0) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Color( != MPI_UNDEFINED) less than 0 !"
                << std::endl;
            rvalue = MARMOT_ERR_COLOR_NEG;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkColor


    // getId
    //========================================================================
    int MPO_Comm_split::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_COMM_SPLIT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Comm_split::checkPost(void)
    {

        our_comms.insert(*my_newcomm);

        our_comm_traces->addAndSetCreation(*my_newcomm,getSourceLocation());
        our_comm_traces->findAndTryAddDependendency(*my_newcomm,my_comm);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Comm_split::checkPrerequisites(void)
    {


        MPO_Communicatorhandler::checkPrerequisites();
        checkColor();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Comm_split::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Comm_split(my_comm, my_color, my_key, my_newcomm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Comm_split::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Comm_split";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_COMM_SPLIT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Comm_split::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", color = "
            << my_color
            << ", key = "
            << my_key
            << ", *newcomm";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Comm_split::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_color;
        out < my_key;
        // out < my_newcomm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Comm_split::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_color;
        in > my_key;
        // in > my_newcomm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/comm_split.c.
/*@

MPI_Comm_split - Creates new communicators based on colors and keys

Input Parameters:
+ comm - communicator (handle)
. color - control of subset assignment (nonnegative integer)
- key - control of rank assigment (integer)

Output Parameter:
. newcomm - new communicator (handle)

Notes:
The 'color' must be non-negative or 'MPI_UNDEFINED'.

.N fortran

Algorithm:

The current algorithm used has quite a few (read: a lot of) inefficiencies
that can be removed.  Here is what we do for now

.vb
1) A table is built of colors, and keys (has a next field also).
2) The tables of all processes are merged using 'MPI_Allreduce'.
3) Two contexts are allocated for all the comms to be created.  These
same two contexts can be used for all created communicators since
the communicators will not overlap.
4) If the local process has a color of 'MPI_UNDEFINED', it can return
a 'NULL' comm.
5) The table entries that match the local process color are sorted
by key/rank.
6) A group is created from the sorted list and a communicator is created
with this group and the previously allocated contexts.
.ve

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Comm_free

int MPI_Comm_split (
MPI_Comm comm,
int color,
int key,
MPI_Comm *comm_out )
@ */

