/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Errhandler_get.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Errhandler_get.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Errhandler_get::MPO_Errhandler_get()
        : my_comm(MPI_COMM_NULL),
          //bk// because of compiler warning
          //: my_errhandler(MPI_ERRHANDLER_NULL)
          my_errhandler(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Errhandler_get::MPO_Errhandler_get(MPO_Comm comm,
                                           MPI_Errhandler *
                                           errhandler)
        : my_comm(comm),
          my_errhandler(errhandler)
    {
    }


    // destructor
    //========================================================================
    MPO_Errhandler_get::~MPO_Errhandler_get()
    {
    }


    // getId
    //========================================================================
    int MPO_Errhandler_get::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ERRHANDLER_GET_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Errhandler_get::checkPrerequisites()
    {


        // check correctness of parameters
        checkComm();
        checkErrhandler();
    
        
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        
        
        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Errhandler_get::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


    // checkErrhandler
    //========================================================================
    int MPO_Errhandler_get::checkErrhandler()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // Check if errhandler is valid.
        if ((our_errcodes.at(MARMOT_ERR_ERRHANDLER_NOT_VALID))&&(our_errhandlers.find(*my_errhandler) == our_errhandlers.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": errhandler is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_ERRHANDLER_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkErrhandler


    // checkPost
    //========================================================================
    int MPO_Errhandler_get::checkPost(void)
    {
        
        our_errhandler_traces->findAndAddUsage(*my_errhandler,getSourceLocation());
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Errhandler_get::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Errhandler_get(my_comm, my_errhandler);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Errhandler_get::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Errhandler_get";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ERRHANDLER_GET";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Errhandler_get::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(comm = " 
            << my_comm 
            << ", *errhandler";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Errhandler_get::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        // out < my_errhandler;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Errhandler_get::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        // in > my_errhandler;

        return in;
    }



}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/errget.c.
/*@
  MPI_Errhandler_get - Gets the error handler for a communicator

  Input Parameter:
  . comm - communicator to get the error handler from (handle) 

  Output Parameter:
  . errhandler - MPI error handler currently associated with communicator
  (handle) 

  .N fortran

  Note on Implementation:

  The MPI Standard was unclear on whether this routine required the user to
  call 'MPI_Errhandler_free' once for each call made to this routine in 
  order to free the error handler.  After some debate, the MPI Forum added 
  an explicit statement that users are required to call 
  'MPI_Errhandler_free' when the return value from this routine is no 
  longer needed.  This behavior is similar to the other MPI routines for 
  getting objects; for example, 
  'MPI_Comm_group' requires that the user call 'MPI_Group_free' when the 
  group returned by 'MPI_Comm_group' is no longer needed.

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_ARG

  int MPI_Errhandler_get( 
  MPI_Comm comm, 
  MPI_Errhandler *errhandler )
  @ */

