/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Finalize.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Finalize.cc 964 2009-06-02 06:38:44Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include "mpo_TCheck_Races.h"
#include <fstream>
#include <stdio.h>

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Finalize::MPO_Finalize()
    {
    }


    // destructor
    //========================================================================
    MPO_Finalize::~MPO_Finalize()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Finalize::checkPrerequisites()
    {

        // check if there's anything left
        checkRequestsLeft();
        checkPendingMessagesLeft();

        // in order to detect resource miss-use we will also
        // check whether there are still Datatypes left
        checkDatatypesLeft();
        checkErrhandlersLeft();
        checkCommsLeft();
        checkGroupsLeft();
        checkOperationsLeft();

        //multi thrad checks and infos
        logMinimalThreadLevel();
        checkIsThreadMain();
        checkOtherThreadsDone();

        //we might also log some information about the ressource usage
        printResourceGraphs();

        //used to let the debugserver wait until we are finished
        if (MPO_Space::isDebugServerEnabled())
        {
            int server_token = -1;
            our_channel.sendId(MPO_FINALIZE_READY,-1,-1);
            our_channel.sendId(MPO_FINALIZE_READY,-1,-1); //send twice as debugserver requires family id and type id !!!
            our_channel.receiveId(server_token,-1,-1); //wait for OK from debugserver
        }

        our_finalize_was_called = true;

        return 0;
    }


    // checkRequestsLeft
    //========================================================================
    int MPO_Finalize::checkRequestsLeft()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int count = our_request_traces->getNumResourcesNotFreed();

        // Check whether there are active or non-freed requests left.
        if ((our_errcodes.at(MARMOT_WARN_REQUESTS_LEFT))&&(count > 0) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": There are still "
                << count
                << " active/non-freed Requests! "
                << std::endl
                << "Listing of information for all remaining Requests:"
                << std::endl;

            our_request_traces->printResourcesNotFreed(stream);

            rvalue = MARMOT_WARN_REQUESTS_LEFT;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        //print a note for the maximum amount of this resource
        int max_used = our_request_traces->getMaxNumResourcesUsed();
        if ((max_used > 0)&&(MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << "the run needed a maximum of "
                << max_used
                << " requests! "
                << std::endl;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text,getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_UNDEFINED,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkRequestsLeft


    // checkPendingMessagesLeft
    //========================================================================
    int MPO_Finalize::checkPendingMessagesLeft()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_WARN_PENDING_MESSAGES_LEFT))
        return rvalue;

        std::set < MPO_Comm >::iterator pos;
        int flag = 0;
        MPI_Status status;

        // Check whether there are pending messages.
        for (pos = our_comms.begin(); pos != our_comms.end(); pos++)
        {
            //If node is not member of communicator: *pos == MPI_COMM_NULL
            if (*pos != MPI_COMM_NULL)
            {
                PMPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, *pos, &flag, &status);

                if ((flag) &&
                    (MSG_ERR_WARN <= DEBUG_MODE))
                {
                    std::stringstream stream;
                    stream
                        << "WARNING: "
                        << getCallname()
                        << ": There are still pending messages!"
                        << std::endl;
                    rvalue = MARMOT_WARN_PENDING_MESSAGES_LEFT;

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
        }

        return rvalue;
    }                           // end checkPendingMessagesLeft

    // checkDatatypesLeft
    //========================================================================
    int MPO_Finalize::checkDatatypesLeft()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int num_crit_datatypes = our_datatype_traces->getNumResourcesNotFreed();

         if ((our_errcodes.at(MARMOT_WARN_TYPE_LEFT))&&(num_crit_datatypes > 0))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": There are still "
                << num_crit_datatypes
                << " datatypes left ! "
                << std::endl
                << "Listing of information for all remaining Datatypes:"
                << std::endl;

            our_datatype_traces->printResourcesNotFreed(stream);
            rvalue = MARMOT_WARN_TYPE_LEFT;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        //print a note for the maximum amount of this resource
        int max_used = our_datatype_traces->getMaxNumResourcesUsed();
        if ((max_used > 0) && (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << "the run needed a maximum of "
                << max_used
                << " Datatypes ! "
                << std::endl;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_UNDEFINED,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkDatatypesLeft

    // checkErrhandlersLeft
    //========================================================================
    int MPO_Finalize::checkErrhandlersLeft()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int count = our_errhandler_traces->getNumResourcesNotFreed();

        if ((our_errcodes.at(MARMOT_WARN_ERRHANDLER_LEFT))&&(count > 0) && (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": There are still "
                << count
                << " errorhandlers left ! "
                << std::endl
                << "Listing of information for all remaining Errhandlers:"
                << std::endl;

            our_errhandler_traces->printResourcesNotFreed(stream);
            rvalue = MARMOT_WARN_ERRHANDLER_LEFT;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        //print a note for the maximum amount of this resource
        int max_used =our_errhandler_traces->getMaxNumResourcesUsed();
        if ((max_used > 0) && (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << "the run needed a maximum of "
                << max_used
                << " Errhandlers ! "
                << std::endl;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_UNDEFINED,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkErrhandlersLeft

    // checkCommsLeft
    //========================================================================
    int MPO_Finalize::checkCommsLeft()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int count = our_comm_traces->getNumResourcesNotFreed();

        if ((our_errcodes.at(MARMOT_WARN_COMM_LEFT))&&(count > 0))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": There are still "
                << count
                << " communicators left ! "
                << std::endl
                << "Listing of information for all remaining Communicators:"
                << std::endl;

            our_comm_traces->printResourcesNotFreed(stream);
            rvalue = MARMOT_WARN_COMM_LEFT;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        //print a note for the maximum amount of this resource
        int max_used = our_comm_traces->getMaxNumResourcesUsed();
        if ((max_used > 0) && (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << "the run needed a maximum of "
                << max_used
                << " Communicators ! "
                << std::endl;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_UNDEFINED,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkCommsLeft

    // checkGroupsLeft
    //========================================================================
    int MPO_Finalize::checkGroupsLeft()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int count = our_group_traces->getNumResourcesNotFreed();

        if ((our_errcodes.at(MARMOT_WARN_GROUP_LEFT))&&(count > 0))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": There are still "
                << count
                << " groups left ! "
                << std::endl
                << "Listing of information for all remaining Groups:"
                << std::endl;

            our_group_traces->printResourcesNotFreed(stream);
            rvalue = MARMOT_WARN_GROUP_LEFT;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        //print a note for the maximum amount of this resource
        int max_used = our_group_traces->getMaxNumResourcesUsed();
        if ((max_used > 0) && (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << "the run needed a maximum of "
                << max_used
                << " Groups ! "
                << std::endl;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_UNDEFINED,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkGroupsLeft

    // checkOperationsLeft
    //========================================================================
    int MPO_Finalize::checkOperationsLeft()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int count = our_op_traces->getNumResourcesNotFreed();

        if ((our_errcodes.at(MARMOT_WARN_OP_LEFT))&&(count > 0))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": There are still "
                << count
                << " operators left ! "
                << std::endl
                << "Listing of information for all remaining Operators:"
                << std::endl;

            our_op_traces->printResourcesNotFreed(stream);
            rvalue = MARMOT_WARN_OP_LEFT;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        //print a note for the maximum amount of this resource
        int max_used = our_op_traces->getMaxNumResourcesUsed();
        if ((max_used) && (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "NOTE: "
                << "the run needed a maximum of "
                << max_used
                << " Operators ! "
                << std::endl;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_UNDEFINED,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkOperationsLeft

    // printResourceGraphs
    //========================================================================
    void MPO_Finalize::printResourceGraphs()
    {
        /** \todo add an environmental to control this. */

        std::fstream out;
        char file_name[128];
        int my_rank;

        PMPI_Comm_rank (MPI_COMM_WORLD,&my_rank);

        sprintf(file_name,"Marmot_ress_datatype_usage.%d.mdg",my_rank);
          our_datatype_traces->printResourcesUsedGraph(file_name);

          sprintf(file_name,"Marmot_ress_op_usage.%d.mdg",my_rank);
        our_op_traces->printResourcesUsedGraph(file_name);

          sprintf(file_name,"Marmot_ress_request_usage.%d.mdg",my_rank);
        our_request_traces->printResourcesUsedGraph(file_name);

          sprintf(file_name,"Marmot_ress_group_usage.%d.mdg",my_rank);
        our_group_traces->printResourcesUsedGraph(file_name);

          sprintf(file_name,"Marmot_ress_comm_usage.%d.mdg",my_rank);
        our_comm_traces->printResourcesUsedGraph(file_name);

          sprintf(file_name,"Marmot_ress_errhandler_usage.%d.mdg",my_rank);
        our_errhandler_traces->printResourcesUsedGraph(file_name);
    }

    // checkIsThreadMain
    //========================================================================
    int MPO_Finalize::checkIsThreadMain (void)
    {
           #ifdef MARMOT_THREADED
               int flag;
               int rvalue = MARMOT_SUCCESS;
               PMPI_Is_thread_main(&flag);
                  if (!flag)
                  {
                      //MPI_FINALIZE is not performed by the main thread
                      //thus we create a warning
                      std::stringstream stream;
                stream
                    << "WARNING: "
                    << getCallname()
                    << " was not performed by the thread that performed MPI_Init or MPI_Init_thread."
                    << std::endl;

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_FINALIZE_NOT_THREAD_MAIN,getSourceLocation());
                our_log->getMsg(msg);
                rvalue = MARMOT_WARN_FINALIZE_NOT_THREAD_MAIN;
                  }
                  return rvalue;
           #else
               return MARMOT_SUCCESS;
           #endif
    }

    // checkOtherThreadsDone
    //========================================================================
    int MPO_Finalize::checkOtherThreadsDone()
    {

#ifdef MARMOT_TCHECK
        //read - write Data race for TCheck in order to detect an MPI_FINALIZE
        //performed while other threads are still doing MPI
        MPO_TCheck_Races::getInstance()->write(APP_MPI_FINALIZE_TO_EARLY);
#endif

        int rvalue = MARMOT_SUCCESS;

        //if no parallel region is present the problem can`t arise
        if (!MPO_Space::getSyncObject()->isInParallelRegion())
            return rvalue;

        //check for other threads waiting to enter MARMOT
        //This check is not complete as it would need a simulation
        //approach.
        if (MPO_Space::getSyncObject()->otherThreadsWaiting())
        {
            //MPI_FINALIZE was performed while other threads were still calling MPI
            rvalue = MARMOT_WARN_FINALIZE_OTHER_THREADS_WORKING;
               std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << " was performed while other threads were calling MPI."
                << std::endl;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }

    // logMinimalThreadLevel
    //========================================================================
    int MPO_Finalize::logMinimalThreadLevel()
    {
        #ifdef MARMOT_THREADED
            int min_ThreadLevel = getSyncObject()->getMinThreadLvl();
            int provided_ThreadLevel = getProvidedThreadLevel();

            //if the provided thread level was already violated thers no need to give additional information
            if (min_ThreadLevel > provided_ThreadLevel)
                return MARMOT_SUCCESS;

            //lets log all violations to thread levels that occured
            std::stringstream stream;
            stream
                << "Note: "
                << "The minimal threadlevel required by this run was: "
                << getThreadLevelName(min_ThreadLevel)
                << std::endl;

            const int level[] = {MPI_THREAD_SINGLE,MPI_THREAD_FUNNELED,MPI_THREAD_SERIALIZED};
            const int num_levels = 3;

            for (int i = 0; i < num_levels;i++)
            {
                if (min_ThreadLevel > level[i])
                {
                    MPO_ThreadViolationInfo info = getSyncObject()->getViolationInfoForThreadLvl(level[i]);
                    stream
                        << std::endl
                        << getThreadLevelName(level[i])
                        << " was violated by: "
                        << std::endl;
                    info.printParticipants(stream);
                    stream << std::endl;
                }
            }//for

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(), MARMOT_ERR_UNDEFINED, getSourceLocation());
            our_log->getMsg(msg);

        #endif

        return MARMOT_SUCCESS;
    }


    // getId
    //========================================================================
    int MPO_Finalize::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_FINALIZE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Finalize::checkPost(void)
    {
        //free the log file
        delete our_log;
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Finalize::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Finalize();
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Finalize::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Finalize";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_FINALIZE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Finalize::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << "ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Finalize::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Finalize::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/finalize.c.
/*@
  MPI_Finalize - Terminates MPI execution environment

  Notes:
  All processes must call this routine before exiting.  The number of
  processes running `after` this routine is called is undefined;
  it is best not to perform much more than a 'return rc' after calling
  'MPI_Finalize'.

  .N fortran

  int MPI_Finalize()
  @ */

