/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Graph_create.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Graph_create.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Graph_create::MPO_Graph_create()
        : my_comm_old(MPI_COMM_NULL),
          my_nnodes(-1),
          my_index(0),
          my_edges(0),
          my_reorder(-1),
          my_comm_graph(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Graph_create::MPO_Graph_create(MPO_Comm comm_old,
                                       int nnodes,
                                       int *index,
                                       int *edges,
                                       int reorder,
                                       MPI_Comm *
                                       comm_graph)
        : my_comm_old(comm_old),
          my_nnodes(nnodes),
          my_index(index),
          my_edges(edges),
          my_reorder(reorder),
          my_comm_graph(comm_graph)
    {
    }


    // destructor
    //========================================================================
    MPO_Graph_create::~MPO_Graph_create()
    {
    }


    /**
     *  We overload MPO_Graph::checkPrerequisites to check other
     *  pre_requisites:
     *  Check if communicator, size and edges fit together.
     */
    // checkPrerequisites2
    //========================================================================
    int MPO_Graph_create::checkPrerequisites()
    {
        // check correctness of parameters
        checkComm();
        checkTopologyNew();
        checkEdgeArray();

        our_comm_traces->findAndAddUsage(my_comm_old,getSourceLocation());

        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Graph_create::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Communicator is MPI_COMM_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm_old");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Communicator is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm_old");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkComm


    // checkTopologyNew
    //========================================================================
    int MPO_Graph_create::checkTopologyNew()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int comm_size = -1;
        PMPI_Comm_size(my_comm_old, &comm_size);

        // (1) Check if nnodes > size of old communicator
        if ((our_errcodes.at(MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD))&&(comm_size < my_nnodes) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream    << "ERROR: "
                      << getCallname()
                      << ": New topology is larger than old communicator:"
                      << std::endl
                      << " comm_size = "
                      << comm_size
                      << ", nnodes = "
                      << my_nnodes
                      << std::endl;
            rvalue = MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm_old");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if nnodes < size of old communicator
        else if ((our_errcodes.at(MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD))&&(comm_size > my_nnodes) &&
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": New topology is smaller than old communicator:"
                << " Some nodes will get MPI_COMM_NULL."
                << std::endl
                << " comm_size = "
                << comm_size
                << ", nnodes = "
                << my_nnodes
                << std::endl;
            rvalue = MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm_old");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkTopologyNew


    // checkEdgeArray
    //========================================================================
    int MPO_Graph_create::checkEdgeArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_ERR_EDGE_INVALID))
        return rvalue;

        int comm_size = -1;
        PMPI_Comm_size(my_comm, &comm_size);

        // Check if all entries in my_edges are valid
        for (int i = 0; i < my_index[my_nnodes - 1]; i++)
        {
            if ((my_edges[i] >= comm_size) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Invalid rank: edges["
                    << i
                    << "] = "
                    << my_edges[i]
                    << std::endl;
                rvalue = MARMOT_ERR_EDGE_INVALID;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm_old");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }
        }

        return rvalue;
    }                           // end checkEdgeArray


    // getId
    //========================================================================
    int MPO_Graph_create::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GRAPH_CREATE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Graph_create::checkPost(void)
    {

        our_comms.insert(*my_comm_graph);

        our_comm_traces->addAndSetCreation(*my_comm_graph,getSourceLocation());
        our_comm_traces->findAndTryAddDependendency(*my_comm_graph,my_comm_old);

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Graph_create::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Graph_create(my_comm_old, my_nnodes, my_index,
                                   my_edges, my_reorder, my_comm_graph);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Graph_create::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Graph_create";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GRAPH_CREATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Graph_create::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm_old = "
            << my_comm_old
            << ", nnodes = "
            << my_nnodes
            << ", *index, "
            << "*edges, "
            << "reorder = "
            << my_reorder
            << ", *comm_graph";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Graph_create::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm_old;
        out < my_nnodes;
        // out < my_index;
        // out < my_edges;
        out < my_reorder;
        // out < my_comm_graph;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Graph_create::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm_old;
        in > my_nnodes;
        // in > my_index;
        // in > my_edges;
        in > my_reorder;
        // in > my_comm_graph;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied
// from mpich-1.2.4/src/topol/graphcreate.c.
/*@

MPI_Graph_create - Makes a new communicator to which topology information
has been attached

Input Parameters:
+ comm_old - input communicator without topology (handle)
. nnodes - number of nodes in graph (integer)
. index - array of integers describing node degrees (see below)
. edges - array of integers describing graph edges (see below)
- reorder - ranking may be reordered (true) or not (false) (logical)

Output Parameter:
. comm_graph - communicator with graph topology added (handle)

Algorithm:
We ignore the 'reorder' info currently.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG

int MPI_Graph_create (
MPI_Comm comm_old,
int nnodes,
int *index,
int *edges,
int reorder,
MPI_Comm *comm_graph )
@ */

