/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Graph_get.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Graph_get.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Graph_get::MPO_Graph_get()
        : MPO_Graph(MPI_COMM_NULL),
          my_maxindex(-1),
          my_maxedges(-1),
          my_index(0),
          my_edges(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Graph_get::MPO_Graph_get(MPO_Comm communicator,
                                 int maxindex,
                                 int maxedges,
                                 int *index,
                                 int *edges)
        : MPO_Graph(communicator),
          my_maxindex(maxindex),
          my_maxedges(maxedges),
          my_index(index),
          my_edges(edges)
    {
    }


    // destructor
    //========================================================================
    MPO_Graph_get::~MPO_Graph_get()
    {
    }


    /**
     *  We overload MPO_Graph::checkPrerequisites to
     *  check other prerequisites:
     *  Check whether index and edges match with communicator.
     */
    // checkPrerequisites2
    //========================================================================
    int MPO_Graph_get::checkPrerequisites2()
    {
        // check correctness of parameters
        checkMaxIndex();
        checkMaxEdges();

        return 0;
    }


    // checkMaxIndex
    //========================================================================
    int MPO_Graph_get::checkMaxIndex()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        // (1) Check if maxindex is negative.
        if ((our_errcodes.at(MARMOT_ERR_INDEX_NEG))&&(my_maxindex < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": maxindex less than 0!"
                << std::endl;
            rvalue = MARMOT_ERR_INDEX_NEG;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if maxindex is less than size of communicator.
        else if ((our_errcodes.at(MARMOT_WARN_INDEX_LT_SIZE))&&(my_maxindex < size) &&
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": maxindex is less than size of communicator!"
                << std::endl;
            rvalue = MARMOT_WARN_INDEX_LT_SIZE;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkMaxIndex


    // checkMaxEdges
    //========================================================================
    int MPO_Graph_get::checkMaxEdges()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_ERR_EDGE_NEG))
        return rvalue;

        if ((my_maxedges < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": maxedges less than 0!"
                << std::endl;
            rvalue = MARMOT_ERR_EDGE_NEG;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkMaxEdges


    // getId
    //========================================================================
    int MPO_Graph_get::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GRAPH_GET_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Graph_get::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Graph_get::checkPrerequisites(void)
    {


        MPO_Graph::checkPrerequisites();
        checkPrerequisites2();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Graph_get::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Graph_get(my_comm, my_maxindex, my_maxedges,
                              my_index, my_edges);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Graph_get::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Graph_get";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GRAPH_GET";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Graph_get::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", maxindex = "
            << my_maxindex
            << ", maxedges = "
            << my_maxedges
            << ", *index, "
            << "*edges";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Graph_get::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_maxindex;
        out < my_maxedges;
        // out < my_index;
        // out < my_edges;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Graph_get::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_maxindex;
        in > my_maxedges;
        // in > my_index;
        // in > my_edges;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/topol/graph_get.c.
/*@

MPI_Graph_get - Retrieves graph topology information associated with a
communicator

Input Parameters:
+ comm - communicator with graph structure (handle)
. maxindex - length of vector 'index' in the calling program  (integer)
- maxedges - length of vector 'edges' in the calling program  (integer)

Output Parameter:
+ index - array of integers containing the graph structure (for details
see the definition of 'MPI_GRAPH_CREATE')
- edges - array of integers containing the graph structure

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG

int MPI_Graph_get (
MPI_Comm comm,
int maxindex,
int maxedges,
int *index,
int *edges )
@ */
