/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Graph_neighbors_count.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Graph_neighbors_count.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Graph_neighbors_count::
    MPO_Graph_neighbors_count()
        : MPO_Graph(MPI_COMM_NULL),
          my_rank(MPI_PROC_NULL),
          my_nneighbors(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Graph_neighbors_count::MPO_Graph_neighbors_count(MPO_Comm communicator,
                                                         int rank,
                                                         int
                                                         *nneighbors)
        : MPO_Graph(communicator),
          my_rank(rank),
          my_nneighbors(nneighbors)
    {
    }


    // destructor
    //========================================================================
    MPO_Graph_neighbors_count::~MPO_Graph_neighbors_count()
    {
    }


    /**
     *  We overload MPO_Graph::checkPrerequisites
     *  to check other prerequisites:
     *  Check whether rank is valid.
     */
    // checkRank
    //========================================================================
    int MPO_Graph_neighbors_count::checkRank()
    {

        /** \todo @creator of this method: what about return values ? */

        if (my_rank < 0)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Rank less than 0 !"
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_NEG,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            int commsize = 0;
            PMPI_Comm_size(my_comm, &commsize);
            if ((my_rank >= commsize) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Rank out of communicator!"
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_TOO_BIG,getSourceLocation());
                our_log->getMsg(msg);
            }
        }

        return 1;
    }


    // getId
    //========================================================================
    int MPO_Graph_neighbors_count::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GRAPH_NEIGHBORS_COUNT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Graph_neighbors_count::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Graph_neighbors_count::checkPrerequisites(void)
    {


        MPO_Graph::checkPrerequisites();
        checkRank();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Graph_neighbors_count::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Graph_neighbors_count(my_comm, my_rank, my_nneighbors);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Graph_neighbors_count::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Graph_neighbors_count";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GRAPH_NEIGHBORS_COUNT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Graph_neighbors_count::debugout(std::ostream
                                                       & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", rank = "
            << my_rank
            << ", *nneighbors";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Graph_neighbors_count::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_rank;
        // out < my_nneighbors;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Graph_neighbors_count::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_rank;
        // in > my_nneighbors;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied
// from mpich-1.2.4/src/topol/grapgnbrcnt.c.
/*@

MPI_Graph_neighbors_count - Returns the number of neighbors of a node
associated with a graph topology

Input Parameters:
+ comm - communicator with graph topology (handle)
- rank - rank of process in group of 'comm' (integer)

Output Parameter:
. nneighbors - number of neighbors of specified process (integer)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
.N MPI_ERR_RANK

int MPI_Graph_neighbors_count (
MPI_Comm comm,
int rank,
int *nneighbors )
@ */


