/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Group_range_excl.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Group_range_excl.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Group_range_excl::MPO_Group_range_excl()
        : MPO_Groupconstructor(MPI_GROUP_NULL,
                               MPI_GROUP_NULL,
                               0,
                               -1)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Group_range_excl::MPO_Group_range_excl(MPO_Group group,
                                               int n,
                                               int ranges[][3],
                                               MPI_Group *
                                               newgroup)
        : MPO_Groupconstructor(group,
                               MPI_GROUP_NULL,
                               newgroup,
                               1),
          my_n(n),
          my_ranges(ranges)
    {
    }


    // destructor
    //========================================================================
    MPO_Group_range_excl::~MPO_Group_range_excl()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Group_range_excl::checkPrerequisites2()
    {
        // check correctness of parameters
        checkRangeArray();

        return 0;
    }


    // checkRangeArray
    //========================================================================
    int MPO_Group_range_excl::checkRangeArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int size = -1;
        int i = 0;
        int j = 0;
        int comprank = -1;
        int valid = 1;
        int *testarray = 0;

        PMPI_Group_size(my_group, &size);

        // Initialize testarray to 0.
        testarray = new int[size];
        for (j = 0; j <= size - 1; j++)
        {
            testarray[j] = 0;
        }

        // Check array of ranges
        for (i = 0; i <= my_n - 1; i++)
        {
            // (1.1) Check if first rank is negative.
            if ((our_errcodes.at(MARMOT_ERR_RANK_FIRST_NEG))&&(my_ranges[i][0] < 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR :"
                    << getCallname()
                    << " : First rank in triplet "
                    << i
                    << " is less than zero!"
                    << std::endl;
                rvalue = MARMOT_ERR_RANK_FIRST_NEG;
                valid = 0;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (1.2) Check if last rank is negative.
            if ((our_errcodes.at(MARMOT_ERR_RANK_LAST_NEG))&&(my_ranges[i][1] < 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR :"
                    << getCallname()
                    << " : Last rank in triplet "
                    << i
                    << " is less than zero!"
                    << std::endl;
                rvalue = MARMOT_ERR_RANK_LAST_NEG;
                valid = 0;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (2.1) Check if first rank is within group.
            if ((our_errcodes.at(MARMOT_ERR_RANK_FIRST_TOO_BIG))&&(my_ranges[i][0] > size) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": First rank in triplet "
                    << i
                    << " is out of Group!"
                    << std::endl;
                rvalue = MARMOT_ERR_RANK_FIRST_TOO_BIG;
                valid = 0;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (2.2) Check if last rank is within group.
            if ((our_errcodes.at(MARMOT_ERR_RANK_LAST_TOO_BIG))&&((my_ranges[i][0] +
                  ((my_ranges[i][1] - my_ranges[i][0]) / my_ranges[i][2]) /
                  my_ranges[i][2]) > size) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Last rank in triplet "
                    << i
                    << " is out of Group!"
                    << std::endl;
                rvalue = MARMOT_ERR_RANK_LAST_TOO_BIG;
                valid = 0;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (3.1) Check if stride is negative.
            if ((our_errcodes.at(MARMOT_ERR_STRIDE_NEG))&&(my_ranges[i][0] < my_ranges[i][1]) &&
                (my_ranges[i][2] < 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": In triplet "
                    << i
                    << ": First rank less than last rank, but stride less than 0 !"
                    << std::endl;
                rvalue = MARMOT_ERR_STRIDE_NEG;
                valid = 0;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (3.2) Check if stride is 0.
            if ((our_errcodes.at(MARMOT_ERR_STRIDE_ZERO))&&(my_ranges[i][2] == 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Stride of triplet "
                    << i
                    << " is zero!"
                    << std::endl;
                rvalue = MARMOT_ERR_STRIDE_ZERO;
                valid = 0;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (3.3) Check if stride is positive.
            if ((our_errcodes.at(MARMOT_ERR_STRIDE_POS))&&(my_ranges[i][0] > my_ranges[i][1]) &&
                (my_ranges[i][2] > 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": In triplet "
                    << i
                    << ": First rank greater than last rank, but stride greater than 0 !"
                    << std::endl;
                rvalue = MARMOT_ERR_STRIDE_POS;
                valid = 0;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (4) Check if all computed ranges are distinct and valid.
            if (valid)
            {
                comprank = my_ranges[i][0];

                while (((my_ranges[i][0] <= my_ranges[i][1]) &&
                        (comprank <= my_ranges[i][1])) ||
                       ((my_ranges[i][0] > my_ranges[i][1]) &&
                        (comprank >= my_ranges[i][1])))
                {
                    // (4.1) Check if computed rank is negative.
                    if ((our_errcodes.at(MARMOT_ERR_RANK_COMPUTED_NEG))&&(comprank < 0) &&
                        (MSG_ERR <= DEBUG_MODE))
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << " Computed rank less than 0"
                            << std::endl;
                        rvalue = MARMOT_ERR_RANK_COMPUTED_NEG;

                        our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
                    }

                    // (4.2) Check if computed rank is out of group.
                    if ((our_errcodes.at(MARMOT_ERR_RANK_COMPUTED_TOO_BIG))&&(comprank > size) &&
                        (MSG_ERR <= DEBUG_MODE))
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << " Computed rank out of group: "
                            << std::endl
                            << "Rank = "
                            << comprank
                            << ", size of comm = "
                            << size
                            << std::endl;
                        rvalue = MARMOT_ERR_RANK_COMPUTED_TOO_BIG;

                        our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
                    }

                    if (testarray[comprank] == 0)
                    {
                        testarray[comprank] = 1;
                    }
                    // (4.3) Check if computed rank appears twice.
                    else if ((our_errcodes.at(MARMOT_ERR_RANK_COMPUTED_TWICE))&&(testarray[comprank] != 0) &&
                             (MSG_ERR <= DEBUG_MODE))
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": Rank "
                            << comprank
                            << " twice to exclude! "
                            << std::endl;
                        rvalue = MARMOT_ERR_RANK_COMPUTED_TWICE;

                        our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
                    }

                    comprank = comprank + (my_ranges[i][2]);
                }               // end while
            }                   // end if (4)
        }                       // end for

        delete[] testarray;

        return rvalue;
    }                           // end checkRangeArray


    // getId
    //========================================================================
    int MPO_Group_range_excl::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GROUP_RANGE_EXCL_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Group_range_excl::checkPost(void)
    {

        insertGroup();

        our_group_traces->addAndSetCreation(*my_newgroup,getSourceLocation());
        our_group_traces->findAndTryAddDependendency(*my_newgroup,my_group);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Group_range_excl::checkPrerequisites(void)
    {


        MPO_Groupconstructor::checkPrerequisites();
        checkPrerequisites2();

        our_group_traces->findAndAddUsage(my_group,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Group_range_excl::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Group_range_excl(my_group, my_n, my_ranges, my_newgroup);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Group_range_excl::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Group_range_excl";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GROUP_RANGE_EXCL";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Group_range_excl::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(group = "
            << my_group
            << ", n = "
            << my_n
            << ", *ranks, "
            << "*newgroup";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Group_range_excl::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_group;
        out < my_n;
        // out < my_ranges;
        // out < my_newgroup;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Group_range_excl::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_group;
        in > my_n;
        // in > my_ranges;
        // in > my_newgroup;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/group_rexcl.c.
/*@

MPI_Group_range_excl - Produces a group by excluding ranges of processes
from an existing group

Input Parameters:
+ group - group (handle)
. n - number of elements in array 'ranks' (integer)
- ranges - a one-dimensional
array of integer triplets of the
form (first rank, last rank, stride), indicating the ranks in
'group'  of processes to be excluded
from the output group 'newgroup' .

Output Parameter:
. newgroup - new group derived from above, preserving the
order in 'group'  (handle)

Note:
Currently, each of the ranks to exclude must be
a valid rank in the group and all elements must be distinct or the
function is erroneous.  This restriction is per the draft.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_RANK
.N MPI_ERR_ARG

.seealso: MPI_Group_free

int MPI_Group_range_excl (
MPI_Group group,
int n,
int ranges[][3],
MPI_Group *newgroup )
@ */

