/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Group_translate_ranks.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Group_translate_ranks.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Group_translate_ranks::
    MPO_Group_translate_ranks()
        : MPO_Grouphandler(MPI_GROUP_NULL,
                           MPI_GROUP_NULL,
                           -1),
          my_num(-1),
          my_ranks(0),
          my_ranks2(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Group_translate_ranks::MPO_Group_translate_ranks(MPO_Group group,
                                                         int num,
                                                         int *ranks,
                                                         MPO_Group group2,
                                                         int
                                                         *ranks2)
        : MPO_Grouphandler(group,
                           group2,
                           2),
          my_num(num),
          my_ranks(ranks),
          my_ranks2(ranks2)
    {
    }


    // destructor
    //========================================================================
    MPO_Group_translate_ranks::~MPO_Group_translate_ranks()
    {
    }


    // checkRankArray
    //========================================================================
    int MPO_Group_translate_ranks::checkRankArray()
    {
        /** \todo @creator of this method: what about return values ? */

        int rvalue = 1;
        int size = -1;

        PMPI_Group_size(my_group, &size);
        if ((my_num > size) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Arrays of ranks are larger than first group !!"
                << std::endl;

            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group1");
            our_group_traces->findAndPrintTraceNamed(my_group2,stream,"group2");

            //bad Errorcode used here - Maybe we need a new Error code here - at least I couldn`t find a fiting one
            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_TOO_BIG,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Group_translate_ranks::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GROUP_TRANSLATE_RANKS_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Group_translate_ranks::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Group_translate_ranks::checkPrerequisites(void)
    {

        MPO_Grouphandler::checkPrerequisites();
        checkRankArray();

        our_group_traces->findAndAddUsage(my_group,getSourceLocation());
        our_group_traces->findAndAddUsage(my_group2,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Group_translate_ranks::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Group_translate_ranks(my_group, my_num, my_ranks,
                                            my_group2, my_ranks2);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Group_translate_ranks::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Group_translate_ranks";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GROUP_TRANSLATE_RANKS";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Group_translate_ranks::debugout(std::ostream
                                                       & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(group1 = "
            << my_group
            << ", n = "
            << my_num
            << ", *ranks1, "
            << "group2 = "
            << my_group2
            << ", *ranks2";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Group_translate_ranks::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_group;
        out < my_num;
        // out < my_ranks;
        out < my_group2;
        // out < my_ranks2;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Group_translate_ranks::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_group;
        in > my_num;
        // in > my_ranks;
        in > my_group2;
        // in > my_ranks2;

        return in;
    }

}                               // end of namespace MPO




// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/group_tranks.c.
/*@

MPI_Group_translate_ranks - Translates the ranks of processes in one group
to those in another group

Input Parameters:
+ group1 - group1 (handle)
. n - number of ranks in  'ranks1' and 'ranks2'  arrays (integer)
. ranks1 - array of zero or more valid ranks in 'group1'
- group2 - group2 (handle)

Output Parameter:
. ranks2 - array of corresponding ranks in group2,  'MPI_UNDEFINED'  when
no correspondence exists.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_ARG
.N MPI_ERR_RANK

int MPI_Group_translate_ranks (
MPI_Group group_a,
int n,
int *ranks_a,
MPI_Group group_b,
int *ranks_b )
@ */

