/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Ibsend.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Ibsend.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Ibsend::MPO_Ibsend()
        : MPO_Initiator(-1,
                        MPI_DATATYPE_NULL,
                        MPI_PROC_NULL,
                        MPI_ANY_TAG,
                        MPI_COMM_NULL,
                        0,
                        REQUEST_NULL),
          my_buf(0)
    {
    }


    /** Constructor taking proper arguments.
     *  The  constructor performs the following tasks:
     *  1. Check whether request is still in our_requests
     *  - no:
     *  - yes: print warning to stderr
     *  2. Store request in our_request
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     *  (it is a opaque type and thus  may contain some internal status etc. )
     */
    // proper constructor
    //========================================================================
    MPO_Ibsend::MPO_Ibsend(void *buf,
                           int count,
                           MPO_Datatype datatype,
                           int rank,
                           int tag,
                           MPO_Comm comm,
                           MPI_Request * request)
        : MPO_Initiator(count,
                        datatype,
                        rank,
                        tag,
                        comm,
                        request,
                        REQUEST_NOT_PERSISTENT),
          my_buf(buf)
    {
    }


    // destructor
    //========================================================================
    MPO_Ibsend::~MPO_Ibsend()
    {
    }


    // getId
    //========================================================================
    int MPO_Ibsend::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_IBSEND_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Ibsend::checkPost(void)
    {

           our_requests.insert(std::make_pair(MPO_Request(my_request_p), my_persistent));
        our_request_traces->addAndSetCreation(my_request_p,getSourceLocation());

        //register the buffer as used
        registerBuffer(my_buf,my_count,my_datatype,my_request_p);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Ibsend::checkPrerequisites(void)
    {


        MPO_Initiator::checkPrerequisites(MPO_IBSEND_ID);
        checkBufferInUse(my_buf,my_count,my_datatype);

        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Ibsend::callNativeMPI()
    {
        int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
#endif
        rvalue = PMPI_Ibsend(my_buf, my_count, my_datatype, my_rank, my_tag,
                           my_comm, my_request_p);

#ifdef MARMOT_TCHECK
        MPO_Space::getSyncObject()->unregisterRequestToIntEntry(my_request_p);
#endif

#ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
#endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Ibsend::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Ibsend";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_IBSEND";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Ibsend::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*buf, "
            << "count = "
            << my_count
            << ", datatype = "
            << my_datatype
            << ", dest = "
            << my_rank
            << ", tag = "
            << my_tag
            << ", comm = "
            << my_comm
            << ", *request";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Ibsend::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_buf;
        out < my_count;
        out < my_datatype;
        out < my_rank;
        out < my_tag;
        out < my_comm;
        // out < my_request_p;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Ibsend::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_buf;
        in > my_count;
        in > my_datatype;
        in > my_rank;
        in > my_tag;
        in > my_comm;
        // in > my_request_p;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/ibsend.c.
/*@
  MPI_Ibsend - Starts a nonblocking buffered send

  Input Parameters:
  + buf - initial address of send buffer (choice)
  . count - number of elements in send buffer (integer)
  . datatype - datatype of each send buffer element (handle)
  . dest - rank of destination (integer)
  . tag - message tag (integer)
  - comm - communicator (handle)

  Output Parameter:
  . request - communication request (handle)

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_COUNT
  .N MPI_ERR_TYPE
  .N MPI_ERR_TAG
  .N MPI_ERR_RANK
  .N MPI_ERR_BUFFER

  int MPI_Ibsend(
  void *buf,
  int count,
  MPI_Datatype datatype,
  int dest,
  int tag,
  MPI_Comm comm,
  MPI_Request *request )
  @ */


