/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Init.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Init.cc 979 2009-07-17 12:54:04Z hpcdiche $
 */


#include <stdlib.h>
#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo.h"
#include "mpo_LogASCII.h"
#include "mpo_LogToServer.h"
#include "mpo_LogToServerDDT.h"
#include "mpo_LogQueue.h"
#include "mpo_GenericLogFilter.h"
#include "mpo-fortran-datatypehelpers.h"
#include "MPO_ResourceGatherer.h"
#include "MPO_ResourceGathererFull.h"
#include "MPO_ResourceGathererDummy.h"
#include "MPO_ResourceGathererGraph.h"
#include "config.h"

namespace MPO
{
    // Initialize static stuff
    //========================================================================
    bool MPO_Init::our_call_init = true;

    // default constructor
    //========================================================================
    MPO_Init::MPO_Init()
    {
    }


    // proper constructor
    //========================================================================
    MPO_Init::MPO_Init(int *argc,
                       char ***argv)
        : my_argc(argc),
          my_argv(argv)
    {
        /* The call to the constructor of MPO_Init is special because this is
         * the place to initialize everything. We call MPI_Init here instead
         * of calling it in callNativeMPI().
         */
        if (our_call_init)
        {
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->enterPMPI();
#endif
            my_rvalue = PMPI_Init(argc, argv);
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->leavePMPI();
#endif
        }/*our_call_init*/

        /**
         * Create a new MPI_COMM_WORLD which excludes the debug server.
         * Init some environmentals.
         */
        initCommAndEnv();

        /* Add all the predefined resources such as errhandlers, datatypes,
         * communicators, operations and groups to our maps.
         */
        initFactories();
        initErrhandlers();
        //initDatatypes(); //Moved to checkPost, needed for UniMCI (Must not call MPI yet, calls MPI_f2c)
        initRequests();
        initCommunicators();
        initOperations();
        initGroups();
        initLogfile();

        //we will call this in callDebugServer()
        //because we need to know wheter source locations are present or not
        //initRessourceGatherers();

        /* Set buffer attached to 'false'. */
        buffer_attached = 0;

#ifdef MARMOT_THREADED
		our_threadlevel_required = MPI_THREAD_SINGLE;
		our_threadlevel_provided = MPI_THREAD_SINGLE;
#endif

		//now we set MPI as initalized !
		our_init_was_called = true;
    }                             // end constructor


    // destructor
    //========================================================================
    MPO_Init::~MPO_Init()
    {
    }

    // initCommAndEnv
    //========================================================================
    int MPO_Init::initCommAndEnv()
    {
		/* We could now check whether argc and argv are the same on all ranks.
         * We need a communicator that replaces MPI_COMM_WORLD for the
         * application.
         * The node with the highest rank is excluded and used for the
         * debug server.
         */
// TODO // node with any rank could be debug server.

        int rank = 0;

        if (MPO_Space::isDebugServerEnabled())
        {
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->enterPMPI();
#endif
            PMPI_Comm_rank(MPI_COMM_WORLD, &rank);
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->leavePMPI();
#endif

            int size = 0;
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->enterPMPI();
#endif
            PMPI_Comm_size(MPI_COMM_WORLD, &size);
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->leavePMPI();
#endif

            int color = 0;
            if (rank == (size - 1))
            {
                color = 1;
            }

#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->enterPMPI();
#endif
            PMPI_Comm_split(MPI_COMM_WORLD, color, rank, &our_mpi_comm_world);
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->leavePMPI();
#endif

#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->enterPMPI();
#endif
            our_channel.setCommunicator(MPI_COMM_WORLD);
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->leavePMPI();
#endif
        }/*MPO_Space::isDebugServerEnabled*/
        else
        {
            our_mpi_comm_world = MPI_COMM_WORLD;
        }

        /* Set environment variables. */

        // We set DEBUG_MODE here in the constructor of MPO_Init and not
        // in the constructor of DebugServer (doesn't work there).
        if (getenv("MARMOT_DEBUG_MODE") == NULL)
        {
            DEBUG_MODE = DEBUG_MODE_DEFAULT;
        }
        else
        {
            DEBUG_MODE = atoi(getenv("MARMOT_DEBUG_MODE"));
        }

        // We set INTERFACE_MODE here in the constructor of MPO_Init and not
        // in the constructor of DebugServer (doesn't work there).
        // It is set automatically in mk_fortran_wrapper.sh for the Fortran
        // interface.
        if (getenv("MARMOT_INTERFACE_MODE") == NULL)
        {
            if (INTERFACE_MODE == LAST_INTERFACE_MODE) //only update if not yet set !
                INTERFACE_MODE = INTERFACE_MODE_DEFAULT;
        }
        else
        {
            INTERFACE_MODE = atoi(getenv("MARMOT_INTERFACE_MODE"));
        }

        return 0;
    }


    /** HINT FOR DEVELOPERS:
     *  ====================
     *  For the setting and mapping, we use the datatypes,
     *  communicators etc. in the order in which they are listed in the
     *  MPI-standard 1.2, pp. 210. Please keep this order, it makes it
     *  easier to check if the standard is implemented properly in MARMOT.
     *
     *  We add the resources and the resources names to our maps.
     *  For the resources, we use e.g.
     *      our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_CHAR),
     *                                          DATATYPE_COMMITTED_C));
     *  For the names of resources, we use e.g.
     *      our_datatype_names.insert(std::pair
     *                                < MPO_Datatype, std::string >
     *                                (MPO_Datatype(MPI_CHAR),
     *                                 "MPI_CHAR"));
     *  Unfortunately it didn't work with "make_pair"...., e.g.
     *      our_datatype_names.insert(std::make_pair
     *                                     (MPO_Datatype(MPI_CHAR),
     *                                      "MPI_CHAR"));
     *
     *  Currently we don't insert the optional Fortran types in our map
     *  because they are defined in mpif.h, not in mpi.h.
     *
     *  The magic numbers that were previously used in these maps and
     *  consequently in the switch statements of our check functions
     *  were replaced by enumeration values defined by enums such as
     *  KIND_OF_DATATYPE or KIND_OF_COLLECTIVE_CALL.
     *  KIND_OF_DATATYPE is very important because we have to consider
     *  the differences between the C and the Fortran language binding
     *  here (the magic numbers used previously didn't reflect all
     *  these differences properly.)
     *
     *  DON'T USE MAGIC NUMBERS, THEY ARE EVIL AND MAKE HORRIBLE CODE!
     *  Use descriptive enums and try to keep the complexity of check
     *  functions low.
     */


    // initErrhandlers
    //========================================================================
    int MPO_Init::initErrhandlers()
    {
        //==========================================
        // Error-handling specifiers (C and Fortran)
        // as defined in MPI-1.2, p. 211
        //==========================================

        /*  Insert predefined errhandlers.
         *  MPI_ERRORS_ARE_FATAL is associated to MPI_COMM_WORLD!
         */
// TODO // remove these magic numbers, grr
        int tmp[2] = { 1, 0 };

        // compilation error: xlC didn't like this
        // our_errhandlers.insert(std::make_pair(MPI_ERRORS_ARE_FATAL,
        //                                       (int*)tmp));
        our_errhandlers.insert(std::
                               make_pair(MPI_Errhandler(MPI_ERRORS_ARE_FATAL),
                                         (int *) tmp));

        tmp[0] = 0;

        // compilation error: xlC didn't like this
        // our_errhandlers.insert(std::make_pair(MPI_ERRORS_RETURN,
        //                                       (int*)tmp));
        our_errhandlers.insert(std::
                               make_pair(MPI_Errhandler(MPI_ERRORS_RETURN),
                                         (int *) tmp));
	//parse the env variable MARMOT_ERRCODES_SET
	ParseErrCodesEnv *ecv;
	if(getenv("MARMOT_ERRCODES_SET")!=NULL)
	{
		ecv=new ParseErrCodesEnv(getenv("MARMOT_ERRCODES_SET"),MARMOT_LAST_ERR_CODE+1,1);
	}
	else
	{
		ecv=new ParseErrCodesEnv("",MARMOT_LAST_ERR_CODE+1,1);
	}
	ecv->evaluate();
	for(int i=0;i<MARMOT_LAST_ERR_CODE;i++)
	{
		our_errcodes.push_back(ecv->getValue(i));
	}
	delete ecv;
	/*for(int j=0;j<our_errcodes.size();j++)
	{
		printf("ecv[%i]=%i ",j,our_errcodes.at(j));
	}*/
// TODO // Why is there no our_errhandler_names.insert?

        return 0;
    }                           // end initErrhandlers


    // initCommunicators
    //========================================================================
    int MPO_Init::initCommunicators()
    {
        //==========================================
        // reserved communicators (C and Fortran)
        // as defined in MPI-1.2, p.212
        //==========================================

        /*  Add *our* MPI_COMM_WORLD to our set of communicators. */

        // our_comms.insert(MPI_COMM_WORLD);
        our_comms.insert(our_mpi_comm_world);
        our_comms.insert(MPI_COMM_SELF);

// TODO // Why is there no our_comms_names.insert?

        return 0;
    }                           // end initCommunicators

    // initRequests
    //========================================================================
    int MPO_Init::initRequests()
    {

        MPI_Request r = MPI_REQUEST_NULL;
        our_requests.insert (std::make_pair(&r,REQUEST_NULL));

        return 0;
    }

    // initDatatypes
    //========================================================================
    int MPO_Init::initDatatypes()
    {
        initDatatypeNull();
        initDatatypeElementaryC();
        initDatatypeElementaryFortran();
        initDatatypeReductionC();
        initDatatypeReductionFortran();
        initDatatypeOptionalC();
        initDatatypeOptionalFortran();
        initDatatypeConstructDerivedDatatypes();
        return 0;
    }                           // end initDatatypes


    // initDatatypeNull
    //========================================================================
    int MPO_Init::initDatatypeNull()
    {
        //==========================================
        // Datatype null handle
        // as defined in MPI-1.2, p. 213
        //==========================================

        /*  Add the null handle to our map. */
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_DATATYPE_NULL),
                                            DATATYPE_NULL));

        /*  Add the null-handle-name to our map. */
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_DATATYPE_NULL),
                                   "MPI_DATATYPE_NULL"));
        return 0;
    }                           // end initDatatypeNull


    // initDatatypeElementaryC
    //========================================================================
    int MPO_Init::initDatatypeElementaryC()
    {
        //==========================================
        // elementary datatypes (C)
        // as defined in MPI-1.2, p. 211
        //==========================================

        /*  Add the elementary datatypes (C) to our map (as committed).
         */

        // The following datatypes are only for C, not for Fortran.
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_CHAR),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_SHORT),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_INT),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_LONG),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_UNSIGNED_CHAR),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_UNSIGNED_SHORT),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_UNSIGNED),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_UNSIGNED_LONG),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_FLOAT),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_DOUBLE),
                                            DATATYPE_COMMITTED_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_LONG_DOUBLE),
                                            DATATYPE_COMMITTED_C));

        // The following datatypes are for C AND for Fortran.
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_BYTE),
                                            DATATYPE_COMMITTED_C_AND_FORTRAN));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_PACKED),
                                            DATATYPE_COMMITTED_C_AND_FORTRAN));


        /*  Add the elementary datatype-names (C) to our map.
         */

        // The following datatypes are only for C, not for Fortran.
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_CHAR),
                                   "MPI_CHAR"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_SHORT),
                                   "MPI_SHORT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_INT),
                                   "MPI_INT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_LONG),
                                   "MPI_LONG"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_UNSIGNED_CHAR),
                                   "MPI_UNSIGNED_CHAR"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_UNSIGNED_SHORT),
                                   "MPI_UNSIGNED_SHORT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_UNSIGNED),
                                   "MPI_UNSIGNED"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_UNSIGNED_LONG),
                                   "MPI_UNSIGNED_LONG"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_FLOAT),
                                   "MPI_FLOAT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_DOUBLE),
                                   "MPI_DOUBLE"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_LONG_DOUBLE),
                                   "MPI_LONG_DOUBLE"));

        // The following datatypes are for C AND for Fortran.
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_BYTE),
                                   "MPI_BYTE"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_PACKED),
                                   "MPI_PACKED"));
        return 0;
    }                           // end initDatatypeElementaryC



    // initDatatypeElementaryFortran
    //========================================================================
    int MPO_Init::initDatatypeElementaryFortran()
    {

        //==========================================
        // elementary datatypes (Fortran)
        // as defined in MPI-1.2, p. 211
        //==========================================

        /*  Add Fortran-Types to the map (because they are not allowed in C).
         */

        //#ifdef USE_LAM_MPI
        // Most MPI implementations have Fortran datatypes in mpi.h,
        // e.g. mpich and implementations based on mpich.
        // In this case, we cannot insert Fortran datatypes here.
        //#else
#ifdef MPI_INTEGER_PRESENT_IN_MPI_H
        // The following datatypes are only for Fortran, not for C.
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_INTEGER),
                                            DATATYPE_COMMITTED_FORTRAN));
#endif
#ifdef MPI_REAL_PRESENT_IN_MPI_H
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_REAL),
                                            DATATYPE_COMMITTED_FORTRAN));
#endif
#ifdef MPI_DOUBLE_PRECISION_PRESENT_IN_MPI_H
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_DOUBLE_PRECISION),
                                            DATATYPE_COMMITTED_FORTRAN));
#endif
#ifdef MPI_COMPLEX_PRESENT_IN_MPI_H
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_COMPLEX),
                                            DATATYPE_COMMITTED_FORTRAN));
#endif
#ifdef MPI_LOGICAL_PRESENT_IN_MPI_H
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_LOGICAL),
                                            DATATYPE_COMMITTED_FORTRAN));
#endif
#ifdef MPI_CHARACTER_PRESENT_IN_MPI_H
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_CHARACTER),
                                            DATATYPE_COMMITTED_FORTRAN));
#endif

        // The following two datatypes are for C AND for Fortran, see above:
        // MPI_BYTE
        // MPI_PACKED

        // According to clarifications of MPI-1.2 in MPI-2.0,
        // MPI_DOUBLE_COMPLEX is an optional datatype. So we don't
        // have it here as elementary datatype, but later as optional
        // datatype.

        /*  Add Fortran-Type-names to the map
         *  (they are not allowed in C).
         */

        // The following datatypes are only for Fortran, not for C.
#ifdef MPI_INTEGER_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_INTEGER),
                                   "MPI_INTEGER"));
#endif
#ifdef MPI_REAL_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_REAL),
                                   "MPI_REAL"));
#endif
#ifdef MPI_DOUBLE_PRECISION_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_DOUBLE_PRECISION),
                                   "MPI_DOUBLE_PRECISION"));
#endif
#ifdef MPI_COMPLEX_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_COMPLEX),
                                   "MPI_COMPLEX"));
#endif
#ifdef MPI_LOGICAL_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_LOGICAL),
                                   "MPI_LOGICAL"));
#endif
#ifdef MPI_CHARACTER_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_CHARACTER),
                                   "MPI_CHARACTER"));
#endif

        // The following two datatypes are for C AND for Fortran, see above:
        // MPI_BYTE
        // MPI_PACKED
        // MPI_DOUBLE_COMPLEX is an optional datatype.

        return 0;
    }                           // end initDatatypeElementaryFortran



    // initDatatypeReductionC
    //========================================================================
    int MPO_Init::initDatatypeReductionC()
    {
        //==========================================
        // datatypes for reduction functions (C)
        // as defined in MPI-1.2, p. 212
        //==========================================

        /*  Add the datatypes for reduction functions (C) to our map.
         */
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_FLOAT_INT),
                                            DATATYPE_REDUCTION_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_DOUBLE_INT),
                                            DATATYPE_REDUCTION_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_LONG_INT),
                                            DATATYPE_REDUCTION_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_2INT),
                                            DATATYPE_REDUCTION_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_SHORT_INT),
                                            DATATYPE_REDUCTION_C));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_LONG_DOUBLE_INT),
                                            DATATYPE_REDUCTION_C));

        /*  Add the datatype-names for reduction functions (C) to our map.
         */
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_FLOAT_INT),
                                   "MPI_FLOAT_INT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_DOUBLE_INT),
                                   "MPI_DOUBLE_INT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_LONG_INT),
                                   "MPI_LONG_INT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_2INT),
                                   "MPI_2INT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_SHORT_INT),
                                   "MPI_SHORT_INT"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_LONG_DOUBLE_INT),
                                   "MPI_LONG_DOUBLE_INT"));

        return 0;
    }                           // end initDatatypeReductionC



    // initDatatypeReductionFortran
    //========================================================================
    int MPO_Init::initDatatypeReductionFortran()
    {
        //==========================================
        // datatypes for reduction functions (Fortran)
        // as defined in MPI-1.2, p. 212
        //==========================================

        //#ifdef USE_LAM_MPI
        // Most MPI implementations have Fortran datatypes in mpi.h,
        // e.g. mpich and implementations based on mpich.
        // In this case, we cannot insert Fortran datatypes here.
        //#else
#ifdef MPI_2REAL_PRESENT_IN_MPI_H
        /*  Add the datatypes for reduction functions (Fortran) to our map.
         */
        our_datatypes.insert(std::
                             make_pair(MPO_Datatype(MPI_2REAL),
                                       DATATYPE_REDUCTION_FORTRAN));
#endif
#ifdef MPI_2DOUBLE_PRECISION_PRESENT_IN_MPI_H
        our_datatypes.insert(std::
                             make_pair(MPO_Datatype(MPI_2DOUBLE_PRECISION),
                                       DATATYPE_REDUCTION_FORTRAN));
#endif
#ifdef MPI_2INTEGER_PRESENT_IN_MPI_H
        our_datatypes.insert(std::
                             make_pair(MPO_Datatype(MPI_2INTEGER),
                                       DATATYPE_REDUCTION_FORTRAN));
#endif

        /*  Add the datatype-names for reduction functions (Fortran)
         *  to our map.
         */
#ifdef MPI_2REAL_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_2REAL),
                                   "MPI_2REAL"));
#endif
#ifdef MPI_2DOUBLE_PRECISION_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_2DOUBLE_PRECISION),
                                   "MPI_2DOUBLE_PRECISION"));
#endif
#ifdef MPI_2INTEGER_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_2INTEGER),
                                   "MPI_2INTEGER"));
#endif
        return 0;
    }                           // end initDatatypeReductionFortran


    // initDatatypeOptionalC
    //========================================================================
    int MPO_Init::initDatatypeOptionalC()
    {
        //==========================================
        // optional datatypes (C)
        // as defined in MPI-1.2, p. 212
        //==========================================

        /*  Add the optional datatypes (C) to our map. */
#ifdef MPI_LONG_LONG_INT_PRESENT_IN_MPI_H
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_LONG_LONG_INT),
                                            DATATYPE_OPTIONAL_C));
#endif

        /*  Add the optional datatype-names (C) to our map. */
#ifdef MPI_LONG_LONG_INT_PRESENT_IN_MPI_H
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_LONG_LONG_INT),
                                   "MPI_LONG_LONG_INT"));
#endif
        return 0;
    }                           // end initDatatypeOptionalC


    // initDatatypeOptionalFortran
    //========================================================================
    int MPO_Init::initDatatypeOptionalFortran()
    {
        //==========================================
        // optional datatypes (Fortran)
        // as defined in MPI-1.2, p. 212
        //==========================================
// TODO //
        /*  Add the optional datatypes (Fortran) to our map.
         *  Currently we don't do this because they are defined in mpif.h,
         *  not in mpi.h.
         * -> first version fixing this Problem the solution uses Fortran
         *    functions to get the type handles. No tests for
         *    compatibility and portability made yet.
         */
#ifndef MARMOT_NO_FORTRAN
        MPI_Fint    f_type_integer1,
                    f_type_integer2,
                    f_type_integer4,
                    f_type_real2,
                    f_type_real4,
                    f_type_real8,
                    f_type_double_complex;
        MPI_Datatype    type_integer1,
                        type_integer2,
                        type_integer4,
                        type_real2,
                        type_real4,
                        type_real8,
                        type_double_complex;


        try
        {
            GET_MPI_INTEGER1_CODE(&f_type_integer1);
            GET_MPI_INTEGER2_CODE(&f_type_integer2);
            GET_MPI_INTEGER4_CODE(&f_type_integer4);
            GET_MPI_REAL2_CODE(&f_type_real2);
            GET_MPI_REAL4_CODE(&f_type_real4);
            GET_MPI_REAL8_CODE(&f_type_real8);
            GET_MPI_DOUBLE_COMPLEX_CODE(&f_type_double_complex);

#ifdef MPI_HANLDE_CONVERT_PRESENT
            type_integer1   = MPI_Type_f2c(f_type_integer1);
            type_integer2   = MPI_Type_f2c(f_type_integer2);
            type_integer4   = MPI_Type_f2c(f_type_integer4);
            type_real2      = MPI_Type_f2c(f_type_real2);
            type_real4      = MPI_Type_f2c(f_type_real4);
            type_real8      = MPI_Type_f2c(f_type_real8);
            type_double_complex = MPI_Type_f2c(f_type_double_complex);
#else
            //we assume that handles are the same in Fortran and C here
            type_integer1   = (MPI_Datatype)f_type_integer1;
            type_integer2   = (MPI_Datatype)f_type_integer2;
            type_integer4   = (MPI_Datatype)f_type_integer4;
            type_real2      = (MPI_Datatype)f_type_real2;
            type_real4      = (MPI_Datatype)f_type_real4;
            type_real8      = (MPI_Datatype)f_type_real8;
            type_double_complex = (MPI_Datatype)f_type_double_complex;
#endif


        }
        catch (...)     {}

        if (type_integer1 != MPI_DATATYPE_NULL)
            our_datatypes.insert(std::make_pair(MPO_Datatype(type_integer1),
                                                DATATYPE_OPTIONAL_FORTRAN));
        if (type_integer2 != MPI_DATATYPE_NULL)
            our_datatypes.insert(std::make_pair(MPO_Datatype(type_integer2),
                                                DATATYPE_OPTIONAL_FORTRAN));
        if (type_integer4 != MPI_DATATYPE_NULL)
            our_datatypes.insert(std::make_pair(MPO_Datatype(type_integer4),
                                                DATATYPE_OPTIONAL_FORTRAN));
        if (type_real2 != MPI_DATATYPE_NULL)
            our_datatypes.insert(std::make_pair(MPO_Datatype(type_real2),
                                                DATATYPE_OPTIONAL_FORTRAN));
        if (type_real4 != MPI_DATATYPE_NULL)
            our_datatypes.insert(std::make_pair(MPO_Datatype(type_real4),
                                                DATATYPE_OPTIONAL_FORTRAN));
        if (type_real8 != MPI_DATATYPE_NULL)
            our_datatypes.insert(std::make_pair(MPO_Datatype(type_real8),
                                                DATATYPE_OPTIONAL_FORTRAN));

        // According to clarifications of MPI-1.2 in MPI-2.0,
        // MPI_DOUBLE_COMPLEX is an optional datatype and not an elementary
        // datatype.
        if (type_double_complex != MPI_DATATYPE_NULL)
            our_datatypes.insert(std::make_pair(MPO_Datatype(
                                                    type_double_complex),
                                                DATATYPE_OPTIONAL_FORTRAN));

        /*  Add the optional datatype-names (Fortran) to our map.
         *  Currently we don't do this because they are defined in mpif.h,
         *  not in mpi.h.
         */
        if (type_integer1 != MPI_DATATYPE_NULL)
            our_datatype_names.insert(std::pair<MPO_Datatype, std::string>
                                      (MPO_Datatype(type_integer1),
                                       "MPI_INTEGER1"));
        if (type_integer2 != MPI_DATATYPE_NULL)
            our_datatype_names.insert(std::pair<MPO_Datatype, std::string>
                                      (MPO_Datatype(type_integer2),
                                       "MPI_INTEGER2"));
        if (type_integer4 != MPI_DATATYPE_NULL)
            our_datatype_names.insert(std::pair<MPO_Datatype, std::string>
                                      (MPO_Datatype(type_integer4),
                                       "MPI_INTEGER4"));
        if (type_real2 != MPI_DATATYPE_NULL)
            our_datatype_names.insert(std::pair<MPO_Datatype, std::string>
                                      (MPO_Datatype(type_real2),
                                       "MPI_REAL2"));
        if (type_real4 != MPI_DATATYPE_NULL)
            our_datatype_names.insert(std::pair<MPO_Datatype, std::string>
                                      (MPO_Datatype(type_real4),
                                       "MPI_REAL4"));
        if (type_real8 != MPI_DATATYPE_NULL)
            our_datatype_names.insert(std::pair<MPO_Datatype, std::string>
                                      (MPO_Datatype(type_real8),
                                       "MPI_REAL8"));
        if (type_double_complex != MPI_DATATYPE_NULL)
            our_datatype_names.insert(std::pair<MPO_Datatype, std::string>
                                      (MPO_Datatype(type_double_complex),
                                       "MPI_DOUBLE_COMPLEX"));
#endif
        return 0;
    }                           // end initDatatypeOptionalFortran



    // initDatatypeConstructDerivedDatatypes
    //========================================================================
    int MPO_Init::initDatatypeConstructDerivedDatatypes()
    {
        //==========================================
        // special datatypes for constructing derived
        // datatypes
        // as defined in MPI-1.2, p. 212
        //==========================================

        /*  Add the special datatypes for constructing derived datatypes
         *  to our map.
         */
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_UB),
                                            DATATYPE_CONSTRUCT_DERIVED_TYPE));
        our_datatypes.insert(std::make_pair(MPO_Datatype(MPI_LB),
                                            DATATYPE_CONSTRUCT_DERIVED_TYPE));

        /*  Add the names of the special datatypes for constructing derived
         *  datatypes to our map.
         */
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_UB),
                                   "MPI_UB"));
        our_datatype_names.insert(std::pair < MPO_Datatype, std::string >
                                  (MPO_Datatype(MPI_LB),
                                   "MPI_LB"));

        return 0;
    }                           // end initDatatypeConstructDerivedDatatypes


    // initOperations
    //========================================================================
    int MPO_Init::initOperations()
    {
        //==========================================
        // collective operations (C and Fortran)
        // as defined in MPI-1.2, p. 213
        //==========================================

        /*  Add the predefined reduction operations to our set. */
        our_ops.insert(MPI_MAX);
        our_ops.insert(MPI_MIN);
        our_ops.insert(MPI_SUM);
        our_ops.insert(MPI_PROD);
        our_ops.insert(MPI_MAXLOC);
        our_ops.insert(MPI_MINLOC);
        our_ops.insert(MPI_BAND);
        our_ops.insert(MPI_BOR);
        our_ops.insert(MPI_BXOR);
        our_ops.insert(MPI_LAND);
        our_ops.insert(MPI_LOR);
        our_ops.insert(MPI_LXOR);

        /*  Add the operator-names for reduction functions (C and Fortran)
         *  to our map.
         */
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_MAX),
                             "MPI_MAX"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_MIN),
                             "MPI_MIN"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_SUM),
                             "MPI_SUM"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_PROD),
                             "MPI_PROD"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_MAXLOC),
                             "MPI_MAXLOC"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_MINLOC),
                             "MPI_MINLOC"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_BAND),
                             "MPI_BAND"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_BOR),
                             "MPI_BOR"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_BXOR),
                             "MPI_BXOR"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_LAND),
                             "MPI_LAND"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_LOR),
                             "MPI_LOR"));
        our_op_names.insert(std::pair < MPO_Op, std::string >
                            (MPO_Op(MPI_LXOR),
                             "MPI_LXOR"));

        return 0;
    }                           // end initOperations


    // initGroups
    //========================================================================
    int MPO_Init::initGroups()
    {
        //==========================================
        // predefined groups
        // as defined in MPI-1.2, p. 213
        //==========================================

        /*  Add the predefined groups to our set. */
        our_groups.insert(MPI_GROUP_EMPTY);

// TODO // Why is there no our_group_names.insert?

        return 0;
    }                           // end initGroups


#ifdef MARMOT_1SIDED

    // initWins
    //========================================================================
    int MPO_Init::initWins()
    {
        //==========================================
        // predefined windows
        //==========================================

        /*  Add the predefined windows to our set. */
        our_wins.insert(std::pair<MPI_Win,mpi_win_marmot_states_t>(MPI_WIN_NULL,MPI_WIN_FREE_ALLOWED));

        return 0;
    }                           // end initGroups
#endif


    // initLogFile
    //========================================================================
    int MPO_Init::initLogfile()
    {
        // Initialize the Global Log filter count by the corresponding
        // environmental variable
        if (getenv("MARMOT_LOG_FILTER_COUNT") == NULL)
        {
            LOGFILTER_COUNT = LOGFILTER_COUNT_DEFAULT;
        }
        else
        {
            LOGFILTER_COUNT = atoi(getenv("MARMOT_LOG_FILTER_COUNT"));
        }

        // Initialize the Global Logfiletype value by the corresponding
        // environmental variable
        if (!MPO_Space::isLogTypeForced())
        {
            if (getenv("MARMOT_LOGFILE_TYPE") == NULL)
            {
                LOGFILE_TYPE = LOGFILE_TYPE_DEFAULT;
            }
            else
            {
                LOGFILE_TYPE = atoi(getenv("MARMOT_LOGFILE_TYPE"));
            }
        }
        else
        {
            LOGFILE_TYPE = MPO_Space::getForcedLogType ();
        }

        //Initialize the global selection for the flushing type
        if (getenv("MARMOT_LOG_FLUSH_TYPE") == NULL)
        {
            LOGFLUSH_TYPE = LOGFLUSH_TYPE_DEFAULT;
        }
        else
        {
            LOGFLUSH_TYPE = atoi(getenv("MARMOT_LOG_FLUSH_TYPE"));
        }

        //Initialize our_log depending on the selected logfile type
        switch (LOGFILE_TYPE)
        {
            case LOGFILE_TYPE_ASCII:
                our_log = new MPO_GenericLogFilter<MPO_LogToServer> ();
                break;
            case LOGFILE_TYPE_HTML:
                our_log = new MPO_GenericLogFilter<MPO_LogToServer> ();
                break;
            case LOGFILE_TYPE_CUBE:
                our_log = new MPO_GenericLogFilter<MPO_LogToServer> ();
                break;
            case LOGFILE_TYPE_CUBE_DDT:
                our_log = new MPO_GenericLogFilter<MPO_LogToServerDDT> ();
                break;
            case LOGFILE_TYPE_QUEUE:
                our_log = new MPO_GenericLogFilter<MPO_LogQueue> ();
                break;
        }

        return (0);
    }// end initLogfile

    // initResourceGatherers
    //========================================================================
    int MPO_Init::initResourceGatherers ()
    {
        // Initialize the Global Logfiletype value by the corresponding
        // environmental variable
        if (getenv("MARMOT_RESOURCE_TRACE_SELECTION") == NULL)
        {
            RESOURCE_TRACE_TYPE = RESOURCE_TRACE_TYPE_DEFAULT;
        }
        else
        {
            RESOURCE_TRACE_TYPE = atoi(getenv("MARMOT_RESOURCE_TRACE_SELECTION"));
        }

        switch (RESOURCE_TRACE_TYPE)
        {
            case RESOURCE_TRACE_ON:
                if (getSourceLocation().isValidLocation())
                {
                    //Use full resource tracing
                    our_datatype_traces = new MPO_ResourceGathererFull<MPO_Datatype> ("MPI_Datatype",1,1,1,1,1);
                    our_op_traces = new MPO_ResourceGathererFull<MPO_Op> ("MPI_Op",1,1,0,1,0);
                    our_comm_traces = new MPO_ResourceGathererFull<MPO_Comm> ("MPI_Comm",1,1,0,1,1);
                    our_errhandler_traces = new MPO_ResourceGathererFull<MPI_Errhandler> ("MPI_Errhandler",1,1,0,1,0);
                    our_request_traces = new MPO_ResourceGathererFull<MPO_Request> ("MPI_Request",1,1,0,1,0);
                    our_group_traces = new MPO_ResourceGathererFull<MPO_Group> ("MPI_Group",1,1,0,1,1);
                }
                else
                {
                    //Use resource graph tracing
                    //(only timed information about the amount of resources used)
                    our_datatype_traces = new MPO_ResourceGathererGraph<MPO_Datatype> ("MPI_Datatype");
                    our_op_traces = new MPO_ResourceGathererGraph<MPO_Op> ("MPI_Op");
                    our_comm_traces = new MPO_ResourceGathererGraph<MPO_Comm> ("MPI_Comm");
                    our_errhandler_traces = new MPO_ResourceGathererGraph<MPI_Errhandler> ("MPI_Errhandler");
                    our_request_traces = new MPO_ResourceGathererGraph<MPO_Request> ("MPI_Request");
                    our_group_traces = new MPO_ResourceGathererGraph<MPO_Group> ("MPI_Group");
                }
                break;
            case RESOURCE_TRACE_OFF:
                //Use dummy implementation for the Gatherer that does nothing at all
                //That enables treating all possible cases of resource tracing equaly.
                //Thus no in-code switches are necessary for logging or using resource
                //information.
                our_datatype_traces = new MPO_ResourceGathererDummy<MPO_Datatype> ();
                our_op_traces = new MPO_ResourceGathererDummy<MPO_Op> ();
                our_comm_traces = new MPO_ResourceGathererDummy<MPO_Comm> ();
                our_errhandler_traces = new MPO_ResourceGathererDummy<MPI_Errhandler> ();
                our_request_traces = new MPO_ResourceGathererDummy<MPO_Request> ();
                our_group_traces = new MPO_ResourceGathererDummy<MPO_Group> ();
                break;
        }

        return 0;
    } //end of method initResourceGatherers

    // callDebugServer
    //========================================================================
    int MPO_Init::callDebugServer()
    {
        /*  The call to MPI_Init was done in the constructor.
         *  The debug process does not return from this call, but will exexute
         *  the server process for debugging.
         */
        if (MPO_Space::isDebugServerEnabled())
        {
            int rank = -1;
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->enterPMPI();
#endif
            PMPI_Comm_rank(MPI_COMM_WORLD, &rank);
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->leavePMPI();
#endif

            int size = -1;
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->enterPMPI();
#endif
            PMPI_Comm_size(MPI_COMM_WORLD, &size);
#ifdef MARMOT_USE_THREADSYNC
            MPO_Space::getSyncObject()->leavePMPI();
#endif

            if (rank == (size - 1))
            {
                // I am the debug server process
                if(*my_argc)
                {
                    MPO::DebugServer server(our_channel,(*my_argv)[0]);
            	    server.runMainLoop();
                }
                else
                {
                    MPO::DebugServer server(our_channel,(char*)NULL);
            	    server.runMainLoop();
                }
            }
            else
            {
                // A client informs the server about this call.
                our_channel.sendId(this->getFamilyId(),-1);
                our_channel.sendId(this->getId(),-1);
                our_channel.sendObject(*this,-1);

                // wait for reply from server
                int server_token = -1;
                our_channel.receiveId(server_token,-1);

                //Get the processor's name and send it to the debugserver
                our_channel.sendId(MPO_MESSAGEOBJECT_ID,-1);
                our_channel.sendId(MPO_STANDARDMESSAGE,-1);
                char *pn=new char [MPI_MAX_PROCESSOR_NAME];
                int pn_length;
                PMPI_Get_processor_name(pn,&pn_length);
                std::string pn_str(pn);
                //pn_str+='\n';	//spater weg damit
                MPO::MPO_MsgObject pn_obj(MSG_TYPE_INFO,0,rank,pn_str,MARMOT_NOTE_PROCESSOR_NAME);
                our_channel.sendObject(pn_obj,-1);
                free(pn);

                //END TEST VALENTIN
                //our_channel.receiveId(server_token,-1);
// TEST errorcodes
//           our_channel.sendId(this->checkPrerequisites());
// end TEST errorcodes

                // now we DO NOT call the MPI call, it was already called in
                // the constructor
                // int rvalue = callNativeMPI();
            }
        }/*MPO_Space::isDebugServerEnabled()*/

        /*Issue pre and post checks*/
        callMPIAndCheck ();

        return MPI_SUCCESS;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Init::checkPrerequisites()
    {
        // currently no checks implemented
        return 0;
    }

    // checkPrerequisites
    //========================================================================
    int MPO_Init::checkPost()
    {
        /**
         * we can`t initalize this earlier as we doesn`t know wheter source-locations
         * are present in the constructor.
         */
        initResourceGatherers();
        initDatatypes();

        return 0;
    }

    // getId
    //========================================================================
    int MPO_Init::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_INIT_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Init::callNativeMPI()
    {
        // we do not call MPI_Init here because it was already called in
        // the constructor
        // return PMPI_Init(my_argc,my_argv);
        checkPrerequisites();
        return my_rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Init::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Init";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_INIT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Init::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        if (C_INTERFACE == INTERFACE_MODE)
        {
            output
                << getCallname()
                << "(*argc, ***argv)"
                << std::endl;
        }

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output
                << getCallname()
                << "(ierror)"
                << std::endl;
        }

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Init::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Init::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        return in;
    }

    // restore
    //========================================================================
    void MPO_Init::doNotCallInit (void)
    {
        our_call_init = false;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/init.c.

/*@
  MPI_Init - Initialize the MPI execution environment

  Input Parameters:
  +  argc - Pointer to the number of arguments
  -  argv - Pointer to the argument vector

  Command line arguments:
  MPI specifies no command-line arguments but does allow an MPI
  implementation to make use of them.

  +  -mpiqueue - print out the state of the message queues when
  'MPI_FINALIZE' is called.  All processors print; the output may be
  hard to decipher.  This is intended as a debugging aid.

  .  -mpiversion - print out the version of the implementation
  (`not` of MPI), including the arguments that were used with configure.

  .  -mpinice nn - Increments the nice value by 'nn' (lowering the
  priority of the program by 'nn').  'nn' must be positive (except for
  root).  Not all systems support this argument; those that do not will
  ignore it.

  .  -mpedbg - Start a debugger in an xterm window if there is an error
  (either detected by MPI or a normally fatal signal).  This works only
  if MPICH was configured with '-mpedbg'.  CURRENTLY DISABLED.  If you
  have TotalView, -mpichtv or mpirun -tv will give you a better
  environment anyway.

  .  -mpimem - If MPICH was built with '-DMPIR_DEBUG_MEM', this checks
  all malloc and free operations (internal to MPICH) for signs of injury
  to the memory allocation areas.

  -  -mpidb options - Activate various debugging options.  Some require
  that MPICH have been built with special options.  These are intended
  for debugging MPICH, not for debugging user programs.  The available
  options include:
  .vb
  mem     - Enable dynamic memory tracing of internal MPI objects
  memall  - Generate output of all memory allocation/deallocation
  ptr     - Enable tracing of internal MPI pointer conversions
  rank n  - Limit subsequent -mpidb options to on the process with
  the specified rank in MPI_COMM_WORLD.  A rank of -1
  selects all of MPI_COMM_WORLD.
  ref     - Trace use of internal MPI objects
  reffile filename - Trace use of internal MPI objects with output
  to the indicated file
  trace   - Trace routine calls
  .ve

  Notes:
  Note that the Fortran binding for this routine has only the error
  return argument ('MPI_INIT(ierror)')

  Because the Fortran and C versions of 'MPI_Init' are different, there
  is a restriction on who can call 'MPI_Init'.  The version (Fortran or
  C) must match the main program.  That is, if the main program is in C,
  then the C version of 'MPI_Init' must be called.  If the main program
  is in Fortran, the Fortran version must be called.

  On exit from this routine, all processes will have a copy of the
  argument list.  This is `not required` by the MPI standard, and truely
  portable codes should not rely on it.  This is provided as a service
  by this  implementation (an MPI implementation is allowed to distribute
  the command line arguments but is not required to).

  Command line arguments are not provided to Fortran programs.  More
  precisely, non-standard Fortran routines such as getarg and iargc
  have undefined behavior in MPI and in this implementation.

  The MPI standard does not say what a program can do before an 'MPI_INIT'
  or after an 'MPI_FINALIZE'.  In the MPICH implementation, you should do
  as little as possible.  In particular, avoid anything that changes the
  external state of the program, such as opening files, reading standard
  input or writing to standard output.

  Signals used:
  The MPI standard requires that all signals used be documented.  The MPICH
  implementation itself uses no signals, but some of the software that
  MPICH relies on may use some signals.  The list below is partial and
  should be independantly checked if you (and any package that you use)
  depend on particular signals.

  IBM POE/MPL for SP2:
  SIGHUP, SIGINT, SIGQUIT, SIGFPE, SIGSEGV, SIGPIPE, SIGALRM, SIGTERM,
  SIGIO

  -mpedbg switch:
  SIGQUIT, SIGILL, SIGFPE, SIGBUS, SIGSEGV, SIGSYS

  Meiko CS2:
  SIGUSR2

  ch_p4 device:
  SIGUSR1

  The ch_p4 device also catches SIGINT, SIGFPE, SIGBUS, and SIGSEGV; this
  helps the p4 device (and MPICH) more gracefully abort a failed program.

  Intel Paragon (ch_nx and nx device):
  SIGUSR2

  Shared Memory (ch_shmem device):
  SIGCHLD

  Note that if you are using software that needs the same signals, you may
  find that there is no way to use that software with the MPI
  implementation.
  The signals that cause the most trouble for applications include
  'SIGIO', 'SIGALRM', and 'SIGPIPE'.  For example, using 'SIGIO' and
  'SIGPIPE' may prevent X11 routines from working.

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_INIT

  int MPI_Init(int *argc, char ***argv)
  @ */


