/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Intercomm_create.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Intercomm_create.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Intercomm_create::MPO_Intercomm_create()
        : my_local_comm(MPI_COMM_NULL),
          my_local_leader(MPI_PROC_NULL),
          my_peer_comm(MPI_COMM_NULL),
          my_remote_leader(MPI_PROC_NULL), 
          my_tag(MPI_ANY_TAG), 
          my_newintercomm(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Intercomm_create::MPO_Intercomm_create(MPO_Comm local_comm,
                                               int local_leader,
                                               MPO_Comm peer_comm,
                                               int remote_leader,
                                               int tag,
                                               MPI_Comm *newintercomm)
        : my_local_comm(local_comm),
          my_local_leader(local_leader),
          my_peer_comm(peer_comm),
          my_remote_leader(remote_leader), 
          my_tag(tag),
          my_newintercomm(newintercomm)
    {
    }


    // destructor
    //========================================================================
    MPO_Intercomm_create::~MPO_Intercomm_create()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Intercomm_create::checkPrerequisites2()
    {
        // check correctness of parameters
        checkComm(my_local_comm,"local_comm");
        checkCommIntracomm(my_local_comm,"local_comm");
        checkComm(my_peer_comm,"peer_comm");
        checkCommIntracomm(my_peer_comm,"peer_comm");
        checkLeaderLocal();
        checkLeaderRemote();
        checkTag(MSG_TAG);

        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Intercomm_create::checkComm(MPO_Comm my_comm,string arg_name)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << "Argument "
                << arg_name
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << "Argument "
                << arg_name 
                << " :Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


    // checkCommIntracomm
    //========================================================================
    int MPO_Intercomm_create::checkCommIntracomm(MPO_Comm my_comm,string arg_name)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_ERR_COMM_INTERCOMM))
        return rvalue;

        int inter = 0;

        PMPI_Comm_test_inter(my_comm, &inter);

        // Check if communicator is intercommunicator.
        if ((inter) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << "Argument "
                << arg_name
                << ": Communicator is inter-communicator,"
                << " but should be intra-communicator!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_INTERCOMM;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkCommIntracomm


    // checkTag
    //========================================================================
    int MPO_Intercomm_create::checkTag(const string KIND_OF_TAG)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if tag is < 0.
        if ((our_errcodes.at(MARMOT_ERR_TAG_NEG))&&(my_tag != MPI_ANY_TAG) && 
            (my_tag < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": "
                << KIND_OF_TAG
                << " = "
                << my_tag
                << " less than 0!" 
                << std::endl;
            rvalue = MARMOT_ERR_TAG_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if tag is within guaranteed range.
        else if ((my_tag != MPI_ANY_TAG) && 
                 (my_tag > 32767) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            int ub = 0;
            int flag = 1;
            PMPI_Attr_get(MPI_COMM_WORLD, MPI_TAG_UB, &ub, &flag);

            if ((our_errcodes.at(MARMOT_ERR_TAG_TOO_BIG))&&(my_tag > ub) && 
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname() 
                    << ": "
                    << KIND_OF_TAG
                    << " = "
                    << my_tag 
                    << " is larger than allowed!" 
                    << std::endl
                    << "THIS implementation allows tags up to "
                    << ub 
                    << "."
                    << std::endl;
                rvalue = MARMOT_ERR_TAG_TOO_BIG;
                
                our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
                our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }

            else if ((our_errcodes.at(MARMOT_WARN_TAG_TOO_BIG))&&(my_tag <= ub) && 
                     (MSG_ERR_WARN <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "WARNING: "
                    << getCallname()
                    << ": "
                    << KIND_OF_TAG
                    << " = "
                    << my_tag
                    << " greater than 32767 ! MPI only guarantees tags up to this." 
                    << std::endl
                    << "THIS implementation allows tags up to "
                    << ub 
                    << "."
                    << std::endl;
                rvalue = MARMOT_WARN_TAG_TOO_BIG;
                
                our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
                our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }                       // end else if (2)

        // (3) Check if Tag is MPI_ANY_TAG.
        else if ((our_errcodes.at(MARMOT_ERR_TAG_ANY_TAG))&&(my_tag == MPI_ANY_TAG) && 
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Tag is MPI_ANY_TAG which is not allowed in this call!"
                << std::endl;
            rvalue = MARMOT_ERR_TAG_ANY_TAG;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkTag



    // checkLeaderLocal
    //========================================================================
    int MPO_Intercomm_create::checkLeaderLocal()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        // local_leader in local_comm:
        int local_size = 0;
        PMPI_Comm_size(my_local_comm, &local_size);

        int local_rank = MPI_PROC_NULL;
        PMPI_Comm_rank(my_local_comm, &local_rank);
           
        // remote_leader in peer_comm 
        // (only significant at local_leader):
        int peer_size = -1;
        PMPI_Comm_size(my_peer_comm, &peer_size);

        // (1) Check local leader
        // (1.1) Check if local leader is MPI_ANY_SOURCE
        if ((our_errcodes.at(MARMOT_ERR_LEADER_ANY_SOURCE))&&(my_local_leader == MPI_ANY_SOURCE) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Local_leader is MPI_ANY_SOURCE! "
                << "Not allowed in this call."
                << std::endl;
            rvalue = MARMOT_ERR_LEADER_ANY_SOURCE;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (1.2) Check if local leader is negative
        else if ((our_errcodes.at(MARMOT_ERR_LEADER_NEG))&&(my_local_leader != MPI_ANY_SOURCE) &&
                 (my_local_leader < 0) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Local_leader less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_LEADER_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (1.3) Check if local leader is out of local communicator
        else if ((our_errcodes.at(MARMOT_ERR_LEADER_OUT_OF_COMM))&&(my_local_leader != MPI_ANY_SOURCE) &&
                 (my_local_leader >= local_size) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Local_leader out of local_comm!" 
                << std::endl;
            rvalue = MARMOT_ERR_LEADER_OUT_OF_COMM;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }
        // both leaders in peer_comm:
        // on nodes that are no member of local_comm: 
        // local_comm==MPI_COMM_NULL
        else if ((our_errcodes.at(MARMOT_ERR_LEADER_OUT_OF_COMM))&&(my_local_comm != MPI_COMM_NULL) &&
                 (local_rank == my_local_leader) &&
// ???
                 (my_peer_comm == MPI_COMM_NULL) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Local_leader does not belong to peer_comm!"
                << std::endl;
            rvalue = MARMOT_ERR_LEADER_OUT_OF_COMM;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }    

        return rvalue;
    }                             // end of checkLeaderLocal


    // checkLeaderRemote
    //========================================================================
    int MPO_Intercomm_create::checkLeaderRemote()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        // local_leader in local_comm:
        int local_size = 0;
        PMPI_Comm_size(my_local_comm, &local_size);

        int local_rank = MPI_PROC_NULL;
        PMPI_Comm_rank(my_local_comm, &local_rank);
           
        // remote_leader in peer_comm 
        // (only significant at local_leader):
        int peer_size = -1;
        PMPI_Comm_size(my_peer_comm, &peer_size);

        // (2) Check remote leader
        // (2.1) Check if remote leader is MPI_ANY_SOURCE
        if ((our_errcodes.at(MARMOT_ERR_LEADER_ANY_SOURCE))&&(my_remote_leader == MPI_ANY_SOURCE) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Remote_leader is MPI_ANY_SOURCE! "
                << "Not allowed in this call."
                << std::endl;
            rvalue = MARMOT_ERR_LEADER_ANY_SOURCE;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2.2) Check if remote leader is negative
        else if ((our_errcodes.at(MARMOT_ERR_LEADER_NEG))&&(my_remote_leader != MPI_ANY_SOURCE) &&
                 (my_remote_leader < 0) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Remote_leader less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_LEADER_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2.3) Check if remote leader is out of peer communicator
        else if ((our_errcodes.at(MARMOT_ERR_LEADER_OUT_OF_COMM))&&(my_remote_leader != MPI_ANY_SOURCE) &&
                 (my_remote_leader >= peer_size) && 
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Remote_leader is out of peer_comm!" 
                << std::endl;
            rvalue = MARMOT_ERR_LEADER_OUT_OF_COMM;
            
            our_comm_traces->findAndPrintTraceNamed(my_local_comm,stream,"local_comm");
            our_comm_traces->findAndPrintTraceNamed(my_peer_comm,stream,"peer_comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                             // end checkLeaderRemote


    // getId
    //========================================================================
    int MPO_Intercomm_create::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_INTERCOMM_CREATE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Intercomm_create::checkPost(void)
    {
                                       
        our_comms.insert(*my_newintercomm);
        
        our_comm_traces->addAndSetCreation(*my_newintercomm,getSourceLocation());
        our_comm_traces->findAndTryAddDependendency(*my_newintercomm,my_local_comm);
    
        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Intercomm_create::checkPrerequisites(void)
    {


        checkPrerequisites2();
        
        our_comm_traces->findAndAddUsage(my_local_comm,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_peer_comm,getSourceLocation());
        
        
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Intercomm_create::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Intercomm_create(my_local_comm, my_local_leader, 
                                       my_peer_comm, my_remote_leader, 
                                       my_tag, my_newintercomm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif                               

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Intercomm_create::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Intercomm_create";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_INTERCOMM_CREATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Intercomm_create::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(local_comm = "
            << my_local_comm
            << ", local_leader = "
            << my_local_leader
            << ", peer_comm = "
            << my_peer_comm
            << ", remote_leader = "
            << my_remote_leader
            << ", tag = " 
            << my_tag 
            << ", *newintercomm";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Intercomm_create::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_local_comm;
        out < my_local_leader;
        out < my_peer_comm;
        out < my_remote_leader;
        out < my_tag;
        // out < my_newintercomm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Intercomm_create::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_local_comm;
        in > my_local_leader;
        in > my_peer_comm;
        in > my_remote_leader;
        in > my_tag;
        // in > my_newintercomm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from 
// mpich-1.2.4/src/context/ic_create.c.

/*@

MPI_Intercomm_create - Creates an intercommuncator from two 
intracommunicators

Input Paramters:
+ local_comm - Local (intra)communicator
. local_leader - Rank in local_comm of leader (often 0)
. peer_comm - Remote communicator
. remote_leader - Rank in peer_comm of remote leader (often 0)
- tag - Message tag to use in constructing intercommunicator; if multiple
'MPI_Intercomm_creates' are being made, they should use different tags 
(more precisely, ensure that the local and remote leaders are using 
different tags for each 'MPI_intercomm_create').

Output Parameter:
. comm_out - Created intercommunicator

Notes:
The MPI 1.1 Standard contains two mutually exclusive comments on the
input intracommunicators.  One says that their repective groups must be
disjoint; the other that the leaders can be the same process.  After
some discussion by the MPI Forum, it has been decided that the groups must
be disjoint.  Note that the `reason` given for this in the standard is
`not` the reason for this choice; rather, the `other` operations on 
intercommunicators (like 'MPI_Intercomm_merge') do not make sense if the
groups are not disjoint.

.N fortran

Algorithm:
+ 1) Allocate a send context, an inter-coll context, and an intra-coll 
context
. 2) Send "send_context" and lrank_to_grank list from local comm group 
if I''m the local_leader.
. 3) If I''m the local leader, then wait on the posted sends and receives
to complete.  Post the receive for the remote group information and
wait for it to complete.
. 4) Broadcast information received from the remote leader.  
. 5) Create the inter_communicator from the information we now have.
-    An inter-communicator ends up with three levels of communicators. 
The inter-communicator returned to the user, a "collective" 
inter-communicator that can be used for safe communications between
local & remote groups, and a collective intra-communicator that can 
be used to allocate new contexts during the merge and dup operations.

For the resulting inter-communicator, 'comm_out'

.vb
comm_out                       = inter-communicator
comm_out->comm_coll            = "collective" inter-communicator
comm_out->comm_coll->comm_coll = safe collective intra-communicator
.ve

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_TAG
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_RANK

.seealso: MPI_Intercomm_merge, MPI_Comm_free, MPI_Comm_remote_group, 
MPI_Comm_remote_size

int MPI_Intercomm_create ( 
MPI_Comm local_comm, 
int local_leader, 
MPI_Comm peer_comm, 
int remote_leader, 
int tag, 
MPI_Comm *comm_out )
@ */

