/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Iprobe.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Iprobe.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Iprobe::MPO_Iprobe()
        : MPO_Probes(-1,
                     -1,
                     MPI_COMM_NULL),
          my_flag(0),
          my_status(0)
    {
    }


    /** Constructor taking proper arguments.
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     *  (it is a opaque type and thus  may contain some internal status etc.)
     */
    // proper constructor
    //========================================================================
    MPO_Iprobe::MPO_Iprobe(int source,
                           int tag,
                           MPO_Comm comm,
                           int *flag,
                           MPI_Status * status)
        : MPO_Probes(source,
                     tag,
                     comm),
          my_flag(flag),
          my_status(status)
    {
    }


    // destructor
    //========================================================================
    MPO_Iprobe::~MPO_Iprobe()
    {
    }


    // getId
    //========================================================================
    int MPO_Iprobe::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_IPROBE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Iprobe::checkPost(void)
    {
#ifdef MARMOT_USE_THREADSYNC
        if (*my_flag)
        {
            registerProbe (my_status->MPI_SOURCE, my_status->MPI_TAG, my_comm, getSourceLocation().getCodeline(),getSourceLocation().getFileName(), MPO_Space::getSyncObject()->getThreadRank(), getId(), getCallname());
        }
#endif

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Iprobe::checkPrerequisites(void)
    {


        MPO_Probes::checkPrerequisites();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());



        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Iprobe::callNativeMPI()
    {
        int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->enterPMPI();
#endif
        rvalue = PMPI_Iprobe(my_source, my_tag, my_comm, my_flag, my_status);

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->leavePMPI();
#endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Iprobe::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Iprobe";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_IPROBE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Iprobe::debugout(std::ostream & output) const
    {
        output
            << getCallname()
            << "(source = "
            << my_source
            << ", tag = "
            << my_tag
            << ", comm = "
            << my_comm
            << ", *flag, "
            << "*status";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Iprobe::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_source;
        out < my_tag;
        out < my_comm;
        // out < my_flag;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Iprobe::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_source;
        in > my_tag;
        in > my_comm;
        // in > my_flag;
        // in > my_status;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/iprobe.c.
/*@
  MPI_Iprobe - Nonblocking test for a message

  Input Parameters:
  + source - source rank, or  'MPI_ANY_SOURCE' (integer)
  . tag - tag value or  'MPI_ANY_TAG' (integer)
  - comm - communicator (handle)

  Output Parameter:
  + flag - (logical)
  - status - status object (Status)

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_TAG
  .N MPI_ERR_RANK


  int MPI_Iprobe(
  int source,
  int tag,
  MPI_Comm comm,
  int *flag,
  MPI_Status *status )
  @ */

