/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This sourcefile implements the MPO_LogCube object.
 * This object recieves MPO_MsgObject objects and creates a logfile from their data.
 * The logfile will be in a CUBE readable format.
 * @author Tobias Hilbrich
 * @date 13.11.2005
 * $Id: mpo_LogCube.cc 744 2007-12-04 10:03:23Z hpczink $ 
 */

#include "mpo_LogCube3.h"
#include <time.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef CUBE

#ifdef _WIN32
#include <direct.h>
#endif

namespace MPO 
{
    //------------------------------------------
    //            lokal enumerations
    //------------------------------------------
    /** \note {
     * For the implementation of the Cube logging we will use ID`s to identify
     * Nodes in a tree. In order to create unique ids we will use the following 
     * enumerations that won`t be needed any where else. }
     */
    
    /** This enumeration is used for identifing nodes of the performance metrics tree.
     * Most important is that those ID`s are not used by the MARMOT_ERR_CODES enum.
     */
    enum MPO_CUBE_PERFORMANCE_METRICS_ID
    {
        PERFORMANCE_METRICS_ID_ROOT = 0xFFFFFFFF,
        PERFORMANCE_METRICS_ID_INFOS = 0xFFFFFFFE,
        PERFORMANCE_METRICS_ID_WARNINGS = 0xFFFFFFFD,
        PERFORMANCE_METRICS_ID_ERRORS = 0xFFFFFFFC,
        PERFORMANCE_METRICS_ID_NOTES = 0xFFFFFFFB
    };
    
    /** This enumeration is used for identifing nodes of the call tree.
     * Most important is that those ID`s are not used by other nodes in the call tree.
     */
    enum MPO_CUBE_CALL_TREE_ID
    {
        CALL_TREE_ID_ROOT = 0xFFFFFFFF,
        CALL_TREE_ID_UNKNOWNLOCATION = 0xFFFFFFFE,
        CALL_TREE_ID_UNKNOWNLOCATION_NOTES = 0xFFFFFFFD,
        CALL_TREE_ID_UNKNOWNLOCATION_INFOS = 0xFFFFFFFC,
        CALL_TREE_ID_UNKNOWNLOCATION_ERRORS = 0xFFFFFFFB,
        CALL_TREE_ID_UNKNOWNLOCATION_WARNINGS = 0xFFFFFFFA
    };
    
    //------------------------------------------
    //            lokal defines
    //------------------------------------------
    /** These defines are used to define URLs and info texts.
     */
    //this define depends on MPO_CALL_ID and equals ceil(ld(MPO_LAST_CALL_ID))
#define CALLTREE_CALLID_SHIFT 9 
    //this define depends on MPO_CUBE_CALL_TREE_ID and equals 0x80000000>>(ceil(ld(NUM_ENTRIES_IN_MPO_CUBE_CALL_TREE_ID)))
#define CALLTREE_UNKNOWNLOCATION_CALLID_MASK 0x10000000
    
    //this define equals 31 - ceil(ld(NUM_ENTRIES_IN_MPO_MSG_TYPE))
#define PERFORMANCETREE_MSGTYPE_SHIFT 29
    
#define HTML_DIR "MARMOT_HTML"
    
#define MESSAGESROOT_INFO "This is the root node of the Messages Tree. To get a log of all the messages during logging use the referenced HTML file."
#define MESSAGESROOT_URL "MARMOT_HTML/PerformanceRoot.html"
#define MESSAGESROOT_NAME "Messages"
    
#define MESSAGESINFOS_INFO "This is the node for all the Infomessages. To get a log of all these messages use the referenced HTML file."
#define MESSAGESINFOS_URL "MARMOT_HTML/PerformanceInfos.html"
#define MESSAGESINFOS_NAME "Infos"
    
#define MESSAGESWARNINGS_INFO "This is the node for all the Warnings. To get a log of all these messages use the referenced HTML file."
#define MESSAGESWARNINGS_URL "MARMOT_HTML/PerformanceWarnings.html"
#define MESSAGESWARNINGS_NAME "Warnings"
    
#define MESSAGESNOTES_INFO "This is the node for all the Notes. To get a log of all these messages use the referenced HTML file."
#define MESSAGESNOTES_URL "MARMOT_HTML/PerformanceNotes.html"
#define MESSAGESNOTES_NAME "Notes"
    
#define MESSAGESERRORS_INFO "This is the node for all the Errors. To get a log of all these messages use the referenced HTML file."
#define MESSAGESERRORS_URL "MARMOT_HTML/PerformanceErrors.html"
#define MESSAGESERRORS_NAME "Errors"
    
#define MESSAGES_MPIERROR_DESCRIPTION "This is a MPI-Error. A reference to the MPI-Standard describing the problem can be found via the Online Description link."
    
#define CALLTREE_MODULENAME "MARMOT"
#define CALLTREE_ROOT "CallTree"
#define CALLTREE_ROOT_DESCRIPTION "This is the root of the Calltree."
#define CALLTREE_UNKNOWNLOCATION "Unknown Location"
#define CALLTREE_UNKNOWNLOCATION_DESCRIPTION "If MARMOT couldn`t get information about linenumber and sourcefile for a call it will be listed here."
#define CALLTREE_UNKNOWNLOCATION_NOTES "Notes"
#define CALLTREE_UNKNOWNLOCATION_NOTES_DESCRIPTION "The URL of this node contains all the notes that couldn`t be assigned to another location."
#define CALLTREE_UNKNOWNLOCATION_NOTES_URL "MARMOT_HTML/UnkownLocationNotes.html"
#define CALLTREE_UNKNOWNLOCATION_INFOS "Info Textmessages"
#define CALLTREE_UNKNOWNLOCATION_INFOS_DESCRIPTION "The URL of this node contains all the infos that couldn`t be assigned to another location."
#define CALLTREE_UNKNOWNLOCATION_INFOS_URL "MARMOT_HTML/UnkownLocationInfos.html"
#define CALLTREE_UNKNOWNLOCATION_ERRORS "Error Textmessages"
#define CALLTREE_UNKNOWNLOCATION_ERRORS_DESCRIPTION "The URL of this node contains all the errors that couldn`t be assigned to another location."
#define CALLTREE_UNKNOWNLOCATION_ERRORS_URL "MARMOT_HTML/UnkownLocationErrors.html"
#define CALLTREE_UNKNOWNLOCATION_WARNINGS "Warning Textmessages"
#define CALLTREE_UNKNOWNLOCATION_WARNINGS_DESCRIPTION "The URL of this node contains all the warnings that couldn`t be assigned to another location."
#define CALLTREE_UNKNOWNLOCATION_WARNINGS_URL "MARMOT_HTML/UnkownLocationWarnings.html"
#define CALLTREE_MPICALLNAME_DESCRIPTION "This is a MPI-Call. Further information about its Actions are available via the Online Reference link."
#define CALLTREE_SOURCEFILE_DESCRIPTION "This node contains all the MPI-Calls made in this Sourcefile."
    
#define SYSTEM_TREE_ROOT_NAME "MPI-Enviroment"
#define SYSTEM_TREE_NODE_NAME "MPI-Processes"
         
    /** Constructor. By Tobias Hilbrich, Date: 18.11.2005.
     * @param logFileName the name of the log file on disk.
     * @param MPIStandardFilename the name or identifing string for the MPI-Standard.
     */
    MPO_LogCube::MPO_LogCube (string logFileName)
    {
        init(logFileName);
    }
    
    /** Standard Destructor. By Tobias Hilbrich, Date: 18.11.2005.
     */
    MPO_LogCube::~MPO_LogCube ()
    {
        if (myCurrentDir)
            delete (myCurrentDir);
        myCurrentDir = NULL;

        if (m_bInitialized)
            closeLog ();
        
        myCharIdMap.deleteAssociatedObjects();
    }
    
    /** Constructor. By Tobias Hilbrich, Date: 18.11.2005.
      * @param logFileName the name of the log file on disk.
     * @param MPIStandardFilename the name or identifing string for the MPI-Standard.
     */
    bool MPO_LogCube::init (string logFileName)
    {
        //preinit all the Members
        mypMetricTree = NULL;
        mypMetricTreeHTML = NULL;
        mypRegionTree = NULL;
        mypCNodeTree = NULL;
        mypCallTreeHTML = NULL;
        
        myNextCharId = 1;
       
        //create a directory for all the HTML sites
        bool bCreatedDir = 0;
        
#ifdef HAVE_SYS_TYPES_H
#ifdef HAVE_SYS_STAT_H
        
        bCreatedDir = 1;
 #ifdef _WIN32
        int iResult = _mkdir(HTML_DIR);
 #else
        int iResult = mkdir (HTML_DIR,0xFFFFFF);
 #endif       
        if (iResult != 0)
        {
//            if (errno != EEXIST)
//                bCreatedDir = 0;
        }
        
#endif
#endif

        if (!bCreatedDir)
        {
            std::cerr   << "MARMOT is unable to create the directory \""
                        << HTML_DIR
                        << "\" please create it yourself if you want to get info HTML files in your Cube logfiles!";
        }
        
        //do the real init stuff
        initMPIStandardReferencer ();
        
        //Store the output file name
        myLogFileName = logFileName;

        //Load the current Directory name
#ifdef HAVE_UNISTD_H        
        //we have unistd.h so get the current directory
        char *ret = NULL;
        myCurrentDirSize = 0;
        myCurrentDir = NULL;
        while (!ret)
        {
            if (myCurrentDir)
                delete (myCurrentDir);
            myCurrentDir = NULL;
            
            myCurrentDirSize += 256;
            
            myCurrentDir = new char[myCurrentDirSize]; //Load it large sized as directory names are often extremly long         
            ret = getcwd (myCurrentDir,myCurrentDirSize);
        }
#endif
#ifndef HAVE_UNISTD_H        
        //sh** we don`t have no unistd.h
        //so we can`t set links to bad 
        /** \todo at Tobias from Tobias: think about this case ! Maybe add a new environmental variable one can use ? */
        
        myCurrentDirSize = 32;
        myCurrentDir = new char[32];
        sprintf (myCurrentDir,"./");
#endif
        
        m_bInitialized = true;
                
        //define informative attributes
        if (!initAttributes())
            m_bInitialized = false;
        
        //create the Performance Metrics
        if (!initPerformanceMetrics ())
            m_bInitialized = false;
        
        //create the Call Tree
        if (!initCallTree ())
            m_bInitialized = false;
        
        //create the System Tree
        if (!initSystemTree ())
            m_bInitialized = false;
        
        //initialize the severity mapping
        if (!initSeverities ())
            m_bInitialized = false;
        
        //initialize the severity mapping
        if (!initMirrors ())
            m_bInitialized = false;
        
        // init socket-logger
        initSocket();

        return m_bInitialized;
    }
    
    
    /** Gets a message object, looks for MPI-Standard reference text and calls logMsg. 
     * By Tobias Hilbrich, Date: 18.11.2005.
     * @param pMsg the Messageobject that will be recieved.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::getMsg (MPO_MsgObject &pMsg)
    {
        //first check whether the msg introduces a new thread
        checkForThreadInMetric(pMsg);

        MARMOT_ERR_CODES err = pMsg.getErrorCode();
        const char *temp = getLocalPathToMPIStandard(err);
        // if there is a socket send msg
        pMsg.writeToSocket(getLogSocket(), temp);
        // wait for ACK from log-host
        if (pMsg.getMsgType() == MSG_TYPE_ERROR)
          CSocketLogger::instance()->waitForAck();

        //we will handle the different message types differently here
        switch (pMsg.getMsgType())
        {
            case MSG_TYPE_INFO:
                return logInfoMsg(pMsg);
                break;
            case MSG_TYPE_WARNING:
                return logWarningMsg(pMsg);
                break;
            case MSG_TYPE_ERROR:
                return logErrorMsg(pMsg);
                break;
            case MSG_TYPE_NOTE:
                return logNoteMsg(pMsg);
                break;
            default:
                //some unknown message was recieved lets print it to the log
                MPO_MsgObject mError (MSG_TYPE_ERROR,pMsg.getTimestamp(),pMsg.getRank(),"ERROR: Unhandled MessageType with id: "+pMsg.getMsgType(),MARMOT_ERR_MARMOT);
                return getMsg(mError);                
                break;
        }

        return false;
    }
    
    /** Closes the Log. By Tobias Hilbrich, Date: 18.11.2005.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::closeLog (void)
    {
		/*make sure we also make our output available to connected clients*/
		CSocketLogger::instance()->waitForAck();
        
        //clear all the own data
        if (mypMetricTree)
            delete (mypMetricTree);
        mypMetricTree = NULL;
        
        mypMetricTreeHTML->deleteAssociatedObjects();
        if (mypMetricTreeHTML)
            delete (mypMetricTreeHTML);
        mypMetricTreeHTML = NULL;
        
        if (mypRegionTree)
            delete (mypRegionTree);
        mypRegionTree = NULL;
        
        if (mypCNodeTree)
            delete (mypCNodeTree);
        mypCNodeTree = NULL;
        
        mypCallTreeHTML->deleteAssociatedObjects();
        if (mypCallTreeHTML)
            delete (mypCallTreeHTML);
        mypCallTreeHTML = NULL;
        
        myProcess.clear();
        myThread.clear();
        
        //write the cube
        std::ofstream out;
        out.open(myLogFileName.c_str(),std::ios_base::out | std::ios_base::trunc);
        out << myCube;
        
        //set logfile as inactive
        m_bInitialized = false;

        return (false);
    }
    
    /** Sets informative attributes like date for the logfile. By Tobias Hilbrich, Date: 18.11.2005.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::initAttributes(void)
    {
        //set a description
        myCube.def_attr("@desc", "MARMOT Cube-Tracefile");
        
        //read the system time and set it as Attribute
        char buf[128];
      
        struct tm *ptr;
        time_t tm;
        tm = time(NULL);
        ptr = localtime(&tm);
        strftime(buf ,128 , "%c.\n",ptr);
        myCube.def_attr("@date",buf);
        
        return true;
    }
    
    /** Initializes the Metrics for all the messages. By Tobias Hilbrich, Date: 18.11.2005.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::initPerformanceMetrics (void)
    {
         Metric    *pMetRoot,
            *pMetInfos, 
            *pMetWarnings, 
            *pMetNotes, 
            *pMetErrors;
                        
        MPO_LogHTML     *pHTMLLog;
        
        char url[256];
        
        //--------------------create the root--------------------
        pHTMLLog = new MPO_LogHTML (MESSAGESROOT_URL,false);
        mypMetricTreeHTML = new MPO_TemplateTree<MPO_LogHTML*> (pHTMLLog,PERFORMANCE_METRICS_ID_ROOT);
                
        sprintf (url,"file:./%s",MESSAGESROOT_URL);
        pMetRoot = myCube.def_met(MESSAGESROOT_NAME,MESSAGESROOT_NAME,"INTEGER", "occ",
                                  "ZERO", 
                                  url, 
                                  MESSAGESROOT_INFO, NULL);
        
        mypMetricTree = new MPO_TemplateTree<Metric*> (pMetRoot,PERFORMANCE_METRICS_ID_ROOT);
        
        //------------create the root->Infos node----------------
        pHTMLLog = new MPO_LogHTML (MESSAGESINFOS_URL,false);
        mypMetricTreeHTML->addChild (pHTMLLog,PERFORMANCE_METRICS_ID_INFOS);
         
        sprintf (url,"file:./%s",MESSAGESINFOS_URL);
        pMetInfos = myCube.def_met( MESSAGESINFOS_NAME,MESSAGESINFOS_NAME,"INTEGER", "occ", 
                                    "ZERO", 
                                    url, 
                                    MESSAGESINFOS_INFO, pMetRoot);
        
        mypMetricTree->addChild (pMetInfos,PERFORMANCE_METRICS_ID_INFOS);
        
        //------------create the root->Warnings node----------------
        pHTMLLog = new MPO_LogHTML (MESSAGESWARNINGS_URL,false);
        mypMetricTreeHTML->addChild (pHTMLLog,PERFORMANCE_METRICS_ID_WARNINGS);
         
        sprintf (url,"file:./%s",MESSAGESWARNINGS_URL);
        pMetWarnings = myCube.def_met(  MESSAGESWARNINGS_NAME, MESSAGESWARNINGS_NAME, "INTEGER","occ", 
                                        "ZERO", 
                                        url, 
                                        MESSAGESWARNINGS_INFO, pMetRoot);
        
        mypMetricTree->addChild (pMetWarnings,PERFORMANCE_METRICS_ID_WARNINGS);
        
        //------------create the root->Notes node----------------
        pHTMLLog = new MPO_LogHTML (MESSAGESNOTES_URL,false);
        mypMetricTreeHTML->addChild (pHTMLLog,PERFORMANCE_METRICS_ID_NOTES);
         
        sprintf (url,"file:./%s",MESSAGESNOTES_URL);
        pMetNotes = myCube.def_met( MESSAGESNOTES_NAME,MESSAGESNOTES_NAME,"INTEGER", "occ", 
                                    "ZERO", 
                                    url, 
                                    MESSAGESNOTES_INFO, pMetRoot);
        
        mypMetricTree->addChild (pMetNotes,PERFORMANCE_METRICS_ID_NOTES);
        
        //------------create the root->Notes node----------------
        pHTMLLog = new MPO_LogHTML (MESSAGESERRORS_URL,false);
        mypMetricTreeHTML->addChild (pHTMLLog,PERFORMANCE_METRICS_ID_ERRORS);
         
        sprintf (url,"file:./%s",MESSAGESERRORS_URL);
        pMetErrors = myCube.def_met(MESSAGESERRORS_NAME,MESSAGESERRORS_NAME,"INTEGER", "occ", 
                                    "ZERO",  
                                    url, 
                                    MESSAGESERRORS_INFO, pMetRoot);
        
        mypMetricTree->addChild (pMetErrors,PERFORMANCE_METRICS_ID_ERRORS);
                
        return true;
    }
    
    /** Initalizes the call locations tree. By Tobias Hilbrich, Date: 18.11.2005.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::initCallTree (void)
    {
         Region    *rgnUknLocation,
            *rgnUknLocationNotes,
            *rgnUknLocationInfos,
            *rgnUknLocationWarnings,
            *rgnUknLocationErrors,
            *rgnRootRegion;
                        
         Cnode     *cnodeUknLocation,
            *cnodeUknLocationNotes,
            *cnodeUknLocationInfos,
            *cnodeUknLocationWarnings,
            *cnodeUknLocationErrors,
            *cnodeRootNode;
        
        char url[256];
        
        // define the one and only module
        //myModule = myCube.def_module(CALLTREE_MODULENAME);
        
        //create the root of the region tree (has no associated Cube-object)
        rgnRootRegion = myCube.def_region(CALLTREE_ROOT, 0, 0, "",CALLTREE_ROOT_DESCRIPTION, CALLTREE_MODULENAME);
        mypRegionTree = new MPO_TemplateTree<const Region*> (rgnRootRegion,CALL_TREE_ID_ROOT);
        
        //create the unkown location region
        rgnUknLocation  = myCube.def_region(CALLTREE_UNKNOWNLOCATION, 0, 0, "",CALLTREE_UNKNOWNLOCATION_DESCRIPTION, CALLTREE_MODULENAME);
        mypRegionTree->addChild(rgnUknLocation,CALL_TREE_ID_UNKNOWNLOCATION);
        
        //create the unkown location->Notes region
        sprintf (url,"file:./%s",CALLTREE_UNKNOWNLOCATION_NOTES_URL);
        rgnUknLocationNotes  = myCube.def_region(CALLTREE_UNKNOWNLOCATION_NOTES, 0, 0,url,CALLTREE_UNKNOWNLOCATION_NOTES_DESCRIPTION, CALLTREE_MODULENAME);
        mypRegionTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(rgnUknLocationNotes,CALL_TREE_ID_UNKNOWNLOCATION_NOTES);
        
        //create the unkown location->Infos region
        sprintf (url,"file:./%s",CALLTREE_UNKNOWNLOCATION_INFOS_URL);
        rgnUknLocationInfos  = myCube.def_region(CALLTREE_UNKNOWNLOCATION_INFOS, 0, 0,url,CALLTREE_UNKNOWNLOCATION_INFOS_DESCRIPTION, CALLTREE_MODULENAME);
        mypRegionTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(rgnUknLocationInfos,CALL_TREE_ID_UNKNOWNLOCATION_INFOS);
        
        //create the unkown location->Warnings region
        sprintf (url,"file:./%s",CALLTREE_UNKNOWNLOCATION_WARNINGS_URL);
        rgnUknLocationWarnings  = myCube.def_region(CALLTREE_UNKNOWNLOCATION_WARNINGS, 0, 0,url,CALLTREE_UNKNOWNLOCATION_WARNINGS_DESCRIPTION, CALLTREE_MODULENAME);
        mypRegionTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(rgnUknLocationWarnings,CALL_TREE_ID_UNKNOWNLOCATION_WARNINGS);
        
        //create the unkown location->Errors region
        sprintf (url,"file:./%s",CALLTREE_UNKNOWNLOCATION_ERRORS_URL);
        rgnUknLocationErrors  = myCube.def_region(CALLTREE_UNKNOWNLOCATION_ERRORS, 0, 0,url,CALLTREE_UNKNOWNLOCATION_ERRORS_DESCRIPTION, CALLTREE_MODULENAME);
        mypRegionTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(rgnUknLocationErrors,CALL_TREE_ID_UNKNOWNLOCATION_ERRORS);
        
        //create the root of the callnode tree (has no associated Cube-object)
        cnodeRootNode = myCube.def_cnode(rgnRootRegion, NULL);
        mypCNodeTree = new MPO_TemplateTree<Cnode*> (cnodeRootNode,CALL_TREE_ID_ROOT);
                
        cnodeUknLocation = myCube.def_cnode(rgnUknLocation, cnodeRootNode);
        mypCNodeTree->addChild(cnodeUknLocation,CALL_TREE_ID_UNKNOWNLOCATION);
        
        //create the Cnode for the UnknownLocation->Notes region
        cnodeUknLocationNotes = myCube.def_cnode(rgnUknLocationNotes, cnodeUknLocation);
        mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(cnodeUknLocationNotes,CALL_TREE_ID_UNKNOWNLOCATION_NOTES);
        
        //create the Cnode for the UnknownLocation->Infos region
        cnodeUknLocationInfos = myCube.def_cnode(rgnUknLocationInfos, cnodeUknLocation);
        mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(cnodeUknLocationInfos,CALL_TREE_ID_UNKNOWNLOCATION_INFOS);
        
        //create the Cnode for the UnknownLocation->Warnings region
        cnodeUknLocationWarnings = myCube.def_cnode(rgnUknLocationWarnings, cnodeUknLocation);
        mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(cnodeUknLocationWarnings,CALL_TREE_ID_UNKNOWNLOCATION_WARNINGS);
        
        //create the Cnode for the UnknownLocation->Errors region
        cnodeUknLocationErrors = myCube.def_cnode(rgnUknLocationErrors, cnodeUknLocation);
        mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(cnodeUknLocationErrors,CALL_TREE_ID_UNKNOWNLOCATION_ERRORS);
        
        //build the HTML tree
        mypCallTreeHTML = new MPO_TemplateTree<MPO_LogHTML*> (NULL,CALL_TREE_ID_ROOT);
        mypCallTreeHTML->addChild (NULL,CALL_TREE_ID_UNKNOWNLOCATION);
        
        //build the HTMLLog for the UnknownLocation->Notes location
        MPO_LogHTML *pHTMLLogUKL_Notes = new MPO_LogHTML (CALLTREE_UNKNOWNLOCATION_NOTES_URL,false);
        mypCallTreeHTML->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild (pHTMLLogUKL_Notes,CALL_TREE_ID_UNKNOWNLOCATION_NOTES);
        
        //build the HTMLLog for the UnknownLocation->Infos location
        MPO_LogHTML *pHTMLLogUKL_Infos = new MPO_LogHTML (CALLTREE_UNKNOWNLOCATION_INFOS_URL,false);
        mypCallTreeHTML->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild (pHTMLLogUKL_Infos,CALL_TREE_ID_UNKNOWNLOCATION_INFOS);
        
        //build the HTMLLog for the UnknownLocation->Warnings location
        MPO_LogHTML *pHTMLLogUKL_Warnings = new MPO_LogHTML (CALLTREE_UNKNOWNLOCATION_WARNINGS_URL,false);
        mypCallTreeHTML->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild (pHTMLLogUKL_Warnings,CALL_TREE_ID_UNKNOWNLOCATION_WARNINGS);
        
        //build the HTMLLog for the UnknownLocation->Errors location
        MPO_LogHTML *pHTMLLogUKL_Errors = new MPO_LogHTML (CALLTREE_UNKNOWNLOCATION_ERRORS_URL,false);
        mypCallTreeHTML->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild (pHTMLLogUKL_Errors,CALL_TREE_ID_UNKNOWNLOCATION_ERRORS);
        
        return true;
    }
    
    /** Initializes the system tree. By Tobias Hilbrich, Date: 18.11.2005.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::initSystemTree (void)
    {
        //define the root nodes
        myMachine = myCube.def_mach(SYSTEM_TREE_ROOT_NAME,"dEsCr");	//debug
        myNode  = myCube.def_node(SYSTEM_TREE_NODE_NAME, myMachine);
        
        //get the number of processes
        int comm_size = 0;
        PMPI_Comm_size(MPI_COMM_WORLD, &comm_size);
        comm_size--; //push away the debugserver process
        
        myThread.resize(comm_size);
        
        char buf[64];
        for (int i = 0; i < comm_size;i++)
        {
            sprintf(buf,"rank %d",i);
            myProcess.push_back(
                myCube.def_proc(buf, i, myNode)  );
            
            myThread[i].insert(
                std::pair<int,Thread*>(
                    0,
                    myCube.def_thrd("Thread 0", 0, myProcess[i])
                    )
                );
        }
        
        return true;
    }
    
    /** Initializes Severities. By Tobias Hilbrich, Date: 18.11.2005.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::initSeverities (void)
    {
        //we must initialize some severities her otherwise cube will crash -> :(
        
        //The only CNode we have
        Cnode *LocUnk = mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->getObject();
        
        //the Performance Root
        Metric *Root = mypMetricTree->searchId(PERFORMANCE_METRICS_ID_ROOT)->getObject();
                
        for (unsigned int i = 0; i < myThread.size();i++)
        {
            myCube.set_sev(Root,LocUnk, myThread[i][0], 0);
        }
        
        //Performance->Infos
        Metric *Infos = mypMetricTree->searchId(PERFORMANCE_METRICS_ID_INFOS)->getObject();
                
        for (unsigned int i = 0; i < myThread.size();i++)
        {
            myCube.set_sev(Infos,LocUnk, myThread[i][0], 0);
        }
        
        //Performance->Warnings
        Metric *Warnings = mypMetricTree->searchId(PERFORMANCE_METRICS_ID_WARNINGS)->getObject();
                
        for (unsigned int i = 0; i < myThread.size();i++)
        {
            myCube.set_sev(Warnings,LocUnk, myThread[i][0], 0);
        }
        
        //Performance->Notes
        Metric *Notes = mypMetricTree->searchId(PERFORMANCE_METRICS_ID_NOTES)->getObject();
                
        for (unsigned int i = 0; i < myThread.size();i++)
        {
            myCube.set_sev(Notes,LocUnk, myThread[i][0], 0);
        }
        
        //Performance->Errors
        Metric *Errors = mypMetricTree->searchId(PERFORMANCE_METRICS_ID_ERRORS)->getObject();
                
        for (unsigned int i = 0; i < myThread.size();i++)
        {
            myCube.set_sev(Errors,LocUnk, myThread[i][0], 0);
        }
        
        return true;
    }
    
    /** Initializes the mirrors. By Tobias Hilbrich, Date: 18.11.2005.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::initMirrors (void)
    {
        /** \note unfortuatly using multiple mirrors for different
         * documents is not what cube was thougt for. 
         * So I will use the mirroring capability for the MPI-
         * Standards. Thus we won`t have the abbility to specify
         * Mirrors for our HTML detail logs.
         */
        
        //mirror for all the local stuff
        const char *buf;
        /*buf = new char[myCurrentDirSize+32];
          sprintf(buf,"file://%s/",myCurrentDir);
          //sprintf(buf,".",myCurrentDir);
          myCube.def_mirror(buf);
          delete(buf);
        */
        
        //mirror for MPI refences
        buf = getLocalDocPath();
        //buf = new char[iRet];
        //getLocalDocPath(buf,iRet);
	int iRet=strlen(buf);
        char *temp = new char[iRet+32];
        sprintf (temp,"file://%s/",buf);
        
        myCube.def_mirror(temp);
        
        //if (buf)
        //    delete(buf);        
        //buf = NULL;
        
        if (temp)
            delete(temp);        
        temp = NULL;
       
        return true;
    }

    /** Logs a message of the MSG_TYPE_INFO type. By Tobias Hilbrich, Date: 25.11.2005.
     * @param pMsg reference to a MPO_MsgObject.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::logInfoMsg(MPO_MsgObject &pMsg)
    {
        //data for the severity logging later on
        Metric *mSev = mypMetricTree->searchId(PERFORMANCE_METRICS_ID_INFOS)->getObject();
        Cnode  *cnSev;
               
        //selct location handling
        if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        {
            //no location can be found -> add it to CALL_TREE_ID_UNKNOWNLOCATION_NOTES
            logMsgToExsitingCallTreeNode (pMsg,CALL_TREE_ID_UNKNOWNLOCATION_INFOS,CALL_TREE_ID_UNKNOWNLOCATION);
            
            //store the Cnode for severity mapping
            cnSev = mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION_INFOS)->getObject();
        }
        else
        {
            //add the location to the Call Tree (if Necessary)
            const MPO_SourceLocation& SourceLocation = pMsg.getSourceLocation();
            unsigned long id = calculateCallTreeId(pMsg.getCallId(),SourceLocation);
                        
            //folowing method creates a new call tree location when necessary and logs the message to its call tree location
            cnSev = logMessageToCallTree (id,pMsg,SourceLocation);
        }//if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        
        
        //Add the severities when possible
        if (mSev && cnSev && (pMsg.getRank() < (int)myThread.size()))
        {
            if (pMsg.getRank() >= 0)
            {
                //normal rank
                myCube.add_sev(mSev,cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
                myCube.add_sev(mypMetricTree->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
            }
            else
            {
                //debugserver rank lets log it to all ranks
                for (unsigned int i = 0; i < myThread.size();i++)
                {
                    for (std::map<int,Thread*>::iterator j = myThread[i].begin(); j != myThread[i].end();j++)
                    {
                        myCube.add_sev(mSev,cnSev, j->second, 1);
                        myCube.add_sev(mypMetricTree->getObject(),cnSev, j->second, 1);
                    }
                }
            }
        }
        
        //Add the logentry to the Messages->Warings and Messages HTMLLog
        MPO_LogHTML *pHTMLLog = mypMetricTreeHTML->searchId(PERFORMANCE_METRICS_ID_INFOS)->getObject();
        if (pHTMLLog)
        {
            pHTMLLog->getMsg(pMsg);
        }
        
        pHTMLLog = mypMetricTreeHTML->getObject();
        if (pHTMLLog)
        {
            pHTMLLog->getMsg(pMsg);
        }
        
        //flush the logfile if need is
        if (LOGFLUSH_TYPE == LOGFILE_FLUSH_TYPE_IMMEDIATE)
            flushOfstream ();
        
        return true;
    }
            
    /** Logs a message of the MSG_TYPE_WARNING type. By Tobias Hilbrich, Date: 25.11.2005.
     * @param pMsg reference to a MPO_MsgObject.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::logWarningMsg(MPO_MsgObject &pMsg)
    {
           
        //find the Performance(Messages) Metrics Node
        unsigned long metricsId = calculatePerformanceTreeId (pMsg.getErrorCode(),MSG_TYPE_WARNING);
        MPO_TemplateTree<Metric*> *mNode = mypMetricTree->searchId(metricsId);
        
        if (!mNode)
            createPerformanceMetric (pMsg.getErrorCode(),metricsId,PERFORMANCE_METRICS_ID_WARNINGS);
        
        //data for the severity logging later on
        Cnode  *cnSev;
               
        //select location handling
        if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        {
            //no location can be found -> add it to CALL_TREE_ID_UNKNOWNLOCATION_WARNINGS
            logMsgToExsitingCallTreeNode (pMsg,CALL_TREE_ID_UNKNOWNLOCATION_WARNINGS,CALL_TREE_ID_UNKNOWNLOCATION);
                        
            //store the Cnode for severity mapping
            cnSev = mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION_WARNINGS)->getObject();
        }
        else
        {
            //add the location to the Call Tree (if Necessary)
            const MPO_SourceLocation& SourceLocation = pMsg.getSourceLocation();
            unsigned long id = calculateCallTreeId(pMsg.getCallId(),SourceLocation);
                        
            //folowing method creates a new call tree location when necessary and logs the message to its call tree location
            cnSev = logMessageToCallTree (id,pMsg,SourceLocation);
        }//if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        
        
        //Add the severities when possible
        if (cnSev && (pMsg.getRank() < (int)myThread.size()))
        {
            if (pMsg.getRank() >= 0)
            {
                //normal rank
                myCube.add_sev(mypMetricTree->searchId(metricsId)->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
                myCube.add_sev(mypMetricTree->searchId(PERFORMANCE_METRICS_ID_WARNINGS)->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
                myCube.add_sev(mypMetricTree->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
            }
            else
            {
                //debugserver rank lets log it to all ranks
                for (unsigned int i = 0; i < myThread.size();i++)
                {
                    for (std::map<int,Thread*>::iterator j = myThread[i].begin(); j != myThread[i].end();j++)
                    {
                        myCube.add_sev(mypMetricTree->searchId(metricsId)->getObject(),cnSev, j->second, 1);
                        myCube.add_sev(mypMetricTree->searchId(PERFORMANCE_METRICS_ID_WARNINGS)->getObject(),cnSev, j->second, 1);
                        myCube.add_sev(mypMetricTree->getObject(),cnSev, j->second, 1);
                    }
                }
            }
        }
        
        //Add the logentry to the Messages->Warnings and Messages HTMLLog
        MPO_LogHTML *pHTMLLog = mypMetricTreeHTML->searchId(PERFORMANCE_METRICS_ID_WARNINGS)->getObject();
        if (pHTMLLog)
        {
            pHTMLLog->getMsg(pMsg);
        }
        
        pHTMLLog = mypMetricTreeHTML->getObject();
        if (pHTMLLog)
        {
            pHTMLLog->getMsg(pMsg);
        }
        
        //flush the logfile if need is
        if (LOGFLUSH_TYPE == LOGFILE_FLUSH_TYPE_IMMEDIATE)
            flushOfstream ();
                
        return true;
    }
            
    /** Logs a message of the MSG_TYPE_ERROR type. By Tobias Hilbrich, Date: 25.11.2005.
     * @param pMsg reference to a MPO_MsgObject.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::logErrorMsg(MPO_MsgObject &pMsg)
    {
        //find the Performance(Messages) Metrics Node
        unsigned long metricsId = calculatePerformanceTreeId (pMsg.getErrorCode(),MSG_TYPE_ERROR);
        MPO_TemplateTree<Metric*> *mNode = mypMetricTree->searchId(metricsId);
        
        if (!mNode)
            createPerformanceMetric (pMsg.getErrorCode(),metricsId,PERFORMANCE_METRICS_ID_ERRORS);
                
        //data for the severity logging later on
        Cnode  *cnSev;
               
        //selct location handling
        if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        {
            //no location can be found -> add it to CALL_TREE_ID_UNKNOWNLOCATION_ERRORS
            logMsgToExsitingCallTreeNode (pMsg,CALL_TREE_ID_UNKNOWNLOCATION_ERRORS,CALL_TREE_ID_UNKNOWNLOCATION);
                        
            //store the Cnode for severity mapping
            cnSev = mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION_ERRORS)->getObject();
        }
        else
        {
            //add the location to the Call Tree (if Necessary)
            const MPO_SourceLocation& SourceLocation = pMsg.getSourceLocation();
            unsigned long id = calculateCallTreeId(pMsg.getCallId(),SourceLocation);
                        
            //folowing method creates a new call tree location when necessary and logs the message to its call tree location
            cnSev = logMessageToCallTree (id,pMsg,SourceLocation);
        }//if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        
        //Add the severities when possible
        if (cnSev && (pMsg.getRank() < (int)myThread.size()))
        {
            if (pMsg.getRank() >= 0)
            {
                //normal rank
                myCube.add_sev(mypMetricTree->searchId(metricsId)->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
                myCube.add_sev(mypMetricTree->searchId(PERFORMANCE_METRICS_ID_ERRORS)->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
                myCube.add_sev(mypMetricTree->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
            }
            else
            {
                //debugserver rank lets log it to all ranks
                for (unsigned int i = 0; i < myThread.size();i++)
                {
                    for (std::map<int,Thread*>::iterator j = myThread[i].begin(); j != myThread[i].end();j++)
                    {
                        myCube.add_sev(mypMetricTree->searchId(metricsId)->getObject(),cnSev, j->second, 1);
                        myCube.add_sev(mypMetricTree->searchId(PERFORMANCE_METRICS_ID_ERRORS)->getObject(),cnSev, j->second, 1);
                        myCube.add_sev(mypMetricTree->getObject(),cnSev, j->second, 1);
                    }
                }
            }
        }
        
        //Add the logentry to the Messages->Errors and Messages HTMLLog
        MPO_LogHTML *pHTMLLog = mypMetricTreeHTML->searchId(PERFORMANCE_METRICS_ID_ERRORS)->getObject();
        if (pHTMLLog)
        {
            pHTMLLog->getMsg(pMsg);
        }
        
        pHTMLLog = mypMetricTreeHTML->getObject();
        if (pHTMLLog)
        {
            pHTMLLog->getMsg(pMsg);
        }
                
        //flush the logfile
        flushOfstream ();
        
        return true;
    }
            
    /** Logs a message of the MSG_TYPE_NOTE type. By Tobias Hilbrich, Date: 25.11.2005.
     * @param pMsg reference to a MPO_MsgObject.
     * @return true when the operation succeds false otherwise.
     */
    bool MPO_LogCube::logNoteMsg(MPO_MsgObject &pMsg)
    {
        
        //data for the severity logging later on
        Metric *mSev = mypMetricTree->searchId(PERFORMANCE_METRICS_ID_NOTES)->getObject();
        Cnode  *cnSev;
               
        //selct location handling
        if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        {
            //no location can be found -> add it to CALL_TREE_ID_UNKNOWNLOCATION_NOTES
            logMsgToExsitingCallTreeNode (pMsg,CALL_TREE_ID_UNKNOWNLOCATION_NOTES,CALL_TREE_ID_UNKNOWNLOCATION);
            
            //store the Cnode for severity mapping
            cnSev = mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION_NOTES)->getObject();
        }
        else
        {
            //add the location to the Call Tree (if Necessary)
            const MPO_SourceLocation& SourceLocation = pMsg.getSourceLocation();
            unsigned long id = calculateCallTreeId(pMsg.getCallId(),SourceLocation);
                        
            //folowing method creates a new call tree location when necessary and logs the message to its call tree location
            cnSev = logMessageToCallTree (id,pMsg,SourceLocation);
        }//if (pMsg.getOutputType() == OUTPUT_TYPE_PURE_TEXT)
        
        
        //Add the severities when possible
        if (mSev && cnSev && (pMsg.getRank() < (int)myThread.size()))
        {
            if (pMsg.getRank() >= 0)
            {
                //normal rank
                myCube.add_sev(mSev,cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
                myCube.add_sev(mypMetricTree->getObject(),cnSev, myThread[pMsg.getRank()][pMsg.getThreadRank()], 1);
            }
            else
            {
                //debugserver rank lets log it to all ranks
                for (unsigned int i = 0; i < myThread.size();i++)
                {
                    for (std::map<int,Thread*>::iterator j = myThread[i].begin(); j != myThread[i].end();j++)
                    {
                        myCube.add_sev(mSev,cnSev, j->second, 1);
                        myCube.add_sev(mypMetricTree->getObject(),cnSev, j->second, 1);
                    }
                }
            }
        }
        
        //Add the logentry to the Messages->Notes and Messages HTMLLog
        //only add if a real message was added
        if (pMsg.getText().length() > 0)
        {
            MPO_LogHTML *pHTMLLog = mypMetricTreeHTML->searchId(PERFORMANCE_METRICS_ID_NOTES)->getObject();
            if (pHTMLLog)
            {
                pHTMLLog->getMsg(pMsg);
            }
            
            pHTMLLog = mypMetricTreeHTML->getObject();
            if (pHTMLLog)
            {
                pHTMLLog->getMsg(pMsg);
            }
        }
        
        //flush the logfile if need is
        if (LOGFLUSH_TYPE == LOGFILE_FLUSH_TYPE_IMMEDIATE)
            flushOfstream ();
                
        return true;
    }
    
    /** Calculates a unique CallTree Id from the given values. By Tobias Hilbrich, Date: 25.11.2005.
     * @param CallID the CallID of the MPI_Call performed.
     * @param Location the SourceLocation (that might also be a invalid location).
     * @return the CallTree unique ID for this Location and CallID.
     */
    unsigned long MPO_LogCube::calculateCallTreeId(MPO_CALL_ID CallID,const MPO_SourceLocation &Location) const
    {
        if (Location.isValidLocation())
        {
            return (unsigned long)CallID | (Location.getCodeline()<<CALLTREE_CALLID_SHIFT);
        }
        else
        {
            return (unsigned long)CallID | CALLTREE_UNKNOWNLOCATION_CALLID_MASK;
        }
    }
    
    /** Calculates a unique PerformanceTree(MessagesTree) Id from the given values. 
     * By Tobias Hilbrich, Date: 25.11.2005.
     * @param ErrorCode the errorcode for which a id has to be calculated.
     * @param MsgType the Messagetype for the id.
     * @return the PerformanceTree wide unique ID for this ErrorCode and MsgType.
     */
    unsigned long MPO_LogCube::calculatePerformanceTreeId(MARMOT_ERR_CODES ErrorCode,MPO_MSG_TYPE MsgType)
    {
        return ErrorCode + (MsgType<<PERFORMANCETREE_MSGTYPE_SHIFT);
    }
    
    /** Calculates an Id from the given char string. 
     * By Tobias Hilbrich, Date: 20.12.2005.
     * @param str the 0 terminated string.
     * @return the id of the string.
     */
    unsigned long MPO_LogCube::calculateIdFromString (const char *str)
    {
        
        /** \todo @Tobias from Tobias: some algorithmic enhancment might speed things up, maybe also use strequal and strcpy standard functions. */
        
        //find out wheter str contains a path rather than a file name 
        //if so extract the filename in the later cases
        std::string filename = MPO_LogCube::getFileNameFromPath(str);
        
        //create an iterator for the known strings array
        MPO_TemplateTree<char*> *node;
        char *cmp;
        MPO_TemplateTreeIterator<char*> iter (&myCharIdMap);
        iter.next();//skip the root node !
                
        //iterate over all the nodes
        while (iter.hasNext())
        {
            node = iter.next();
            cmp = node->getObject();
            
            if (cmp == NULL)
                continue;
            
            //make a strequal
            if (filename.compare(cmp) == 0)
                return node->getId();
        }
        
        //create a copy of the string
        cmp = new char[filename.length()+1];
        strcpy(cmp,filename.c_str());
        
        //add the node to the tree
        myCharIdMap.addChild(cmp,myNextCharId);
        cmp = NULL;
        
        //return the id
        unsigned long temp = myNextCharId;
        myNextCharId++;
        return temp;        
    }
    
    /** Logs the message to the Cnode with the given id or creates the new Cnode and logs to it. 
     * Should only be used for MPI_Call locations not for special location like UnknownLocations->Notes. By Tobias Hilbrich, Date: 26.11.2005.
     * @param id the id of the location computed with see MPO_LogCube::calculateCallTreeId.
     * @param pMsg reference to the Message object.
     * @param SourceLocation the source location of the message.
     * @return the Cnode of the CallTree node (return value needed for severity mapping).
     */
    Cnode* MPO_LogCube::logMessageToCallTree (unsigned long id, MPO_MsgObject &pMsg,const MPO_SourceLocation& SourceLocation)
    {
        unsigned long fileId = 0;
        std::string fileName = "";
        
        if (SourceLocation.isValidLocation())
        {
            fileId = calculateIdFromString (SourceLocation.getFileName());
            fileName = getFileNameFromPath(SourceLocation.getFileName());
        }
        
        MPO_TemplateTree<Cnode*> *pFileNameNode = NULL;
        
        //test whether a sourcefilename is given
        if (fileId != 0)
        {
            pFileNameNode = mypCNodeTree->searchId(fileId);
            
            //test whether the location has to be created
            if (pFileNameNode == NULL)
            {
                Region *rgnFileName = myCube.def_region(fileName.c_str(), 0, 0,"",CALLTREE_SOURCEFILE_DESCRIPTION, CALLTREE_MODULENAME);
                mypRegionTree->addChild(rgnFileName,fileId);
                    
                Cnode *nodeSourceFile = myCube.def_cnode(rgnFileName, mypCNodeTree->searchId(CALL_TREE_ID_ROOT)->getObject());
                mypCNodeTree->addChild(nodeSourceFile,fileId);
                    
                mypCallTreeHTML->addChild(NULL,fileId);                
                
                /** \todo Think whether we want HTML`s here !!! */
//                mypCallTreeHTML->addChild(pHTMLLog,id);
//                pHTMLLog->getMsg(pMsg);
//                mypCNodeTree->searchId(fileId)
                
                //lets load it correctly
                pFileNameNode = mypCNodeTree->searchId(fileId);
            }
        }
        else
        {
            fileId = CALL_TREE_ID_UNKNOWNLOCATION;
        }
        
        //we have to select the correct node for searching a call id (call ids are only unique within a sourcefile) !
        MPO_TemplateTree<Cnode*> *pNode = NULL;
        if (pFileNameNode)
            pNode = pFileNameNode->searchId(id);
        else pNode = mypCNodeTree->searchId(id);
        
        //Does the location exists
        if (pNode == NULL)
        {
            //the location does not exsits   
            //common stuff for given location and for unknown location
            char    url[256], name[256];
            
            if (!pFileNameNode)
              sprintf (name,"%s/MPI_Call_Nr%ld.html",HTML_DIR,id);
            else
              sprintf (name,"%s/MPI_Call_Nr%ld_%s.html",HTML_DIR,id,fileName.c_str());
            
            MPO_LogHTML *pHTMLLog = new MPO_LogHTML (name,false);
                    
            //sprintf (url,"@mirror@%s",name);
            sprintf (url,"file:./%s",name);
                            
            Cnode *node;
                
            if (SourceLocation.isValidLocation())
            {
                //create known location
                sprintf (name,"%s@line: %d",pMsg.getCallinfo().c_str(),SourceLocation.getCodeline());
                Region *rgnCall  = myCube.def_region(name, 0, 0,url,CALLTREE_MPICALLNAME_DESCRIPTION, CALLTREE_MODULENAME);
                mypRegionTree->searchId(fileId)->addChild(rgnCall,id);
                    
                node = myCube.def_cnode(rgnCall, mypCNodeTree->searchId(fileId)->getObject());
                pFileNameNode->addChild(node,id);
                    
                mypCallTreeHTML->searchId(fileId)->addChild(pHTMLLog,id);
                
                if (pMsg.getText().length() > 0)
                    pHTMLLog->getMsg(pMsg);
            }
            else
            {
                //create unknown location
                sprintf (name,"%s",pMsg.getCallinfo().c_str());
                Region *rgnCall  = myCube.def_region(name, 0, 0,url,CALLTREE_MPICALLNAME_DESCRIPTION, CALLTREE_MODULENAME);
                mypRegionTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(rgnCall,id);
                    
                node = myCube.def_cnode(rgnCall, mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->getObject());
                mypCNodeTree->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(node,id);
                    
                mypCallTreeHTML->searchId(CALL_TREE_ID_UNKNOWNLOCATION)->addChild(pHTMLLog,id);
                if (pMsg.getText().length() > 0)
                    pHTMLLog->getMsg(pMsg);
            }
                
            //return the Cnode for severity mapping
            return node;
                
        }//if (pNode == NULL)
        else
        {
            //log to the node
            logMsgToExsitingCallTreeNode (pMsg,id,fileId);                          
            //store the Cnode for severity mapping
            return pNode->getObject();
        }// if (pNode == NULL) -> else
        
        
        //theoreticaly impossible because of if ... else ...
        return NULL;
    }
    
    /** Logs the message to the node with the given id. By Tobias, Date 26.11.2005.
     * @param pMsg reference to the MessageObject that will be logged to the node.
     * @param nodeId the id of the to which the message will be logged.
     */
    bool MPO_LogCube::logMsgToExsitingCallTreeNode (MPO_MsgObject &pMsg,unsigned int nodeId,unsigned int subtreeId)
    {
        /** \todo handle enhanced id mangling -> get an Location ID and only search ID in subtree of this id !!! */
        MPO_TemplateTree<MPO_LogHTML*> *SubTreeNode = mypCallTreeHTML->searchId(subtreeId);
        
        if (!SubTreeNode)
            return false;
        
        MPO_TemplateTree<MPO_LogHTML*> *HTMLNode = SubTreeNode->searchId(nodeId);
        
        if (HTMLNode != NULL)
        {
            MPO_LogHTML *HTMLLog = HTMLNode->getObject();
            if (HTMLLog)
            {
                if (pMsg.getText().length() > 0)
                    HTMLLog->getMsg(pMsg);
                return true;
            }
        }
        else
        {
            std::cerr << "MARMOT ERROR: MPO_LogCube: failed at logMsgToExsitingCallTreeNode";
        }
        
        return false;
    }
    
    /** Creates a child of type err for the node with idFather. 
     * By Tobias Hilbrich, Date: 26.11.2005.
     * @param err the error code of the new node.
     * @param metricsId the id of the new metrics.
     * @param idFather the id of the node which will be the father of the new node.
     * @return true when the operation succeeds false otherwise.
     */
    bool MPO_LogCube::createPerformanceMetric (MARMOT_ERR_CODES err,unsigned int metricsId,unsigned int idFather)
    {
        //load a reference to the MPI Standard
        const char *temp = getRelativPathToMPIStandard(err);
        //int iLength = getLocalPathToMPIStandard(err,temp,0);
        int iLength=strlen(temp);
        //temp = new char[iLength];
        //getRelativPathToMPIStandard(err,temp,iLength);
        
        char *url = new char[iLength + 32];
                
        sprintf (url,"@mirror@%s",temp);
        
        //search for the father node
        MPO_TemplateTree<Metric*> *nodeFather = mypMetricTree->searchId(idFather);
        
        if (!nodeFather)
        {
            std::cerr << "Error within "<< __FILE__ << " line: " << __LINE__ << std::endl;
            return false;
        }
        
        //get a description for the error
        std::stringstream stream;
        DebugServer::outputErrorMessages (err,stream);
        MPO_String str;
        str.assign(&stream);
        
        Metric *pMet = myCube.def_met( str,  str, "INTEGER","occ", 
                                               "ZERO", 
                                               url, 
                                               MESSAGES_MPIERROR_DESCRIPTION, 
                                               nodeFather->getObject() );
        
        nodeFather->addChild (pMet,metricsId);
        
        //free the dynamic memory we use here
        if (url)
            delete (url);
        url = NULL;
        
        //if (temp)
        //    delete (temp);
        //temp = NULL;
        
        return true; 
    }
    
    /** Flushs the Cube logfile. 
     * By Tobias Hilbrich, Date: 29.11.2005.
     * @return true when the operation succeeds false otherwise.
     */
    bool MPO_LogCube::flushOfstream (void)
    {
        //write the cube
        //this operation might be pretty expensive so we should realy avoid it whenever it is possible
        std::ofstream out;
        out.open(myLogFileName.c_str(),std::ios_base::out | std::ios_base::trunc);
        out << myCube;
        out.close();
        
        //flush all the HTML log files :(
        MPO_TemplateTreeIterator<MPO_LogHTML*> i(mypMetricTreeHTML);
        MPO_TemplateTree<MPO_LogHTML*>  *node;
        MPO_LogHTML *log;
        
        while (i.hasNext())
        {
            node = i.next();
            log = node->getObject();
            
            if (log)
                log->flushOfstream();
        }
        
        MPO_TemplateTreeIterator<MPO_LogHTML*> j(mypCallTreeHTML);
        
        while (j.hasNext())
        {
            node = j.next();
            log = node->getObject();
            
            if (log)
                log->flushOfstream();
        }
                
        return true;
    }

    /** Extracts the filename from a string possibly also containing a path.
     * @param str the filename that might include a path.
     * @return the filename without of path info (the characters after the last '/').
     */
    std::string MPO_LogCube::getFileNameFromPath(const char* str)
    {
        std::string ret;
        int start = 0;
        int i=0;
        while ((str[i] != '\0'))
        {
            if ((str[i] == '/') || (str[i] == '\\'))
                start = i+1;
            i++;
        }    
        
        const char* fileName = &(str[start]);
        ret.assign(fileName);
        return ret;
    }
    
    /**
     * Checks whether the message introduces an unknown thread, if so it is created.
     * @param the message that might conatin a new thread.
     */
    void MPO_LogCube::checkForThreadInMetric (MPO_MsgObject &pMsg)
    {
        int rank = pMsg.getRank();
        int thread = pMsg.getThreadRank();
        
        if (rank < 0)
            return;
            
        std::map<int,Thread*>::iterator i = myThread[rank].find(thread);
        
        if (i == myThread[rank].end())
        {
            //thread not known so add it
            char buf[128];
            sprintf(buf,"Thread %d",thread);
            
            myThread[rank].insert(
                std::pair<int,Thread*>(
                    thread,
                    myCube.def_thrd(buf, thread, myProcess[rank])
                    )
                );
        }
    }            
    
}//end of namespace MPO
#endif







