/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file 
 * This Sourcefile implements the MPO_LogHTML object.
 * This object recieves MPO_MsgObject objects and creates a logfile from their data.
 * The logfile will be a html file.
 * @author Tobias Hilbrich
 * @date 18.11.2005
 * $Id: mpo_LogHTML.cc 886 2008-11-10 15:21:03Z hpczink $
 */

#include "mpo_LogHTML.h"
#include "mpo-binaryIO.h"
#include <ostream>

#include "config.h"
#include <time.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#define TRAILERSTRING     "  </table></body></html>"
#define TRAILERSTRINGLEN  23

namespace MPO 
{
    /**
     * Constructor. By Tobias, Date 18.11.2005.
     * @param logFileName name of the output file.
     * @param MPIStandardFilename name of a file that contains information about the MPI-Standard.
     * @param bUseStandardReference determines whether referncing the MPI-Standard for background information is allowed or not.
     */
    MPO_LogHTML::MPO_LogHTML (string logFileName,bool bUseStandardReference)
    {
        init(logFileName,bUseStandardReference);
    }
    
    /**
     * init call. By Tobias, Date 18.11.2005.
     * @param logFileName name of the output file.
     * @param MPIStandardFilename name of a file that contains information about the MPI-Standard.
     * @param bUseStandardReference determines whether referncing the MPI-Standard for background information is allowed or not.
     */
    bool MPO_LogHTML::init (string logFileName,bool bUseStandardReference)
    {
        m_bUseStandardReference = bUseStandardReference;
        
        //even if we don`t use the referencer we will init it cause it costs almost nothing
        initMPIStandardReferencer();
        
        m_bPrintedTrailer = false;
        m_bInitialized = false;

        initSocket();
      
        if (!createOfstream(logFileName) || !printHeader ())
            return m_bInitialized;
      
        m_bInitialized = true;
        m_bLineEven = true;
        
        return m_bInitialized;
    } 
   
    /**
     * Standard Destructor. By Tobias, Date 03.11.2005.
     */
    MPO_LogHTML::~MPO_LogHTML ()
    {
        if (m_bInitialized) 
        {
            closeLog (); //sets m_bInitialized to false and makes a call to PrintTrailer
        }
    }
   
    /**
     * Gets a message object. By Tobias, Date 03.11.2005.
     * @param pMsg the Message that will be logged.
     * @return true when the operation succeeds false otherwise.
     */
    bool MPO_LogHTML::getMsg (MPO_MsgObject &pMsg)
    {
        if (!m_bInitialized)
            return (false);
        
        bool bWasTrailerPrinted = m_bPrintedTrailer;
      
        if (bWasTrailerPrinted)
            unflushOfstream();
        
        //select even or odd char
        char even_or_odd = 'e';
        if (!m_bLineEven)
            even_or_odd = 'o';
        
        //select even or odd char
        char info_warn_error = 'i';
        if (pMsg.getMsgType() == MSG_TYPE_WARNING)
            info_warn_error = 'w';
        else if (pMsg.getMsgType() == MSG_TYPE_ERROR)
            info_warn_error = 'e';
      
        //do the html output
        (*m_pLog) << "<tr>"
                  << "<td class="<< info_warn_error << even_or_odd << "1>"
                  << pMsg.getTimestamp ()
                  << "</td>"
                  << "<td class="<< info_warn_error << even_or_odd << "2>";
                
        if (pMsg.getRank() != -1)
            (*m_pLog) << pMsg.getRank();
        else
            (*m_pLog) << "Global";
        
        (*m_pLog) << "</td>"
      	          << "<td class="<< info_warn_error << even_or_odd << "3>"
      	          << pMsg.getThreadRank()
                  << "</td>"
                  << "<td class="<< info_warn_error << even_or_odd << "4>"
                  << "<b>"
                  << pMsg.getMessageTypeString()
                  << "</b>"
                  << "</td>"
                  << "<td class="<< info_warn_error << even_or_odd << "5>";
      
        if (pMsg.getOutputType() != OUTPUT_TYPE_PURE_CALLINFO)
        {
            //replace all '\n' characters with "<br>" 
            std::string::size_type p = 0;
            string temp (pMsg.getText()); //copy the text
          
            do {
                p = temp.find ('\n', p);
                if (p == std::string::npos)
                    break;
                temp.replace (p,1,"<br>");              
            }while (p != std::string::npos);
          
            (*m_pLog) << "Text: " 
                      << temp;
        }
      
        if (pMsg.getOutputType() == OUTPUT_TYPE_TEXT_AND_CALLINFO)
        {
            (*m_pLog) << "<br>"; 
        }

        if (pMsg.getOutputType() != OUTPUT_TYPE_PURE_TEXT)
        {
            (*m_pLog) << "Call: " 
                      << pMsg.getCallinfo();
        }       
      
        (*m_pLog)  << "</td>"
                   << "<td class="<< info_warn_error << even_or_odd << "6>";
      
        pMsg.getSourceLocation().debugout(*m_pLog);
      
        (*m_pLog)  << "</td>";
        
        const char *temp = NULL;
        if (m_bUseStandardReference)
        {
            (*m_pLog) << "<td class="<< info_warn_error << even_or_odd << "7>";
                    
            MARMOT_ERR_CODES err = pMsg.getErrorCode ();
          
            if ( (err != MARMOT_SUCCESS) &&
                 (err != MARMOT_ERR_UNDEFINED) &&
                 (err != MARMOT_ERR_MARMOT) &&
                 (err != MARMOT_LAST_ERR_CODE)    )
            {
                temp = getLocalPathToMPIStandard(err);
                std::string url("<a href=\"file://");
                url.append(temp);
                url.append("\">Infos see MPI-Standard</a>");
                //char *url = new char[iLength + 128];
                //sprintf (url,"<a href=\"file://%s\">Infos see MPI-Standard</a>",temp);
              
                (*m_pLog) << url.c_str();
            }
            else 
                (*m_pLog) << "&nbsp";
          
            (*m_pLog) << "</td>";
        }
      
        (*m_pLog)  << "</tr>" << std::endl;

        // if there is a socket send msg
        pMsg.writeToSocket(getLogSocket(), temp);
        // wait for ACK from log-host
        if (pMsg.getMsgType() == MSG_TYPE_ERROR)
          CSocketLogger::instance()->waitForAck();
      
        if ((pMsg.getMsgType() == MSG_TYPE_ERROR) || bWasTrailerPrinted || (LOGFLUSH_TYPE == LOGFILE_FLUSH_TYPE_IMMEDIATE))
            flushOfstream();
        
        //invert even/odd flag
        m_bLineEven = !m_bLineEven;
      
        return (true);
    }
   
    /**
     * Prints the Header of the Logfile like date and so on. By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogHTML::printHeader (void)
    {
        if (!m_pLog)
            return (false);
      
        //read the system time
        char buf[128];
      
        struct tm *ptr;
        time_t tm;
        tm = time(NULL);
        ptr = localtime(&tm);
        strftime(buf ,128 , "%c.\n",ptr);
        
        //print the header
        (*m_pLog) << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">" << std::endl
                  << "<html>" << std::endl
                  << "<head>" << std::endl
                  << "<title>MARMOT HTML Logfile</title>" << std::endl
                  << "<style type=\"text/css\">" << std::endl
                  << "td,td,table {border:thin solid black}" << std::endl
                  << "td.ie1{ background-color:#DDFFDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ie2{ background-color:#EEFFEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ie3{ background-color:#DDFFDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ie4{ background-color:#EEFFEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ie5{ background-color:#DDFFDD;}" << std::endl
                  << "td.ie6{ background-color:#EEFFEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ie7{ background-color:#DDFFDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.io1{ background-color:#CCEECC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.io2{ background-color:#DDEEDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.io3{ background-color:#CCEECC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.io4{ background-color:#DDEEDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.io5{ background-color:#CCEECC;}" << std::endl
                  << "td.io6{ background-color:#DDEEDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.io7{ background-color:#CCEECC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.we1{ background-color:#FFFFDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.we2{ background-color:#FFFFEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.we3{ background-color:#FFFFDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.we4{ background-color:#FFFFEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.we5{ background-color:#FFFFDD;}" << std::endl
                  << "td.we6{ background-color:#FFFFEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.we7{ background-color:#FFFFDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.wo1{ background-color:#EEEECC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.wo2{ background-color:#EEEEDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.wo3{ background-color:#EEEECC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.wo4{ background-color:#EEEEDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.wo5{ background-color:#EEEECC;}" << std::endl
                  << "td.wo6{ background-color:#EEEEDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.wo7{ background-color:#EEEECC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ee1{ background-color:#FFDDDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ee2{ background-color:#FFEEEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ee3{ background-color:#FFDDDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ee4{ background-color:#FFEEEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ee5{ background-color:#FFDDDD;}" << std::endl
                  << "td.ee6{ background-color:#FFEEEE; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.ee7{ background-color:#EEDDDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.eo1{ background-color:#EECCCC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.eo2{ background-color:#EEDDDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.eo3{ background-color:#EECCCC; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.eo4{ background-color:#EEDDDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.eo5{ background-color:#EECCCC;}" << std::endl
                  << "td.eo6{ background-color:#EEDDDD; text-align:center; vertical-align:middle;}" << std::endl
                  << "td.eo7{ background-color:#EECCCC; text-align:center; vertical-align:middle;}" << std::endl
                  << "</style>" << std::endl
                  << "</head>" << std::endl
                  << "<body>" << std::endl
                  << "<p> <b>MARMOT Logfile</b>, date: "
                  << buf
                  << "</p>" << std::endl
                  << "<table border=\"0\" width=\"100%\" cellspacing=\"0\" cellpadding=\"0\">" << std::endl
                  << "<tr>" << std::endl
                  << "<td align=\"center\" bgcolor=#7777BB>"
                  << "<b>Timestamp</b>"
                  << "</td>" << std::endl
                  << "<td align=\"center\" bgcolor=#9999DD>"
                  << "<b>Rank</b>"
                  << "</td>" << std::endl
                  << "<td align=\"center\" bgcolor=#7777BB>"
                  << "<b>Thread</b>"
                  << "</td>" << std::endl
                  << "<td align=\"center\" bgcolor=#9999DD>"
                  << "<b>Type</b>"
                  << "</td>" << std::endl
                  << "<td align=\"center\" bgcolor=#7777BB>"
                  << "<b>Message</b>"
                  << "</td>" << std::endl
                  << "<td align=\"center\" bgcolor=#9999DD>"
                  << "<b>Location</b>"
                  << "</td>" << std::endl;
                  
        if (m_bUseStandardReference)
        {
            (*m_pLog) << "<td align=\"center\" bgcolor=#7777BB>"
                      << "<b>MPI-Standard Reference</b>"
                      << "</td>" << std::endl;
        }
      
        (*m_pLog) << "</tr>" << std::endl;
      
        return (true);
    }
   
    /**
     * Prints the Trailer of the Logfile. By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogHTML::printTrailer (void)
    {
        if (!m_bInitialized)
            return (false);
      
        (*m_pLog) << TRAILERSTRING << std::endl;
      
        return (true);
    }
    
    /**
     * Flushs the ofstream in order to prevent lose of data. By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogHTML::flushOfstream (void)
    {
        if (!m_bInitialized)
            return (false);
        
        if (!m_bPrintedTrailer)
        {
            printTrailer();
            m_bPrintedTrailer = true;   
        }
        
        m_pLog->flush();
        
        return true;
    }
    
    /**
     * Used before printing a message when the ofstream was already flushed. By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogHTML::unflushOfstream (void)
    {
        if (!m_bInitialized)
            return (false);
        
        if (m_bPrintedTrailer)
        {
            long pos=m_pLog->tellp();
            m_pLog->seekp (pos-TRAILERSTRINGLEN-1);
            //m_pLog->seekp(TRAILERSTRINGLEN,std::ios::end);
            m_bPrintedTrailer = false;
            //Falsch -> /** IMPORTANT -> as a result a message must have at leaset TRAILERSTRINGLEN otherwise a rest of the trailer might survive ! */
        }
        
        return true;
    }
    
}//end of namespace MPO







