/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Pack.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Pack.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Pack::MPO_Pack()
        : my_inbuf(0),
          my_incount(-1),
          my_datatype(MPI_DATATYPE_NULL),
          my_outbuf(0), 
          my_outsize(-1), 
          my_position(0), 
          my_comm(MPI_COMM_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Pack::MPO_Pack(void *inbuf,
                       int incount,
                       MPO_Datatype datatype,
                       void *outbuf,
                       int outsize,
                       int *position,
                       MPO_Comm comm)
        : my_inbuf(inbuf),
          my_incount(incount),
          my_datatype(datatype),
          my_outbuf(outbuf),
          my_outsize(outsize),
          my_position(position),
          my_comm(comm)
    {
    }


    // destructor
    //========================================================================
    MPO_Pack::~MPO_Pack()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Pack::checkPrerequisites()
    {


        // check correctness of parameters
        checkCountIn(MSG_INCOUNT);
        checkSizeOut(MSG_OUTSIZE);
        checkComm();
        checkDatatype();
    
        
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        
        
        return 0;
    }


    // checkCountIn
    //========================================================================
    int MPO_Pack::checkCountIn(const string KIND_OF_COUNT)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if count is negative.
        if ((our_errcodes.at(MARMOT_ERR_COUNT_NEG))&&(my_incount < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << " = "
                << my_incount
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_COUNT_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if count is 0.
        if ((our_errcodes.at(MARMOT_WARN_COUNT_ZERO))&&(my_incount == 0) && 
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: " 
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << "= 0!" 
                << std::endl;
            rvalue = MARMOT_WARN_COUNT_ZERO;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkCountIn


    // checkSizeOut
    //========================================================================
    int MPO_Pack::checkSizeOut(const string KIND_OF_SIZE)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        // (1) Check if size is negative.
        if ((our_errcodes.at(MARMOT_ERR_SIZE_NEG))&&(my_outsize < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: " 
                << getCallname() 
                << ": "
                << KIND_OF_SIZE
                << " = "
                << my_outsize
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_SIZE_NEG;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if size is 0.
        else if ((our_errcodes.at(MARMOT_WARN_SIZE_ZERO))&&(my_outsize == 0) && 
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: " 
                << getCallname() 
                << ": "
                << KIND_OF_SIZE
                << " = 0 !" 
                << std::endl;
            rvalue = MARMOT_WARN_SIZE_ZERO;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkSizeOut


    // checkComm
    //========================================================================
    int MPO_Pack::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


// TODO
    // checkDatatype
    //========================================================================
    int MPO_Pack::checkDatatype()
    {
        /** \todo @creator of this method: what about return values ? */
        
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;
        if (my_datatype == MPI_DATATYPE_NULL)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname() 
                    << ": Datatype is MPI_DATATYPE_NULL!" 
                    << std::endl;
                    
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }
        else
        {
            if (our_datatypes.find(my_datatype) == our_datatypes.end())
            {
                if (MSG_ERR <= DEBUG_MODE)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname() 
                        << ": Datatype is not valid!" 
                        << std::endl;
                        
                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
                    MPO_String text;
                    text.assign(&stream);      
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation()); 
                    our_log->getMsg(msg);   
                }
            }
            else
            {
                std::stringstream stream;
                MPO_MSG_TYPE type = MSG_TYPE_ERROR;
                                        
                switch (our_datatypes.find(my_datatype)->second)
                {
                    case (DATATYPE_CREATED_NOT_COMMITTED):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_COMMITTED))&&(MSG_ERR <= DEBUG_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname() 
                                << ": Datatype is not committed!" 
                                << std::endl;
                            
                            //check this !
                            rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;
                        }
                        break;
                    case (DATATYPE_REDUCTION_C):
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_C))&&(MSG_ERR_WARN <= DEBUG_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": Datatype is for reduction functions only!" 
                                << std::endl;
                            
                            type = MSG_TYPE_WARNING;
                            
                            //check this !
                            rvalue = MARMOT_WARN_TYPE_REDUCTION_C;
                        }
                        break;
                    case (DATATYPE_OPTIONAL_C):
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_C))&&(MSG_ERR_WARN <= DEBUG_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname() 
                                << ": Datatype is only optional!" 
                                << std::endl;
                            
                            type = MSG_TYPE_WARNING;
                            
                            //check this !
                            rvalue = MARMOT_WARN_TYPE_OPTIONAL_C;
                        }
                        break;
                    case (DATATYPE_COMMITTED_FORTRAN):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_F))&&(MSG_ERR <= DEBUG_MODE) && 
                            (C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname() 
                                << ": Datatype is Fortran-Type!" 
                                << std::endl;
                            
                            //check this !
                            rvalue = MARMOT_ERR_TYPE_F;
                        }
                        break;
                    default:
                        // do nothing
                        break;
                }//end switch
                
                if (rvalue != MARMOT_SUCCESS)
                {
                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                    
                    MPO_String text;
                    text.assign(&stream);      
                    MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                    our_log->getMsg(msg);   
                }
                
            }//end if else
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Pack::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_PACK_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Pack::checkPost(void)
    {
                               
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Pack::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Pack(my_inbuf, my_incount, my_datatype, my_outbuf,
                           my_outsize, my_position, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Pack::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Pack";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_PACK";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Pack::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(*inbuf, "
            << "incount = "
            << my_incount
            << ", datatype = "
            << my_datatype
            << ", *outbuf, "
            << "outsize = "
            << my_outsize 
            << ", *position, "
            << "comm = " 
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")" 
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Pack::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_inbuf;
        out < my_incount;
        out < my_datatype;
        // out < my_outbuf;
        out < my_outsize;
        // out < my_position;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Pack::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_inbuf;
        in > my_incount;
        in > my_datatype;
        // in > my_outbuf;
        in > my_outsize;
        // in > my_position;
        in > my_comm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/pack.c.
/*@
  MPI_Pack - Packs a datatype into contiguous memory

  Input Parameters:
  + inbuf - input buffer start (choice) 
  . incount - number of input data items (integer) 
  . datatype - datatype of each input data item (handle) 
  . outcount - output buffer size, in bytes (integer) 
  . position - current position in buffer, in bytes (integer) 
  - comm - communicator for packed message (handle) 

  Output Parameter:
  . outbuf - output buffer start (choice) 

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_TYPE
  .N MPI_ERR_COUNT
  .N MPI_ERR_ARG

  .seealso: MPI_Unpack, MPI_Pack_size

  int MPI_Pack ( 
  void *inbuf, 
  int incount, 
  MPI_Datatype datatype, 
  void *outbuf, 
  int outcount, 
  int *position, 
  MPI_Comm comm )

  @ */
