/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Reduce.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Reduce.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Reduce::MPO_Reduce()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_recvbuf(0),
          my_count(-1),
          my_datatype(MPI_DATATYPE_NULL),
          my_op(MPI_OP_NULL),
          my_root(MPI_PROC_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Reduce::MPO_Reduce(void *sendbuf,
                           void *recvbuf,
                           int count,
                           MPO_Datatype datatype,
                           MPO_Op op,
                           int root,
                           MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_recvbuf(recvbuf),
          my_count(count),
          my_datatype(datatype),
          my_op(op),
          my_root(root)
    {
    }


    // destructor
    //========================================================================
    MPO_Reduce::~MPO_Reduce()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Reduce::checkPrerequisites2()
    {
        // check correctness of parameters
        checkCount(my_count, MSG_COUNT);
        checkDatatypeSendRecv(my_count, my_datatype,
                              IS_COUNT, IS_REDUCE_ALLREDUCE);
        checkRoot(my_root);
        checkOp(my_op);

        //check buffers
        int rank;
        PMPI_Comm_rank(my_comm,&rank);
        checkBufferInUse(my_sendbuf,my_count,my_datatype);
        if (rank == my_root)
        {
            checkBufferInUse(my_recvbuf,my_count,my_datatype);
        }

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Reduce::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_REDUCE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Reduce::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Reduce::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkPrerequisites2();

        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        our_op_traces->findAndAddUsage(my_op,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Reduce::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Reduce(my_sendbuf, my_recvbuf, my_count, my_datatype,
                           my_op, my_root, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Reduce::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Reduce";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_REDUCE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Reduce::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "*recvbuf, "
            << "count = "
            << my_count
            << ", datatype = "
            << my_datatype
            << ", op = "
            << my_op
            << ", root = "
            << my_root
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Reduce::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        // out < my_recvbuf;
        out < my_count;
        out < my_datatype;
        out < my_op;
        out < my_root;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Reduce::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        // in > my_recvbuf;
        in > my_count;
        in > my_datatype;
        in > my_op;
        in > my_root;
        in > my_comm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/reduce.c.
/*@

MPI_Reduce - Reduces values on all processes to a single value

Input Parameters:
+ sendbuf - address of send buffer (choice)
. count - number of elements in send buffer (integer)
. datatype - data type of elements of send buffer (handle)
. op - reduce operation (handle)
. root - rank of root process (integer)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice,
significant only at 'root')

Algorithm:
This implementation currently uses a simple tree algorithm.

.N fortran

.N collops

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER
.N MPI_ERR_BUFFER_ALIAS

int MPI_Reduce (
void *sendbuf,
void *recvbuf,
int count,
MPI_Datatype datatype,
MPI_Op op,
int root,
MPI_Comm comm )
@ */

