/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Reduce_scatter.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Reduce_scatter.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Reduce_scatter::MPO_Reduce_scatter()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_recvbuf(0),
          my_recvcounts(0),
          my_datatype(MPI_DATATYPE_NULL),
          my_op(MPI_OP_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Reduce_scatter::MPO_Reduce_scatter(void *sendbuf,
                                           void *recvbuf,
                                           int *recvcounts,
                                           MPO_Datatype datatype,
                                           MPO_Op op,
                                           MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_recvbuf(recvbuf),
          my_recvcounts(recvcounts),
          my_datatype(datatype),
          my_op(op)
    {
    }


    // destructor
    //========================================================================
    MPO_Reduce_scatter::~MPO_Reduce_scatter()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Reduce_scatter::checkPrerequisites2()
    {
        // check correctness of parameters
        checkCountArray(my_recvcounts, MSG_RECVCOUNT);
        checkOp(my_op);

        //check buffers
        int count = 0,size;
        PMPI_Comm_size(my_comm,&size);
        for (int i = 0;i < size;i++)
        {
            count += my_recvcounts[i];
        }
        checkBufferInUse(my_sendbuf,count,my_datatype);

        int rank;
        PMPI_Comm_rank(my_comm,&rank);
        checkBufferInUse(my_recvbuf,my_recvcounts[rank],my_datatype);

        return 0;
    }


    // checkDatatypeArray
    //========================================================================
    int MPO_Reduce_scatter::checkDatatypeArray()
    {

        /** \todo @creator of this method: what about return values ? */

        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        if (my_datatype == MPI_DATATYPE_NULL)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Datatype is MPI_DATATYPE_NULL!"
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                our_op_traces->findAndPrintTraceNamed(my_op,stream,"op");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            if (our_datatypes.find(my_datatype) == our_datatypes.end())
            {
                if (MSG_ERR <= DEBUG_MODE)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname()
                        << ": Datatype is not valid!"
                        << std::endl;

                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                    our_op_traces->findAndPrintTraceNamed(my_op,stream,"op");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
            else
            {
                std::stringstream stream;
                MPO_MSG_TYPE type = MSG_TYPE_ERROR;

                switch (our_datatypes.find(my_datatype)->second)
                {
                    case (DATATYPE_CREATED_NOT_COMMITTED):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_COMMITTED))&&(MSG_ERR <= DEBUG_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": Datatype is not committed!"
                                << std::endl;

                            //check this !
                            rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;
                        }
                        break;
                    case (DATATYPE_OPTIONAL_C):
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_C))&&(MSG_ERR_WARN <= DEBUG_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": Datatype is only optional!"
                                << std::endl;

                            type = MSG_TYPE_WARNING;

                            //check this !
                            rvalue = MARMOT_WARN_TYPE_OPTIONAL_C;
                        }
                        break;
                    case (DATATYPE_COMMITTED_FORTRAN):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_F))&&(MSG_ERR <= DEBUG_MODE) &&
                            (C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": Datatype is Fortran-Type!"
                                << std::endl;

                            //check this !
                            rvalue = MARMOT_ERR_TYPE_F;
                        }
                        break;
                    default:
                        // do nothing
                        break;
                }

                if (rvalue != MARMOT_SUCCESS)
                {
                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                    our_op_traces->findAndPrintTraceNamed(my_op,stream,"op");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Reduce_scatter::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_REDUCE_SCATTER_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Reduce_scatter::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Reduce_scatter::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkPrerequisites2();

        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_op_traces->findAndAddUsage(my_op,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Reduce_scatter::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Reduce_scatter(my_sendbuf, my_recvbuf, my_recvcounts,
                                   my_datatype, my_op, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Reduce_scatter::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Reduce_scatter";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_REDUCE_SCATTER";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Reduce_scatter::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "*recvbuf, "
            << "*recvcounts, "
            << "datatype = "
            << my_datatype
            << ", op = "
            << my_op
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Reduce_scatter::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        // out < my_recvbuf;
        // out < my_recvcounts;
        out < my_datatype;
        out < my_op;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Reduce_scatter::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        // in > my_recvbuf;
        // in > my_recvcounts;
        in > my_datatype;
        in > my_op;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/red_scat.c.
/*@

MPI_Reduce_scatter - Combines values and scatters the results

Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. recvcounts - integer array specifying the
number of elements in result distributed to each process.
Array must be identical on all calling processes.
. datatype - data type of elements of input buffer (handle)
. op - operation (handle)
- comm - communicator (handle)

Output Parameter:
. recvbuf - starting address of receive buffer (choice)

.N fortran

.N collops

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER
.N MPI_ERR_OP
.N MPI_ERR_BUFFER_ALIAS

int MPI_Reduce_scatter (
void *sendbuf,
void *recvbuf,
int *recvcnts,
MPI_Datatype datatype,
MPI_Op op,
MPI_Comm comm )
@ */

