/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Request_free.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Request_free.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Request_free::MPO_Request_free()
        : MPO_Completion(0, 0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Request_free::MPO_Request_free(MPI_Request * request)
        : MPO_Completion(1,
                         request)
    {
    }


    // destructor
    //========================================================================
    MPO_Request_free::~MPO_Request_free()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Request_free::checkPrerequisites()
    {


        checkRequest();
    
        eraseRequest();
        
        our_request_traces->findAndSetFree(my_request_array,getSourceLocation());
        unregisterBuffer(my_request_array);
        
        
        return 0;
    }


    // checkRequest
    //========================================================================
    int MPO_Request_free::checkRequest()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Request, enum KIND_OF_REQUEST >::iterator pos;
        pos = our_requests.find(&my_request_array[0]);

        // (1) Request is not valid, i.e. not registered in the map
        // of our_requests.  
        if ((our_errcodes.at(MARMOT_ERR_REQUEST_NOT_VALID))&&(pos == our_requests.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Request does not exist, is not / no more valid! " 
                << std::endl;
            rvalue = MARMOT_ERR_REQUEST_NOT_VALID;
            
            if (my_request_array)
                our_request_traces->findAndPrintTraceNamed(my_request_array,stream,"request");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Request is registered in the map of our_requests.
        // In this case, we have to check different violations 
        // in the switch statement.
        else if ((pos != our_requests.end()) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            
            std::stringstream stream; 
            MPO_MSG_TYPE type = MSG_TYPE_ERROR;                      
            
            // all the cases from enum KIND_OF_REQUEST
            switch (pos->second)
            {
                case (REQUEST_NULL):
                    if ((our_errcodes.at(MARMOT_WARN_REQUEST_NULL))&&(MSG_ERR_WARN <= DEBUG_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname()
                            << ": Request is MPI_REQUEST_NULL! " 
                            << std::endl;
                        rvalue = MARMOT_WARN_REQUEST_NULL;
                        
                        type = MSG_TYPE_WARNING;         
                    }
                    break;
                case (REQUEST_NOT_PERSISTENT):
                    if ((our_errcodes.at(MARMOT_WARN_REQUEST_ACTIVE_NOT_PERSISTENT))&&(MSG_ERR_WARN <= DEBUG_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname() 
                            << ": Active non-persistent request! "
                            << std::endl;
                        rvalue = MARMOT_WARN_REQUEST_ACTIVE_NOT_PERSISTENT;
                        
                        type = MSG_TYPE_WARNING;         
                    }         
                    break;
                case (REQUEST_ACTIVE_PERSISTENT):
                    if ((our_errcodes.at(MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT))&&(MSG_ERR_WARN <= DEBUG_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname() 
                            << ": Active persistent request! "
                            << std::endl;
                        rvalue = MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT;
                        
                        type = MSG_TYPE_WARNING;         
                    }
                    break;
                case (REQUEST_INACTIVE_PERSISTENT):
                    // Everything is fine.
                    break;
                default:
            if(our_errcodes.at(MARMOT_ERR_MARMOT))
            {
                        stream
                        << "Check mapping of requests."
                        << std::endl
                        << "Check implementation of switch statement."
                        << std::endl;
                        rvalue = MARMOT_ERR_MARMOT;
            }
                    break;
            }               // end switch
            
            if (rvalue != MARMOT_SUCCESS)
            {
                if (my_request_array)
                    our_request_traces->findAndPrintTraceNamed(my_request_array,stream,"request");
                 
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }                   // end else if

        return rvalue;
    }                       // end checkRequest


    // eraseRequest
    //========================================================================
    int MPO_Request_free::eraseRequest()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Request, enum KIND_OF_REQUEST >::iterator pos;
        pos = our_requests.find(&my_request_array[0]);

        // (1) Request is not valid, i.e. not registered in the map
        // of our_requests.  
        if ((our_errcodes.at(MARMOT_ERR_REQUEST_NOT_VALID))&&(pos == our_requests.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Request does not exist, is not / no more valid! " 
                << std::endl;
            rvalue = MARMOT_ERR_REQUEST_NOT_VALID;
            
            if (my_request_array)
                our_request_traces->findAndPrintTraceNamed(my_request_array,stream,"request");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Request is registered in the map of our_requests.
        // In this case, we erase the request.
        else if (pos != our_requests.end())
        {
            our_requests.erase(pos);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Request_free::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_REQUEST_FREE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Request_free::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Request_free::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Request_free(&my_request_array[0]);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Request_free::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Request_free";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_REQUEST_FREE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Request_free::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname() 
            << "(*request";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Request_free::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_request_array;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Request_free::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_request_array;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================

// The following comment is copied from
// mpich-1.2.4/src/pt2pt/commreq_free.c.

/*@
  MPI_Request_free - Frees a communication request object

  Input Parameter:
  . request - communication request (handle) 

  Notes:
  This routine is normally used to free persistent requests created with 
  either 'MPI_Recv_init' or 'MPI_Send_init' and friends.  However, it can 
  be used to free a request created with 'MPI_Irecv' or 'MPI_Isend' and
  friends; in that case the use can not use the test/wait routines on the
  request.

  It `is` permitted to free an active request.  However, once freed, you
  can not use the request in a wait or test routine (e.g., 'MPI_Wait').

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_REQUEST
  .N MPI_ERR_ARG

  .see also: MPI_Isend, MPI_Irecv, MPI_Issend, MPI_Ibsend, MPI_Irsend,
  MPI_Recv_init, MPI_Send_init, MPI_Ssend_init, MPI_Rsend_init, MPI_Wait,
  MPI_Test, MPI_Waitall, MPI_Waitany, MPI_Waitsome, MPI_Testall, 
  MPI_Testany, MPI_Testsome

  int MPI_Request_free( MPI_Request *request )
  @ */
