/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Sendrecv_replace.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Sendrecv_replace.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Sendrecv_replace::MPO_Sendrecv_replace()
        : MPO_Sendrecvs(-1,
                        MPI_DATATYPE_NULL,
                        MPI_PROC_NULL,
                        MPI_ANY_TAG,
                        MPI_PROC_NULL,
                        MPI_ANY_TAG,
                        MPI_COMM_NULL),
          my_buf(0),
          my_status(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Sendrecv_replace::MPO_Sendrecv_replace(void *buf,
                                               int count,
                                               MPO_Datatype datatype,
                                               int dest,
                                               int sendtag,
                                               int source,
                                               int recvtag,
                                               MPO_Comm comm,
                                               MPI_Status *
                                               status)
        : MPO_Sendrecvs(count,
                        datatype,
                        dest,
                        sendtag,
                        source,
                        recvtag,
                        comm),
          my_buf(buf),
          my_status(status)
    {
    }


    // destructor
    //========================================================================
    MPO_Sendrecv_replace::~MPO_Sendrecv_replace()
    {
    }


    // getId
    //========================================================================
    int MPO_Sendrecv_replace::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_SENDRECV_REPLACE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Sendrecv_replace::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Sendrecv_replace::checkPrerequisites(void)
    {


        MPO_Sendrecvs::checkPrerequisites();
        checkBufferInUse(my_buf,my_count,my_datatype);

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());

        #ifdef MARMOT_USE_THREADSYNC
            testIfProbed (my_source, my_recvtag, my_comm, getSourceLocation().getCodeline(), getSourceLocation().getFileName(), MPO_Space::getSyncObject()->getThreadRank(), getId(), getCallname());
        #endif


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Sendrecv_replace::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Sendrecv_replace(my_buf, my_count, my_datatype, my_dest,
                                     my_sendtag, my_source, my_recvtag,
                                     my_comm, my_status);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Sendrecv_replace::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Sendrecv_replace";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_SENDRECV_REPLACE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Sendrecv_replace::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*buf, "
            << "count = "
            << my_count
            << ", datatype = "
            << my_datatype
            << ", dest = "
            << my_dest
            << ", sendtag = "
            << my_sendtag
            << ", source = "
            << my_source
            << ", recvtag = "
            << my_recvtag
            << ", comm = "
            << my_comm
            << ", *status";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Sendrecv_replace::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_buf;
        out < my_count;
        out < my_datatype;
        out < my_dest;
        out < my_sendtag;
        out < my_source;
        out < my_recvtag;
        out < my_comm;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Sendrecv_replace::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_buf;
        in > my_count;
        in > my_datatype;
        in > my_dest;
        in > my_sendtag;
        in > my_source;
        in > my_recvtag;
        in > my_comm;
        // in > my_status;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/pt2pt/sendrecv_rep.c.
/*@
  MPI_Sendrecv_replace - Sends and receives using a single buffer

  Input Parameters:
  + count - number of elements in send and receive buffer (integer)
  . datatype - type of elements in send and receive buffer (handle)
  . dest - rank of destination (integer)
  . sendtag - send message tag (integer)
  . source - rank of source (integer)
  . recvtag - receive message tag (integer)
  - comm - communicator (handle)

  Output Parameters:
  + buf - initial address of send and receive buffer (choice)
  - status - status object (Status)

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_COUNT
  .N MPI_ERR_TYPE
  .N MPI_ERR_TAG
  .N MPI_ERR_RANK
  .N MPI_ERR_TRUNCATE
  .N MPI_ERR_EXHAUSTED

  int MPI_Sendrecv_replace(
  void *buf,
  int count,
  MPI_Datatype datatype,
  int dest,
  int sendtag,
  int source,
  int recvtag,
  MPI_Comm comm,
  MPI_Status *status )
  @ */

