/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Testall.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Testall.cc 964 2009-06-02 06:38:44Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Testall::MPO_Testall()
        : MPO_Tests(0,
                    0,
                    0),
          my_status(0),
          my_temp_request_array(NULL)
    {
    }


    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether request is in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove request from our_requests if test was successful
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     *  (it is a opaque type and thus  may contain some internal status etc.)
     */
    // proper constructor
    //========================================================================
    MPO_Testall::MPO_Testall(int count,
                             MPI_Request * request_array,
                             int *flag,
                             MPI_Status * status)
        : MPO_Tests(count,
                    request_array,
                    flag),
          my_status(status)
    {
        //copy requests into temporary array
        /**
         * TODO
         * I think this does not works with requests as
         * pointers. Though thats something that we don't
         * use anymore ....
         * (Address will change, thus the issue)
         */
        if (!handlePreDataConstructor ())
        {
            my_temp_request_array = NULL;
            if (count >= 0)
            {
                if (request_array)
                {
                    my_temp_request_array = new MPI_Request[count];
                    for (int i = 0; i < count; i++)
                        my_temp_request_array[i] = request_array[i];
                }
            }
        }
    }


    // destructor
    //========================================================================
    MPO_Testall::~MPO_Testall()
    {
        if (!handlePreDataDestructor ())
        {
            if (my_temp_request_array)
                delete my_temp_request_array;
            my_temp_request_array = NULL;
        }
    }

    // setPreData
    //========================================================================
    void MPO_Testall::setPreData (MPO_Callobject *obj)
    {
        my_temp_request_array = ((MPO_Testall*)obj)->my_temp_request_array;
        ((MPO_Testall*)obj)->my_temp_request_array = NULL;
    }

    // clonePreData
    //========================================================================
    MPO_Callobject* MPO_Testall::clonePreData (void)
    {
        MPO_Testall *obj = new MPO_Testall();
        obj->my_temp_request_array = my_temp_request_array;
        return obj;
    }

    // getId
    //========================================================================
    int MPO_Testall::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TESTALL_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Testall::checkPost(void)
    {
		int outcount[1] = { 0 };
        int *index_array = 0;

        if (my_flags)
        {
            index_array = new int[my_count];
            // don't use malloc in C++
            //            index_array = (int *) malloc(sizeof(int) * my_count);

            outcount[0] = my_count;
            for (int i = 0; i <= my_count; i++)
            {
                index_array[i] = i;
            }
            eraseRequest(outcount, index_array, my_temp_request_array);

            delete[]index_array;
            // don't use malloc/free in C++
            // free (index_array);
        }

        for (int i = 0; i < my_count;i++)
        {
            if (my_flags[i])
            {
                our_request_traces->findAndSetFree(&(my_temp_request_array[i]),getSourceLocation());
                unregisterBuffer(&(my_temp_request_array[i]));
            }
        }

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Testall::checkPrerequisites(void)
    {

        MPO_Tests::checkPrerequisites();

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Testall::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Testall(my_count, my_request_array, my_flags, my_status);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Testall::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Testall";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TESTALL";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Testall::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", *array_of_requests, "
            << "*flag, "
            << "*array_of_statuses";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Testall::save(std::ostream & out) const
    {
        MPO_Callobject::save(out);
        out < my_count;
        // we don't save and restore pointers
        // out < my_request_array;
        // out < my_flag;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Testall::restore(std::istream & in)
    {
        MPO_Callobject::restore(in);
        in > my_count;
        // we don't save and restore pointers
        // in > my_request_array;
        // in > my_flag;
        // in > my_status;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/testall.c.
/*@
  MPI_Testall - Tests for the completion of all previously initiated
  communications

  Input Parameters:
  + count - lists length (integer)
  - array_of_requests - array of requests (array of handles)

  Output Parameters:
  + flag - (logical)
  - array_of_statuses - array of status objects (array of Status).  May be
  'MPI_STATUSES_NULL'.

  Notes:
  'flag' is true only if all requests have completed.  Otherwise, flag is
  false and neither the 'array_of_requests' nor the 'array_of_statuses' is
  modified.

  .N waitstatus

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_IN_STATUS


  int MPI_Testall(
  int count,
  MPI_Request array_of_requests[],
  int *flag,
  MPI_Status array_of_statuses[] )
  @ */

