/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Testany.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Testany.cc 964 2009-06-02 06:38:44Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Testany::MPO_Testany()
        : MPO_Tests(0,
                    0,
                    0),
          my_index(0),
          my_status(0),
          my_temp_request_array(NULL)
    {
    }


    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether request is in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove request from our_requests if test was successful
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     *  (it is a opaque type and thus  may contain some internal status etc.)
     */
    // proper constructor
    //========================================================================
    MPO_Testany::MPO_Testany(int count,
                             MPI_Request * request_array,
                             int *index,
                             int *flag,
                             MPI_Status * status)
        : MPO_Tests(count,
                    request_array,
                    flag),
          my_index(index),
          my_status(status)
    {
        //copy requests into temporary array
        /**
         * TODO
         * I think this does not works with requests as
         * pointers. Though thats something that we don't
         * use anymore ....
         * (Address will change, thus the issue)
         */
        if (!handlePreDataConstructor ())
        {
            my_temp_request_array = NULL;
            if (count >= 0)
            {
                if (request_array)
                {
                    my_temp_request_array = new MPI_Request[count];
                    for (int i = 0; i < count; i++)
                        my_temp_request_array[i] = request_array[i];
                }
            }
        }
    }


    // destructor
    //========================================================================
    MPO_Testany::~MPO_Testany()
    {
        if (!handlePreDataDestructor ())
        {
            if (my_temp_request_array)
                delete my_temp_request_array;
            my_temp_request_array = NULL;
        }
    }

    // setPreData
    //========================================================================
    void MPO_Testany::setPreData (MPO_Callobject *obj)
    {
        my_temp_request_array = ((MPO_Testany*)obj)->my_temp_request_array;
        ((MPO_Testany*)obj)->my_temp_request_array = NULL;
    }

    // clonePreData
    //========================================================================
    MPO_Callobject* MPO_Testany::clonePreData (void)
    {
        MPO_Testany *obj = new MPO_Testany();
        obj->my_temp_request_array = my_temp_request_array;
        return obj;
    }

    // getId
    //========================================================================
    int MPO_Testany::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TESTANY_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Testany::checkPost(void)
    {
		int my_outcount[1], my_index_array[1];

        if (*my_index != MPI_UNDEFINED)
        {
            my_outcount[0] = 1;
        }
        else
        {
            my_outcount[0] = MPI_UNDEFINED;
        }
        my_index_array[0] = *my_index;
        eraseRequest(my_outcount, my_index_array, my_temp_request_array);

        if (*my_index != MPI_UNDEFINED)
        {
            our_request_traces->findAndSetFree(&(my_temp_request_array[*my_index]),getSourceLocation());
            unregisterBuffer(&(my_temp_request_array[*my_index]));
        }

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Testany::checkPrerequisites(void)
    {

        MPO_Tests::checkPrerequisites();

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Testany::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Testany(my_count, my_request_array, my_index, my_flags,
                              my_status);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Testany::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Testany";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TESTANY";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Testany::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", *array_of_requests, "
            << "*index, "
            << "*flag, "
            << "*status";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Testany::save(std::ostream & out) const
    {
        MPO_Callobject::save(out);
        out < my_count;
        // we don't save and restore pointers
        // out < my_request_array;
        // out < my_index;
        // out < my_flag;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Testany::restore(std::istream & in)
    {
        MPO_Callobject::restore(in);
        in > my_count;
        // we don't save and restore pointers
        // in > my_request_array;
        // in > my_index;
        // in > my_flag;
        // in > my_status;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/testany.c.
/*@
  MPI_Testany - Tests for completion of any previdously initiated
  communication

  Input Parameters:
  + count - list length (integer)
  - array_of_requests - array of requests (array of handles)

  Output Parameters:
  + index - index of operation that completed, or 'MPI_UNDEFINED'  if none
  completed (integer)
  . flag - true if one of the operations is complete (logical)
  - status - status object (Status).  May be 'MPI_STATUS_NULL'.

  .N waitstatus

  .N fortran

  .N Errors
  .N MPI_SUCCESS

  int MPI_Testany(
  int count,
  MPI_Request array_of_requests[],
  int *index,
  int *flag,
  MPI_Status *status )
  @ */

