/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Testsome.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Testsome.cc 964 2009-06-02 06:38:44Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Testsome::MPO_Testsome()
        : MPO_Tests(0,
                    0,
                    0),
          my_outcount(0),
          my_index_array(0),
          my_status(0),
          my_temp_request_array(NULL)
    {
    }


    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether request is in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove request from our_requests if test was successful
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     * (it is a opaque type and thus  may contain some internal status etc.)
     */
    // proper constructor
    //========================================================================
    MPO_Testsome::MPO_Testsome(int count,
                               MPI_Request * request_array,
                               int *outcount,
                               int *index_array,
                               MPI_Status * status)
        : MPO_Tests(count,
                    request_array,
                    outcount),
          my_outcount(outcount),
          my_index_array(index_array),
          my_status(status)
    {
        //copy requests into temporary array
        /**
         * TODO
         * I think this does not works with requests as
         * pointers. Though thats something that we don't
         * use anymore ....
         * (Address will change, thus the issue)
         */
        if (!handlePreDataConstructor ())
        {
            my_temp_request_array = NULL;
            if (count >= 0)
            {
                if (request_array)
                {
                    my_temp_request_array = new MPI_Request[count];
                    for (int i = 0; i < count; i++)
                        my_temp_request_array[i] = request_array[i];
                }
            }
        }
    }


    // destructor
    //========================================================================
    MPO_Testsome::~MPO_Testsome()
    {
        if (!handlePreDataDestructor ())
        {
            if (my_temp_request_array)
                delete my_temp_request_array;
            my_temp_request_array = NULL;
        }
    }

    // setPreData
    //========================================================================
    void MPO_Testsome::setPreData (MPO_Callobject *obj)
    {
        my_temp_request_array = ((MPO_Testsome*)obj)->my_temp_request_array;
        ((MPO_Testsome*)obj)->my_temp_request_array = NULL;
    }

    // clonePreData
    //========================================================================
    MPO_Callobject* MPO_Testsome::clonePreData (void)
    {
        MPO_Testsome *obj = new MPO_Testsome();
        obj->my_temp_request_array = my_temp_request_array;
        return obj;
    }

    // getId
    //========================================================================
    int MPO_Testsome::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TESTSOME_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Testsome::checkPost(void)
    {

        eraseRequest(my_outcount, my_index_array, my_temp_request_array);

        for (int i = 0; i < *my_outcount;i++)
        {
            our_request_traces->findAndSetFree(&(my_temp_request_array[my_index_array[i]]),getSourceLocation());
            unregisterBuffer(&(my_temp_request_array[my_index_array[i]]));
        }

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Testsome::checkPrerequisites(void)
    {


        MPO_Tests::checkPrerequisites();


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Testsome::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Testsome(my_count, my_request_array, my_outcount,
                               my_index_array, my_status);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Testsome::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Testsome";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TESTSOME";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Testsome::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(incount = "
            << my_count
            << ", *array_of_requests, "
            << "*outcount, "
            << "*array_of_indeces, "
            << "*array_of_statuses";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Testsome::save(std::ostream & out) const
    {
        MPO_Callobject::save(out);
        out < my_count;
        // we don't save and restore pointers
        // out < my_request_array;
        // out < my_outcount;
        // out < my_index_array;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Testsome::restore(std::istream & in)
    {
        MPO_Callobject::restore(in);
        in > my_count;
        // we don't save and restore pointers
        // in > my_request_array;
        // in > outcount;
        // in > my_index_array;
        // in > my_status;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/testsome.c.
/*@
  MPI_Testsome - Tests for some given communications to complete

  Input Parameters:
  + incount - length of array_of_requests (integer)
  - array_of_requests - array of requests (array of handles)

  Output Parameters:
  + outcount - number of completed requests (integer)
  . array_of_indices - array of indices of operations that
  completed (array of integers)
  - array_of_statuses - array of status objects for
  operations that completed (array of Status).  May be 'MPI_STATUSES_NULL'.

  .N waitstatus

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_IN_STATUS

  int MPI_Testsome(
  int incount,
  MPI_Request array_of_requests[],
  int *outcount,
  int array_of_indices[],
  MPI_Status array_of_statuses[] )
  @ */

