/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_free.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Type_free.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Type_free::MPO_Type_free():my_datatype(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Type_free::MPO_Type_free(MPI_Datatype * datatype)
        : my_datatype(datatype)
    {
    }


    // destructor
    //========================================================================
    MPO_Type_free::~MPO_Type_free()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Type_free::checkPrerequisites()
    {


        // check correctness of parameters
        checkDatatype();


        // remove datatype from our bookkeeping
        our_datatypes.erase(*my_datatype);
        our_datatype_traces->findAndSetFree(*my_datatype,getSourceLocation());


        return 0;
    }


    // checkDatatype
    //========================================================================
    int MPO_Type_free::checkDatatype()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        if (*my_datatype == MPI_DATATYPE_NULL)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Datatype is MPI_DATATYPE_NULL (already freed)!"
                    << std::endl;

                if (my_datatype)
                    our_datatype_traces->findAndPrintTraceNamed(*my_datatype,stream,"datatype");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else if ((our_datatypes.find(*my_datatype) == our_datatypes.end()) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Datatype is NOT valid!"
                << std::endl;

            if (my_datatype)
                our_datatype_traces->findAndPrintTraceNamed(*my_datatype,stream,"datatype");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Type_free::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TYPE_FREE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Type_free::checkPost(void)
    {
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Type_free::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Type_free(my_datatype);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Type_free::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Type_free";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TYPE_FREE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Type_free::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*datatype";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Type_free::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_datatype;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Type_free::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_datatype;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/type_free.c.
/*@
  MPI_Type_free - Frees the datatype

  Input Parameter:
  . datatype - datatype that is freed (handle)

  Predefined types:

  The MPI standard states that (in Opaque Objects)
  .vb
  MPI provides certain predefined opaque objects and predefined, static
  handles to these objects. Such objects may not be destroyed.
  .ve

  Thus, it is an error to free a predefined datatype.  The same section
  makes it clear that it is an error to free a null datatype.

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_TYPE
  .N MPI_ERR_ARG

  int MPI_Type_free ( MPI_Datatype *datatype )
  @ */

