/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_hindexed.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Type_hindexed.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Type_hindexed::MPO_Type_hindexed()
        : MPO_Typeconstructors(-1,
                               MPI_DATATYPE_NULL,
                               -1),
          my_blocklength(0),
          my_displ_array(0),
          my_newtype(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Type_hindexed::MPO_Type_hindexed(int count,
                                         int *blocklength,
                                         MPI_Aint * displ_array,
                                         MPO_Datatype oldtype,
                                         MPI_Datatype *
                                         newtype)
        : MPO_Typeconstructors(count,
                               oldtype,
                               MPO_TYPE_HINDEXED_ID),
          my_blocklength(blocklength),
          my_displ_array(displ_array),
          my_newtype(newtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Type_hindexed::~MPO_Type_hindexed()
    {
    }


    // checkBlocklengthArray
    //========================================================================
    int MPO_Type_hindexed::checkBlocklengthArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(our_errcodes.at(MARMOT_ERR_BLOCKLENGTH_NEG)||our_errcodes.at(MARMOT_WARN_BLOCKLENGTH_ZERO))
    {
        for (int i = 0; i < my_count; i++)
        {
            // (1) Check if blocklength is negative.
            if ((our_errcodes.at(MARMOT_ERR_BLOCKLENGTH_NEG))&&(my_blocklength[i] < 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": blocklength["
                    << i
                    << "] less than 0!"
                    << std::endl;
                rvalue = MARMOT_ERR_BLOCKLENGTH_NEG;

                  our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (2) Check if blocklength is 0.
            else if ((our_errcodes.at(MARMOT_WARN_BLOCKLENGTH_ZERO))&&(my_blocklength[i] == 0) &&
                     (MSG_ERR_WARN <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "WARNING: "
                    << getCallname()
                    << ": blocklength["
                    << i
                    << "] = 0!"
                    << std::endl;
                rvalue = MARMOT_WARN_BLOCKLENGTH_ZERO;

                our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
    // end for
    }
        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Type_hindexed::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TYPE_HINDEXED_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Type_hindexed::checkPost(void)
    {

        our_datatypes.insert(std::make_pair(MPO_Datatype(*my_newtype),
                                            DATATYPE_CREATED_NOT_COMMITTED));

        our_datatype_traces->addAndSetCreation(*my_newtype,getSourceLocation());
        our_datatype_traces->findAndTryAddDependendency(*my_newtype,my_oldtype);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Type_hindexed::checkPrerequisites(void)
    {


        MPO_Typeconstructors::checkPrerequisites();
        checkBlocklengthArray();

        our_datatype_traces->findAndAddUsage(my_oldtype,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Type_hindexed::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Type_hindexed(my_count, my_blocklength, my_displ_array,
                                    my_oldtype, my_newtype);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Type_hindexed::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Type_hindexed";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TYPE_HINDEXED";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Type_hindexed::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", *array_of_blocklengths, "
            << "*array_of_displacements, "
            << "oldtype = "
            << my_oldtype
            << ", *newtype";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Type_hindexed::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_count;
        // out < my_blocklength;
        // out < my_displ_array;
        out < my_oldtype;
        // out < my_newtype;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Type_hindexed::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_count;
        // in > my_blocklength;
        // in > my_displ_array;
        in > my_oldtype;
        // in > my_newtype;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/type_hind.c.
/*@
  MPI_Type_hindexed - Creates an indexed datatype with offsets in bytes

  Input Parameters:
  + count - number of blocks -- also number of entries in indices and
  blocklens
  . blocklens - number of elements in each block (array of nonnegative
  integers)
  . indices - byte displacement of each block (array of MPI_Aint)
  - old_type - old datatype (handle)

  Output Parameter:
  . newtype - new datatype (handle)

  .N fortran

  Also see the discussion for MPI_Type_indexed about the 'indices' in
  Fortran.

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_TYPE
  .N MPI_ERR_COUNT
  .N MPI_ERR_EXHAUSTED
  .N MPI_ERR_ARG

  int MPI_Type_hindexed(
  int count,
  int blocklens[],
  MPI_Aint indices[],
  MPI_Datatype old_type,
  MPI_Datatype *newtype )
  @ */
