/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_hvector.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Type_hvector.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Type_hvector::MPO_Type_hvector()
        : MPO_Typeconstructors(-1,
                               MPI_DATATYPE_NULL,
                               -1),
          my_blocklength(-1),
          my_stride(0),
          my_newtype(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Type_hvector::MPO_Type_hvector(int count,
                                       int blocklength,
                                       MPI_Aint stride,
                                       MPO_Datatype oldtype,
                                       MPI_Datatype *newtype)
        : MPO_Typeconstructors(count,
                               oldtype,
                               MPO_TYPE_HVECTOR_ID),
          my_blocklength(blocklength),
          my_stride(stride),
          my_newtype(newtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Type_hvector::~MPO_Type_hvector()
    {
    }


    // checkBlocklength
    //========================================================================
    int MPO_Type_hvector::checkBlocklength()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if blocklength is negative.
        if ((our_errcodes.at(MARMOT_ERR_BLOCKLENGTH_NEG))&&(my_blocklength < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": blocklength less than 0!"
                << std::endl;
            rvalue = MARMOT_ERR_BLOCKLENGTH_NEG;

            our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if blocklength is 0.
        else if ((our_errcodes.at(MARMOT_WARN_BLOCKLENGTH_ZERO))&&(my_blocklength == 0) &&
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": blocklength = 0!"
                << std::endl;
            rvalue = MARMOT_WARN_BLOCKLENGTH_ZERO;

            our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Type_hvector::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TYPE_HVECTOR_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Type_hvector::checkPost(void)
    {

        our_datatypes.insert(std::make_pair(MPO_Datatype(*my_newtype),
                                            DATATYPE_CREATED_NOT_COMMITTED));

        our_datatype_traces->addAndSetCreation(*my_newtype,getSourceLocation());
        our_datatype_traces->findAndTryAddDependendency(*my_newtype,my_oldtype);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Type_hvector::checkPrerequisites(void)
    {


        MPO_Typeconstructors::checkPrerequisites();
        checkBlocklength();

        our_datatype_traces->findAndAddUsage(my_oldtype,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Type_hvector::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Type_hvector(my_count, my_blocklength, my_stride,
                                   my_oldtype, my_newtype);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Type_hvector::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Type_hvector";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TYPE_HVECTOR";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Type_hvector::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", blocklength = "
            << my_blocklength
            << ", stride = "
            << my_stride
            << ", oldtype = "
            << my_oldtype
            << ", *newtype";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Type_hvector::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_count;
        out < my_blocklength;
        out < my_stride;
        out < my_oldtype;
        // out < my_newtype;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Type_hvector::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_count;
        in > my_blocklength;
        in > my_stride;
        in > my_oldtype;
        // in > my_newtype;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/type_hvec.c.
/*@
  MPI_Type_hvector - Creates a vector (strided) datatype with offset in
  bytes

  Input Parameters:
  + count - number of blocks (nonnegative integer)
  . blocklength - number of elements in each block
  (nonnegative integer)
  . stride - number of bytes between start of each block (integer)
  - old_type - old datatype (handle)

  Output Parameter:
  . newtype - new datatype (handle)

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_TYPE
  .N MPI_ERR_COUNT
  .N MPI_ERR_ARG
  .N MPI_ERR_EXHAUSTED

  int MPI_Type_hvector(
  int count,
  int blocklen,
  MPI_Aint stride,
  MPI_Datatype old_type,
  MPI_Datatype *newtype )
  @ */
