/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Waitall class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Waitall.cc 963 2009-05-27 12:37:15Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Waitall::MPO_Waitall()
        : MPO_Completion(-1,
                         0),
          my_status_array(0)
    {
    }


    // proper constructor
    //========================================================================
    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *
     *  1. Check whether requests are in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *
     *  2. Remove requests from our_request
     *
     */
    MPO_Waitall::MPO_Waitall(int count,
                             MPI_Request * request_array,
                             MPI_Status * status_array)
        : MPO_Completion(count,
                         request_array),
          my_status_array(status_array)
    {
    }


    // destructor
    //========================================================================
    MPO_Waitall::~MPO_Waitall()
    {
    }


    // getId
    //========================================================================
    int MPO_Waitall::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WAITALL_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Waitall::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Waitall::checkPrerequisites(void)
    {

        int outcount = 1;
        int *index_array = 0;
        index_array = new int[my_count];

        MPO_Completion::checkPrerequisites();
        outcount = my_count;
        for (int i = 0; i < my_count; i++)
        {
            index_array[i] = i;
        }
        eraseRequest(&outcount, index_array, my_request_array);

        for (int i = 0;i < my_count;i++)
        {
            our_request_traces->findAndSetFree(&(my_request_array[i]),getSourceLocation());
            unregisterBuffer(&(my_request_array[i]));
        }

        delete[]index_array;

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Waitall::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Waitall(my_count, my_request_array, my_status_array);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Waitall::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Waitall";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WAITALL";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Waitall::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", *array_of_requests, "
            << "*array_of_statuses";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Waitall::save(std::ostream & out) const
    {
        MPO_Callobject::save(out);
        out < my_count;
        // we don't save and restore pointers
        // out < my_request_array;
        // out < my_status_array;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Waitall::restore(std::istream & in)
    {
        MPO_Callobject::restore(in);
        in > my_count;
        // we don't save and restore pointers
        // in > my_request_array;
        // in > my_status_array;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/waitall.c.
/*@

MPI_Waitall - Waits for all given communications to complete

Input Parameters:
+ count - lists length (integer)
- array_of_requests - array of requests (array of handles)

Output Parameter:
. array_of_statuses - array of status objects (array of Status).  May be
'MPI_STATUSES_NULL'

.N waitstatus

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_REQUEST
.N MPI_ERR_ARG
.N MPI_ERR_IN_STATUS
.N MPI_ERR_PENDING

int MPI_Waitall(
int count,
MPI_Request array_of_requests[],
MPI_Status array_of_statuses[] )
@ */

