/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Implementation of public interface of Marmot.
 *
 *  @author Julian Stecklina
 *
 *  $Id: mpo_interface.cc 942 2009-03-27 13:22:07Z tobias $
 */

#include "mpo.h"
#include "mpo-enums.h"
#include "marmot-interface.h"
#include "MPO_LogFile.h"
#include "mpo_LogQueue.h"
#include <assert.h>
#if !defined(__APPLE__)
#include <malloc.h>
#endif
#include <string.h>
#include <stdio.h>
#include "config.h"
#include "MPO_IThreadSynchronisation.h"

int MPO_in_profile = 0;

extern "C" void
marmot_set_state(marmot_state state)
{
	switch (state) {
	case MARMOT_STATE_NORMAL:
#ifdef MARMOT_USE_THREADSYNC
           MPO::MPO_Space::getSyncObject()->wrappingOn();
#else
           MPO_in_profile = 0;
#endif
		break;
	case MARMOT_STATE_IGNORE:
#ifdef MARMOT_USE_THREADSYNC
	    MPO::MPO_Space::getSyncObject()->wrappingOff();
#else
            MPO_in_profile = 1;
#endif
		break;
	default:
		// Throw an error? Ignore?
		;
	}
}

//==================================
// marmot_has_msg
//==================================
extern "C" int marmot_has_msg (void)
{
    int retval=0;

#ifdef MARMOT_USE_THREADSYNC
    MPO::MPO_Space::getSyncObject()->enterPMPI();
#endif

    MPO::MPO_LogFile *log = MPO::MPO_Space::getLogFile ();

    //Ok, I will just assume that the logger is a
    //MPO_LogQueue here and cast it such !
    if (log)
    {
        MPO::MPO_LogQueue *queue = (MPO::MPO_LogQueue*) log;

        if (queue->hasMsg ())
            retval=1;
        else
            retval=0;
    }

#ifdef MARMOT_USE_THREADSYNC
    MPO::MPO_Space::getSyncObject()->leavePMPI();
#endif
    return retval;
}

//==================================
// marmot_get_num_msgs
//==================================
int marmot_get_num_msgs (void)
{
    int retval=0;

#ifdef MARMOT_USE_THREADSYNC
    MPO::MPO_Space::getSyncObject()->enterPMPI();
#endif

    MPO::MPO_LogFile *log = MPO::MPO_Space::getLogFile ();

    //Ok, I will just assume that the logger is a
    //MPO_LogQueue here and cast it such !
    if (log)
    {
        MPO::MPO_LogQueue *queue = (MPO::MPO_LogQueue*) log;

        retval = queue->numMsgs ();
    }

#ifdef MARMOT_USE_THREADSYNC
    MPO::MPO_Space::getSyncObject()->leavePMPI();
#endif
    return retval;
}

//==================================
// marmot_pop_msg
//==================================
extern "C" int marmot_pop_msg (MARMOT_MSG **marmot_msg)
{
    int retval = 1;

#ifdef MARMOT_USE_THREADSYNC
    MPO::MPO_Space::getSyncObject()->enterPMPI();
#endif

    //Get the error from the logger
    MPO::MPO_LogFile *log = MPO::MPO_Space::getLogFile ();
    //Ok, I will just assume that the logger is a
    //MPO_LogQueue here and cast it such !

    MPO::MPO_LogQueue *queue = NULL;

    if (log)
    {
        queue = (MPO::MPO_LogQueue*) log;
        if (!queue->hasMsg ())
        {
#ifdef MARMOT_USE_THREADSYNC
            MPO::MPO_Space::getSyncObject()->leavePMPI();
#endif
            return 1;
        }
    }
    else
    {
#ifdef MARMOT_USE_THREADSYNC
        MPO::MPO_Space::getSyncObject()->leavePMPI();
#endif
        return 1;
    }

    std::pair<MPO::MPO_MsgObject,std::string> pair;
    pair = queue->popMsg ();

    MARMOT_MSG* p;
    //We use malloc here as we return this object to the user,
    //who can't use delete if he is C only
    p = (MARMOT_MSG*) malloc (sizeof (MARMOT_MSG));
    assert (p);

    //Message type
    switch (pair.first.getMsgType())
    {
    case MPO::MSG_TYPE_INFO:
    case MPO::MSG_TYPE_NOTE:
        p->msgType = MARMOT_MSG_TYPE_INFO;
        break;
    case MPO::MSG_TYPE_WARNING:
        p->msgType = MARMOT_MSG_TYPE_WARNING;
        break;
    case MPO::MSG_TYPE_ERROR:
        p->msgType = MARMOT_MSG_TYPE_ERROR;
        break;
    }

    //The three strings
    int len;

    len = strlen (pair.first.getText ().c_str());
    p->strText = (char*) malloc (sizeof(char*) * (len+1));
    assert (p->strText);
    sprintf (p->strText, "%s", pair.first.getText ().c_str());

    len = strlen (pair.first.getCallinfo ().c_str());
    p->strCallInfo = (char*) malloc (sizeof(char*) * (len+1));
    assert (p->strCallInfo);
    sprintf (p->strCallInfo, "%s", pair.first.getCallinfo ().c_str());

    len = strlen (pair.second.c_str());
    p->strHTMLReference = (char*) malloc (sizeof(char*) * (len+1));
    assert (p->strHTMLReference);
    sprintf (p->strHTMLReference, "%s", pair.second.c_str());

    //Global message parameters
    //CURRENTLY NOT SUPPORTED BY MARMOT, so this is easy
    p->isGlobal = 0;
    p->numParticipatingRanks = 0;
    p->participatingRanks = NULL;

    p->eventDescriptor = pair.first.getSourceLocation().getExtraID();

    *marmot_msg = p;
    retval = 0;

#ifdef MARMOT_USE_THREADSYNC
    MPO::MPO_Space::getSyncObject()->leavePMPI();
#endif

    return retval;
}

//==================================
// marmot_free_msg
//==================================
extern "C" int marmot_set_fortran (int yes)
{
    if (yes == 1)
        MPO::INTERFACE_MODE = MPO::F_INTERFACE;
    else
        MPO::INTERFACE_MODE = MPO::C_INTERFACE;

    return 0;
}

//==================================
// marmot_free_msg
//==================================
extern "C" int marmot_free_msg (MARMOT_MSG **marmot_msg)
{
    /*Needs no synchronization ...*/

    if (*marmot_msg)
    {
        if ((*marmot_msg)->strCallInfo)
            free ((*marmot_msg)->strCallInfo);
        (*marmot_msg)->strCallInfo = NULL;

        if ((*marmot_msg)->strText)
            free ((*marmot_msg)->strText);
        (*marmot_msg)->strText = NULL;

        if ((*marmot_msg)->strHTMLReference)
            free ((*marmot_msg)->strHTMLReference);
        (*marmot_msg)->strHTMLReference = NULL;

        if ((*marmot_msg)->isGlobal && (*marmot_msg)->participatingRanks)
            free ((*marmot_msg)->participatingRanks);
        (*marmot_msg)->participatingRanks = NULL;

        free (*marmot_msg);
    }

    *marmot_msg = NULL;

    return 0;
}

//  EOF
