package de.hlrs.marmoclipse.views;

import java.io.File;

import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Table;
import org.eclipse.ui.part.*;
import org.eclipse.jface.viewers.*;
import org.eclipse.jface.action.*;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.*;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.SWT;

import de.hlrs.marmoclipse.log.*;
import de.hlrs.marmoclipse.log.ui.*;

public class MarmotLogView extends ViewPart {
	/**
	 * The ID of the view as specified by the extension.
	 */
	public static final String ID = "de.hlrs.marmoclipse.views.MarmotLogView";
	public static MarmotLogList logs = new MarmotLogList();

	private TableViewer logViewer;
	private LogViewerFilter logFilter;
	private Action clearLogAction;
	private Action showWarningAction;
	private Action showErrorAction;
	private Action showInfoAction;
	private Action showNoteAction;
	private Action saveLogAction;
	
	private String tmpFile;
	private MarmotLogFormat tmpFileFormat;

	private static final String[] tableTitles = { "", "Time", "Rank", "Thread",
		"Messages", "File", "Line", "Call", "Location", "Reference", "11" };
	private static final int[] tableWidths = { 20, 40, 60, 60, 300, 40, 50, 50,
		70, 120, 20 };

	public void setTmpFile(String tmpFile) {
		this.tmpFile = tmpFile;
	}

	public void setTmpFileFormat(MarmotLogFormat tmpFileFormat) {
		this.tmpFileFormat = tmpFileFormat;
	}
	
	
	@Override
	public void createPartControl(Composite parent) {
		createViewer(parent);
		logViewer.setInput(logs);
		makeActions();
		hookContextMenu();
		contributeToActionBars();

		reset();

		// Create the help context id for the viewer's control
		PlatformUI.getWorkbench().getHelpSystem().setHelp(logViewer
			.getControl(),
			"de.hlrs.marmoclipse.viewer");
	}

	private void createViewer(Composite parent) {
		logViewer = new TableViewer(parent, SWT.MULTI | SWT.H_SCROLL
			| SWT.V_SCROLL | SWT.FULL_SELECTION);
		createColumns(logViewer);
		logViewer.setContentProvider(new LogContentProvider(logViewer, logs));
		logViewer.setLabelProvider(new LogLabelProvider());
		logFilter = new LogViewerFilter();
		logViewer.addFilter(logFilter);
	}

	private void createColumns(TableViewer logViewer) {
		for (int i = 0; i < tableTitles.length; i++) {
			TableViewerColumn col = new TableViewerColumn(logViewer, SWT.NONE);
			col.getColumn().setText(tableTitles[i]);
			col.getColumn().setWidth(tableWidths[i]);
			col.getColumn().setResizable(true);
			col.getColumn().setMoveable(true);
			col.setEditingSupport(new LogEditingSupport(logViewer, i));
		}
		Table logTable = logViewer.getTable();
		logTable.setHeaderVisible(true);
		logTable.setLinesVisible(true);
	}

	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		fillLocalPullDown(bars.getMenuManager());
		fillLocalToolBar(bars.getToolBarManager());
	}

	private void fillLocalPullDown(IMenuManager manager) {
		manager.add(clearLogAction);
		manager.add(saveLogAction);
		manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
		manager.add(showInfoAction);
		manager.add(showWarningAction);
		manager.add(showErrorAction);
		manager.add(showNoteAction);
	}

	private void fillLocalToolBar(IToolBarManager manager) {
		manager.add(clearLogAction);
		manager.add(saveLogAction);
	}

	private void hookContextMenu() {
		MenuManager menuMgr = new MenuManager("#PopupMenu");
		menuMgr.setRemoveAllWhenShown(true);
		menuMgr.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				MarmotLogView.this.fillContextMenu(manager);
			}
		});
		Menu menu = menuMgr.createContextMenu(logViewer.getControl());
		logViewer.getControl().setMenu(menu);
		getSite().registerContextMenu(menuMgr, logViewer);
	}

	private void fillContextMenu(IMenuManager manager) {
		manager.add(clearLogAction);
		manager.add(saveLogAction);
		manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
		manager.add(showInfoAction);
		manager.add(showWarningAction);
		manager.add(showErrorAction);
		manager.add(showNoteAction);
	}

	private void makeActions() {
		showInfoAction = new Action("Show informations", IAction.AS_CHECK_BOX) {
			@Override
			public void run() {
				logFilter.setShowInfo(showInfoAction.isChecked());
				logViewer.refresh();
			}
		};
		showInfoAction.setChecked(true);

		showWarningAction = new Action("Show warnings", IAction.AS_CHECK_BOX) {
			@Override
			public void run() {
				logFilter.setShowWarning(showWarningAction.isChecked());
				logViewer.refresh();
			}
		};
		showWarningAction.setChecked(true);

		showErrorAction = new Action("Show errors", IAction.AS_CHECK_BOX) {
			@Override
			public void run() {
				logFilter.setShowError(showErrorAction.isChecked());
				logViewer.refresh();
			}
		};
		showErrorAction.setChecked(true);

		showNoteAction = new Action("Show notes", IAction.AS_CHECK_BOX) {
			@Override
			public void run() {
				logFilter.setShowNote(showNoteAction.isChecked());
				logViewer.refresh();
			}
		};
		showNoteAction.setChecked(true);

		clearLogAction = new Action("Clear Marmot Log") {
			@Override
			public void run() {
				reset();
			}
		};
		clearLogAction.setToolTipText("Clear all Marmot log entries");
		clearLogAction.setImageDescriptor(PlatformUI.getWorkbench()
			.getSharedImages()
			.getImageDescriptor(ISharedImages.IMG_ETOOL_CLEAR));

		saveLogAction = new Action("Save Marmot Log") {
			@Override
			public void run() {
				String ext = tmpFileFormat.getExtention();
				File f = new File(getFullTmpFileName());
				if (f.exists()) {
					FileDialog fd = new FileDialog(getSite().getShell(),
						SWT.SAVE);
					fd.setText("Save Marmot Log ...");
					fd.setOverwrite(true);
					fd.setFilterExtensions(new String[] { ext });
					fd.setFilterNames(new String[] { tmpFileFormat.getName() });
					fd.setFileName(f.getName());
					String newFileName = fd.open();
					if (newFileName != null) {
						f.renameTo(new File(newFileName));
						DeleteTmpFile();
					}
				} else {
					System.out.println("Temporary file not exists.");
				}
			}
		};
		saveLogAction.setImageDescriptor(PlatformUI.getWorkbench()
			.getSharedImages()
			.getImageDescriptor(ISharedImages.IMG_ETOOL_SAVE_EDIT));
	}

	/**
	 * Passing the focus request to the viewer's control.
	 */
	@Override
	public void setFocus() {
		logViewer.getControl().setFocus();

	}

	public boolean reset() {
		if (!logs.getLogs().isEmpty() && tmpFile != null) {
			boolean ret = MessageDialog
				.openQuestion(getSite().getShell(),
					"Log unsaved",
					"The current Marmot log is not saved, do you want to disard it and continue?");
			if (!ret)
				return false;
			DeleteTmpFile();			
		}
		logs.clearLogs();
		clearLogAction.setEnabled(false);
		saveLogAction.setEnabled(false);
		return true;
	}

	public void enableActions() {
		if (!logs.getLogs().isEmpty() && tmpFile != null) {
			saveLogAction.setEnabled(true);
			clearLogAction.setEnabled(true);
		}
	}

	private String getFullTmpFileName() {
		return tmpFile + tmpFileFormat.getExtention().substring(1);
	}
	
	private void DeleteTmpFile() {
		File f = new File(getFullTmpFileName());
		if (f.exists())
			f.delete();
		tmpFile = null;
		saveLogAction.setEnabled(false);
	}
}